// ancova2e.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-29 14:40:33 tac>
//
// 最初の項目を群項目、以後の項目を共変数Xと集計項目Yにし、
// 時期をダミー変数化した共変数とする二元配置型共分散分析
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W ancova2e ancova2e.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "ancova2e  Version 7.41e      Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:ancova2 [<OPTION> ...]  [<DATA FILE> ...]",
  "  The two-way layout ANCOVA with the first item as group, subsequent items",
  "  as covariate and analysis item.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -nx=<NUMBER>:specify the number of covariates",
  "  -res=[-][{def|all|atab|bt|cc[<C-COEF>]|gstat|m[{a|d}]{b|f|s|t}|pd",
  "      |tstat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    atab:ANCOVA table  bt:test of regression coefficient",
  "    cc[<C-COEF>]:confidence intervals(0<=C-COEF<1)  gstat:group statistics",
  "    m[{a|d}]{b|f|s|t}:multiple comparisons",
  "      a:adjusted mean  d:nonparallelism",
  "      b:Bonferroni type  f:Fisher type  s:Scheffe type  t:Tukey type",
  "    pd:ANCOVA when nonparallelism is pooled into residuals",
  "    tstat:overall statistics",
  "    default:auto judgment according tod data",
  "  -t=<TIME1>,<TIME2>[,<TIME3>,...]:specify the relative time of items",
  "    default:period is dummy variable"
};

const string Title = "=== Two-way Layout ANCOVA ==="; // 表題

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  mctype vma;                   // 修正群差の多重比較法
  mctype vmd;                   // 非平行性の多重比較法
  lstr vreslst;                 // 解析結果出力リスト
  map<string, bool> vresmap;    // 解析結果出力の有無マップ
  int vnx;                      // 共変数の個数
  lstr tlst;                    // 時期変数の値リスト
  vecdbl vtvec;                 // 時期変数の値ベクトル
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  mctype ma() const             // 修正群差の多重比較法を返す
  { return vma; }
  mctype md() const             // 非平行性の多重比較法を返す
  { return vmd; }
  bool res(const string& code); // 結果出力フラッグを返す
  int nx() const                // 共変数の個数を返す
  { return vnx; }
  const vecdbl& tvec() const    // 時期変数の値ベクトルを返す
  { return vtvec; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vma(mctuk), vmd(mcbon), vreslst(), vresmap(), vnx(0), tlst(), vtvec()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "nx")      // 共変数の個数指定
  {
    char* ptr = nullptr;
    vnx = strtol(opt[1].c_str(), &ptr, 0);
    if (vnx < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "res")     // 結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 時期変数の値指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  int p = ni - 1 - vnx;
  if (p < 1) throw runtime_error("There are not enough items !");
  if (set_res() < 0) return -3; // -resオプションのエラー
  if (tlst.size() < 1) return 0; // 時期はダミー変数
  vtvec.resize(p);
  if (settvec(vtvec, tlst, p) < p) return disperr(vpname, "Wrong peiods of items !");
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["gstat"] = vresmap["tstat"] = true;
  vresmap["bt"] = vresmap["ma"] = vresmap["md"] = vresmap["pd"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["gstat"] = vresmap["tstat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["gstat"] = vresmap["tstat"] = true;
      }
      else if (res == "all")    // 全ての結果を出力する
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["gstat"] = vresmap["tstat"] = vresmap["bt"] = vresmap["ma"] = vresmap["md"] = vresmap["pd"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "m")      // -res=mと指定された時
      { vresmap["ma"] = vresmap["md"] = true; }
      else if (res[0] == 'm' && res[1] == 'a' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vma = mcbon;
          break;
        case 'f' : vma = mcfis;
          break;
        case 's' : vma = mcsch;
          break;
        case 't' : vma = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["ma"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'd' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["md"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群別統計量クラス
{
private:
  option* voptptr;              // オプションへのポインター
  int vp;                       // 時期数
  int vdim;                     // ベクトルと行列の次数(x+時期変数の個数+2)
  vector<statv<double>> vxstatvec; // xの基礎統計量ベクトル
  vector<statv<double>> vystatvec; // yの基礎統計量ベクトル
  mat<double> smat;             // 積和行列
  vecdbl vmevec;                // xとyの平均ベクトル
  vecdbl vssvec;                // xとyの平方和ベクトル
public:
  gstat() : voptptr(0), vp(0), vdim(0), vxstatvec(), vystatvec(), smat(), vmevec(), vssvec()
  // デフォールトコンストラクタ
  {}
  gstat(option& opt0, int p0); // コンストラクタ
  gstat(const gstat& gstat0) : voptptr(gstat0.voptptr), vp(gstat0.vp), vdim(gstat0.vdim), vxstatvec(gstat0.vxstatvec), vystatvec(gstat0.vystatvec), smat(gstat0.smat), vmevec(gstat0.vmevec), vssvec(gstat0.vssvec)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  const gstat& operator+=(const gstat& gstat0); // 加算代入演算子
  int sigma(const vecdbl& fdatavec); // データを累積してデータ数を返す
  bool setsmat(bool sets); // 平均と積和行列をセットして計算可能性を返す
  void calreg();           // 回帰式を計算する
  int p() const            // 時期数を返す
  { return vp; }
  int nx() const                // 共変数xの個数を返す
  { return voptptr->nx(); }
  int dim() const    // ベクトルと行列の次数を返す(x+時期変数の個数+2)
  { return vdim; }
  const vecdbl& tvec() const    // 時期変数の値ベクトルを返す
  { return voptptr->tvec(); }
  const statv<double>& xstat(int idx) const // xの基礎統計量を返す
  { return vxstatvec[idx]; }
  const statv<double>& ystat(int idx) const // yの基礎統計量を返す
  { return vystatvec[idx]; }
  int n() const                 // 全例数を返す
  { return vystatvec[vp].n(); }
  double me(int idx) const      // 平均値を返す
  { return (0 <= idx && idx < vdim) ? vmevec[idx] : 0.0; }
  double my() const             // yの平均値を返す
  { return vmevec[vdim - 1]; }
  double ss(int idx) const      // 平方和を返す
  { return (0 <= idx && idx < vdim) ? vssvec[idx] : 0.0; }
  double syy() const            // yの平方和を返す
  { return vssvec[vdim - 1]; }
  double vy() const             // yの全分散を返す
  { return vystatvec[vp].vr(); }
  double b(int idx) const       // 偏回帰係数を返す
  { return (0 <= idx && idx < vdim - 1) ? smat(idx, vdim - 1) : 0.0; }
  double vbinv(int idx) const   // 回帰係数の分散要素を返す
  { return (0 <= idx && idx < vdim - 1) ? smat(idx, idx) : 0.0; }
  double sr() const             // 回帰残差を返す
  { return smat(vdim - 1, vdim - 1); }
  double vxr(const vec<double>& dxv) const; // 多重比較用dV^(-1)dを返す
};

gstat::gstat(option& opt0, int p0) : voptptr(&opt0), vp(p0)
// コンストラクタ
{
  if (voptptr == 0) return;       // オプションポインターなし
  int dxp = voptptr->nx() + p0 + 1; // 時期がダミー変数の時の次数
  vdim = ((voptptr->tvec()).size() > 0) ? voptptr->nx() + 3 : dxp;
  vxstatvec.resize(voptptr->nx());  vystatvec.resize(p0 + 1);
  smat.resize(dxp, dxp);  vmevec.resize(dxp);  vssvec.resize(dxp);
}

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  voptptr = gstat0.voptptr;  vp = gstat0.vp;  vdim = gstat0.vdim;
  vxstatvec = gstat0.vxstatvec;  vystatvec = gstat0.vystatvec;
  smat = gstat0.smat;  vmevec = gstat0.vmevec;  vssvec = gstat0.vssvec;
  return *this;                 // 自分自身を返す
}

const gstat& gstat::operator+=(const gstat& gstat0) // 加算代入演算子
{
  for (int j = 0; j < voptptr->nx(); ++j) vxstatvec[j] += gstat0.vxstatvec[j];
  for (int j = 0; j < vp; ++j) vystatvec[j] += gstat0.vystatvec[j];
  smat += gstat0.smat;
  return *this;                 // 自分自身を返す
}

int gstat::sigma(const vecdbl& fdatavec) // データを累積してデータ数を返す
{
  vmevec[0] = 1.0;
  for (int i = 0; i < voptptr->nx(); ++i)
  {
    vxstatvec[i].sigma(fdatavec[i]);
    vmevec[1 + i] = fdatavec[i];
  }
  int nd = 0;
  for (int j = 0; j < vp; ++j)
  {
    int i = voptptr->nx() + j;
    if (fdatavec[i] != BLANKf)
    {
      vystatvec[j].sigma(fdatavec[i]);  ++nd;
      if ((voptptr->tvec()).size() > 0) // 時期は連続変数
      { vmevec[1 + voptptr->nx()] = (voptptr->tvec())[j]; }
      else
      {
        for (int k = 1; k < vp; ++k)
        {
          if (voptptr->nx() + k == i) vmevec[voptptr->nx() + k] = 1.0;
          else vmevec[voptptr->nx() + k] = 0.0;
        }
      }
      vmevec[vdim - 1] = fdatavec[i];
      for (int k = 0; k < vdim; ++k)
      {
        double xk = vmevec[k];
        for (int l = k; l < vdim; ++l) smat.add(k, l, xk * vmevec[l]);
      }
    }
  }
  return nd;                    // 累積したyのデータ数を返す
}

bool gstat::setsmat(bool sets) // 平均と積和行列をセットして例数を返す
{
  for (int i = 0; i < vp; ++i) vystatvec[vp] += vystatvec[i];
  if (sets)                     // 単純積和行列を積和行列にする
  {
    for (int i = 0; i < vdim; ++i)
    { for (int j = i + 1; j < vdim; ++j) smat.assign(j, i, smat(i, j)); }
    smat.sweep(0);
  }
  for (int i = 0; i < vdim; ++i)
  { vmevec[i] = smat(0, i);  vssvec[i] = smat(i, i); }
  return (vystatvec[vp].n() < vdim - 1) ? false : true; // 計算可能性を返す
}

void gstat::calreg()            // 回帰式を計算する
{
  for (int i = 1; i < vdim - 1; ++i) smat.sweep(i);
}

double gstat::vxr(const vec<double>& dxv) const // 多重比較用dV^(-1)dを返す
{
  int dd = dxv.dim();
  vec tv = smat.sub(1, 1, dd, dd) * dxv;
  return inp(dxv, tv);
}

using mapsg = map<string, gstat>;  // 群別統計量マップの別名定義
using mapsg_itr = mapsg::iterator; // 反復子の別名定義
using mapsg_citr = mapsg::const_iterator; // const反復子の別名定義

class varclass                 // 変数クラス
{
private:
  int vni;                     // 項目数
  mapsg vgsmap;                // 群統計量マップ
  gstat vgstatall;             // 全体統計量オブジェクト
  gstat vgstatcom;             // 共通統計量オブジェクト
  int vdfb;                    // 回帰自由度
  int vdfa;                    // 群自由度
  int vdfd;                    // 非平行性自由度
  int vdfr;                    // 残差自由度
  int vdft;                    // 全体自由度
  double vsbc;                 // 共通回帰平方和
  double vsa;                  // 群差平方和
  double vsbt;                 // 全体回帰平方和
  double vsaa;                 // 修正群差平方和
  double vsd;                  // 非平行性平方和
  double vsr;                  // 残差平方和
  double vst;                  // 全体平方和
  void cal_ancova2();          // 共分散分析用各種パラメターを計算する
public:
  varclass(option& opt0, size_t ni); // コンストラクタ
  ~varclass()                   // デストラクタ
  {}
  int ni() const                // 項目数を返す
  { return vni; }
  const mapsg& gsmap() const    // 群統計量マップを返す
  { return vgsmap; }
  const gstat& gstatall() const // 全体統計量オブジェクトを返す
  { return vgstatall; }
  const gstat& gstatcom() const // 共通統計量オブジェクトを返す
  { return vgstatcom; }
  int dfb() const               // 回帰自由度を返す
  { return vdfb; }
  int dfa() const               // 群自由度を返す
  { return vdfa; }
  int dfd() const               // 非平行性自由度を返す
  { return vdfd; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  double sbc() const            // 共通回帰平方和を返す
  { return vsbc; }
  double sa() const             // 群差平方和を返す
  { return vsa; }
  double sbt() const            // 全体回帰平方和を返す
  { return vsbt; }
  double saa() const            // 修正群差平方和を返す
  { return vsaa; }
  double sd() const             // 非平行性平方和を返す
  { return vsd; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int cal(option& opt, ldata& datalst);
  // データを集計して各種統計量を求める
};

varclass::varclass(option& opt0, size_t ni0) : vni(ni0), vgsmap(), vgstatall(), vgstatcom(), vdfb(0), vdfa(0), vdfd(0), vdfr(0), vdft(0), vsbc(0.0), vsa(0.0), vsbt(0.0), vsaa(0.0), vsd(0.0), vsr(0.0), vst(0.0)
// コンストラクタ
{
  int p = vni - 1 - opt0.nx();
  gstat tgstat(opt0, p);        // 一時的統計量オブジェクト
  vgstatall = tgstat;  vgstatcom = tgstat;
}

void varclass::cal_ancova2()   // 共分散分析用各種パラメターを計算する
{
  int g = vgsmap.size();
  vst = vgstatall.syy();
  vsbt = vst - vgstatall.sr();  vsa = 0.0;
  double ssb = 0.0;
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    vsa += gitr->second.n() * gitr->second.my() * gitr->second.my();
    ssb += gitr->second.syy() - gitr->second.sr();
  }
  vsa -= vgstatall.n() * vgstatall.my() * vgstatall.my();
  const double Seps = DBL_EPSILON * fmax(5.0, vst);
  if (vsa < Seps) vsa = 0.0;
  vsbc = vgstatcom.syy() - vgstatcom.sr();
  vsaa = vsa + vsbc - vsbt;  vsd = ssb - vsbc;  vsr = vst - vsa - ssb;
  if (vsaa < Seps) vsaa = 0.0;
  if (vsd < Seps) vsd = 0.0;
  if (vsr < Seps) vsr = 0.0;
  vdft = vgstatall.n() - 1;  vdfb = vgstatall.dim() - 2;
  vdfa = g - 1;
  vdfd = vdfb * vdfa;  vdfr = vdft - vdfa - g * vdfb;
  if (vdfr < 1) vsr = 0.0;
}

int varclass::cal(option& opt, ldata& datalst)
// データを集計して各種統計量を求める
{
  int nxp = vni - 1, p = vni - 1 - opt.nx();
  vecdbl fdatavec(nxp);
  for (auto& cdata : datalst)
  {
    bool xok = true;
    for (int j = 0; xok && j < opt.nx(); ++j) // 共変数のチェック
    {
      if (cdata.fdata(j + 1) == BLANKf) xok = false;
      else fdatavec[j] = cdata.fdata(j + 1);
    }
    if (cdata.sdata(0) != BLANKs && xok)
    {
      for (int j = opt.nx(); j < nxp; ++j) fdatavec[j] = cdata.fdata(j + 1);
      mapsg_itr gitr = vgsmap.find(cdata.sdata(0));
      if (gitr == vgsmap.end()) // 新規群コード
      {
        gstat gstati(opt, p);
        gstati.sigma(fdatavec);
        vgsmap.insert(make_pair(cdata.sdata(0), gstati));
      }
      else                      // すでにある群
      { gitr->second.sigma(fdatavec); }
    }
  }
  for (auto& gmapi : vgsmap)
  {
    vgstatall += gmapi.second;
    bool setsmat = gmapi.second.setsmat(true);
    vgstatcom += gmapi.second;
    if (setsmat) gmapi.second.calreg();
  }
  vgstatall.setsmat(true);  vgstatcom.setsmat(false);
  vgstatall.calreg();  vgstatcom.calreg();
  cal_ancova2();
  return (int)(vgsmap.size());  // 群数を返す
}

class ancova2class              // ancova2メインクラス
{
private:
  option opt;                   // オプション
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout);
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_mreg(ostream& fout, const varclass& var);
  // 群別基礎統計量と回帰式を出力する
  void put_mregtab(ostream& fout, const gstat& gstati, const gstat& gstatcom);
  // 基礎統計量と回帰式の表を出力する
  void put_bt(ostream& fout, const varclass& var);
  // void put_bt(ostream& fout, gstat& gstatcom, int dfr, double sr);
  // 回帰係数の検定結果を出力する
  void put_ancova2tab(ostream& fout,double sbc, int dfb, double sa, int dfa, double sbt, double saa, double sd, int dfd, double sr, int dfr, double st, int dft) const;
  // 共分散分析表を出力する
  void put_macmp(ostream& fout, double sr, int dfr, const varclass& var);
  // 修正群差の多重比較を出力する
  void put_maci(ostream& fout, double sr, int dfr, const varclass& var);
  // 修正群差の信頼区間を出力する
  void put_mdcmp(ostream& fout, double sr, int dfr, const varclass& var);
  // 非平行性の多重比較を出力する
public:
  ancova2class(const char* c_pname = nullptr) : opt(c_pname), fname(), id(), datalst()
  // コンストラクタ
  {}
  ~ancova2class()               // デストラクタ
  {}
  const string& pname() const   // プログラム名を返す
  { return opt.pname(); }
  void init()      // 初期設定を行う(コンストラクタ中ではできない処理)
  { opt.setdef(); }
  void setopt(const string& optstr, char devc = OPTdev) // オプションを設定する
  { opt.set(optstr, devc); }
  int get(const char* c_fname); // データを読み込み累積する
  void put(int nc);             // 集計結果を出力する
};

void ancova2class::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM     :" << id.item(0) << "\n";
  for (int i = 1; i < int(id.ni()); ++i)
  {
    if (i <= opt.nx()) fout << "COVARIATE" << setw(6) << i;
    else fout << "ANALYSIS ITEM" << setw(2) << i - opt.nx();
    fout << ":" << id.item(i) << "\n";
  }
}

void ancova2class::put_mreg(ostream& fout, const varclass& var)
// 群別基礎統計量と回帰式を出力する
{
  int nt = (var.gstatall()).n(), ng = (var.gsmap()).size();
  int p = id.ni() - 1 - opt.nx();
  if (opt.res("gstat"))
  {
    int gno = 1;
    for (mapsg_citr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr, ++gno)
    {
      fout << "\nGROUP" << setw(2) << gno;
      fout << ":" << id.item(0) << "=" << gitr->first << "\n";
      put_mregtab(fout, gitr->second, var.gstatcom());
    }
  }
  if (opt.res("tstat"))
  {
    fout << "\n<OVERALL>\n";
    put_mregtab(fout, var.gstatall(), var.gstatcom());
  }
}

void ancova2class::put_mregtab(ostream& fout, const gstat& gstati, const gstat& gstatcom)
// 基礎統計量と回帰式の表を出力する
{
  constexpr int LENline = 79;   // 区切り線の長さ
  constexpr int Nterm = 4;      // 1行に出力する項数
  putline(fout, LENline);
  for (int i = 0; i < gstati.nx(); ++i)
  {
    fout << "X" << left << setw(5) << i + 1 << right << ":";
    putmsd(fout, gstati.xstat(i).n(), gstati.xstat(i).mean(), gstati.xstat(i).vr());
  }
  for (int j = 0; j < gstati.p(); ++j)
  {
    fout << "Y:t" << left << setw(3);
    if ((gstati.tvec()).size() > 0) fout << (gstati.tvec())[j];
    else fout << j;
    fout << right << ":";
    putmsd(fout, gstati.ystat(j).n(), gstati.ystat(j).mean(), gstati.ystat(j).vr());
  }
  putline(fout, LENline);
  fout << "Y     :";
  putmsd(fout, gstati.n(), gstati.my(), gstati.vy());
  double bc0 = gstati.my(), syy = gstati.syy();
  double sb = syy - gstati.sr();
  double r2 = (syy > 0.0) ? sb / syy : 0.0;
  fout << "Group Regression Equation :y=" << setw(14) <<  gstati.b(0);
  for (int j = 1; j < gstati.dim() - 1; ++j)
  {
    if ((j - 3) % Nterm == 0) fout << "\n" << setw(2) << "";
    fout << showpos << setw(11) << gstati.b(j) << noshowpos;
    if (j <= gstati.nx())
    { fout << "x" << left << setw(2) << j << right; }
    else
    {
      if ((gstati.tvec()).size() > 0) fout << "t";
      else fout << "t" << left << setw(2) << j - gstati.nx();
      fout << right;
    }
    bc0 -= gstatcom.b(j) * gstati.me(j);
  }
  fout << "\nCommon Regression Equation:y=" << setw(14) << bc0;
  for (int j = 1; j < gstatcom.dim() - 1; ++j)
  {
    if ((j - 3) % Nterm == 0) fout << "\n" << setw(2) << "";
    fout << showpos << setw(11) << gstatcom.b(j) << noshowpos;
    if (j <= gstatcom.nx())
    { fout << "x" << left << setw(2) << j << right; }
    else
    {
      if ((gstatcom.tvec()).size() > 0) fout << "t";
      else fout << "t" << left << setw(2) << j - gstatcom.nx();
      fout << right;
    }
  }
  double r = sqrt(r2);
  int dfb = gstati.dim() - 2;
  int dfr = gstati.n() - 1 - dfb;
  double f = (gstati.sr() > 0.0) ? sb / (double)dfb / gstati.sr() * (double)dfr : 0.0;
  ldbl pr = ftop(f, dfb, dfr);
  fout << "\nGroup Contribution Rate:r-squared=" << r2 << "  r=" << r << "  p=" << pr << pind(pr) << endl;
}

void ancova2class::put_bt(ostream& fout, const varclass& var)
// 回帰係数の検定結果を出力する
{
  int p = id.ni() - 2;
  double vr = (var.dfr() > 0) ? var.sr() / var.dfr() : 0.0;
  fout << "\n";
  constexpr int LENline = 79;   // 区切り線の長さ
  fout << "<TEST AND CI OF COMMON REGRESSION COEFFICIENT>\n";
  fout << "VAR    Partial b     SE of b" << setw(3) << opt.cc() * 100.0 << "%CI Lower       Upper     F-value     p-value\n";
  putline(fout, LENline);
  for (int j = 1; j < (var.gstatcom()).dim() - 1; ++j)
  {
    double bj = (var.gstatcom()).b(j), vbj = vr * (var.gstatcom()).vbinv(j);
    double fj = bj * bj / vbj;
    double pr = ftop(fj, 1, var.dfr());
    double ci = calmcci(mcfis, opt.cc(), 1, 1, var.dfr(), 1, vbj);
    if (j <= (var.gstatcom()).nx())
    { fout << "X" << left << setw(3) << j; }
    else
    {
      if ((var.gstatcom().tvec()).size() > 0) fout << left << setw(4) << "t";
      else fout << "t" << left << setw(3) << j - (var.gstatcom()).nx();
    }
    fout << right << setw(12) << bj;
    fout << setw(12) << sqrt(vbj) << setw(12) << bj - ci << setw(12) << bj + ci;
    fout << setw(12) << fj << setw(12) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
}

void ancova2class::put_ancova2tab(ostream& fout,double sbc, int dfb, double sa, int dfa, double sbt, double saa, double sd, int dfd, double sr, int dfr, double st, int dft) const
// 共分散分析表を出力する
{
  fout << "\n";
  constexpr int LENline = 78;   // 区切り線の長さ
  string tstr;
  if (dfd > 0) tstr = "ANCOVA2 Table";
  else tstr = "ANCOVA2 Table(Nonparallelism Pooled)";
  int w = (LENline - strclen(tstr, CHARwide)) / 2 + tstr.size();
  fout << setw(w) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0, fd = 0.0;
  double vr = calvfp(v, f, pr, sa, dfa, sr, dfr);
  tstr = "GROUP";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, sbc, dfb, sr, dfr);
  tstr = "COMMON-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sbc;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, saa, dfa, sr, dfr);
  tstr = "ADJ-GROUP";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << saa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, sbt, dfb, sr, dfr);
  tstr = "OVERALL-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(11 + npad) << tstr << right << setw(13) << sbt;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  if (dfd > 0)
  {
    calvfp(v, f, pr, sd, dfd, sr, dfr);
    fd = f;
    tstr = "NONPRALLEL";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(10 + npad) << tstr << right << setw(14) << sd;
    fout << setw(10) << dfd << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
  }
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void ancova2class::put_macmp(ostream& fout, double sr, int dfr, const varclass& var)
// 修正群差の多重比較を出力する
{
  if (sr <= 0.0) return;        // 出力なし
  int g = (var.gsmap()).size(), d = (var.gstatcom()).dim() - 1;
  int df1 = 1, dc = 1;
  string method = "TUKEY";
  switch (opt.ma())
  {
  case mcbon : method = "BONFERRONI";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON>\n";
  fout << "  Covariate=Overall Mean  ";
  if (opt.ma() == mctuk)
  {
    fout << "GROUPS=" << g << "  DF=" << dfr << "\n";
    fout << " G:    Adj-Mean -  G:    Adj-Mean    Adj-Diff     q-value       p-value\n";
  }
  else
  {
    fout << "DF1=" << df1 << "  DF2=" << dfr << "\n";
    fout << " G:    Adj-Mean -  G:    Adj-Mean    Adj-Diff     F-value       p-value\n";
  }
  constexpr int LENline = 75;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = sr / dfr;
  vec<double> dxv(d - 1);
  double yc = 0.0;
  for (int k = 1; k < d; ++k) yc += (var.gstatcom()).b(k) * (var.gstatall()).me(k);
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    double ni = (double)(gitri->second.n());
    double bc0i = gitri->second.my();
    for (int k = 1; k < d; ++k) bc0i -= (var.gstatcom()).b(k) * gitri->second.me(k);
    double yi = bc0i + yc;
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      double bc0j = gitrj->second.my();
      for (int k = 1; k < d; ++k)
      {
        bc0j -= (var.gstatcom()).b(k) * gitrj->second.me(k);
        dxv.assign(k - 1, gitri->second.me(k) - gitrj->second.me(k));
      }
      double yj = bc0j + yc;
      double dy = bc0j - bc0i;
      double vd = (1.0 / ni + 1.0 / (double)(gitrj->second.n()) + (var.gstatcom()).vxr(dxv)) * vr;
      double f = dy * dy / vd / (double)df1;
      double pr = calmcpr(opt.ma(), f, g, df1, dfr, dc);
      fout << setw(2) << j + 1 << ":" << setw(12) << yj << " -";
      fout << setw(3) << i + 1 << ":" << setw(12) << yi << setw(12) << dy;
      fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void ancova2class::put_maci(ostream& fout, double sr, int dfr, const varclass& var)
// 修正群差の信頼区間を出力する
{
  if (sr <= 0.0) return;        // 出力なし
  int g = (var.gsmap()).size(), d = (var.gstatcom()).dim() - 1;
  int df1 = 1, dc = 1;
  string method = "TUKEY";
  switch (opt.ma())
  {
  case mcbon : method = "BONFERRONI";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "\n<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVAL>\n";
  fout << " G -  G    Adj-Mean   Precision   Lower-Lim   Upper-Lim\n";
  const int LENline = 55;       // 区切り線の長さ
  putline(fout, LENline);
  double vr = sr / dfr;
  vec<double> dxv(d - 1);
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    double ni = (double)(gitri->second.n());
    double bc0i = gitri->second.my();
    for (int k = 1; k < d; ++k) bc0i -= (var.gstatcom()).b(k) * gitri->second.me(k);
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      double bc0j = gitrj->second.my();
      for (int k = 1; k < d; ++k)
      {
        bc0j -= (var.gstatcom()).b(k) * gitrj->second.me(k);
        dxv.assign(k - 1, gitri->second.me(k) - gitrj->second.me(k));
      }
      double dy = bc0j - bc0i;
      double vd = (1.0 / ni + 1.0 / gitrj->second.n() + (var.gstatcom()).vxr(dxv)) * vr;
      double ci = calmcci(opt.ma(), opt.cc(), g, df1, dfr, dc, vd);
      fout << setw(2) << j + 1 << " -" << setw(3) << i + 1 << setw(12) << dy;
      fout << setw(12) << ci << setw(12) << dy - ci << setw(12) << dy + ci << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void ancova2class::put_mdcmp(ostream& fout, double sr, int dfr, const varclass& var)
// 非平行性の多重比較を出力する
{
  if (sr <= 0.0) return;        // 出力なし
  int g = (var.gsmap()).size(), d = (var.gstatcom()).dim() - 1;
  int df1 = d - 1, dc = dc = g * (g - 1) / 2;
  string method = "BONFERRONI";
  switch (opt.md())
  {
  case mcfis : method = "FISHER";  dc = 1;
    break;
  case mcsch : method = "SCHEFFE";  df1 = (g - 1) * df1;  dc = 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON>\n";
  fout << "  DF1=" << df1 << "  DF2=" << var.dfr() << "\n";
  fout << " G -  G   SS of Partial b     F-value       p-value\n";
  const int LENline = 55;       // 区切り線の長さ
  putline(fout, LENline);
  double vr = sr / dfr;
  vec<double> dbv(d - 1);
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      gstat gstatc(opt, (var.gstatcom()).p());
      gstatc += gitri->second;  gstatc += gitrj->second;
      gstatc.calreg();
      for (int k = 1; k < d; ++k)
      { dbv.assign(k - 1, gitrj->second.b(k) - gitri->second.b(k)); }
      double sdb = gstatc.vxr(dbv);
      double f = sdb / vr / df1;
      double pr = calmcpr(opt.md(), f, g, df1, dfr, dc);
      fout << setw(2) << j + 1 << " -" << setw(3) << i + 1;
      fout << setw(18) << sdb << setw(12) << f;
      fout << setw(14) << pr << pind(pr) << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

int ancova2class::get(const char* c_fname) // データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void ancova2class::put(int nc)  // 集計結果を出力する
{
  varclass var(opt, id.ni());     // 変数オブジェクト
  int ng = var.cal(opt, datalst); // 各種統計量を求める
  if (nc < 1 || ng < 1)           // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
  }
  double vd = 0.0, fd = 0.0, pr = 0.0;
  calvfp(vd, fd, pr, var.sd(), var.dfd(), var.sr(), var.dfr());
  if (opt.res("res"))
  {
    put_resid(cout);
    put_mreg(cout, var);
    if (opt.res("bt")) put_bt(cout, var);
    if (opt.res("atab")) put_ancova2tab(cout, var.sbc(), var.dfb(), var.sa(), var.dfa(), var.sbt(), var.saa(), var.sd(), var.dfd(), var.sr(), var.dfr(), var.st(), var.dft());
    if (opt.res("ma"))
    {
      put_macmp(cout, var.sr(), var.dfr(), var);
      if (opt.res("cc")) put_maci(cout, var.sr(), var.dfr(), var);
    }
    if (opt.res("md")) put_mdcmp(cout, var.sr(), var.dfr(), var);
    if ((opt.res("def") && fd < 1.0) || opt.res("pd"))
    {
      double srt = var.st() - var.saa() - var.sbt();
      int dfrt = var.dft() - var.dfa() - var.dfb();
      put_ancova2tab(cout, var.sbc(), var.dfb(), var.sa(), var.dfa(), var.sbt(), var.saa(), 0.0, 0, srt, dfrt, var.st(), var.dft());
      if (opt.res("ma"))
      {
        put_macmp(cout, srt, dfrt, var);
        if (opt.res("cc")) put_maci(cout, srt, dfrt, var);
      }
    }
    cout << "\n" << endl;
  }
}

int main(int argc, char *argv[]) // メインルーチン
{
  ancova2class ancova2(argv[0]); // メインクラス変数
  try
  {
    ancova2.init();             // ancova2の初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nc = 0;                 // 読み込んだケース数
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) nc += ancova2.get(nullptr); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { ancova2.setopt(argv[i] + 1);  fok = true; }
      else
      { nc += ancova2.get(argv[i]);  fok = false; }
    }
    ancova2.put(nc);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(ancova2.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(ancova2.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
