// ancovae.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-29 14:35:44 tac>
//
// 最初の項目を群項目、次の項目を集計項目Y、以後の項目を共変数Xとした共分散分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W ancove ancovae.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "ancovae  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:ancovae [<OPTION> ...]  [<DATA FILE> ...]",
  "  The analysis of covariance with the first item as group, the second item",
  "  as analysis item Y, subsequent items as covariate X.",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=ancovae_dns.d",
  "    output data:group regression coefficient",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|atab|bt|cc[<C-COEF>]|gstat|m[{a|d}]{b|f|s|t}|pd",
  "      |tstat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    atab:ANCOVA table  bt:test of regression coefficient",
  "    cc[<C-COEF>]:confidence intervals(0<=C-COEF<1)  gstat:group statistics",
  "    m[{a|d}]{b|f|s|t}:multiple comparisons",
  "      a:adjusted mean  d:nonparallelism",
  "      b:Bonferroni type  f:Fisher type  s:Scheffe type  t:Tukey type",
  "    pd:ANCOVA when nonparallelism is pooled into residuals",
  "    tstat:overall statistics",
  "    default:auto judgment according tod data"
};

const string Title = "=== Aanalysis of Covariance(ANCOVA) ==="; // 表題
const string DNSNAMEdef = "ancovae_dns.d"; // 省略時DANSファイル名

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  mctype vma;                   // 修正群差の多重比較法
  mctype vmd;                   // 非平行性の多重比較法
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  mctype ma() const             // 修正群差の多重比較法を返す
  { return vma; }
  mctype md() const             // 非平行性の多重比較法を返す
  { return vmd; }
  bool res(const string& code); // 結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdnsname(), vma(mctuk), vmd(mcbon), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if (set_res() < 0) return -3; // -resオプションのエラー
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["gstat"] = vresmap["tstat"] = true;
  vresmap["bt"] = vresmap["ma"] = vresmap["md"] = vresmap["pd"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["gstat"] = vresmap["tstat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["gstat"] = vresmap["tstat"] = true;
      }
      else if (res == "all")    // 全ての結果を出力する
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["gstat"] = vresmap["tstat"] = vresmap["bt"] = vresmap["ma"] = vresmap["md"] = vresmap["pd"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "m")      // -res=mと指定された時
      { vresmap["ma"] = vresmap["md"] = true; }
      else if (res[0] == 'm' && res[1] == 'a' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vma = mcbon;
          break;
        case 'f' : vma = mcfis;
          break;
        case 's' : vma = mcsch;
          break;
        case 't' : vma = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["ma"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'd' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["md"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群別統計量クラス
{
private:
  mat<double> smat;             // 積和行列
  int vn;                       // 例数
  vecdbl vmevec;                // xとyの平均ベクトル
  vecdbl vssvec;                // xとyの平方和ベクトル
public:
  gstat() : smat(), vn(0), vmevec(), vssvec() // デフォールトコンストラクタ
  {}
  explicit gstat(size_t nd) : smat(nd), vn(0), vmevec(nd), vssvec(nd)
  // コンストラクタ
  {}
  gstat(const gstat& gstat0) : smat(gstat0.smat), vn(gstat0.vn), vmevec(gstat0.vmevec), vssvec(gstat0.vssvec)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  const gstat& operator+=(const gstat& gstat0); // 加算代入演算子
  void sigma(const vecdbl& fdata);              // データを累積する
  int setsmat(bool sets);       // 平均と積和行列を設定して例数を返す
  void calreg();                // 回帰式を計算する
  int n() const                 // 例数を返す
  { return vn; }
  double me(size_t idx) const   // 平均値を返す
  { return (idx < vmevec.size()) ? vmevec[idx] : 0.0; }
  double my() const             // yの平均値を返す
  { return vmevec[vmevec.size() - 1]; }
  double ss(size_t idx) const   // 平方和を返す
  { return (idx < vssvec.size()) ? vssvec[idx] : 0.0; }
  double vr(size_t idx) const   // 分散を返す
  { return (idx < vssvec.size() && vn > 1) ? vssvec[idx] / (vn - 1) : 0.0; }
  double syy() const            // yの平方和を返す
  { return vssvec[vssvec.size() - 1]; }
  double vyy() const            // yの分散を返す
  { return (vn > 1) ? vssvec[vssvec.size() - 1] / (vn - 1) : 0.0; }
  double sr() const             // 回帰残差を返す
  { return smat(smat.r() - 1, smat.c() - 1); }
  double b(size_t idx) const    // 回帰係数を返す
  { return (idx < smat.r() - 1) ? smat(idx, smat.c() - 1) : 0.0; }
  double vbinv(size_t idx) const // 回帰係数の分散要素を返す
  { return (idx < smat.r() - 1) ? smat(idx, idx) : 0.0; }
  double vxr(const vec<double>& dxv) const; // 多重比較用dV^(-1)dを返す
};

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  smat = gstat0.smat;  vn = gstat0.vn;
  vmevec = gstat0.vmevec;  vssvec = gstat0.vssvec;
  return *this;
}

const gstat& gstat::operator+=(const gstat& gstat0) // 加算代入演算子
{
  smat += gstat0.smat;  vn += gstat0.vn;
  return *this;
}

void gstat::sigma(const vecdbl& fdatavec) // データを累積する
{
  vmevec[0] = 1.0;
  for (size_t i = 0; i < fdatavec.size(); ++i) vmevec[i + 1] = fdatavec[i];
  for (size_t i = 0; i <= fdatavec.size(); ++i)
  {
    double x = vmevec[i];
    for (size_t j = i; j <= fdatavec.size(); ++j) smat.add(i, j, x * vmevec[j]);
  }
}

int gstat::setsmat(bool sets)  // 平均と積和行列をセットして例数を返す
{
  if (sets)                     // 単純積和行列を積和行列にする
  {
    vn = (int)(smat(0, 0));
    for (size_t i = 0; i < smat.r() - 1; ++i)
    { for (size_t j = i + 1; j < smat.c(); ++j) smat.assign(j, i, smat(i, j)); }
    smat.sweep(0);
  }
  for (size_t i = 0; i < smat.r(); ++i)
  { vmevec[i] = smat(0, i);  vssvec[i] = smat(i, i); }
  return vn;                    // 例数を返す
}

void gstat::calreg()            // 回帰式を計算する
{
  for (size_t i = 1; i < smat.r() - 1; ++i) smat.sweep(i);
}

double gstat::vxr(const vec<double>& dxv) const // 多重比較用dV^(-1)dを返す
{
  int p = dxv.dim();
  vec<double> tv = smat.sub(1, 1, p, p) * dxv;
  return inp(dxv, tv);
}

using mapsg = map<string, gstat>;  // 群別統計量マップの別名定義
using mapsg_itr = mapsg::iterator; // 反復子の別名定義
using mapsg_citr = mapsg::const_iterator; // const反復子の別名定義

class varclass                 // 変数クラス
{
private:
  int vni;                     // 項目数
  mapsg vgsmap;                // 群統計量マップ
  gstat vgstatall;             // 全体統計量オブジェクト
  gstat vgstatcom;             // 共通統計量オブジェクト
  int vdfb;                    // 回帰自由度
  int vdfa;                    // 群自由度
  int vdfd;                    // 非平行性自由度
  int vdfr;                    // 残差自由度
  int vdft;                    // 全体自由度
  double vsbc;                 // 共通回帰平方和
  double vsa;                  // 群差平方和
  double vsbt;                 // 全体回帰平方和
  double vsaa;                 // 修正群差平方和
  double vsd;                  // 非平行性平方和
  double vsr;                  // 残差平方和
  double vst;                  // 全体平方和
  void cal_ancova();           // 共分散分析用各種パラメターを計算する
public:
  varclass(size_t ni) : vni(ni), vgsmap(), vgstatall(ni), vgstatcom(ni), vdfb(0), vdfa(0), vdfd(0), vdfr(0), vdft(0), vsbc(0.0), vsa(0.0), vsbt(0.0), vsaa(0.0), vsd(0.0), vsr(0.0), vst(0.0)
  // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int ni() const                // 項目数を返す
  { return vni; }
  const mapsg& gsmap() const    // 群統計量マップを返す
  { return vgsmap; }
  const gstat& gstatall() const // 全体統計量オブジェクトを返す
  { return vgstatall; }
  const gstat& gstatcom() const // 共通統計量オブジェクトを返す
  { return vgstatcom; }
  int dfb() const               // 回帰自由度を返す
  { return vdfb; }
  int dfa() const               // 群自由度を返す
  { return vdfa; }
  int dfd() const               // 非平行性自由度を返す
  { return vdfd; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  double sbc() const            // 共通回帰平方和を返す
  { return vsbc; }
  double sa() const             // 群差平方和を返す
  { return vsa; }
  double sbt() const            // 全体回帰平方和を返す
  { return vsbt; }
  double saa() const            // 修正群差平方和を返す
  { return vsaa; }
  double sd() const             // 非平行性平方和を返す
  { return vsd; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int cal(const option& opt, ldata& datalst);
  // データを集計して各種統計量を求める
};

void varclass::cal_ancova()    // 共分散分析用各種パラメターを計算する
{
  int g = vgsmap.size();
  vst = vgstatall.syy();
  vsbt = vst - vgstatall.sr();  vsa = 0.0;
  double ssb = 0.0;
  for (auto& gmapi : vgsmap)
  {
    vsa += gmapi.second.n() * gmapi.second.my() * gmapi.second.my();
    ssb += gmapi.second.syy() - gmapi.second.sr();
  }
  vsa -= vgstatall.n() * vgstatall.my() * vgstatall.my();
  const double Seps = DBL_EPSILON * fmax(5.0, vst);
  if (vsa < Seps) vsa = 0.0;
  vsbc = vgstatcom.syy() - vgstatcom.sr();
  vsaa = vsa + vsbc - vsbt;  vsd = ssb - vsbc;  vsr = vst - vsa - ssb;
  if (vsaa < Seps) vsaa = 0.0;
  if (vsd < Seps) vsd = 0.0;
  if (vsr < Seps) vsr = 0.0;
  vdft = vgstatall.n() - 1;
  vdfb = vni - 2;  vdfa = g - 1;
  vdfd = vdfb * vdfa;  vdfr = vdft - vdfa - g * vdfb;
  if (vdfr < 1) vsr = 0.0;
}

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して各種統計量を求める
{
  int p = vni - 2;
  vecdbl fdatavec(p + 1);
  for (auto& cdata : datalst)
  {
    bool blank = false;
    fdatavec[p] = cdata.fdata(1);
    for (int j = 0; blank == false && j < p; ++j)
    {
      fdatavec[j] = cdata.fdata(j + 2);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (cdata.sdata(0) != BLANKs && blank == false && fdatavec[p] != BLANKf)
    {
      mapsg_itr gitr = vgsmap.find(cdata.sdata(0));
      if (gitr == vgsmap.end()) // 新規群コード
      {
        gstat gstati(vni);
        gstati.sigma(fdatavec);
        vgsmap.insert(make_pair(cdata.sdata(0), gstati));
      }
      else                      // すでにある群
      { gitr->second.sigma(fdatavec); }
    }
  }
  for (auto& gmapi : vgsmap)
  {
    vgstatall += gmapi.second;
    gmapi.second.setsmat(true);
    vgstatcom += gmapi.second;
    gmapi.second.calreg();
  }
  vgstatall.setsmat(true);  vgstatcom.setsmat(false);
  vgstatall.calreg();  vgstatcom.calreg();
  cal_ancova();
  return (int)(vgsmap.size());  // 群数を返す
}

class ancovaclass               // ancovaメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout); // 結果のタイトルとデータ識別情報を出力する
  void put_mreg(ostream& fout, option& opt, const varclass& var);
  // 群別平均と回帰式を出力する
  void put_mregtab(ostream& fout, const gstat& gstati, const gstat& gstatcom) const;
  // 平均と回帰式の表を出力する
  void put_bg(ostream& fout, const option& opt, const varclass& var) const;
  // 群別回帰係数の検定結果を出力する
  void put_bc(ostream& fout, const option& opt, const varclass& var) const;
  // 共通回帰係数の検定結果を出力する
  void put_ancovatab(ostream& fout,double sbc, int dfb, double sa, int dfa, double sbt, double saa, double sd, int dfd, double sr, int dfr, double st, int dft) const;
  // 共分散分析表を出力する
  void put_macmp(ostream& fout, const option& opt, double sr, int dfr, const varclass& var) const;
  // 修正群差の多重比較を出力する
  void put_maci(ostream& fout, const option& opt, double sr, int dfr, const varclass& var) const;
  // 修正群差の信頼区間を出力する
  void put_mdcmp(ostream& fout, const option& opt, double sr, int dfr, const varclass& var) const;
  // 非平行性の多重比較を出力する
  void put_dnsid(ofstream& fout);
  // DANSデータのタイトルとデータ識別情報を出力する
  void put_dns(option& opt, const varclass& var);
  // 群別回帰係数をデータファイル形式で出力する
public:
  ancovaclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~ancovaclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

void ancovaclass::put_resid(ostream& fout)
// 結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM     :" << id.item(0) << "\n";
  fout << "ANALYSIS ITEM Y:" << id.item(1) << "\n";
  for (size_t i = 2; i < id.ni(); ++i)
  {
    fout << "COVARIATE  X" << setw(3) << i - 1;
    fout << ":" << id.item(i) << "\n";
  }
}

void ancovaclass::put_mreg(ostream& fout, option& opt, const varclass& var)
// 群別平均と回帰式を出力する
{
  if (opt.res("gstat"))
  {
    int gno = 1;
    for (const auto& gmapi : var.gsmap())
    {
      fout << "\nGROUP" << setw(2) << gno;
      fout << ":" << id.item(0) << "=" << gmapi.first << "\n";
      put_mregtab(fout, gmapi.second, var.gstatcom());
      ++gno;
    }
  }
  if (opt.res("tstat"))
  {
    fout << "\n<OVERALL>\n";
    put_mregtab(fout, var.gstatall(), var.gstatcom());
  }
}

void ancovaclass::put_mregtab(ostream& fout, const gstat& gstati, const gstat& gstatcom) const
// 平均と回帰式の表を出力する
{
  constexpr int LENline = 79;   // 区切り線の長さ
  constexpr int Nterm = 5;      // 1行あたりの項数
  putline(fout, LENline);
  int p = id.ni() - 2;
  double bc0 = gstati.my(), syy = gstati.syy();
  fout << "Y     :";
  putmsd(fout, gstati.n(), bc0, gstati.vyy());
  for (int j = 1; j <= p; ++j)
  {
    fout << "X" << left << setw(5) << j << right << ":";
    putmsd(fout, gstati.n(), gstati.me(j), gstati.vr(j));
  }
  putline(fout, LENline);
  double sb = syy - gstati.sr();
  double r2 = (syy > 0.0) ? sb / syy : 0.0;
  fout << "Group Regression Equation :y=" << setw(14) << gstati.b(0);
  for (int j = 1; j <= p; ++j)
  {
    if ((j - 3) % Nterm == 0) fout << "\n" << setw(2) << "";
    fout << showpos << setw(11) << gstati.b(j) << noshowpos;
    fout << "x" << left << setw(2) << j << right;
    bc0 -= gstatcom.b(j) * gstati.me(j);
  }
  fout << "\nCommon Regression Equation:y=" << setw(14) << bc0;
  for (int j = 1; j <= p; ++j)
  {
    if ((j - 3) % Nterm == 0) fout << "\n" << setw(2) << "";
    fout << showpos << setw(11) << gstatcom.b(j) << noshowpos;
    fout << "x" << left << setw(2) << j << right;
  }
  double r = sqrt(r2);
  if (p == 1) r = sgn(gstati.b(1)) * r;
  int dfr = gstati.n() - 1 - p;
  double f = (gstati.sr() > 0.0) ? sb / (double)p / gstati.sr() * (double)dfr : 0.0;
  ldbl pr = ftop(f, p, dfr);
  fout << "\nGroup Contribution Rate:r-squared=" << r2 << "  r=" << r << "  p=" << pr << pind(pr) << endl;
}

void ancovaclass::put_bg(ostream& fout, const option& opt, const varclass& var) const
// 群別回帰係数の検定結果を出力する
{
  const int LENline = 74;       // 区切り線の長さ
  int p = id.ni() - 2, g = (var.gsmap()).size();
  double vr = (var.dfr() > 0) ? var.sr() / (double)(var.dfr()) : 0.0;
  int df1 = 1, dc = g * (g - 1) / 2;
  string method = "BONFERRONI";
  switch (opt.md())
  {
  case mcfis : method = "FISHER";  dc = 1;
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;  dc = 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON OF GROUP REGRESSION COEFFICIENT>\n";
  fout << "G -G  VAR     Diff fo b" << setw(3) << opt.cc() * 100.0 << "%CI Lower       Upper     F-value     p-value\n";
  putline(fout, LENline);
  int gno1 = 1;
  for (mapsg_citr gitr1 = (var.gsmap()).begin(); gitr1 != (var.gsmap()).end(); ++gitr1, ++gno1)
  {
    mapsg_citr gitr2 = gitr1;
    ++gitr2;
    if (gitr2 != (var.gsmap()).end())
    {
      int gno2 = gno1 + 1;
      while (gitr2 != (var.gsmap()).end())
      {
        for (int i = 1; i <= p; ++i)
        {
          double bdi = (gitr2->second).b(i) - (gitr1->second).b(i);
          double vbdi = vr * ((gitr1->second).vbinv(i) + (gitr2->second).vbinv(i));
          double fi = bdi * bdi / vbdi / (double)df1;
          double pr = calmcpr(opt.md(), fi, g, df1, var.dfr(), dc);
          double ci = calmcci(opt.md(), opt.cc(), g, df1, var.dfr(), dc, vbdi);
          fout << setw(2) << gno2 << "-" << setw(2) << gno1 << "  x" << left << setw(3) << i << right << setw(12) << bdi;
          fout << setw(12) << bdi - ci << setw(12) << bdi + ci;
          fout << setw(12) << fi << setw(12) << pr << pind(pr) << "\n";
        }
        ++gitr2;  ++gno2;
      }
    }
  }
  putline(fout, LENline);
}

void ancovaclass::put_bc(ostream& fout, const option& opt, const varclass& var) const
// 共通回帰係数の検定結果を出力する
{
  int p = id.ni() - 2;
  double vr = (var.dfr() > 0) ? var.sr() / (double)(var.dfr()) : 0.0;
  fout << "\n";
  constexpr int LENline = 67;   // 区切り線の長さ
  fout << "<TEST AND CI OF COMMON REGRESSION COEFFICIENT>\n";
  fout << "VAR    Partial b" << setw(3) << opt.cc() * 100.0 << "%CI Lower       Upper     F-value     p-value\n";
  putline(fout, LENline);
  for (int i = 1; i <= p; ++i)
  {
    double bi = (var.gstatcom()).b(i), vbi = vr * (var.gstatcom()).vbinv(i);
    double fi = bi * bi / vbi;
    double pr = ftop(fi, 1, var.dfr());
    double ci = calmcci(mcfis, opt.cc(), 1, 1, var.dfr(), 1, vbi);
    fout << "X" << left << setw(3) << i << right << setw(12) << bi;
    fout << setw(12) << bi - ci << setw(12) << bi + ci;
    fout << setw(12) << fi << setw(12) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
}

void ancovaclass::put_ancovatab(ostream& fout,double sbc, int dfb, double sa, int dfa, double sbt, double saa, double sd, int dfd, double sr, int dfr, double st, int dft) const
// 共分散分析表を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  fout << "\n";
  string tstr;
  if (dfd > 0) tstr = "ANCOVA Table";
  else tstr = "ANCOVA Table(Nonparallelism Pooled)";
  fout << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, sa, dfa, sr, dfr);
  tstr = "GROUP";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, sbc, dfb, sr, dfr);
  tstr = "COMMON-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sbc;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, saa, dfa, sr, dfr);
  tstr = "ADJ-GROUP";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << saa;
  fout << setw(10) << dfa << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, sbt, dfb, sr, dfr);
  tstr = "OVERALL-REG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(11 + npad) << tstr << right << setw(13) << sbt;
  fout << setw(10) << dfb << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  if (dfd > 0)
  {
    calvfp(v, f, pr, sd, dfd, sr, dfr);
    tstr = "NONPRALLEL";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(10 + npad) << tstr << right << setw(14) << sd;
    fout << setw(10) << dfd << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
  }
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << st;
  fout << setw(10) << dft << endl;
}

void ancovaclass::put_macmp(ostream& fout, const option& opt, double sr, int dfr, const varclass& var) const
// 修正群差の多重比較を出力する
{
  if (sr <= 0.0) return;        // 出力なし
  int g = (var.gsmap()).size(), p = id.ni() - 2;
  int df1 = 1, dc = 1;
  string method = "TUKEY";
  switch (opt.ma())
  {
  case mcbon : method = "BONFERRONI";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON>\n";
  fout << "  Covariate=Overall Mean  ";
  if (opt.ma() == mctuk)
  {
    fout << "GROUPS=" << g << "  DF=" << dfr << "\n";
    fout << " G:    Adj-Mean -  G:    Adj-Mean    Adj-Diff     q-value       p-value\n";
  }
  else
  {
    fout << "DF1=" << df1 << "  DF2=" << dfr << "\n";
    fout << " G:    Adj-Mean -  G:    Adj-Mean    Adj-Diff     F-value       p-value\n";
  }
  constexpr int LENline = 75;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = sr / dfr;
  vec<double> dxv(p);
  double yc = 0.0;
  for (int k = 1; k <= p; ++k) yc += (var.gstatcom()).b(k) * (var.gstatall()).me(k);
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    double ni = (double)(gitri->second.n());
    double bc0i = gitri->second.my();
    for (int k = 1; k <= p; ++k) bc0i -= (var.gstatcom()).b(k) * gitri->second.me(k);
    double yi = bc0i + yc;
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      double bc0j = gitrj->second.my();
      for (int k = 1; k <= p; ++k)
      {
        bc0j -= (var.gstatcom()).b(k) * gitrj->second.me(k);
        dxv.assign(k - 1, gitri->second.me(k) - gitrj->second.me(k));
      }
      double yj = bc0j + yc;
      double dy = bc0j - bc0i;
      double vd = (1.0 / ni + 1.0 / (double)(gitrj->second.n()) + (var.gstatcom()).vxr(dxv)) * vr;
      double f = dy * dy / vd / (double)df1;
      double pr = calmcpr(opt.ma(), f, g, df1, dfr, dc);
      fout << setw(2) << j + 1 << ":" << setw(12) << yj << " -";
      fout << setw(3) << i + 1 << ":" << setw(12) << yi << setw(12) << dy;
      fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void ancovaclass::put_maci(ostream& fout, const option& opt, double sr, int dfr, const varclass& var) const
// 修正群差の信頼区間を出力する
{
  if (sr <= 0.0) return;        // 出力なし
  int g = (var.gsmap()).size(), p = id.ni() - 2;
  int df1 = 1, dc = 1;
  string method = "TUKEY";
  switch (opt.ma())
  {
  case mcbon : method = "BONFERRONI";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "\n<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVAL>\n";
  fout << " G -  G    Adj-Mean   Precision   Lower-Lim   Upper-Lim\n";
  constexpr int LENline = 55;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = sr / dfr;
  vec<double> dxv(p);
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    double ni = (double)(gitri->second.n());
    double bc0i = gitri->second.my();
    for (int k = 1; k <= p; ++k) bc0i -= (var.gstatcom()).b(k) * gitri->second.me(k);
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      double bc0j = gitrj->second.my();
      for (int k = 1; k <= p; ++k)
      {
        bc0j -= (var.gstatcom()).b(k) * gitrj->second.me(k);
        dxv.assign(k - 1,  gitri->second.me(k) - gitrj->second.me(k));
      }
      double dy = bc0j - bc0i;
      double vd = (1.0 / ni + 1.0 / (double)(gitrj->second.n()) + (var.gstatcom()).vxr(dxv)) * vr;
      double ci = calmcci(opt.ma(), opt.cc(), g, df1, dfr, dc, vd);
      fout << setw(2) << j + 1 << " -" << setw(3) << i + 1 << setw(12) << dy;
      fout << setw(12) << ci << setw(12) << dy - ci << setw(12) << dy + ci << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void ancovaclass::put_mdcmp(ostream& fout, const option& opt, double sr, int dfr, const varclass& var) const
// 非平行性の多重比較を出力する
{
  if (sr <= 0.0) return;        // 出力なし
  int g = (var.gsmap()).size(), p = id.ni() - 2;
  int df1 = p, dc = g * (g - 1) / 2;
  string method = "BONFERRONI";
  switch (opt.md())
  {
  case mcfis : method = "FISHER";  dc = 1;
    break;
  case mcsch : method = "SCHEFFE";  df1 = (g - 1) * p;  dc = 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON>\n";
  fout << "  DF1=" << df1 << "  DF2=" << var.dfr() << "\n";
  fout << " G -  G   SS of Reg-Coeff     F-value       p-value\n";
  constexpr int LENline = 55;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = sr / dfr;
  vec<double> dbv(p);
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < g - 1; ++i, ++gitri)
  {
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      gstat gstatc(p + 2);
      gstatc += gitri->second;  gstatc += gitrj->second;
      gstatc.calreg();
      for (int k = 1; k <= p; ++k)
      { dbv.assign(k - 1, gitri->second.b(k) - gitrj->second.b(k)); }
      double sdb = gstatc.vxr(dbv);
      double f = sdb / vr / df1;
      double pr = calmcpr(opt.md(), f, g, df1, dfr, dc);
      fout << setw(2) << j + 1 << " -" << setw(3) << i + 1;
      fout << setw(18) << sdb << setw(12) << f;
      fout << setw(14) << pr << pind(pr) << "\n";
      ++gitrj;  ++j;
    }
  }
  putline(fout, LENline);
}

void ancovaclass::put_dnsid(ofstream& fout)
// DANSデータのタイトルとデータ識別情報を出力する
{
  idd tid(id.title(), id.ni());
  tid.assign_item(0, id.item(0));
  ostringstream ostr;
  ostr << "Group Regression Intercept";
  tid.assign_item(1, ostr.str());
  for (size_t i = 2; i < id.ni(); ++i)
  {
    ostr.str("");
    ostr << "Group Regresion Coefficient of " << id.item(i);
    tid.assign_item(i, ostr.str());
  }
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  tid.put_item(fout, formtag);
}

void ancovaclass::put_dns(option& opt, const varclass& var)
// 群別回帰係数をデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout);
  datac cdata(id.ni());
  for (const auto& gmapi : var.gsmap())
  {
    cdata.assign(0, gmapi.first);
    cdata.assign(1, (gmapi.second).b(0));
    for (size_t i = 2; i < id.ni(); ++i) cdata.assign(i, (gmapi.second).b(i - 1));
    cdata.put(fout);
  }
}

int ancovaclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void ancovaclass::put(option& opt) // 集計結果を出力する
{
  varclass var(id.ni());          // 変数オブジェクト
  int ng = var.cal(opt, datalst); // 各種統計量を求める
  if (id.nc() < 1 || ng < 1)      // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  double vd = 0.0, fd = 0.0, pr = 0.0;
  calvfp(vd, fd, pr, var.sd(), var.dfd(), var.sr(), var.dfr());
  if (opt.res("res"))
  {
    put_resid(cout);
    put_mreg(cout, opt, var);
    if (opt.res("bt"))
    {
      put_bg(cout, opt, var);
      put_bc(cout, opt, var);
    }
    if (opt.res("atab")) put_ancovatab(cout, var.sbc(), var.dfb(), var.sa(), var.dfa(), var.sbt(), var.saa(), var.sd(), var.dfd(), var.sr(), var.dfr(), var.st(), var.dft());
    if (opt.res("ma"))
    {
      put_macmp(cout, opt, var.sr(), var.dfr(), var);
      if (opt.res("cc")) put_maci(cout, opt, var.sr(), var.dfr(), var);
    }
    if (opt.res("md")) put_mdcmp(cout, opt, var.sr(), var.dfr(), var);
    if (var.dfd() > 0)          // 非平行性をプールした結果出力
    {
      if ((opt.res("def") && fd < 1.0) || opt.res("pd"))
      {
        double srt = var.sd() + var.sr();
        int dfrt = var.dfd() + var.dfr();
        if (opt.res("atab")) put_ancovatab(cout, var.sbc(), var.dfb(), var.sa(), var.dfa(), var.sbt(), var.saa(), 0.0, 0, srt, dfrt, var.st(), var.dft());
        if (opt.res("ma"))
        {
          put_macmp(cout, opt, srt, dfrt, var);
          if (opt.res("cc")) put_maci(cout, opt, srt, dfrt, var);
        }
      }
    }
    cout << "\n" << endl;
  }
  if (opt.dnsname() != "") put_dns(opt, var);
}

int main(int argc, char *argv[]) // メインルーチン
{
  ancovaclass ancova;           // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) ancova.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { ancova.get(argv[i], opt);  fok = false; }
    }
    ancova.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
