// ccae.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-29 14:24:01 tac>
//
// 指定した2つの項目グループ間で正準相関分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W ccae ccae.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "ccae  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:cca [<OPTION> ...]  [<DATA FILE> ...]",
  "  The canonical correlation analysis with 2 item groups.",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=ccae_dns.d",
  "    output:canonical score",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|ccc|ccl|cl|cvc|stat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    ccc:canonical correlation coefficient  ccl:cross loading",
  "    cl:canonical loading  cvc:coefficient of canonical variate",
  "    stat:basic statistics  default:all",
  "  -x=[<START NUMBER>][-[<END NUMBER>]]:specify the item number of X",
  "    default:-x=1",
  "  -y=[<START NUMBER>][-[<END NUMBER>]]:specify the item number of Y",
  "    default:-y=2"
};

const string Title = "=== Canonical Correlation Analysis ==="; // 表題
const string DNSNAMEdef = "ccae_dns.d"; // 省略時DANSファイル名

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  bool vdns;                    // 正準スコアの出力フラッグ
  string vdnsname;              // DANSファイル名
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  bool vstat;                   // 基礎統計量の出力フラッグ
  pairint xno;                  // 変数xの開始・終了項目番号
  pairint yno;                  // 変数yの開始・終了項目番号
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  bool dns() const              // 正準スコアの出力フラッグを返す
  { return vdns; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  bool stat() const             // 基礎統計量出力フラッグを返す
  { return vstat; }
  int xs() const                // 変数xの開始番号を返す
  { return xno.first; }
  int xe() const                // 変数xの終了番号を返す
  { return xno.second; }
  int nx() const                // 変数xの数を返す
  { return xno.second - xno.first + 1; }
  int ys() const                // 変数yの開始番号を返す
  { return yno.first; }
  int ye() const                // 変数yの終了番号を返す
  { return yno.second; }
  int ny() const                // 変数yの数を返す
  { return yno.second - yno.first + 1; }
  bool res(const string& code); // 結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vdns(false), vdnsname(), vreslst(), vresmap(), vstat(true), xno(1, 1), yno(2, 2)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "x")       // 変数Xの項目番号指定
  {
    xno.first = 1;  xno.second = INT_MAX;
    if (getrange(xno.first, xno.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (xno.first < 1 || xno.first > xno.second) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "y")       // 変数yの項目番号指定
  {
    yno.first = 1;  yno.second = INT_MAX;
    if (getrange(yno.first, yno.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (yno.first < 1 || yno.first > yno.second) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if ((int)ni < xno.second || (int)ni < yno.second) return disperr(vpname, "Wrong item number !");
  if (set_res() < 0) return -3; // -resオプションのエラー
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["ccc"] = vresmap["ccl"] = vresmap["cl"] = vresmap["cvc"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["ccc"] = vresmap["ccl"] = vresmap["cl"] = vresmap["cvc"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all") // 全ての結果を出力する
      {
        vresmap["def"] = vresmap["ccc"] = vresmap["ccl"] = vresmap["cl"] = vresmap["cvc"] = vresmap["stat"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class ccaclass                  // ccaメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_smat(const option& opt, mat<ldbl>& smat); // 積和行列を求める
  void put_resid(ostream& fout, const option& opt);
  // 解析結果のタイトルとデータ識別情報を出力する
  int cal_cca(const option& opt, vecldbl& evalvec, mat<ldbl>& amat, mat<ldbl>& bmat, const mat<ldbl>& smat) const;
  // 正準相関を計算する
  void put_dnscca(option& opt, const mat<ldbl>& amat, const mat<ldbl>& bmat, int n, const mat<ldbl>& smat);
  // 正準スコアをデータファイル形式で出力する
  void put_dnsid(ostream& fout);
  // DANSデータのタイトルとデータ識別情報を出力する
  void amat_cmat(int is, int ie, mat<ldbl>& cmat, const mat<ldbl>& smat, int n) const;
  // 固有ベクトルを正準変量係数に変換する
  void cal_score(const option& opt, datac& score, datac& cdata, const mat<ldbl>& amat, const mat<ldbl>& bmat, int q) const;
  // 正準スコアを計算し、データとして返す
  void put_ccatab(ostream& fout, option& opt, int n, const mat<ldbl>& smat, const vecldbl& evalvec, const mat<ldbl>& amat, const mat<ldbl>& bmat);
  // 正準相関分析結果を出力する
  void amat_lmat(int is, int ie, mat<ldbl>& amat, const mat<ldbl>& smat) const;
  // 固有ベクトルを正準変量との相関係数に変換する
  void put_cl(ostream& fout, const string& mtitle, const string& rhed, const string& ched, mat<ldbl>& lmat) const;
  // 正準負荷量または交差負荷量を出力する。
public:
  ccaclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~ccaclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int ccaclass::cal_smat(const option& opt, mat<ldbl>& smat) // 積和行列を求める
{
  size_t dim = smat.r();
  smat = 0.0L;
  vecdbl fdatavec(dim);
  fdatavec[0] = 1.0L;
  for (auto& cdata : datalst)
  {
    bool blank = false;
    for (int j = opt.xs(); blank == false && j <= opt.xe(); ++j)
    {
      int k = 1 + j - opt.xs();
      fdatavec[k] = cdata.fdata(j - 1);
      if (fdatavec[k] == BLANKf) blank = true;
    }
    for (int j = opt.ys(); blank == false && j <= opt.ye(); ++j)
    {
      int k = 1 + opt.nx() + j - opt.ys();
      fdatavec[k] = cdata.fdata(j - 1);
      if (fdatavec[k] == BLANKf) blank = true;
    }
    if (blank == false)
    {
      for (size_t i = 0; i < dim; ++i)
      {
        double x = fdatavec[i];
        for (size_t j = i; j < dim; ++j) smat.add(i, j, (ldbl)(x * fdatavec[j]));
      }
    }
  }
  for (size_t i = 0; i < smat.r() - 1; ++i)
  { for (size_t j = i + 1; j < smat.c(); ++j) smat.assign(j, i, smat(i, j)); }
  int n = (int)(smat(0, 0));
  smat.sweep(0);                // 単純積和行列を積和行列に変換する
  return n;                     // 例数を返す
}

void ccaclass::put_resid(ostream& fout, const option& opt)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (int i = opt.xs(); i <= opt.xe(); ++i)
  {
    fout<< "VARIABLE X" << setw(3) << i - opt.xs() + 1;
    fout << ":" << id.item(i - 1) << "\n";
  }
  for (int i = opt.ys(); i <= opt.ye(); ++i)
  {
    fout << "VARIABLE Y" << setw(3) << i - opt.ys() + 1;
    fout << ":" << id.item(i - 1) << "\n";
  }
}

int ccaclass::cal_cca(const option& opt, vecldbl& evalvec, mat<ldbl>& amat, mat<ldbl>& bmat, const mat<ldbl>& smat) const
// 正準相関を計算する
{
  int p = opt.nx() + opt.ny();
  mat<ldbl> rmat = smat.sub(1, 1, smat.r() - 1, smat.c() - 1);
  smat2rmat(rmat);
  mat<ldbl> rxxmat = rmat.sub(0, 0, opt.nx() - 1, opt.nx() - 1);
  mat<ldbl> rxxrmat = rxxmat;
  rxxrmat.inv();
  mat<ldbl> rxymat = rmat.sub(0, opt.nx(), opt.nx() - 1, p - 1);
  mat<ldbl> ryyrmat = rmat.sub(opt.nx(), opt.nx(), p - 1, p - 1);
  ryyrmat.inv();
  mat<ldbl> ryxmat = rmat.sub(opt.nx(), 0, p - 1, opt.nx() - 1);
  mat<ldbl> tmat = rxxrmat * rxymat * ryyrmat * ryxmat;
  evalvec[0] = 0.0;
  for (int i = 0; i < opt.nx(); i++) evalvec[0] += tmat(i, i);
  int qmax = min(opt.nx(), opt.ny()), q = 0;
  vec<ldbl> axv(opt.nx()), ayv(opt.nx());
  while (q < qmax && (evalvec[q + 1] = eigenud(tmat, axv, ayv)) > 0.0L)
  {
    ++q;
    vec<ldbl> tv = rxxmat * axv;
    ldbl w = inp(axv, tv);
    axv *= (w > 0.0L) ? sqrt(1.0L / w) : 0.0L;
    vec<ldbl> bv = ryyrmat * ryxmat * axv;
    bv *= sqrt(1.0L / evalvec[q]);
    for (int i = 0; i < opt.nx(); ++i) amat.assign(i + 1, q, axv(i));
    for (int i = 0; i < opt.ny(); ++i) bmat.assign(i + 1, q, bv(i));
  }
  return q;                     // 正準相関数を返す
}

void ccaclass::put_dnscca(option& opt, const mat<ldbl>& amat, const mat<ldbl>& bmat, int n, const mat<ldbl>& smat)
// 正準スコアをデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout);
  size_t q = amat.c() - 1;
  idd tid(id.title(), q * 2);
  ostringstream ostr;
  for (size_t i = 0; i < q; ++i)
  {
    ostr << "Canonical Score U" << i + 1;
    tid.assign_item(i, ostr.str());
    ostr.str("");
  }
  for (size_t i = 0; i < q; ++i)
  {
    ostr << "Canonical Score V" << i + 1;
    tid.assign_item(q + i, ostr.str());
    ostr.str("");
  }
  tid.put_item(fout, formtag);
  datac score(q * 2);
  mat<ldbl> camat = amat;
  amat_cmat(1, opt.nx(), camat, smat, n);
  mat<ldbl> cbmat = bmat;
  amat_cmat(opt.nx() + 1, opt.nx() + opt.ny(), cbmat, smat, n);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    cal_score(opt, score, *ditr, camat, cbmat, q);
    score.put(fout);
  }
}

void ccaclass::put_dnsid(ostream& fout)
// DANSデータのタイトルとデータ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
}

void ccaclass::amat_cmat(int is, int ie, mat<ldbl>& cmat, const mat<ldbl>& smat, int n) const
// 固有ベクトルを正準変量係数に変換する
{
  int p = ie - is + 1, q = cmat.c() - 1;
  vecldbl sdrvec(p, 0.0L);
  for (int i = is; i <= ie; ++i)
  { sdrvec[i - is] = (smat(i, i) > 0.0L) ? sqrt((ldbl)(n - 1) / smat(i, i)) : 0.0L; }
  for (int j = 0; j <= q; ++j) cmat.assign(0, j, 0.0L);
  for (int i = 1; i <= p; ++i)
  {
    ldbl w = sdrvec[i - 1], m = smat(0, i + is - 1);
    for (int j = 1; j <= q; ++j)
    {
      cmat.assign(i, j, cmat(i, j) * w);
      cmat.add(0, j, -cmat(i, j) * m);
    }
  }
}

void ccaclass::cal_score(const option& opt, datac& score, datac& cdata, const mat<ldbl>& amat, const mat<ldbl>& bmat, int q) const
// 正準スコアを計算し、データとして返す
{
  vecldbl fdatavec(opt.nx() + opt.ny(), 0.0L);
  bool blank = false;
  for (int j = opt.xs(); blank == false && j <= opt.xe(); ++j)
  {
    int k = j - opt.xs();
    fdatavec[k] = cdata.fdata(j - 1);
    if (fdatavec[k] == BLANKf) blank = true;
  }
  for (int j = opt.ys(); blank == false && j <= opt.ye(); ++j)
  {
    int k = opt.nx() + j - opt.ys();
    fdatavec[k] = cdata.fdata(j - 1);
    if (fdatavec[k] == BLANKf) blank = true;
  }
  if (blank)                    // スコアの計算ができない
  { score = BLANKs; }
  else
  {
    for (int i = 1; i <= q; ++i)
    {
      ldbl z = amat(0, i);
      for (int j = 0; j < opt.nx(); ++j) z += amat(j + 1, i) * fdatavec[j];
      score.assign(i - 1, z);
    }
    for (int i = 1; i <= q; ++i)
    {
      ldbl z = bmat(0, i);
      for (int j = 0; j < opt.ny(); ++j) z += bmat(j + 1, i) * fdatavec[opt.nx() + j];
      score.assign(q + i - 1, z);
    }
  }
}

void ccaclass::put_ccatab(ostream& fout, option& opt, int n, const mat<ldbl>& smat, const vecldbl& evalvec, const mat<ldbl>& amat, const mat<ldbl>& bmat)
// 正準相関分析結果を出力する
{
  constexpr int LENline = 72;   // 区切り線の長さ
  int p = opt.nx() + opt.ny(), q = amat.c() - 1;
  if (opt.res("ccc"))
  {
    fout << "\nNo.   CC-Coefficient     Eigen Value  Normal z-value         p-value\n";
    putline(fout, LENline);
    ldbl plamda = 1.0;
    for (int i = 1; i <= q; ++i)
    {
      ldbl lamda = evalvec[i], z = 0.0L;
      if (lamda >= 1.0)
      { lamda = 1.0; }
      else
      { plamda *= (1.0L - lamda);  z = sqrt((ldbl)n * lamda / (1.0L - lamda)); }
      ldbl pr = ztop(z);
      fout << "No" << setw(2) << i << setw(16) << sqrt(lamda);
      fout << setw(16) << lamda << setw(16) << z;
      fout << setw(16) << pr << pind(pr) << "\n";
    }
    putline(fout, LENline);
    ldbl csqr = - ((ldbl)(n - 1) - (ldbl)(p + 1) * 0.5L) * log(plamda);
    int df = opt.nx() * opt.ny();
    ldbl pr = ctop(csqr, df);
    fout << "Test of All CC-Coefficients=0:ch-squared=" << csqr << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  }
  mat<ldbl> cmat = amat;
  amat_cmat(1, opt.nx(), cmat, smat, n);
  if (opt.res("cvc")) putcoef(fout, "<COEFFICIENT OF CANONICAL VARIATE>", "U", "X", cmat);
  cmat = bmat;
  amat_cmat(opt.nx() + 1, p, cmat, smat, n);
  if (opt.res("cvc")) putcoef(fout, "<COEFFICIENT OF CANONICAL VARIATE>", "V", "Y", cmat);
  cmat = amat;
  amat_lmat(1, opt.nx(), cmat, smat);
  if (opt.res("cl")) put_cl(fout, "<CANONINICAL LOADING(CORRELATION OF X AND U)>", "U", "X", cmat);
  for (size_t i = 1; i < cmat.r(); ++i)
  {
    for (size_t j = 1; j < cmat.c(); ++j) cmat.assign(i, j, cmat(i, j) * sqrt(evalvec[j]));
  }
  if (opt.res("ccl")) put_cl(fout, "<CROSS LOADING(CORRELATION OF X AND V)>", "V", "X", cmat);
  cmat = bmat;
  amat_lmat(opt.nx() + 1, p, cmat, smat);
  if (opt.res("cl")) put_cl(fout, "<CANONICAL LOADING(CORRELATION OF Y AND V)>", "V", "Y", cmat);
  for (size_t i = 1; i < cmat.r(); ++i)
  {
    for (size_t j = 1; j < cmat.c(); ++j) cmat.assign(i, j, cmat(i, j) * sqrt(evalvec[j]));
  }
  if (opt.res("ccl")) put_cl(fout, "<CROSS LOADING(CORRELATION OF Y AND U)>", "U", "Y", cmat);
}

void ccaclass::amat_lmat(int is, int ie, mat<ldbl>& amat, const mat<ldbl>& smat) const
// 固有ベクトルを正準負荷量(正準変量との相関係数)に変換する
{
  mat<ldbl> rmat = smat.sub(is, is, ie, ie);
  smat2rmat(rmat);
  mat<ldbl> tmat = amat.sub(1, 1, amat.r() - 1, amat.c() - 1);
  mat<ldbl> lmat = rmat * tmat;
  for (size_t i = 1; i < amat.r(); ++i)
  {
    for (size_t j = 1; j < amat.c(); ++j) amat.assign(i, j, lmat(i - 1, j - 1));
  }
}

void ccaclass::put_cl(ostream& fout, const string& mtitle, const string& rhed, const string& ched, mat<ldbl>& lmat) const
// 正準負荷量または交差負荷を量出力する。
{
  constexpr size_t Cmax = 6;    // 行列表横の最大数
  size_t pidx = lmat.r() - 1, qidx = lmat.c() - 1, low = 1, up = 0;
  vecldbl l2vec(lmat.c(), 0.0L); // 負荷量平方和
  bool eop = false;             // 表出力終了フラッグ
  for (size_t i = 0; i <= pidx; ++i) lmat.assign(i, 0, 0.0L);
  fout << "\n" << mtitle;
  while (eop == false)
  {
    size_t nc = low + Cmax - 1;
    if (low > qidx) eop = true;
    else if (nc > qidx) { up = qidx;  eop = true; }
    else if (nc == qidx) up = qidx;
    else up = nc;
    fout << "\nVAR ";
    for (size_t j = low; j <= up; ++j)
    {
      int npad = rhed.size() - strclen(rhed, CHARwide);
      fout << setw(9 + npad) << rhed << setw(3) << j;
    }
    if (eop) fout << "         r^2";
    fout << "\n";
    size_t llen = 4 + 12 * (up - low + 1 + (eop ? 1 : 0));
    putline(fout, llen);
    for (size_t i = 1; i <= pidx; ++i)
    {
      fout << setw(1) << ched << setw(3) << i;
      int prec0 = fout.precision();
      fout.precision(6);
      fout.setf(ios::fixed);
      for (size_t j = low; j <= up; ++j)
      {
        fout << setw(12) << lmat(i, j);
        ldbl l2 = lmat(i, j) * lmat(i, j);
        lmat.assign(i, 0, lmat(i, 0) + l2);
        l2vec[j] += l2;
      }
      if (eop) fout << setw(12) << lmat(i, 0);
      fout << "\n";
      fout.unsetf(ios::fixed);
      fout.precision(prec0);
    }
    putline(fout, llen);
    int prec0 = fout.precision();
    fout.precision(6);
    fout.setf(ios::fixed);
    fout << "r^2 ";
    for (size_t j = low; j <= up; ++j)
    {
      ldbl r2 = l2vec[j] / (ldbl)pidx;
      fout << setw(12) << r2;
      l2vec[0] += r2;
    }
    if (eop) fout << setw(12) << l2vec[0];
    fout << "\n";
    low = up + 1;
  }
}

int ccaclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void ccaclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, opt);
      cout << "\n" << endl;
    }
    return;
  }
  int dim = 1 + opt.nx() + opt.ny(); // 次元数
  mat<ldbl> smat(dim);               // 積和行列
  int n = cal_smat(opt, smat);       // 例数
  if (opt.res("res"))
  {
    put_resid(cout, opt);
    if (opt.res("stat"))
    {
      vecldbl mevec(smat.r(), 0.0), vrvec(smat.r(), 0.0);
      for (size_t i = 0; i < smat.r(); ++i) // 平均と分散をセットする
      {
        mevec[i] = smat(0, i);
        vrvec[i] = (n > 1) ? smat(i, i) / (ldbl)(n - 1) : 0.0;
      }
      putmsdvec(cout, n, mevec, vrvec, opt.nx(), opt.ny());
      putmat(cout, "<CORRELATIN MATRIX>", matr, n, smat, opt.nx(), opt.ny());
    }
  }
  int qmax = min(opt.nx(), opt.ny());
  vecldbl evalvec(qmax + 1, 0.0L);
  mat<ldbl> amat(opt.nx() + 1, qmax + 1), bmat(opt.ny() + 1, qmax + 1);
  int q = cal_cca(opt, evalvec, amat, bmat, smat);
  if (q > 0)                    // 正順相関あり
  {
    amat.resize(opt.nx() + 1, q + 1);  bmat.resize(opt.ny() + 1, q + 1);
    if (opt.res("res"))
    {
      put_ccatab(cout, opt, n, smat, evalvec, amat, bmat);
      cout << "\n" << endl;
    }
    if (opt.dnsname() != "") put_dnscca(opt, amat, bmat, n, smat);
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  ccaclass cca;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) cca.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { cca.get(argv[i], opt);  fok = false; }
    }
    cca.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
