// crae.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-29 14:47:29 tac>
//
// 曲線回帰分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W crae crae.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "crae  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:cra  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The regression curve with the first item as explanatory variable X and",
  "  the second item as criterion variable Y.",
  "<OPTION>",
  "  -f=<FILE名>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -n=[<START ORDER>][-[<END ORDER>]][,...]:specify the order of explanatory",
  "    variable",
  "  -ngp={c|cl|data|pl}[,...]:select the data for Ngraph",
  "    c:regression curve  cl:confidence limit  data:raw data",
  "    pl:prediction limit",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=crae_ngp.dat",
  "  -res=[-][{def|all|reg|stat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    reg:regression analysis  stat:basic statistics",
  "  -xoffset[=<OFFSET>]:specify the offset value of explanatory variable"
};

const string Title = "=== Regression Curve ==="; // 表題
const string NGPNAMEdef = "crae_ngp.dat"; // 省略時Ngraphファイル名
constexpr int RDdef = 50;       // グラフの座標範囲の分割数省略時値

enum ngptype                    // Ngraph用出力データの種類
{
  ngpnon,                       // 出力なし
  ngpc,                         // 回帰曲線
  ngpcl,                        // 信頼限界
  ngpdata,                      // 個々の生データ
  ngppl,                        // 予測限界
  ngpetc                        // その他
};

using listint = list<int>;      // 整数リストの別名定義

class option                    // オプション集合体
{
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  list<pairint> vnplst;         // 説明変数の次数ペアリスト
  listint vnlst;                // 説明変数の次数リスト
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  bool vxoffset;                // 説明変数のオフセットフラッグ
  double vx0;                   // 説明変数のオフセット値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  list<pairint>& nplst()        // 説明変数の次数ペアリストを返す
  { return vnplst; }
  listint& nlst()               // 説明変数の次数リストを返す
  { return vnlst; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  bool xoffset() const          // 説明変数のオフセットフラッグを返す
  { return vxoffset; }
  double x0() const             // 説明変数のオフセット値を返す
  { return vx0; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vnplst(), vnlst(), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vxoffset(false), vx0(0.0)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "n")       // 説明変数の次数ペア指定
  {
    if (getrangelst(vnplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "xoffset") // 説明変数のオフセット値指定
  {
    vxoffset = true;
    vx0 = opt[1].empty() ? 0.0 : strtodf(opt[1]);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni != 2) return disperr(vpname, "Number of items must be 2 !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (vnplst.size() < 1)        // 次数ペア指定なし
  {
    vnlst.push_back(1);         // 1次式にする
    return 0;
  }
  bool nerr = false;
  list<pairint>::const_iterator ipitr = vnplst.begin();
  while (ipitr != vnplst.end() && nerr == false)
  {
    int nstart = ipitr->first, nend = ipitr->second;
    if (nstart < 1 || nend < 1)
    { nerr = true; }
    else
    {
      int nn = abs(nend - nstart) + 1;
      int inc = (nstart < nend) ? 1 : -1;
      int ni = nstart;
      for (int i = 0; i < nn; ++i)
      {
        vnlst.push_back(ni);
        ni += inc;
      }
      ++ipitr;
    }
  }
  if (nerr) return disperr(vpname, "Order must be greater then 0 !");
  for (listint::const_iterator iitr = vnlst.begin(); iitr != vnlst.end(); ++iitr)
  {
    int ni = *iitr;
    listint::const_iterator jitr = iitr;
    ++jitr;
    while (jitr != vnlst.end() && nerr == false)
    {
      if (ni == *jitr) nerr = true;
      else ++jitr;
    }
  }
  if (nerr) return disperr(vpname, "Duplicated orders !");
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["c"] = vngpmap["cl"] = vngpmap["data"] = vngpmap["pl"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["c"] = vngpmap["cl"] = vngpmap["data"] = vngpmap["pl"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["reg"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["reg"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      { vresmap["def"] = vresmap["reg"] = vresmap["stat"] = true; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class craclass                  // craメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_reg(option& opt, pairdbl& range, statcor<double>& statc, mat<double>& smat);
  // 曲線回帰分析を行って例数を返す
  void put_resid(ostream& fout);
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_reg(ostream& fout, option& opt, const statcor<double>& statc, const mat<double>& smat);
  // 曲線回帰分析結果を出力する
  void plot_ngraph(option& opt, const pairdbl& range, const statcor<double>& statc, const mat<double>& smat);
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, option& opt);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpcurve(ofstream& fout, option& opt, const pairdbl& range, const mat<double>& smat);
  // Ngraph用に回帰曲線を出力する
  double cal_estimate(option& opt, double x, const mat<double>& smat);
  // Yの推定値を計算して返す
  void plot_ngpdata(ofstream& fout);
  // Ngraph用に個々のデータを出力する
  void plot_ngplim(ofstream& fout, option& opt, ngptype ltype, const pairdbl& range, const statcor<double>& statc, const mat<double>& smat);
  // Ngraph用に信頼・予測限界を出力する
  double cal_cr(option& opt, double x, double t, double vr, ngptype ltype, const mat<double>& smat);
  // Yの限界幅を計算して返す
public:
  craclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~craclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int craclass::cal_reg(option& opt, pairdbl& range, statcor<double>& statc, mat<double>& smat)
// 曲線回帰分析を行って例数を返す
{
  int nx = (int)((opt.nlst()).size()); // 説明変数の数
  int dim = ((opt.xoffset()) ? 1 : 2) + nx;
  int yidx = dim - 1;
  vecdbl fdatavec(dim, 0.0);
  int offset = 1;
  if (opt.xoffset()) offset = 0;
  else fdatavec[0] = 1.0;
  ldata_itr ditr = datalst.begin();
  while (ditr != datalst.end())
  {
    double x = ditr->fdata(0), y = ditr->fdata(1);
    if (statc.sigma(x, y) < 1) // 欠損データはリストから削除する
    { ditr = datalst.erase(ditr); }
    else
    {
      if (x < range.first) range.first = x;
      if (x > range.second) range.second = x;
      if (opt.xoffset()) x -= opt.x0();
      listint::const_iterator iitr = (opt.nlst()).begin();
      for (int i = 0; i < nx; ++i, ++iitr) fdatavec[offset + i] = pow(x, *iitr);
      fdatavec[yidx] = y;
      for (int i = 0; i < dim; ++i)
      {
        x = fdatavec[i];
        for (int j = i; j < dim; ++j) smat.add(i, j, x * fdatavec[j]);
      }
      ++ditr;
    }
  }
  int nt = statc.n();
  if (nt < 1) return nt;        // 計算できず
  for (int i = 0; i < yidx; ++i)
  {
    for (int j = i + 1; j < dim; ++j) smat.assign(j, i, smat(i, j));
  }
  if (nt > nx)                  // 計算可能
  { for (int i = 0; i < yidx; ++i) smat.sweep(i); }
  return nt;                    // 例数を返す
}

void craclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "EXPLANATORY VARIABLE X:" << id.item(0) << "\n";
  fout << "CRITERION VARIABLE   Y:" << id.item(1) << "\n";
}

void craclass::put_reg(ostream& fout, option& opt, const statcor<double>& statc, const mat<double>& smat)
// 曲線回帰分析結果を出力する
{
  fout << "\n<REGRESSION CURVE>\n";
  if (opt.xoffset()) fout << "  No Intercept:x=(x-" << opt.x0() << ")\n";
  int nt = statc.n();
  int nx = (int)((opt.nlst()).size());
  if (nt <= nx)
  {
    fout << "->CALCULATION NOT POSSIBLE DUE TO INSUFFICIENT NUMBER OF CASE !\n";
    return;
  }
  constexpr int Llen = 74;      // 区切り線の長さ
  fout << "VAR      Partial b       SE of b";
  fout << setw(4) << opt.cc() * 100.0 << "%Precision  CI:Lower-Lim     Upper-Lim\n";
  putline(fout, Llen);
  int dft = nt - 1, offset = 0;
  int yidx = smat.r() - 1;
  double syy = statc.syy(), sr = smat(yidx, yidx);
  const double Seps = DBL_EPSILON * fmax(5.0, syy);
  if (sr < Seps) sr = 0.0;
  if (opt.xoffset())
  { dft = nt;  offset = 1;  syy += (double)nt * statc.my() * statc.my(); }
  int dfr = dft - nx;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  listint::const_iterator iitr = (opt.nlst()).begin();
  for (int i = 0; i < yidx; ++i)
  {
    double bi = smat(i, yidx), sii = smat(i, i);
    double vbi = vr * sii;
    double ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vbi);
    if (i == 0)
    {
      if (opt.xoffset())
      {
        fout << "X^" << left << setw(2) << *iitr;
        ++iitr;
      }
      else
      { fout << "ICPT"; }
    }
    else
    {
      fout << "X^" << left << setw(2) << *iitr;
      ++iitr;
    }
    fout << right << setw(14) << bi << setw(14) << sqrt(vbi);
    fout << setw(14) << ci << setw(14) << bi - ci;
    fout << setw(14) << bi + ci << "\n";
  }
  putline(fout, Llen);
  double sb = syy - sr;
  if (sb < Seps) sb = 0.0;
  double r2 = (syy > 0.0) ? sb / syy : 0.0;
  double r2a = (syy > 0.0) ? fabs(1.0 - vr * (double)dft / syy) : 0.0;
  double aic = (sr > 0.0) ? (double)nt * log(sr / (double)nt) + (double)(2 * (nx + 2 - offset)) : 0.0;
  fout << "Multiple Contribution Ratio:R^2=" << r2;
  fout << "  DF-adjusted:R'^2=" << r2a << "\n";
  fout << "Multiple Correlation Coefficient:R=" << sqrt(r2) << "  DF-adjusted:R'=" << sqrt(r2a) << "\n";
  fout << "AIC=" << aic << "\n\n";
  putatab(fout, "REGRESSION", sb, nx, sr, dfr, syy, dft);
}

void craclass::plot_ngraph(option& opt, const pairdbl& range, const statcor<double>& statc, const mat<double>& smat)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("data"))          // 個々のデータ
  {
    fout << "-s1 -dmark,00\n";
    plot_ngpid(fout, opt);
    plot_ngpdata(fout);
    delml = true;
  }
  if (opt.ngp("c"))             // 回帰曲線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_ngpcurve(fout, opt, range, smat);
  }
  if (opt.ngp("cl"))            // 回帰曲線の信頼限界
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_ngplim(fout, opt, ngpcl, range, statc, smat);
  }
  if (opt.ngp("pl"))            // 回帰曲線の予測限界
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline\n";
    plot_ngpid(fout, opt);
    plot_ngplim(fout, opt, ngppl, range, statc, smat);
  }
}

void craclass::plot_ngpid(ofstream& fout, option& opt)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%EXPLANATORY VARIABLE X:" << id.item(0) << "\n";
  fout << "%CRITERION VARIABLE   Y:" << id.item(1) << "\n";
  fout << "%Order of Explanatory Variable:";
  for (listint::const_iterator iitr = (opt.nlst()).begin(); iitr != (opt.nlst()).end(); ++iitr) fout << *iitr << " ";
  if (opt.xoffset()) fout << " No Intercept:x=(x-" << opt.x0() << ")";
  fout << "\n";
}

void craclass::plot_ngpdata(ofstream& fout)
// Ngraph用に個々のデータを出力する
{
  fout << "%<RAW DATA>\n";
  for (auto& cdata : datalst) fout << cdata.fdata(0) << " " << cdata.fdata(1) << "\n";
}

void craclass::plot_ngpcurve(ofstream& fout, option& opt, const pairdbl& range, const mat<double>& smat)
// Ngraph用に回帰曲線を出力する
{
  fout << "%<REGRESSION CURVE>\n";
  double xmin = range.first, xmax = range.second;
  double xrange = xmax - xmin;
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart, y = 0.0;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    y = cal_estimate(opt, x, smat);
    fout << x << " " << y << "\n";
  }
}

double craclass::cal_estimate(option& opt, double x, const mat<double>& smat)
// Yの推定値を計算して返す
{
  if (opt.xoffset()) x -= opt.x0();
  int yidx = (int)(smat.r()) - 1;
  double yest = smat(0, yidx);
  listint::const_iterator iitr = (opt.nlst()).begin();
  if (opt.xoffset())
  {
    yest = smat(0, yidx) * pow(x, *iitr);
    ++iitr;
  }
  for (int i = 1; i < yidx; ++i, ++iitr) yest += smat(i, yidx) * pow(x, *iitr);
  return yest;                  // 推定値を返す
}

void craclass::plot_ngplim(ofstream& fout, option& opt, ngptype ltype, const pairdbl& range, const statcor<double>& statc, const mat<double>& smat)
// Ngraph用に信頼・予測限界を出力する
{
  fout << "%<" << opt.cc() * 100;
  if (ltype == ngpcl) fout << "% CONFIDENCE LIMIT OF REGRESSION CURVE:UPPER>\n";
  else fout << "% PREDICTION LIMIT OF REGRESSION CURVE:UPPER>\n";
  int yidx = smat.r() - 1, nt = statc.n(), nx = (int)((opt.nlst()).size());
  int dft = (opt.xoffset()) ? nt : nt - 1;
  int dfr = dft - nx;
  double sr = smat(yidx, yidx);
  const double Seps = DBL_EPSILON * fmax(5.0, statc.syy());
  if (sr < Seps) sr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double t = ptot(1.0 - opt.cc(), dfr);
  double xmin = range.first, xmax = range.second;
  double xrange = xmax - xmin;
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart, y = 0.0;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    y = cal_estimate(opt, x, smat) + cal_cr(opt, x, t, vr, ltype, smat);
    fout << x << " " << y << "\n";
  }
  fout << "= =\n";
  fout << "%<" << opt.cc() * 100;
  if (ltype == ngpcl) fout << "% CONFIDENCE LIMIT OF REGRESSION CURVE:LOWER>\n";
  else fout << "% PREDICTION LIMIT OF REGRESSION CURVE:LOWER>\n";
  x = xstart;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    y = cal_estimate(opt, x, smat) - cal_cr(opt, x, t, vr, ltype, smat);
    fout << x << " " << y << "\n";
  }
}

double craclass::cal_cr(option& opt, double x, double t, double vr, ngptype ltype, const mat<double>& smat)
// Yの限界幅を計算して返す
{
  int yidx = smat.r() - 1;
  vec<double> xv(yidx + 1);
  if (opt.xoffset()) x -= opt.x0();
  int offset = 1;
  if (opt.xoffset()) offset = 0;
  else xv.assign(0, 1.0);
  listint::const_iterator iitr = (opt.nlst()).begin();
  for (int i = offset; i < yidx; ++i, ++iitr) xv.assign(i, pow(x, double(*iitr)));
  xv.assign(yidx, 0.0);
  vec<double> tv(xv.trn() * smat);
  double w = inp(tv, xv);
  if (ltype == ngppl) w += 1.0; // 予測限界幅
  return t * sqrt(vr * w);      // Yの限界幅を返す
}

int craclass::get(const char* c_fname, option& opt)
// データを読み込み累積集計する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void craclass::put(option& opt) // 集計結果を出力する
{
  constexpr int LENline = 75;   // 区切り線の長さ
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  int nx = (int)((opt.nlst()).size()); // 説明変数の数
  int dim = ((opt.xoffset()) ? 1 : 2) + nx;
  pairdbl range(DBL_MAX, -DBL_MAX); // 最小値と最大値
  statcor<double> statc;        // 相関基礎統計量
  mat<double> smat(dim);        // 回帰計算用積和行列
  int nt = cal_reg(opt, range, statc, smat);
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(cout);
    if (opt.res("stat"))
    {
      cout << "\n<BASIC STATISTICS>\n";
      putline(cout, LENline);
      double mx = statc.mx(), my = statc.my();
      cout << "X:";  putmsd(cout, nt, mx, statc.vxx());
      cout << "Y:";  putmsd(cout, nt, my, statc.vyy());
      putline(cout, LENline);
    }
    if (opt.res("reg")) put_reg(cout, opt, statc, smat);
    cout << "\n" << endl;
  }
  if (statc.n() > (int)((opt.nlst()).size()) && opt.ngpname() != "") plot_ngraph(opt, range, statc, smat);
  // Ngraph用データ出力
}

int main(int argc, char* argv[]) // メインルーチン
{
  craclass cra;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) cra.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { cra.get(argv[i], opt);  fok = false; }
    }
    cra.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
