// disane.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-29 15:35:10 tac>
//
// 最初の項目を判別項目、以後の項目を説明変数とした逐次変数選択法による
// 判別分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W disane disane.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "disane  Version 7.41e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:disan [<OPTION> ...] [<DATA FILE> ...]",
  "  The discriminant analysis with the first item as discriminant variable",
  "  and subsequent items as explanatory variables.",
  "<OPTION>",
  "  -comp[={-|+}]:complement deficiencies in the calculation of",
  "    discriminantin scores(-comp:-comp=+)",
  "  -dns[={coef|score}]:select the kind of DANS data",
  "    coef:discriminatn coefficient  score:discriminatn score",
  "    -dns:-dns=score",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=disane_dns.d",
  "  -f=<FILE>:reads options from the specified file",
  "  -fin=<Fin>:specify the inclusion criteria Fin(Fin>=0)",
  "  -fout=<Fout>:specify the expulsion criteria Fout(0=<Fout=<Fin)",
  "  -g{1|2}[=<CODE>]:specify the code of group1 or group2",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={lci|log|score}[,...]:select the data for Ngraph",
  "    lci:CI of logistic curve  log:logistic curve",
  "    score:disciminatn scores  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=disane_ngp.dat",
  "  -pi=<PROBABILITY>:specify the prior probability of group1(0<PROBABILITY<1)",
  "  -pin=<Pin>:specify the inclusion criteria Pin(0<Pin=<1)",
  "  -pout=<Pout>:specify the expulsion criteria Pout(1>=Pout>=Pin)",
  "  -res=[-][{def|all|allx|cc[<C-COEF>]|selx|stat}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    allx:result of selecting all variables",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1 cc:cc0.95)",
  "    selx:result of selecting variables  stat:baisc statistics",
  "  -x<NUMBER>=<VALUE>:specify the value of explanatory variable",
  "  -xno=<NUMBER>:specify the number of explanatory variable"
};

const string Title = "=== Discriminant Analysis ==="; // 表題
const string DNSNAMEdef = "disane_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "disane_ngp.dat"; // 省略時Ngraphファイル名
constexpr double FIOdef = 2.0;  // 省略時変数F値むの取り込み・追い出し基準
constexpr double PIOdef = 0.1;  // 省略時変数p値の取り込み・追い出し基準
constexpr int RDdef = 50;       // ロジスティック曲線の分解数省略時値

enum dnstype                    // DANSデータの種類
{
  dnsnon,                       // 出力なし
  dnscoef,                      // 判別スコア計算用判別係数
  dnsscore,                     // 判別スコア
  dnsetc                        // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  bool vcomp;                   // 欠損補間の有無
  dnstype vdns;                 // DANS出力データの種類
  string vdnsname;              // DANSファイル名
  pairdbl flevel;               // F値の取り込み基準と追い出し基準
  pairdbl plevel;               // P値の取り込み基準と追い出し基準
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  vecstr vgvec;                 // 群コード
  double vpi;                   // ロジスティック曲線における群1の事前確率
  lstr vreslst;                 // 解析結果出力リスト
  map<string, bool> vresmap;    // 解析結果出力の有無マップ
  lstr xlst;                    // 説明変数リスト
  vecdbl vxvec;                 // 説明変数ベクトル
  int vxno;                     // 変化説明変数の番号
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  bool comp() const             // 欠損補間の有無を返す
  { return vcomp; }
  dnstype dns() const           // DANS出力データの種類を返す
  { return vdns; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  double fin() const            // F値の取り込み基準を返す
  { return flevel.first; }
  double fout() const           // F値の追い出し基準を返す
  { return flevel.second; }
  const string& g(int idx) const // 群コードを返す
  { return vgvec[idx]; }
  double pi() const   // ロジスティック曲線における群1の事前確率を返す
  { return vpi; }
  double pin() const            // P値の取り込み基準を返す
  { return plevel.first; }
  double pout() const           // P値の追い出し基準を返す
  { return plevel.second; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
   double x(size_t no) const    // 説明変数の値を返す
  { return (0 < no && no <= vxvec.size()) ? vxvec[no - 1] : BLANKf; }
  int xno() const               // 変化説明変数の番号を返す
  { return vxno; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vcomp(false), vdns(dnsscore), vdnsname(), flevel(FIOdef, FIOdef), plevel(1.0, 1.0), vgvec(2), vngpname(), vpi(0.0), vreslst(), vresmap(), xlst(), vxvec(), vxno(0)
// コンストラクタ
{
  if (c_pname == 0) return;     // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
  vgvec[0] = "1";
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "comp")         // 欠損補完の有無の指定
  {
    if (opt[1].empty() || opt[1] == "+") vcomp = true;
    else if (opt[1] == "-") vcomp = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dns")     // DANSデータの種類指定
  {
    if (opt[1].empty() || opt[1] == "score") vdns = dnsscore;
    else if (opt[1] == "coef") vdns = dnscoef;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "fin")     // F値の取り込み基準指定
  {
    flevel.first = opt[1].empty() ? FIOdef : strtodf(opt[1]);
    if (flevel.first < 0.0) err = badopt(vpname, optstr, USAGE, 0.0, DBL_MAX);
    if (flevel.second > flevel.first) flevel.second = flevel.first;
  }
  else if (opt[0] == "fout")    // F値の追い出し基準指定
  {
    flevel.second = opt[1].empty() ? flevel.first : strtodf(opt[1]);
    if (flevel.second < 0.0 || flevel.first < flevel.second) err = badopt(vpname, optstr, USAGE, 0.0, flevel.first);
  }
  else if (opt[0] == "g1")      // 群1のコード指定
  { vgvec[0] = opt[1]; }
  else if (opt[0] == "g2")      // 群2のコード指定
  { vgvec[1] = opt[1]; }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pi")      // ロジスティック曲線における群1の事前確率指定
  {
    vpi = opt[1].empty() ? 0.0 : strtodf(opt[1]);
    if (vpi < 0.0 || 1.0 <= vpi) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "pin")     // p値の取り込み基準指定
  {
    plevel.first = opt[1].empty() ? PIOdef : strtodf(opt[1]);
    if (plevel.first <= 0.0 || 1.0 < plevel.first) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "pout")    // p値の追い出し基準指定
  {
    plevel.second = opt[1].empty() ? plevel.first : strtodf(opt[1]);
    if (plevel.second < plevel.first || 1.0 < plevel.second) err = badopt(vpname, optstr, USAGE, plevel.first, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "xno")     // 変化説明変数の番号指定
  {
    char* ptr = nullptr;
    vxno = strtol(opt[1].c_str(), &ptr, 0);
    if (vxno < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0][0] == 'x')    // 説明変数の値指定
  {
    if (strsplit(xlst, optstr, CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are requied !");
  if (set_ngp() < 0) return -2; // -ngpオプションのエラー
  if (set_res() < 0) return -3; // -resオプションのエラー
  if ((vgvec[0]).empty() && (vgvec[1]).empty()) return disperr(vpname, "The group code is not specified !");
  int p = ni - 1;
  if (vxno < 1) return 0;       // 説明変数指定なし
  if (p < vxno) return disperr(vpname, "Wrong number of the exlpanatory variable !");
  if (xlst.empty()) return 0;   // 説明変数指定なし
  if ((int)(xlst.size()) > p) return disperr(vpname, "Too many values of the explanatory variables !");
  vxvec.resize(p);
  vxvec.assign(p, BLANKf);
  for (lstr_citr xitr = xlst.begin(); xitr != xlst.end(); ++xitr)
  {
    string opt[2];
    if (getopt(opt, *xitr, OPTdev) > 0)
    {
      char* ptr = nullptr;
      int xno = strtol(opt[0].substr(1).c_str(), &ptr, 0);
      if (xno < 1 || p < xno || xno == vxno) return disperr(vpname, "Wrong number of the explanatory variable->", *xitr);
      double fdata = strtodf(opt[1]);
      if (fdata == BLANKf) return disperr(vpname, "Wrong value of the explanatory variable->", *xitr);
      vxvec[xno - 1] = fdata;
    }
  }
  if (vcc <= 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong confidence coefficient !");
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["lci"] = vngpmap["log"] = vngpmap["score"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["lci"] = vngpmap["log"] = vngpmap["score"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["allx"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["allx"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["allx"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class disanclass                // disanメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_stat(const option& opt, vector<mat<double>>& smatvec, vecint& nvec, mat<double>& dmat);
  // 統計量を計算して総例数を返す
  void put_resid(ostream& fout);
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, option& opt, const vector<mat<double>>& smatvec, const vecint& nvec);
  // 統計量を出力する
  void set_dmat(mat<double>& dmat, vecdbl& midvec, vecdbl& vrvec, const vector<mat<double>>& smatvec, const vecint& nvec);
  // 変数選択用積和行列を準備する
  int step_sel(const vecint& nvec, mat<double>& dmat, vecint& checkvec, double fin, double fout, double pin, double pout);
  // 変数増減法による判別関数の計算
  int sel_var(bool in, double fin, double fout, double pin, double pout, int nt, double w, const mat<double>& dmat, const vecint& checkvec);
  // 変数増減法による変数の選択
  void put_distab(ostream& fout, option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& vrvec, double fin, double flout, double pin, double pout);
  // 判別分析結果を出力する
  void put_arci(ostream& fout, const option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec) const;
  // 判別係数の信頼区間を出力する
  void put_hotest(ostream& fout, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec) const;
  // HotellingのT2乗検定結果と推定結果を出力する
  void put_dnsid(ofstream& fout, const option& opt);
  // DANSデータのタイトルとデータ識別情報を出力する
  void put_coef(ofstream& fout, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec);
  // 判別スコア計算用判別係数をデータファイル形式で出力する
  void put_score(ofstream& fout, const option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec);
  // 判別スコアと群1に属す確率をデータファイル形式で出力する
  int cal_score(const option& opt, double& z, datac& cdata, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec);
  // 判別スコアを計算し、群インデックスを返す
  void plot_disngp(option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec);
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, const option& opt, const vecint& checkvec);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_log(ofstream& fout, const option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec);
  // Ngraph用ロジスティック曲線を出力する
  void plot_lci(ofstream& fout, const option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec);
  // Ngraph用ロジスティック曲線の信頼区間を出力する
  void plot_score(ofstream& fout, const option& opt, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec);
  // Ngraph用判別スコアを出力する
public:
  disanclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~disanclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int disanclass::cal_stat(const option& opt, vector<mat<double>>& smatvec, vecint& nvec, mat<double>& dmat)
// 統計量を計算して総例数を返す
{
  size_t dim = id.ni() + 1, didx = id.ni();
  vecdbl fdatavec(didx, 0.0);
  fdatavec[0] = 1.0;
  for (auto& cdata : datalst)
  {
    bool blank = false;
    for (size_t j = 1; blank == false && j < id.ni(); ++j)
    {
      fdatavec[j] = cdata.fdata(j);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (cdata.sdata(0) != BLANKs && blank == false)
    {
      int idx = -1;
      if (opt.g(0).empty())     // 群1コード指定なし
      { idx = (cdata.sdata(0) == opt.g(1)) ? 1 : 0; }
      else if (cdata.sdata(0) == opt.g(0))
      { idx = 0; }
      else if (opt.g(1).empty() || cdata.sdata(0) == opt.g(1))
      { idx = 1; }
      if (idx >= 0)             // 群1・2以外は集計せず
      {
        for (size_t i = 0; i < didx; ++i)
        {
          double x = fdatavec[i];
          for (size_t j = i; j < didx; ++j) smatvec[idx].add(i, j, x * fdatavec[j]);
        }
      }
    }
  }
  nvec[0] = (int)(smatvec[0](0, 0));  nvec[1] = (int)(smatvec[1](0, 0));
  int nt = 0;
  for (int k = 0; k < 2; ++k)
  {
    if (nvec[k] > 0)
    {
      nt += nvec[k];
      for (size_t i = 0; i < dim - 1; ++i)
      {
        for (size_t j = i + 1; j < dim; ++j) smatvec[k].assign(j, i, smatvec[k](i, j));
      }
      smatvec[k].sweep(0);
    }
  }
  dmat = smatvec[0] + smatvec[1];
  return nt;                    // 総例数を返す
}

void disanclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "DISCRIMINANT ITEM:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i;
    fout << ":" << id.item(i) << "\n";
  }
}

void disanclass::put_stat(ostream& fout, option& opt, const vector<mat<double>>& smatvec, const vecint& nvec)
// 統計量を出力する
{
  size_t didx = id.ni();
  vecdbl mkvec(didx, 0.0), vkvec(didx, 0.0);
  for (int k = 0; k < 2; ++k)
  {
    if (nvec[k] > 0)
    {
      for (size_t j = 0; j < didx; ++j)
      {
        mkvec[j] = smatvec[k](0, j);
        vkvec[j] = (nvec[k] > 1) ? smatvec[k](j, j) / (double)(nvec[k] - 1) : 0.0;
      }
      if (opt.res("stat"))
      {
        fout << "\nGROUP" << k + 1 << ":" << id.item(0) << "=";
        if (opt.g(k).empty()) fout << "other than " << opt.g((k == 0) ? 1 : 0);
        else fout << opt.g(k);
        putmsdvec(fout, nvec[k], mkvec, vkvec);
      }
    }
  }
}

void disanclass::set_dmat(mat<double>& dmat, vecdbl& midvec, vecdbl& vrvec, const vector<mat<double>>& smatvec, const vecint& nvec)
// 変数選択用積和行列を準備する
{
  int df = nvec[0] + nvec[1] - 2;
  dmat *= 1.0 / (double)df;
  size_t didx = id.ni();
  midvec[0] = (double)(df + 2);
  dmat.assign(0, 0, 0.0);  dmat.assign(0, didx, 0.0);
  dmat.assign(didx, 0, 0.0);  dmat.assign(didx, didx, 0.0);
  vrvec[0] = dmat(0, 0);
  for (size_t i = 1; i < didx; ++i)
  {
    midvec[i] = (smatvec[0](0, i) + smatvec[1](0, i)) * 0.5;
    dmat.assign(0, i, (smatvec[0](0, i) + smatvec[1](0, i)) * 0.5);
    dmat.assign(i, 0, -dmat(i, 0));
    dmat.assign(i, didx, smatvec[0](0, i) - smatvec[1](0, i));
    dmat.assign(didx, i, dmat(i, didx));
    vrvec[i] = dmat(i, i);
  }
}

int disanclass::step_sel(const vecint& nvec, mat<double>& dmat, vecint& checkvec, double fin, double fout, double pin, double pout)
// 変数増減法による判別関数の計算
{
  int nt = nvec[0] + nvec[1], i = 0;
  bool in = true;
  double w = (double)(nt - 2) / (double)(nvec[0]) * (double)nt / (double)(nvec[1]);
  do {
    in = true;
    while ((i = sel_var(in, fin, fout, pin, pout, nt, w, dmat, checkvec)) > 0)
    {
      dmat.sweep(i);  checkvec[i] = in ? 1 : 0;
      checkvec[0] += in ? 1 : -1;
      in = !in;
    }
  } while (in == false);        // 取り込む変数がなくなるまで続ける
  return checkvec[0];           // 取り込んだ変数の数を返す
}

int disanclass::sel_var(bool in, double fin, double fout, double pin, double pout, int nt, double w, const mat<double>& dmat, const vecint& checkvec)
// 変数増減法による変数の選択
{
  size_t didx = id.ni();
  int idx = 0, inout = in ? 1 : 0;
  double dsel = in ? DBL_MIN : DBL_MAX;
  bool fsel = true;             // F値による変数選択
  if (pin < 1.0) fsel = false;  // 有意確率pによる変数選択
  const double Seps = DBL_EPSILON * fmax(5.0, w);
  for (size_t j = 1; j < didx; ++j)
  {
    if (checkvec[j] != inout && fabs(dmat(j, j)) > Seps)
    {
      double dj = dmat(j, didx) * dmat(j, didx) / dmat(j, j);
      if ((in && dj > dsel) || (in == false && dj < dsel))
      { dsel = dj;  idx = j; }
    }
  }
  if (idx < 1) return idx;      // 選択すべき変数なし
  if (in)                       // 変数取り込み
  {
    int df = nt - checkvec[0] - 2;
    double f = dsel * (double)df / (w - dmat(didx, didx));
    if (fsel)
    {
      if (f < fin) idx = 0;
    }
    else
    {
      double pr = ftop(f, 1, df);
      if (pr > pin) idx = 0;
    }
  }
  else                          // 変数追い出し
  {
    int df = nt - checkvec[0] - 1;
    double f = dsel * (double)df / (w - dmat(didx, didx) - dsel);
    if (fsel)
    {
      if (f >= fout) idx = 0;
    }
    else
    {
      double pr = ftop(f, 1, df);
      if (pr <= pout) idx = 0;
    }
  }
  return idx;                   // 選択すべき変数のインデックスを返す
}

void disanclass::put_distab(ostream& fout, option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& vrvec, double fin, double flout, double pin, double pout)
// 判別分析結果を出力する
{
  if (fin <= 0.0 && pin >= 1.0)
  { fout << "\n<RESULT OF SELECTING ALL VARIABLES>\n"; }
  else
  {
    fout << "\n<RESULT OF STEPWISE FORWARD SELECTION METHOD>\n";
    if (pin < 1.0) fout << "  Inclusion Criteria:p=<Pin=" << pin << "  Expulsion Criteria:p>Pout=" << pout << "\n";
    else fout << "  Inclusion Criteria:F>=Fin=" << fin << "  Expulsion Criteria:F<Fout=" << flout << "\n";
  }
  fout << "VAR   a:Dis-Coef     SE of a      exp(a)  Standard a     F-value     p-value\n";
  constexpr int LENline = 79;   // 区切り線の長さ
  putline(fout, LENline);
  int didx = id.ni(), nt = nvec[0] + nvec[1];
  int df1 = checkvec[0], df2 = nt - 1 - checkvec[0];
  double d2 = -dmat(didx, didx);
  double w = (double)(nvec[0]) * (double)(nvec[1]) / (double)nt;
  fout << "ICPT" << setw(12) << dmat(0, didx) << "\n";
  for (int i = 1; i < didx; ++i)
  {
    if (checkvec[i])
    {
      double ai = dmat(i, didx);
      double di = ai * ai / dmat(i, i);
      double f = (double)df2 * di / ((double)(nt - 2) / w + d2 - di);
      double va = ai * ai / f;
      ldbl pr = ftop(f, 1, df2);
      fout << "X" << setw(3) << i << setw(12) << ai;
      fout << setw(12) << sqrt(va) << setw(12) << exp(ai);
      fout << setw(12) << ai * sqrt(vrvec[i]) << setw(12) << f;
      fout << setw(12) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
  if (checkvec[0] < 1) return;  // 選択変数なし
  if (opt.res("cc")) put_arci(fout, opt, nvec, dmat, checkvec);
  put_hotest(fout, nvec, dmat, checkvec);
}

void disanclass::put_arci(ostream& fout, const option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec) const
// 判別係数の信頼区間を出力する
{
  fout << "\nVAR   a:Dis-Coef" << setw(3) << opt.cc() * 100.0;
  fout << "%CI:Lower       Upper      exp(a)" << setw(3) << opt.cc() * 100.0;
  fout << "%CI:Lower       Upper\n";
  constexpr int LENline = 76;   // 区切り線の長さ
  putline(fout, LENline);
  int didx = id.ni(), nt = nvec[0] + nvec[1];
  int df2 = nt - 1 - checkvec[0];
  double d2 = - dmat(didx, didx);
  double w = (double)(nvec[0]) * (double)(nvec[1]) / (double)nt;
  for (int i = 1; i < didx; ++i)
  {
    if (checkvec[i])
    {
      double ai = dmat(i, didx);
      double di = ai * ai / dmat(i, i);
      double va = dmat(i, i) * ((double)(nt - 2) / w + d2 - di) / (double)df2;
      double ci = calmcci(mcfis, opt.cc(), 2, 1, df2, 1, va);
      double al = ai - ci, au = ai + ci;
      fout << "X" << setw(3) << i << setw(12) << ai;
      fout << setw(12) << al << setw(12) << au;
      fout << setw(12) << exp(ai) << setw(12) << exp(al);
      fout << setw(12) << exp(au) << "\n";
    }
  }
  putline(fout, LENline);
}

void disanclass::put_hotest(ostream& fout, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec) const
// HotellingのT2乗検定結果と推定結果を出力する
{
  int didx = id.ni(), nt = nvec[0] + nvec[1];
  int df1 = checkvec[0], df2 = nt - 1 - checkvec[0];
  double d2 = -dmat(didx, didx); // Mahalanobisの平方距離
  double w = (double)(nvec[0]) * (double)(nvec[1]) / (double)nt;
  double cf = (double)df2 / (double)(nt - 2) / (double)df1;
  double t2 = w * d2;
  double f = t2 * cf;
  double r2 = f * (double)df1 / (f * (double)df1 + (double)df2);
  double z = sqrt(d2) * 0.5;
  ldbl pr = sndis(z);
  fout << "Mahalanobis's Square Distance:D^2=" << d2 << "  Correlation Ratio:R=" << sqrt(r2) << "\n";
  fout << "  D/2=" << z << "  Discriminant Probability:p=" << pr << "(1-p=" << 1.0L - pr << ")\n";
  pr = ftop(f, df1, df2);
  fout << "Hotelling's T-squared Test:T^2=D^2 x " << nvec[0] << " x " << nvec[1] << " / " << nt << "=" << t2 << "\n";
  fout << "  F=" << f << "  DF1=" << df1 << "  DF2=" << df2 << "  p=" << pr << pind(pr) << endl;
}

void disanclass::put_dnsid(ofstream& fout, const option& opt)
// DANSデータのタイトルとデータ識別情報を出力する
{
  idd tid(id.title(), 3);
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  if (opt.dns() == dnscoef)     // 判別係数出力
  {
    tid.assign_item(0, "ITEM NAME");
    if (opt.fin() > 0.0 || opt.pin() < 1.0) tid.assign_item(1, "Discriminant Coefficient(Selection Method)");
    else tid.assign_item(1, "Discriminatn Coefficient(All Selection)");
    tid.assign_item(2, "Mean for Missing Complement");
    tid.put_item(fout, formtag);
  }
  else                          // 判別スコア出力
  {
    tid.assign_item(0, id.item(0));
    if (opt.fin() > 0.0 || opt.pin() < 1.0)
    {
      tid.assign_item(1, "Discriminant Score(Selection Method)");
      tid.assign_item(2, "Probability of Belonging to Group1(Selection Method)");
    }
    else
    {
      tid.assign_item(1, "Discriminant Score(All Selection)");
      tid.assign_item(2, "Probability of Belonging to Group1(All Selection)");
    }
    tid.put_item(fout, formtag);
  }
}

void disanclass::put_coef(ofstream& fout, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec)
// 判別スコア計算用判別係数をデータファイル形式で出力する
{
  datac cdata(3);
  int didx = id.ni();
  cdata.assign(0, "ICPT");
  cdata.assign(1, dmat(0, didx));
  cdata.assign(2, 0.0);
  cdata.put(fout);
  for (int i = 1; i < didx; ++i)
  {
    if (checkvec[i])
    {
      cdata.assign(0, id.item(i));
      cdata.assign(1, dmat(i, didx));
      cdata.assign(2, midvec[i]);
      cdata.put(fout);
    }
  }
}

void disanclass::put_score(ofstream& fout, const option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec)
// 判別スコアと群1に属す確率をデータファイル形式で出力する
{
  double pi1 = (double)(nvec[0]) / (double)(nvec[0] + nvec[1]);
  double cf = (double)(nvec[1]) / (double)(nvec[0]);
  if (opt.pi() != 0.0)
  { pi1 = opt.pi();  cf = (1.0 - opt.pi()) / opt.pi(); }
  datac cdata(3);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    cdata.assign(0, ditr->sdata(0));
    double z = BLANKf;
    int gindex = cal_score(opt, z, *ditr, dmat, checkvec, midvec);
    if (gindex < 0)
    { cdata.assign(1, BLANKs);  cdata.assign(2, BLANKs); }
    else
    { cdata.assign(1, z);  cdata.assign(2, 1.0 / (1.0 + cf * exp(-z))); }
    cdata.put(fout);
  }
}

int disanclass::cal_score(const option& opt, double& z, datac& cdata, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec)
// 判別スコアを計算し、群インデックスを返す
{
  if (cdata.sdata(0) == BLANKs) return -1; // 群コード欠損のため計算できず
  int gindex= -1;
  if (opt.g(0).empty()) gindex = (cdata.sdata(0) == opt.g(1)) ? 1 : 0;
  else if (cdata.sdata(0) == opt.g(0)) gindex = 0;
  else if (opt.g(1).empty() || cdata.sdata(0) == opt.g(1)) gindex = 1;
  if (gindex < 0) return -1;    // 群1・2以外は計算せず
  int didx = id.ni();
  z = dmat(0, didx);
  bool blank = false;
  for (int i = 1; blank == false && i < didx; ++i)
  {
    if (cdata.fdata(i) == BLANKf)
    {
      if (opt.comp())           // 欠損を2群の平均値の中央値で補完する場合
      {
        if (checkvec[i]) z += dmat(i, didx) * midvec[i];
      }
      else
      { blank = true; }
    }
    else
    {
      if (checkvec[i]) z += dmat(i, didx) * cdata.fdata(i);
    }
  }
  return blank ? -1 : gindex;   // 計算できたら群インデックスを返す
}

void disanclass::plot_disngp(option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("log"))           // ロジスティック曲線
  {
    fout << "-s1 -dline -w80 -c255,0,0 -zy-0.1,1.1,0.2\n";
    plot_ngpid(fout, opt, checkvec);
    plot_log(fout, opt, nvec, dmat, checkvec, midvec);
    delml = true;
  }
  if (opt.ngp("lci"))           // ロジスティック曲線の信頼区間
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -c255,0,0 -zy-0.1,1.1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, opt, checkvec);
    plot_lci(fout, opt, nvec, dmat, checkvec, midvec);
  }
  if (opt.ngp("score"))         // 判別スコア
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dmark,00 -o300 -zy-0.1,1.1,0.2\n";
    plot_ngpid(fout, opt, checkvec);
    plot_score(fout, opt, dmat, checkvec, midvec);
  }
}

void disanclass::plot_ngpid(ofstream& fout, const option& opt, const vecint& checkvec)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%DISCRIMINANT ITEM:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i])
    {
      fout << "%EXPLANATORY VARIABLE X" << setw(3) << i;
      fout << ":" << id.item(i) << "\n";
    }
  }
  if (opt.pin() < 1.0) fout << "%Inclusion Criteria:p=<Pin=" << opt.pin() << "  Expulsion Criteria:p>Pout=" << opt.pout() << "\n";
  else fout << "%Inclusion Criteria:F>=Fin=" << opt.fin() << "  Expulsion Criteria:F<Fout=" << opt.fout() << "\n";
  if (opt.xno() > 0)          // 変化説明変数指定あり
  {
    fout << "%Explanatory Variables with changing values:" << id.item(opt.xno()) << "\n";
    for (size_t i = 1; i < id.ni(); ++i)
    {
      if (checkvec[i])
      {
        if (opt.x(i) != BLANKf) fout << "%Explanatory Variable X:" << id.item(i) << " Specified Value=" << opt.x(i) << "\n";
      }
    }
  }
}

void disanclass::plot_log(ofstream& fout, const option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec)
// Ngraph用ロジスティック曲線を出力する
{
  fout << "%<LOGISTIC CURVE>\n";
  double xmin = -5.0, xmax = 5.0;
  if (opt.xno() > 0) { xmin = FLT_MAX;  xmax = -FLT_MAX; }
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = 0.0;
    if (opt.xno() > 0)
    { x = ditr->fdata(opt.xno()); }
    else
    {
      double z = BLANKf;
      int gindex = cal_score(opt, z, *ditr, dmat, checkvec, midvec);
      x = (gindex >= 0) ? z : BLANKf;
    }
    if (x != BLANKf)
    {
      if (x < xmin) xmin = x;
      if (x > xmax) xmax = x;
    }
  }
  double pi1 = (double)(nvec[0]) / (double)(nvec[0] + nvec[1]);
  double cf = (double)(nvec[1]) / (double)(nvec[0]);
  if (opt.pi() != 0.0)
  { pi1 = opt.pi();  cf = (1.0 - opt.pi()) / opt.pi(); }
  fout << "%Prior Probability of Group1=" << pi1 << "\n";
  double range = xmax - xmin;
  double xstart = xmin - range * 0.1, xend = xmax + range * 0.1;
  double dx = (xend - xstart) / (double)(RDdef);
  datac cdata(id.ni());
  cdata.assign(0, opt.g(0));
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (opt.x(i) == BLANKf) cdata.assign(i, midvec[i]);
    else cdata.assign(i, opt.x(i));
  }
  double x = xstart;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    double z = BLANKf;
    if (opt.xno() < 1)
    { z = x; }
    else
    {
      cdata.assign(opt.xno(), x);
      cal_score(opt, z, cdata, dmat, checkvec, midvec);
    }
    double pr = 1.0 / (1.0 + cf * exp(-z));
    fout << x << " " << pr << "\n";
  }
}

void disanclass::plot_lci(ofstream& fout, const option& opt, const vecint& nvec, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec)
// Ngraph用ロジスティック曲線の信頼区間を出力する
{
  fout << "%<" << opt.cc() * 100.0 << "% CI OF LOGISTIC CURVE>\n";
  double xmin = -5.0, xmax = 5.0;
  if (opt.xno() > 0) { xmin = FLT_MAX;  xmax = -FLT_MAX; }
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = 0.0;
    if (opt.xno() > 0)
    { x = ditr->fdata(opt.xno()); }
    else
    {
      double z = BLANKf;
      int gindex = cal_score(opt, z, *ditr, dmat, checkvec, midvec);
      x = (gindex >= 0) ? z : BLANKf;
    }
    if (x != BLANKf)
    {
      if (x < xmin) xmin = x;
      if (x > xmax) xmax = x;
    }
  }
  double pi1 = (double)(nvec[0]) / (double)(nvec[0] + nvec[1]);
  double cf = (double)(nvec[1]) / (double)(nvec[0]);
  if (opt.pi() != 0.0)
  { pi1 = opt.pi();  cf = (1.0 - opt.pi()) / opt.pi(); }
  fout << "%Prior Probability of Group1=" << pi1 << "\n";
  double range = xmax - xmin;
  double xstart = xmin - range * 0.1, xend = xmax + range * 0.1;
  double dx = (xend - xstart) / (double)(RDdef);
  datac cdata(id.ni());
  cdata.assign(0, opt.g(0));
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (opt.x(i) == BLANKf) cdata.assign(i, midvec[i]);
    else cdata.assign(i, opt.x(i));
  }
  double vr = -dmat(id.ni(), id.ni());
  double t = ptoz(1.0 - opt.cc());
  double ci = t * sqrt((1.0 / (double)(nvec[0]) + 1.0 / (double)(nvec[1])) * vr);
  fout << "%<CONFICENCE INTERVAL:LOWER>\n";
  double x = xstart;
  for (int i = 0; i <= RDdef; ++i, x += dx) // 信頼区間下限データ出力
  {
    double z = BLANKf;
    if (opt.xno() < 1)
    { z = x; }
    else
    {
      cdata.assign(opt.xno(), x);
      cal_score(opt, z, cdata, dmat, checkvec, midvec);
    }
    double pr = 1.0 / (1.0 + cf * exp(-z + ci));
    fout << x << " " << pr << "\n";
  }
  fout << "= =\n";
  fout << "%<CONFICENCE INTERVAL:UPPER>\n";
  x = xstart;
  for (int i = 0; i <= RDdef; ++i, x += dx) // 信頼区間上限データ出力
  {
    double z = BLANKf;
    if (opt.xno() < 1)
    { z = x; }
    else
    {
      cdata.assign(opt.xno(), x);
      cal_score(opt, z, cdata, dmat, checkvec, midvec);
    }
    double pr = 1.0 / (1.0 + cf * exp(-z - ci));
    fout << x << " " << pr << "\n";
  }
}

void disanclass::plot_score(ofstream& fout, const option& opt, const mat<double>& dmat, const vecint& checkvec, const vecdbl& midvec)
// Ngraph用判別スコアを出力する
{
  fout << "%<DISCRIMINANT SCORE:" << id.item(0) << "=";
  if (opt.g(0).empty()) fout << "other than " << opt.g(1);
  else fout << opt.g(0);
  fout << ">\n";
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = BLANKf, z = BLANKf;
    int gindex = cal_score(opt, z, *ditr, dmat, checkvec, midvec);
    if (gindex == 0)
    {
      if (opt.xno() < 1) x = z;
      else x = ditr->fdata(opt.xno());
      if (x != BLANKf) fout << x << " 1\n";
    }
  }
  fout << "= =\n";
  fout << "%<DISCRIMINANT SCORE:" << id.item(0) << "=";
  if (opt.g(1).empty()) fout << "other than " << opt.g(0);
  else fout << opt.g(1);
  fout << ">\n";
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = BLANKf, z = BLANKf;
    int gindex = cal_score(opt, z, *ditr, dmat, checkvec, midvec);
    if (gindex == 1)
    {
      if (opt.xno() < 1) x = z;
      else x = ditr->fdata(opt.xno());
      if (x != BLANKf) fout << x << " 0\n";
    }
  }
}

int disanclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void disanclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  int dim = id.ni() + 1, didx = id.ni();
  vector<mat<double>> smatvec(2); // 積和行列ベクトル
  vecint nvec(2, 0);              // 例数ベクトル
  mat<double> dmat(dim);          // 群合計積和行列
  vecint checkvec(id.ni(), 0);    // 変数選択用チェックベクトル
  vecdbl midvec(dim, 0.0);        // 2群の平均値の中央値
  vecdbl vrvec(didx, 0.0);        // 分散ベクトル
  smatvec[0].resize(dim, dim);  smatvec[1].resize(dim, dim);
  smatvec[0] = 0.0;  smatvec[1] = 0.0;
  int nt = cal_stat(opt, smatvec, nvec, dmat);
  if (opt.res("res")) put_resid(cout);
  put_stat(cout, opt, smatvec, nvec);
  if (nt < didx)                // 判別関数計算できず
  {
    if (opt.res("res")) cout << "\n" << endl;
    return;
  }
  if (opt.res("stat")) putmat(cout, "<OVERALL COMMON CORRELATION MATIX>", matr, nt - 1, dmat, didx - 1);
  set_dmat(dmat, midvec, vrvec, smatvec, nvec);
  step_sel(nvec, dmat, checkvec, opt.fin(), opt.fout(), opt.pin(), opt.pout());
  if (opt.res("res") && opt.res("selx")) put_distab(cout, opt, nvec, dmat, checkvec, vrvec, opt.fin(), opt.fout(), opt.pin(), opt.pout());
  if (opt.dnsname() != "")
  {
    ofstream fout;
    if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
    put_dnsid(fout, opt);
  if (opt.dns() == dnscoef) put_coef(fout, dmat, checkvec, midvec);
  else put_score(fout, opt, nvec, dmat, checkvec, midvec);
  }
  if (opt.ngpname() != "") plot_disngp(opt, nvec, dmat, checkvec, midvec);
  if (opt.res("res"))
  {
    if (opt.res("allx") && (opt.res("selx") == false || checkvec[0] < didx - 1))
    {
      if (checkvec[0] < didx - 1) step_sel(nvec, dmat, checkvec, 0.0, 0.0, 1.0, 1.0);
      put_distab(cout, opt, nvec, dmat, checkvec, vrvec, 0.0, 0.0, 1.0, 1.0);
    }
    cout << "\n" << endl;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  disanclass disan;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) disan.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { disan.get(argv[i], opt);  fok = false; }
    }
    disan.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
