// facte.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-29 15:50:48 tac>
//
// 指定した項目について因子分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W facte facte.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "facte  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:fact  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The factor analysis for all items.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -d2[={max|q|<COEF>}]:specify the initial value of uniqueness(0<COEF<1)",
  "  -dnsname[=<FILE>]:specify the file to output DANS dataします。",
  "    default:no output  -dnsname:-dnsname=facte_dns.d",
  "    output data:factor score",
  "  -eps=<EPS>:specify the convergence error(0<EPS)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -lamda=<VALUE>:specify the eigenvalue of reference(0=<VALUE)",
  "  -meth={ml|pf}:select the factor extraction method",
  "    ml:maximum likelihood method  pf:principal factor analysis(default)",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=facte_ngp.dat",
  "    output data:factor loading",
  "  -pm[=<NUMBER>]:specify the number of exponent of power for promax rotation",
  "    (0<NUMBER  -pm:-pm=4)",
  "  -q=<NUMBER>:specify the maximum number of factors",
  "    (0<NUMBER=<number of variables)",
  "  -res=[-][{def|all|fa|fc|fl|fr|lt|stat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    fa:factor  fc:coefficient for calculating factor scores",
  "    fl:factor loading  fr:correlation of factor scores",
  "    lt:likelihood ratio test  stat:basic statistics",
  "  -rmax=<NUMBER>:specify the maximum number of repetitions(0=<NUMBER)",
  "  -sort[={-|+}]:specify whether factor loadings should be sorted by factor",
  "    order(-sort:-sort=+)",
  "  -vm[={-|+}]:specify whether varimax rotation should be output",
  "    (-vm:-vm=+)"
};

const string Title = "=== Factor Analysis ==="; // 表題
const string DNSNAMEdef = "facte_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "facte_ngp.dat"; // 省略時Ngraphファイル名
constexpr ldbl EPSdef = 1.0e-5L;          // 収束誤差省略時値
constexpr int RMAXdef = 200;              // 省略時最大繰り返し回数
constexpr int PMdef = 4;        // プロマックス回転の省略時次数

enum d2type                     // 独自性初期値の種類
{
  d2max,                        // 上限値
  d2q,                          // 最大因子数に基づく推定値
  d2coef,                       // 上限値にかける係数指定
  d2etc                         // その他
};

enum methtype                   // 因子抽出法の種類
{
  methnon,                      // 指定なし
  methml,                       // 最尤法
  methpf,                       // 主因子法
  methetc                       // その他
};

enum rottype                    // 因子軸回転の種類
{
  rotnon,                       // 回転なし
  rotpm,                        // promax回転
  rotvm,                        // 規準Varimax回転
  rotetc                        // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  d2type vd2;                   // 独自性初期値の種類
  ldbl vd2c;                    // 独自性初期値の係数
  string vdnsname;              // DANSファイル名
  ldbl veps;                    // 収束誤差
  ldbl vlamda;                  // 因子抽出用因子負荷量
  methtype vmeth;               // 因子抽出法
  string vngpname;              // Ngraphファイル名
  int vpm;                      // promax回転のベキ乗数
  int vq;                       // 因子抽出用因子数
  lstr vreslst;                 // 解析結果出力リスト
  map<string, bool> vresmap;    // 解析結果出力の有無マップ
  int vrmax;                    // 最大反復回数
  bool vsort;                   // 因子負荷量並べ替えフラッグ
  bool vvm;                     // 規準varimax回転結果出力フラッグ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  d2type d2() const             // 独自性初期値の種類を返す
  { return vd2; }
  ldbl d2c() const              // 独自性初期値の係数を返す
  { return vd2c; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  ldbl eps() const              // 収束誤差を返す
  { return veps; }
  ldbl lamda() const            // 因子抽出用因子負荷量を返す
  { return vlamda; }
  methtype meth() const         // 因子抽出法を返す
  { return vmeth; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  int pm() const                // promax回転のベキ乗数を返す
  { return vpm; }
  int q() const                 // 因子抽出用因子数を返す
  { return vq; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 最大反復回数を返す
  { return vrmax; }
  bool sort() const             // 因子負荷量並べ替えフラッグを返す
  { return vsort; }
  bool vm() const             // 規準varimax回転結果出力フラッグを返す
  { return vvm; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vd2(d2max), vd2c(1.0L), vdnsname(), veps(EPSdef), vlamda(1.0L), vmeth(methpf), vpm(0), vq(0), vngpname(), vreslst(), vresmap(), vrmax(RMAXdef), vvm(true), vsort(false)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "d2")      // 独自性初期値の種類指定
  {
    if (opt[1].empty() || opt[1] == "max")
    { vd2 = d2max; }
    else if (opt[1] == "q")
    { vd2 = d2q; }
    else
    {
      vd2c = strtodf(opt[1]);
      if (vd2c <= 0.0L || 1.0L <= vd2c) err = badopt(vpname, optstr, USAGE);
      else vd2 = d2coef;
    }
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < LDBL_EPSILON) err = badopt(vpname, optstr, USAGE, LDBL_EPSILON, LDBL_MAX);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "lamda")   // 因子抽出用因子負荷量の指定
  {
    vlamda = opt[1].empty() ? 1.0L : strtodf(opt[1]);
    if (vlamda < 0.0L || vlamda == (ldbl)BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "meth")    // 因子抽出法指定
  {
    if (opt[1].empty() || opt[1] == "pf") vmeth = methpf;
    else if (opt[1] == "ml") vmeth = methml;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pm")      // promax回数のベキ乗数指定
  {
    char* ptr = nullptr;
    vpm = opt[1].empty() ? PMdef : strtol(opt[1].c_str(), &ptr, 0);
    if (vpm < 0) err = badopt(vpname, optstr, USAGE, 0.0, 10.0);
  }
  else if (opt[0] == "q")       // 最大因子数指定
  {
    char* ptr = nullptr;
    vq = strtol(opt[1].c_str(), &ptr, 0);
    if (vq < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 最大反復回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "sort")    // 因子負荷量並べ替え指定
  {
    if (opt[1].empty() || opt[1] == "+") vsort = true;
    else if (opt[1] == "-") vsort = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "vm")      // 規準varimax回転結果出力指定
  {
    if (opt[1].empty() || opt[1] == "+") vvm = true;
    else if (opt[1] == "-") vvm = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if (set_res() < 0) return -3; // -resオプションのエラー
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["fa"] = vresmap["fl"] = vresmap["fc"] = vresmap["fr"] = vresmap["stat"] = true;
  vresmap["lt"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["fa"] = vresmap["fl"] = vresmap["fc"] = vresmap["fr"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["fa"] = vresmap["fl"] = vresmap["fc"] = vresmap["fr"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["fa"] = vresmap["fl"] = vresmap["fc"] = vresmap["fr"] = vresmap["lt"] = vresmap["stat"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

void vrot(mat<ldbl>& lmat)    // varimax回転主要部
{
  int p = lmat.r();             // 変数の数
  int q = lmat.c();             // 因子数
  for (int j = 0; j < q - 1; ++j)
  {
    for (int k = j + 1; k < q; ++k)
    {
      ldbl a = 0.0, b = 0.0, c = 0.0, d = 0.0;
      for (int i = 0; i < p; ++i)
      {
        ldbl lj = lmat(i, j), lk = lmat(i, k);
        ldbl ljjkk = lj * lj - lk * lk, ljk = lj * lk;
        a += ljjkk;  b += ljk;
        c += ljjkk * ljjkk - 4.0 * ljk * ljk;  d += ljjkk * ljk;
      }
      b *= 2.0;  d *= 4.0;
      d -= 2.0 * a * b / (ldbl)p;
      b = c - (a * a - b * b) / (ldbl)p;
      ldbl h = atan2(d, b) / 4.0;
      for (int i = 0; i < p; ++i)
      {
        ldbl lj = lmat(i, j), lk = lmat(i, k);
        lmat.assign(i, j, lj * cos(h) + lk * sin(h));
        lmat.assign(i, k, lk * cos(h) - lj * sin(h));
      }
    }
  }
}

int varimax(mat<ldbl>& lmat, ldbl eps = EPSdef, int rmax = RMAXdef)
// 因子負荷量の規準varimax回転
{
  int p = lmat.r() - 1;         // 変数の数
  int q = lmat.c() - 1;         // 因子数
  if (p < 2 || q < 2) return 0;
  mat<ldbl> amat(lmat.sub(1, 1, lmat.r() - 1, lmat.c() - 1));
  for (int i = 0; i < p; ++i)
  {
    ldbl norm = 0.0;
    for (int j = 0; j < q; ++j) norm += amat(i, j) * amat(i, j);
    norm = sqrt(norm);
    lmat.assign(i + 1, 0, norm);
    for (int j = 0; j < q; ++j) amat.assign(i, j, (norm > 0.0) ? amat(i, j) / norm : 0.0);
  }
  int r = 0;
  ldbl sign = 0.0L, sigo = 0.0L, sigm = 0.0L, sigm2 = 0.0L, m = 0.0L;
  do {
    ++r;
    sigo = sign;
    vrot(amat);
    sign = 0.0L;
    for (int j = 0; j < q; ++j)
    {
      sigm = sigm2 = 0.0L;
      for (int i = 0; i < p; ++i)
      {
        m = amat(i, j) * amat(i, j);
        sigm += m;  sigm2 += m * m;
      }
      sign += sigm2 - sigm * sigm / (ldbl)p;
    }
  } while (sgndiff(sign, sigo) != 0 && fabs(sign - sigo) > eps * (fabs(sign) + fabs(sigo)) && r < rmax);
  for (int i = 1; i <= p; ++i)
  {
    ldbl norm = lmat(i, 0);
    for (int j = 1; j <= q; ++j) lmat.assign(i, j, amat(i - 1, j - 1) * norm);
  }
  return r;                     // 繰り返し数を返す
}

int promax(mat<ldbl>& lmat, mat<ldbl>& tprmat, int pm)
// 因子負荷量のpromax回転
{
  int p = (int)(lmat.r()) - 1;  // 変数の数
  int q = (int)(lmat.c()) - 1;  // 因子数
  if (p < 2 || q < 2) return 0;
  mat<ldbl> amat(lmat.sub(1, 1, lmat.r() - 1, lmat.c() - 1));
  mat<ldbl> cmat(p, q);       // ターゲット行列
  ldbl pmf = (ldbl)pm;
  for (int j = 0; j < p; ++j)
  {
    ldbl norm = 0.0L;
    for (int k = 0; k < q; ++k) norm += amat(j, k) * amat(j, k);
    norm = sqrt(norm);
    for (int k = 0; k < q; ++k)
    {
      ldbl ljk = (norm > 0.0) ? amat(j, k) / norm : 0.0L;
      cmat.assign(j, k, sgn(ljk) * pow(fabs(ljk), pmf));
    }
  }
  mat<ldbl> tmat(amat);
  mat<ldbl> aamat(tmat.trn() * amat);
  aamat.inv();
  mat<ldbl> acmat(tmat * cmat);
  mat<ldbl> trmat(aamat * acmat);
  mat<ldbl> ttmat(trmat);
  mat<ldbl> dmat(ttmat.trn() * trmat);
  dmat.inv();
  for (size_t j = 0; j < dmat.c(); ++j)
  {
    for (size_t i = 0; i < dmat.r(); ++i)
    {
      if (i == j) dmat.assign(i, j, (dmat(i, j) > 0.0) ? sqrt(dmat(i, j)) : 0.0L);
      else dmat.assign(i, j, 0.0L);
    }
  }
  tprmat = trmat * dmat;
  tmat.trn();
  tmat = amat * tprmat;
  for (int j = 1; j <= p; ++j)
  { for (int k = 1; k <= q; ++k) lmat.assign(j, k, tmat(j - 1, k - 1)); }
  return 1;                     // 回転成功
}

class factclass                 // factメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_smat(mat<ldbl>& smat, vecldbl& mevec, vecldbl& vrvec);
  // データを集計して積和行列を求める
  void put_resid(ostream& fout);
  // 解析結果のタイトルとデータ識別情報を出力する
  int cal_fact(const option& opt, int& r, ldbl& llr, ldbl& csqr, int& df, vecldbl& evalvec0, vecldbl& evalvec, mat<ldbl>& lmat, int n, const mat<ldbl>& smat) const;
  // 因子を抽出する
  ldbl init_dv(const option& opt, vec<ldbl>& dv, const mat<ldbl>& rmat) const;
  // 独自性初期値を設定する
  int cal_lmat(const option& opt, vecldbl& evalvec, mat<ldbl>& qmat, mat<ldbl>& lmat, int q, ldbl lamda, const vec<ldbl>& dv, const mat<ldbl>& rmat) const;
  // 因子負荷行列を計算する
  int fact_ml(const option& opt, int& r, vecldbl& evalvec0, vecldbl& evalvec, mat<ldbl>& lmat, vec<ldbl>& dv, const mat<ldbl>& rmat) const;
  // 最尤法による因子の抽出
  int fact_pf(const option& opt, int& r, vecldbl& evalvec0, vecldbl& evalvec, mat<ldbl>& lmat, vec<ldbl>& dv, const mat<ldbl>& rmat) const;
  // 主因子法による因子の抽出
  ldbl fsllr(const mat<ldbl>& lmat, const mat<ldbl>& rmat, const vec<ldbl>& dv) const;
  // 特殊対数尤度比を返す関数
  void put_dns(const option& opt, const mat<ldbl>& amat, rottype rtype);
  // 因子スコアをデータファイル形式で出力する
  void lmat2amat(mat<ldbl>& amat, int n, const mat<ldbl>& smat) const;
  // 因子負荷量を因子スコア計算用係数(回帰係数)に変換する
  void cal_score(datac& score, datac& cdata, const mat<ldbl>& amat) const;
  // 因子スコアを計算し、データとして返す
  void put_factor(ostream& fout, option& opt, int r, ldbl llr, ldbl csqr, int df, const vecldbl& evalvec0, const vecldbl& evalvec, mat<ldbl>& lmat, mat<ldbl>& amat, int n, const mat<ldbl>& smat);
  // 因子分析結果を出力する
  rottype put_rot(ostream& fout, option& opt, int p , int q, mat<ldbl>& lmat, mat<ldbl>& amat, int n, const mat<ldbl>& smat);
  // 回転後の因子分析結果を出力する
  void put_eval(ostream& fout, int q, const vecldbl& evalvec0, const vecldbl& evalvec, const mat<ldbl>& lmat) const;
  // 固有値を出力する。
  void put_lt(ostream& fout, ldbl llr, ldbl csqr, int df) const;
  // 尤度比検定結果を出力する。
  void put_fload(ostream& fout, const option& opt, const string& mtitle, const mat<ldbl>& lmat, bool fl) const;
  // 因子負荷量行列を表形式で出力する。
  void sort_lvecvec(vector<vecldbl >& lvecvec, vector<vecldbl*>& lvecpvec) const;
  // 因子負荷量ベクトルを因子順に並べ替える(Schellソート)
  int cmp_vec(const vecldbl* lvecp1, const vecldbl* lvecp2) const;
  // 因子負荷量ベクトルの比較
  void put_pmat(ostream& fout, const string& mtitle, const mat<ldbl>& pmat) const;
  // 回転用因子変換行列を出力する。
  void put_rfact(ostream& fout, const string& mtitle, const mat<ldbl>& amat0, int n, const mat<ldbl>& smat) const;
  // 因子スコア間相関係数とCronbachのα係数を出力する。
  void plot_flplot(option& opt, const mat<ldbl>& lmat, rottype rtype);
  // Ngraph用因子負荷量のプロットデータを出力する
public:
  factclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~factclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み保存する
  void put(option& opt);                     // 集計結果を出力する
};

int factclass::cal_smat(mat<ldbl>& smat, vecldbl& mevec, vecldbl& vrvec)
// データを集計して積和行列を求める
{
  size_t dim = id.ni() + 1, p = id.ni();
  vecdbl fdatavec(dim, 0.0);
  fdatavec[0] = 1.0;
  for (auto& cdata : datalst)
  {
    bool blank = false;
    for (size_t j = 1; blank == false && j < dim; ++j)
    {
      fdatavec[j] = cdata.fdata(j - 1);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (blank)
    { cdata.sdata(0, BLANKs); } // 後の計算のために最初の項目をブランクにする
    else
    {
      for (size_t i = 0; i < dim; ++i)
      {
        double x = fdatavec[i];
        for (size_t j = i; j < dim; ++j) smat.add(i, j, (ldbl)(x * fdatavec[j]));
      }
    }
  }
  int n = (int)(smat(0, 0));
  for (size_t i = 0; i < dim - 1; ++i)
  {
    for (size_t j = i + 1; j < dim; ++j) smat.assign(j, i, smat(i, j));
  }
  smat.sweep(0);              // 単純積和行列→積和行列
  for (size_t i = 0; i < dim; ++i)
  {
    mevec[i] = smat(0, i);
    vrvec[i] = (n > 1) ? smat(i, i) / (ldbl)(n - 1) : 0.0L;
  }
  return n;                     // 例数を返す
}

void factclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    fout << "VARIABLE X" << setw(3) << i + 1;
    fout << ":" << id.item(i) << "\n";
  }
}

int factclass::cal_fact(const option& opt, int& r, ldbl& llr, ldbl& csqr, int& df, vecldbl& evalvec0, vecldbl& evalvec, mat<ldbl>& lmat, int n, const mat<ldbl>& smat) const
// 因子を抽出する
{
  int p = id.ni(), q = 0;
  mat<ldbl> rmat(smat.sub(1, 1, p, p));
  smat2rmat(rmat);
  vec<ldbl> dv(p);
  ldbl d2c = init_dv(opt, dv, rmat);
  switch (opt.meth())
  {
  case methml : q = fact_ml(opt, r, evalvec0, evalvec, lmat, dv, rmat);
    break;
  default : q = fact_pf(opt, r, evalvec0, evalvec, lmat, dv, rmat);
  }
  lmat.resize(p + 1, q + 1);
  ldbl sllr = fsllr(lmat, rmat, dv);
  llr = -0.5L * (ldbl)n * sllr;
  csqr = ((ldbl)n - (ldbl)(2 * p + 5) / 6.0L - 2.0L * (ldbl)q / 3.0L) * sllr;
  df = ((p - q) * (p - q) - (p + q)) / 2;
  return q;                     // 因子数を返す
}

ldbl factclass::init_dv(const option& opt, vec<ldbl>& dv, const mat<ldbl>& rmat) const
// 独自性初期値を設定する
{
  int p = id.ni();
  int q = p;
  ldbl d2c = 1.0;
  switch (opt.d2())             // 独自性初期値の種類
  {
  case d2coef : d2c = opt.d2c(); // 係数指定
    break;
  case d2q : d2c = (0 < opt.q() && opt.q() < p) ? (ldbl)(opt.q()) : 1.0L;
    d2c = 1.0L - d2c / (ldbl)(2 * p); // 因子数に基づく推定値
  }
  mat<ldbl> rtmat(rmat);
  rtmat.inv();
  const ldbl Deps = LDBL_EPSILON * fmax(10.0L, rtmat.maxabse());
  for (int i = 0; i < p; ++i)
  {
    ldbl di = (rtmat(i, i) > Deps) ? d2c * sqrt(1.0L / rtmat(i, i)) : 0.9L;
    dv.assign(i, di);
  }
  return d2c;                   // 上限値にかける係数を返す
}

int factclass::fact_ml(const option& opt, int& r, vecldbl& evalvec0, vecldbl& evalvec, mat<ldbl>& lmat, vec<ldbl>& dv, const mat<ldbl>& rmat) const
// 最尤法による因子の抽出
{
  int p = id.ni();
  int q = p;
  mat<ldbl> rtmat(rmat);
  mat<ldbl> qmat(p);
  vec<ldbl> ev(p);
  ldbl lamda = 1.0L;
  if (opt.lamda() != BLANKf && opt.lamda() > 1.0) lamda = opt.lamda();
  ldbl res = 0.0L;
  r = 0;                        // 繰り返し計算数
  do
  {
    ++r;
    q = cal_lmat(opt, evalvec, qmat, lmat, q, lamda, dv, rmat);
    if (r == 1)
    { for (int i = 0; i <= q; ++i) evalvec0[i] = evalvec[i]; }
    res = 0.0L;
    for (int i = 0; i < p; ++i)
    {
      ldbl hi2 = 0.0L;
      for (int k = 0; k < q; ++k) hi2 += lmat(i + 1, k + 1) * lmat(i + 1, k + 1);
      ldbl hi2pre = 1.0 - dv(i) * dv(i);
      ldbl er = hi2pre - hi2;
      res += er * er;
      dv.assign(i, sqrt(1.0L - hi2));
    }
  } while (res >= opt.eps() && r < opt.rmax());
  return q;                     // 因子数を返す
}

int factclass::cal_lmat(const option& opt, vecldbl& evalvec, mat<ldbl>& qmat, mat<ldbl>& lmat, int q, ldbl lamda, const vec<ldbl>& dv, const mat<ldbl>& rmat) const
// 因子負荷行列を計算する
{
  int p = rmat.r();
  mat<ldbl> rtmat(rmat);
  evalvec[0] = 0.0L;
  for (int i = 0; i < p; ++i)
  {
    ldbl di = dv(i);
    for (int j = 0; j < p; ++j) rtmat.assign(i, j, rmat(i, j) / di / dv(j));
    evalvec[0] += rtmat(i, i);
  }
  vec<ldbl> ev(p);
  int k = 0;
  if (0 < opt.q() && opt.q() <= p) q = opt.q(); // 最大因子数指定の時
  while (k < q && (evalvec[k + 1] = eigenud(rtmat, ev)) > lamda)
  {
    for (int j = 0; j < p; ++j) qmat.assign(j, k, ev(j));
    ++k;
  }
  if (k < 1) return 0;
  q = k;                    // level指定の時、因子数は初期値以下になる
  vecldbl lamvec(q, 0.0);
  for (k = 0; k < q; ++k) lamvec[k] = sqrt(evalvec[k + 1] - 1.0);
  for (int i = 0; i < p; ++i)
  {
    ldbl di = dv(i);
    for (k = 0; k < q; ++k) lmat.assign(i + 1, k + 1, di * qmat(i, k) * lamvec[k]);
  }
  return q;                     // 抽出した固有値の数を返す
}

int factclass::fact_pf(const option& opt, int& r, vecldbl& evalvec0, vecldbl& evalvec, mat<ldbl>& lmat, vec<ldbl>& dv, const mat<ldbl>& rmat) const
// 主因子法による因子の抽出
{
  int p = id.ni();
  int q = p;
  ldbl lamda = 1.0L;
  vec<ldbl> ev(p);
  if (0 < opt.q() && opt.q() <= p) { q = opt.q();  lamda = 0.0L; }
  else if (opt.lamda() != BLANKf) lamda = opt.lamda();
  evalvec0[0] = evalvec[0] = (ldbl)p;
  mat<ldbl> rtmat(rmat);
  int k = 0;
  while (k < q && (evalvec0[k + 1] = eigenud(rtmat, ev)) > lamda) ++k;
  if (k < 1) return 0;
  q = k;                        // 因子数決定
  mat<ldbl> rlmat(rmat);
  if (opt.rmax() > 0)
  {
    for (int i = 0; i < p; ++i) rlmat.assign(i, i, 1.0 - dv(i) * dv(i));
  }
  ldbl res = 0.0L;
  r = 0;                        // 繰り返し計算数
  do {
    rtmat = rlmat;
    for (int k = 1; k <= q; ++k)
    {
      evalvec[k] = eigenud(rtmat, ev);
      ldbl w = sqrt(evalvec[k]);
      for (int i = 1; i <= p; ++i) lmat.assign(i, k, ev(i - 1) * w);
    }
    res = 0.0L;
    for (int i = 1; i <= p; ++i)
    {
      ldbl hi2 = 0.0L;
      for (int k = 1; k <= q; ++k) hi2 += lmat(i, k) * lmat(i, k);
      dv.assign(i - 1, (hi2 < 1.0) ? sqrt(1.0 - hi2) : 0.0L);
      ldbl er = rlmat(i - 1, i - 1) - hi2;
      res += er * er;
      rlmat.assign(i - 1, i - 1, hi2);
    }
    ++r;
  } while (res >= opt.eps() && r < opt.rmax());
  return q;                     // 因子数を返す
}

ldbl factclass::fsllr(const mat<ldbl>& lmat, const mat<ldbl>& rmat, const vec<ldbl>& dv) const
// 特殊対数尤度比を返す関数
{
  int p = lmat.r() - 1, q = lmat.c() - 1;
  mat<ldbl> ltmat(lmat.sub(1, 1, p, q));
  mat<ldbl> lttmat(ltmat);
  lttmat.trn();
  mat<ldbl> rlmat(ltmat * lttmat);
  for (int i = 0; i < p; ++i) rlmat.assign(i, i, rlmat(i, i) + dv(i) * dv(i));
  ldbl det = rmat.det();
  const ldbl Deps = LDBL_EPSILON * fmax(10.0L, rmat.maxabse());
  ldbl lnrr = (fabs(det) > Deps) ? log(rlmat.det() / det) : 0.0L;
  rlmat.inv();
  mat<ldbl> tmat(rlmat * rmat);
  ldbl tl = 0.0;
  for (int i = 0; i < p; ++i) tl += tmat(i, i);
  return lnrr + tl - (ldbl)p; // 特殊対数尤度を返す
}

void factclass::put_dns(const option& opt, const mat<ldbl>& amat, rottype rtype)
// 因子スコアをデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  int q = amat.c() - 1;
  idd tid(id.title(), q);
  ostringstream ostr;
  for (int i = 0; i < q; ++i)
  {
    ostr << "FACTOR SCORE F" << i + 1;
    if (rtype == rotpm) ostr << "(after Promax Rotation)";
    else if (rtype == rotvm) ostr << "(after Varimax Rotation)";
    tid.assign_item(i, ostr.str());
    ostr.str("");
  }
  tid.put_item(fout, formtag);
  datac score(q);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    cal_score(score, *ditr, amat);
    score.put(fout);
  }
}

void factclass::lmat2amat(mat<ldbl>& amat, int n, const mat<ldbl>& smat) const
// 因子負荷量を因子スコア計算用係数(回帰係数)に変換する
{
  int p = amat.r() - 1;
  mat<ldbl> rmat(smat.sub(1, 1, smat.r() - 1, smat.c() - 1));
  smat2rmat(rmat);
  rmat.inv();
  mat<ldbl> tmat(rmat * amat.sub(1, 1, amat.r() - 1, amat.c() - 1));
  vecldbl wvec(p, 0.0);
  int df = n - 1;
  for (int i = 0; i < p; ++i)
  {
    ldbl ss = smat(i + 1, i + 1);
    wvec[i] = (ss > 0.0L) ? sqrt((ldbl)df / ss) : 0.0L;
  }
  int q = amat.c() - 1;
  for (int j = 1; j <= q; ++j)
  {
    amat.assign(0, j, 0.0L);
    for (int i = 1; i <= p; ++i)
    {
      amat.assign(i, j, tmat(i - 1, j - 1) * wvec[i - 1]);
      amat.assign(0, j, amat(0, j) - amat(i, j) * smat(0, i));
    }
  }
}

void factclass::cal_score(datac& score, datac& cdata, const mat<ldbl>& amat) const
// 因子スコアを計算し、データとして返す
{
  if (cdata.sdata(0) == BLANKs) // 欠損あり
  { score = BLANKs; }
  else
  {
    vecdbl fdatavec(id.ni(), 0.0);
    for (size_t i = 0; i < id.ni(); ++i) fdatavec[i] = cdata.fdata(i);
    for (size_t i = 1; i <= score.size(); ++i)
    {
      ldbl z = amat(0, i);
      for (size_t j = 0; j < id.ni(); ++j) z += amat(j + 1, i) * (ldbl)(fdatavec[j]);
      score.assign(i - 1, z);
    }
  }
}

void factclass::put_factor(ostream& fout, option& opt, int r, ldbl llr, ldbl csqr, int df, const vecldbl& evalvec0, const vecldbl& evalvec, mat<ldbl>& lmat, mat<ldbl>& amat, int n, const mat<ldbl>& smat)
// 因子分析結果を出力する
{
  int p = id.ni(), q = lmat.c() - 1;
  if (opt.res("res") && opt.res("fa"))
  {
    fout << "\n<FACTOR ANALYSIS:";
    switch (opt.meth())
    {
    case methml : fout << "Maximum Likelihood Method(repetitions=" << r << ")";
      break;
    default :
      if (opt.rmax() < 1) fout << "Principal Component Method";
      else fout << "Principal Factor Analysis(repetitions=" << r << ")";
    }
    fout << ">\n  Conditions:";
    if (0 < opt.q() && opt.q() <= p) fout << "Maximum Factors=" << opt.q();
    else fout << "Initial Eigenvalue>" << opt.lamda();
    fout << "  Initial Uniqueness=";
    switch (opt.d2())
    {
    case d2coef : fout << "Upper Limit X" << opt.d2c();
      break;
    case d2q : fout << "Estimates based on Factors";
      break;
    default : fout << "Upper Limit";
    }
    fout << "\n";
    put_eval(fout, q, evalvec0, evalvec, lmat);
    if (opt.res("lt") || (opt.res("def") && opt.meth() == methml)) put_lt(fout, llr, csqr, df);
  }
  amat = lmat;
  lmat2amat(amat, n, smat);
  if (opt.res("res"))
  {
    if (opt.res("fl")) put_fload(fout, opt, "<FACTOR LOADING>", lmat, true);
    if (opt.res("fc")) putcoef(fout, "<COEFFICIENT FOR CALCULATING FACTOR SCORES>", "FACTOR", "X", amat);
    if (opt.res("fr")) put_rfact(fout, "<CORRETATION OF FACTOR SCORES AND CRONBACH'S ALPHA>", amat, n, smat);
  }
}

rottype factclass::put_rot(ostream& fout, option& opt, int p , int q, mat<ldbl>& lmat, mat<ldbl>& amat, int n, const mat<ldbl>& smat)
// 回転後の因子分析結果を出力する
{
  mat<ldbl> lmat0(lmat.sub(1, 1, lmat.r() - 1, lmat.c() - 1));
  int rvm = varimax(lmat, opt.eps(), opt.rmax());
  mat<ldbl> mmat(lmat.sub(1, 1, lmat.r() - 1, lmat.c() - 1));
  ostringstream ostr;
  rottype rtype = rotnon;       // 回転の種類
  if (opt.vm() || opt.pm() > 0) // promax回転もVarimax回転を実行する
  {
    ostr << "<FACTOR LOADINGS AFTER VARIMAX ROTATION(repetitions=" << rvm << ")>";
    if (opt.vm() && opt.res("fl")) put_fload(fout, opt, ostr.str(), lmat, true);
    amat = lmat;
    lmat2amat(amat, n, smat);
    if (opt.vm() && opt.res("fc")) putcoef(fout, "<COEFFICIENT FOR CALCULATING FACTOR SCORES AFTER VARIMAX ROTATION>", "FACTOR", "X", amat);
    rtype = rotvm;
  }
  mat<ldbl> tmat(lmat0);
  mat<ldbl> llmat(tmat.trn() * lmat0);
  mat<ldbl> lmmat(tmat * mmat);
  mat<ldbl> pmat(llmat.inv() * lmmat);
  if (opt.vm())
  {
    if (opt.res("def")) put_pmat(fout, "<FACTOR TRANSFORMATION MATRIX FOR VARIMAX ROTATION>", pmat);
    if (opt.res("fr")) put_rfact(fout, "<CORRETATION OF FACTOR SCORES AND CRONBACH'S ALPHA AFTER VARIMAX ROTATION>", amat, n, smat);
  }
  if (opt.pm() < 1) return rtype; // promax回転実行せず
  mat<ldbl> tprmat(q, q);
  promax(lmat, tprmat, opt.pm());
  mmat = lmat.sub(1, 1, lmat.r() - 1, lmat.c() - 1);
  ostr.str("");
  ostr << "<FACTOR LOADINGS AFTER PROMAX ROTAION(Exponet of power=" << opt.pm() << ")>";
  if (opt.res("fl")) put_fload(fout, opt, ostr.str(), lmat, false);
  mat<ldbl> trmat(tprmat);
  mat<ldbl> ttmat(trmat.trn() * tprmat);
  ttmat.inv();
  mat<ldbl> fcormat(q + 1, q + 1);
  for (int i = 1; i <= q; ++i)
  {
    for (int j = 1; j <= q; ++j) fcormat.assign(i, j, ttmat(i - 1, j - 1));
  }
  amat = lmat * fcormat;
  if (opt.res("fc")) put_fload(fout, opt, "<FACTOR STRUCTURE MATRIX AFTER PROMAX ROTATION>", amat, false);
  lmat2amat(amat, n, smat);
  if (opt.res("fc")) putcoef(fout, "<COEFFICIENT FOR CALCULATING FACTOR SCORES AFTER PROMAX ROTAION>", "FACTOR", "X", amat);
  if (opt.res("def")) put_pmat(fout, "<FACTOR TRANSFORMATION MATRIX FOR PROMAX ROTATION>", tprmat);
  if (opt.res("fr"))
  {
    put_pmat(fout, "<CORRELATIN MATRIX OF FACTOR AXIS AFTER PROMAX ROTATION(r=cos(h))>", ttmat);
    put_rfact(fout, "<CORRETATION OF FACTOR SCORES AND CRONBACH'S ALPHA AFTER PROMAX ROTATION>", amat, n, smat);
  }
  rtype = rotpm;
  return rtype;                 // 回転の種類を返す
}
  
void factclass::put_eval(ostream& fout, int q, const vecldbl& evalvec0, const vecldbl& evalvec, const mat<ldbl>& lmat) const
// 固有値を出力する。
{
  fout << "FACT  Initial EV    Final EV  Loading SS Cum-Loading         r^2     Cum-r^2\n";
  constexpr int LENline = 76;   // 区切り線の長さ
  putline(fout, LENline);
  int p = id.ni();
  ldbl slss = 0.0L;
  for (int k = 1; k <= q; ++k)
  {
    fout << "F" << setw(3) << k << setw(12) << evalvec0[k];
    fout << setw(12) << evalvec[k];
    ldbl lss = 0.0L;
    for (int i = 1; i <= p; ++i) lss += lmat(i, k) * lmat(i, k);
    slss += lss;
    fout << setw(12) << lss << setw(12) << slss;
    fout << setw(12) << lss / (ldbl)p;
    fout << setw(12) << slss / (ldbl)p << "\n";
  }
  putline(fout, LENline);
}

void factclass::put_lt(ostream& fout, ldbl llr, ldbl csqr, int df) const
// 尤度比検定結果を出力する。
{
  fout << "Test of LOF for Number of Factors  LR=" << exp(llr) << "  ln(LR)=" << llr << "\n";
  ldbl pr = 0.0L;
  if (csqr < 0.0 || df < 1)
  { fout << "  UNTESTABLE !\n"; }
  else
  {
    pr = ctop(csqr, df);
    fout << "  chi-squared=" << csqr << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  }
}

void factclass::put_fload(ostream& fout, const option& opt, const string& mtitle, const mat<ldbl>& lmat, bool fl) const
// 因子負荷量行列を表形式で出力する。
{
  int p = lmat.r() - 1, q = lmat.c() - 1;
  vecldbl lvec(q + 3, 0.0);
  vector<vecldbl> lvecvec(p, lvec); // ソート用因子負荷量ベクトル
  vector<vecldbl*> lvecpvec(p); // ソート用ポインターベクトル
  for (int i = 0; i < p; ++i)   // 因子負荷量に変数番号と因子番号を付加する
  {
    ldbl lmax = 0.0L;
    int jmax = 0;
    (lvecvec[i])[2] = 0.0L;
    for (int j = 3; j < q + 3; ++j)
    {
      (lvecvec[i])[j] = lmat(i + 1, j - 2);
      ldbl labs = fabs(lmat(i + 1, j - 2));
      if (labs > lmax) { lmax = labs;  jmax = j; }
    }
    (lvecvec[i])[0] = (ldbl)i;
    (lvecvec[i])[1] = (ldbl)jmax;
    lvecpvec[i] = &(lvecvec[i]);
  }
  if (opt.sort()) sort_lvecvec(lvecvec, lvecpvec);
  vecldbl l2vec(q + 1, 0.0L);
  for (int j = 0; j <= q; ++j) l2vec[j] = 0.0L;
  constexpr int Cmax = 6;       // 行列表横の最大数
  int low = 1, up = 0, eopl = (fl) ? 2 : 0;
  bool eop = false;             // 表の最後出力フラッグ
  fout << "\n" << mtitle;
  int prec0 = fout.precision(); // 現在の浮動小数点精度を保存する
  while (eop == false)
  {
    int nc = low + Cmax - 1;
    if (low > q) eop = true;
    else if (nc >= q + eopl) { up = q;  eop = true; }
    else if (nc >= q) up = q;
    else up = nc;
    fout << "\nVAR ";
    for (int j = low; j <= up; j++) fout << "    FACTOR" << setw(2) << j;
    if (eop && fl) fout << " COMMUNALITY  UNIQUENESS";
    fout << "\n";
    int llen = 4 + 12 * (up - low + 1 + ((eop) ? eopl : 0));
    putline(fout, llen);
    for (int i = 0; i < p; ++i)
    {
      vecldbl* lvecpi = lvecpvec[i];
      fout << "X" << setw(3) << (*lvecpi)[0] + 1;
      fout.precision(6);  fout.setf(ios::fixed);
      for (int j = low; j <= up; ++j)
      {
        fout << setw(12) << (*lvecpi)[j + 2];
        ldbl l2 = (*lvecpi)[j + 2] * (*lvecpi)[j + 2];
        (*lvecpi)[2] += l2;
        l2vec[j] += l2;
      }
      if (eop && fl) fout << setw(12) << (*lvecpi)[2] << setw(12) << 1.0L - (*lvecpi)[2];
      fout.precision(prec0);  fout.unsetf(ios::fixed);
      fout << "\n";
    }
    putline(fout, llen);
    if (fl)
    {
      fout << "r^2   ";
      fout.precision(6);  fout.setf(ios::fixed);
      ldbl r2 = 0.0L;
      for (int j = low; j <= up; ++j)
      {
        r2 = l2vec[j] / (ldbl)p;
        if (j == low) fout << setw(10) << r2;
        else fout << setw(12) << r2;
        l2vec[0] += l2vec[j];
      }
      if (eop)
      {
        r2 = l2vec[0] / (ldbl)p;
        if (low > q) fout << setw(10) << r2;
        else fout << setw(12) << r2;
        fout << setw(12) << 1.0L - r2;
      }
      fout.precision(prec0);  fout.unsetf(ios::fixed);
      fout << "\n";
    }
    low = up + 1;
  }
}

void factclass::sort_lvecvec(vector<vecldbl >& lvecvec, vector<vecldbl*>& lvecpvec) const
// 因子負荷量ベクトルを因子順に並べ替える(Schellソート)
{
  int p = lvecvec.size();
  for (int gap = p / 2; gap > 0; gap /=2)
  {
    for (int i = gap; i < p; ++i)
    {
      vecldbl* lvecpi = lvecpvec[i];
      int j = 0;
      for (j = i - gap; j >= 0 && cmp_vec(lvecpi, lvecpvec[j]) < 0; j -= gap) lvecpvec[j + gap] = lvecpvec[j];
      lvecpvec[j + gap] = lvecpi;
    }
  }
}

int factclass::cmp_vec(const vecldbl* lvecp1, const vecldbl* lvecp2) const
// 因子負荷量ベクトルの比較
{
  int jmax1 = (int)((*lvecp1)[1]), jmax2 = (int)((*lvecp2)[1]);
  if (jmax1 < jmax2) return -1; // 因子番号は小さい順
  if (jmax1 > jmax2) return 1;
  ldbl labs1 = fabs((*lvecp1)[jmax1]), labs2 = fabs((*lvecp2)[jmax2]);
  if (labs1 > labs2) return -1; // 因子負荷量は絶対値の大きい順
  if (labs1 < labs2) return 1;
  return 0;
}

void factclass::put_pmat(ostream& fout, const string& mtitle, const mat<ldbl>& pmat) const
// 回転用因子変換行列を出力する。
{
  constexpr int Cmax = 5;       // 行列表横の最大個数
  int p = pmat.r(), q = pmat.c(), low = 0, up = 0;
  fout << "\n" << mtitle;
  int prec0 = fout.precision(); // 現在の浮動小数点精度を保存する
  while (low < q)
  {
    up = (low + Cmax < q) ? low + Cmax : q;
    fout << "\n" << setw(8) << "";
    for (int j = low; j < up; j++) fout << "    FACTOR" << setw(2) << j + 1;
    fout <<  "\n";
    int llen = 8 + 12 * (up - low);
    putline(fout, llen);
    for (int i = 0; i < q; ++i)
    {
      fout << "FACTOR" << setw(2) << i + 1;
      fout.precision(6);  fout.setf(ios::fixed);
      for (int j = low; j < up; ++j) fout << setw(12) << pmat(i, j);
      fout.precision(prec0);  fout.unsetf(ios::fixed);
      fout << "\n";
    }
    putline(fout, llen);
    low = up;
  }
}

void factclass::put_rfact(ostream& fout, const string& mtitle, const mat<ldbl>& amat0, int n, const mat<ldbl>& smat) const
// 因子スコア間相関係数とCronbachのα係数を出力する。
{
  constexpr int Cmax = 5;       // 行列表横の最大個数
  mat<ldbl> amat(amat0);
  int p = amat.r() - 1, q = amat.c() - 1, low = 1, up = 0, eopl = 1;
  ldbl df = (ldbl)(n - 1);
  for (int i = 0; i <= q; ++i)
  { amat.assign(i, 0, 0.0L);  amat.assign(0, i, 0.0L); }
  for (int i = q + 1; i <= p; ++i) amat.assign(i, 0, 0.0L);
  mat<ldbl> atmat(amat);
  mat<ldbl> attmat(atmat.trn() * smat);
  mat<ldbl> rmat(attmat * amat);
  bool eop = false;             // 表の最後の出力フラッグ
  fout << "\n" << mtitle;
  int prec0 = fout.precision(); // 現在の浮動小数点精度を保存する
  while (eop == false)
  {
    int nc = low + Cmax - 1;
    if (low > q) eop = true;
    else if (nc >= q + eopl) { up = q;  eop = true; }
    else if (nc >= q) up = q;
    else up = nc;
    fout << "\n" << setw(8) << "";
    for (int j = low; j <= up; j++) fout << "    FACTOR" << setw(2) << j;
    if (eop) fout << "    VARIANCE";
    fout <<  "\n";
    int llen = 8 + 12 * (up - low + 1 + ((eop) ? eopl : 0));
    putline(fout, llen);
    for (int i = 1; i <= q; ++i)
    {
      fout << "FACTOR" << setw(2) << i;
      fout.precision(6);  fout.setf(ios::fixed);
      for (int j = low; j <= up; ++j)
      {
        if (i == j)
        { fout << setw(12) << 1.0; }
        else
        {
          ldbl r = rmat(i, i) * rmat(j, j);
          r = (r > 0.0L) ? rmat(i, j) / sqrt(r) : 0.0L;
          fout << setw(12) << r;
        }
      }
      if (eop) fout << setw(12) << rmat(i, i) / df;
      fout.precision(prec0);  fout.unsetf(ios::fixed);
      fout << "\n";
    }
    putline(fout, llen);
    fout << "ALPHA   ";
    fout.precision(6);  fout.setf(ios::fixed);
    for (int j = low; j <= up; ++j)
    {
      ldbl sxx = 0.0;
      for (int i = 1; i <= p; ++i) sxx += amat(i, j) * smat(i, i) * amat(i, j);
      ldbl alpha = (rmat(j, j) > 0.0L) ? (ldbl)p / (ldbl)(p - 1) * (1.0L - sxx / rmat(j, j)) : 0.0L;
      fout << setw(12) << alpha;
    }
    fout.precision(prec0);  fout.unsetf(ios::fixed);
    fout << "\n";
    low = up + 1;
  }
}

void factclass::plot_flplot(option& opt, const mat<ldbl>& lmat, rottype rtype)
// Ngraph用因子負荷量のプロットデータを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  fout << "-s1 -zx-1,1,0.5 -zy-1,1,0.5\n";
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    fout << "%VARIABLE X" << setw(3) << i + 1;
    fout << ":" << id.item(i) << "\n";
  }
  fout << "%When drawing a circle:-dline, When drawing a plot:-dmark,00 -o300\n";
  constexpr uint RDdef = 60;    // 座標軸の分割数
  ldbl dh = 2.0L * M_PI / (ldbl)RDdef;
  ldbl h = 0.0L;
  for (uint i = 0; i <= RDdef; ++i, h += dh) fout << cos(h) << " " <<  sin(h) << "\n";
  int p = lmat.r() - 1, q = lmat.c() - 1;
  for (int j = 1; j < q; ++j)
  {
    for (int k = j + 1; k <= q; ++k)
    {
      fout << NGPdelm << "\n";
      fout << "%X:F" << j << "  Y:F" << k;
      if (rtype == rotpm) fout << " (after Promax Rotation)";
      else if (rtype == rotvm) fout << " (after Varimax Rotation)";
      fout << "\n";
      for (int i = 1; i <= p; ++i) fout << lmat(i, j) << " " << lmat(i, k) << "\n";
    }
  }
}

int factclass::get(const char* c_fname, option& opt) // データを読み込み保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void factclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  int p = id.ni(), dim = id.ni() + 1;
  mat<ldbl> smat(dim);          // 積和行列
  vecldbl mevec(dim, 0.0), vrvec(dim, 0.0); // 平均値ベクトルと分散ベクトル
  int n = cal_smat(smat, mevec, vrvec); // 例数
  if (opt.res("res"))
  {
    put_resid(cout);
    if (opt.res("stat"))
    {
      putmsdvec(cout, n, mevec, vrvec);
      putmat(cout, "<CORRELATION MATIRX>", matr, n, smat);
    }
  }
  vecldbl evalvec0(dim, 0.0L), evalvec(dim, 0.0L); // 固有値ベクトル
  mat<ldbl> lmat(dim);          // 因子負荷量行列
  mat<ldbl> amat(dim);          // 因子スコア計算用係数行列
  ldbl llr = 0.0L, csqr = 0.0L;
  int r = 0, df = 0;
  int q = cal_fact(opt, r, llr, csqr, df, evalvec0, evalvec, lmat, n, smat);
  if (q < 1)                    // 因子なし
  {
    if (opt.res("res")) cout << "\n" << endl;
    return;
  }
  rottype rtype = rotnon;       // 回転の種類
  put_factor(cout, opt, r, llr, csqr, df, evalvec0, evalvec, lmat, amat, n, smat);
  if (q > 1 && (opt.vm() || opt.pm() > 0)) rtype = put_rot(cout, opt, p, q, lmat, amat, n, smat);
  if (opt.res("res")) cout << "\n" << endl;
  if (opt.dnsname() != "") put_dns(opt, amat, rtype);
  if (opt.ngpname() != "") plot_flplot(opt, lmat, rtype);
}

int main(int argc, char* argv[]) // メインルーチン
{
  factclass fact;               // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) fact.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { fact.get(argv[i], opt);  fok = false; }
    }
    fact.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
