// ldacve.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-29 16:03:17 tac>
//
// 最初の項目を判別項目、以後の項目を説明変数とした判別分析の交差検証を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W ldacve ldacve.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "ldacve  Version 7.41e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:ldacv [<OPTION> ...] [<DATA FILE> ...]",
  "  The cross validation of discriminant analysis with the first item as",
  "  discriminant variable and subsequent items as explanatory variables.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -g{1|2}[=<CODE>]:specify the code of group1 or group2",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -r=<NUMBER>:specify the number of cases in validation group"
};

using lditr = list<ldata_itr>; // データリスト反復子リストの別名定義
using lditr_itr = list<ldata_itr>::iterator; // 反復子リスト反復子の別名定義
using ldvec = list<vecdbl>;     // ベクトルデータの別名定義
using ldvec_itr = list<vecdbl>::iterator; // 反復子の別名定義

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  vecstr vgvec;                 // 群コード
  int vr;                       // 検証群のデータ数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const string& g(int idx) const // 群コードを返す
  { return vgvec[idx]; }
  int r() const                 // 検証群データ数を返す
  { return vr; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vr(1), vgvec(2)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
  vgvec[0] = "1";
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g1")      // 群1のコード指定
  { vgvec[0] = opt[1]; }
  else if (opt[0] == "g2")      // 群2のコード指定
  { vgvec[1] = opt[1]; }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "r")       // 検証群データ数の指定
  {
    char* ptr = nullptr;
    if (opt[1].empty()) vr = 1;
    else vr = int(strtol(opt[1].c_str(), &ptr, 10));
    if (vr < 0) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

class ldacvclass                // ldacvメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int search_g(const option& opt, const string& gcode) const;
  // 群コードを検索する
  int cal_lda(mat<double>& dmat, lditr& lditrlstt);
  // 判別分析の計算をして総例数を返す
  int cv_lda(int& nd, int& tp, int& nn, int& fn, mat<double>& dmat, lditr& lditrlstv);
  // 判別分析結果でデータリストを検証して検証例数を返す
  int dis_score(double& z, datac& cdata, mat<double>& dmat);
  // 判別スコアを計算し、群インデックスを返す
  void set_cvdlst(ldvec& cvdveclst, int vno, int nd, int tp, int nn, int fn, mat<double>& dmat);
  // 検証結果を検証結果リストに追加する
  int cal_cv(ldvec& cvdveclst, int r, int k); // 判別分析の交差検証を行う
  void put_cv(ostream& fout, ldvec& cvdveclst);
  // 判別分析の交差検証結果をデータファイル形式で出力する
public:
  ldacvclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~ldacvclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int ldacvclass::search_g(const option& opt, const string& gcode) const
// 群コードを検索する
{
  if (gcode == BLANKs) return -1; // 欠損データはヒットさせない
  int idx = -1;
  if ((opt.g(0)).empty())       // 群1のコード指定がない場合
  { idx = (gcode == opt.g(1)) ? 1 : 0;  }
  else if (gcode == opt.g(0))   // 群1のコード
  { idx = 0; }
  else if ((opt.g(1)).empty())  // 群2のコード指定がない場合
  { idx = 1; }
  else if (gcode == opt.g(1))   // 群2のコード
  { idx = 1; }
  return idx;                   // -1:ヒットせず,0:群1,1:群2
}

int ldacvclass::cal_lda(mat<double>& dmat, lditr& lditrlstt)
// 判別分析の計算をして総例数を返す
{
  size_t ni = id.ni();
  size_t d = ni + 1;
  vector<mat<double>> smatvec(2); // 積和行列ベクトル
  smatvec[0].resize(d, d);  smatvec[1].resize(d, d);
  smatvec[0] = 0.0;  smatvec[1] = 0.0;
  vecint nvec(2, 0);            // 例数ベクトル
  vecdbl fdatavec(ni, 0.0);
  fdatavec[0] = 1.0;
  for (lditr_itr lditr = lditrlstt.begin(); lditr != lditrlstt.end(); ++lditr)
  {
    int idx = (int)((*lditr)->fdata(0));
    for (size_t i = 1; i < ni; ++i) fdatavec[i] = (*lditr)->fdata(i);
    for (size_t i = 0; i < ni; ++i)
    {
      double x = fdatavec[i];
      for (size_t j = i; j < id.ni(); ++j) smatvec[idx].add(i, j, x * fdatavec[j]);
    }
  }
  nvec[0] = (int)(smatvec[0](0, 0));  nvec[1] = (int)(smatvec[1](0, 0));
  int nt = nvec[0] + nvec[1];
  if (nt < 1) return 0;         // データなし
  for (size_t k = 0; k < 2; ++k)
  {
    if (nvec[k] > 0)
    {
      for (size_t i = 0; i < d - 1; ++i)
      {
        for (size_t j = i + 1; j < d; ++j) smatvec[k].assign(j, i, smatvec[k](i, j));
      }
      smatvec[k].sweep(0);
    }
  }
  if (nvec[0] < 1 || nvec[1] < 1 || nt < (int)ni) return 0; // 判別関数計算できず
  dmat = smatvec[0] + smatvec[1];
  int df = nt - 2;
  dmat *= 1.0 / (double)df;
  dmat.assign(0, 0, 0.0);  dmat.assign(0, ni, 0.0);
  dmat.assign(ni, 0, 0.0);  dmat.assign(ni, ni, 0.0);
  for (size_t i = 1; i < ni; ++i)
  {
    dmat.assign(0, i, (smatvec[0](0, i) + smatvec[1](0, i)) * 0.5);
    dmat.assign(i, 0, -dmat(i, 0));
    dmat.assign(i, ni, smatvec[0](0, i) - smatvec[1](0, i));
    dmat.assign(ni, i, dmat(i, ni));
  }
  for (size_t i = 1; i < ni; ++i) dmat.sweep(i); // 行列を掃き出す
  return nt;                    // 総例数を返す
}

int ldacvclass::cv_lda(int& nd, int& tp, int& nn, int& fn, mat<double>& dmat, lditr& lditrlstv)
// 判別分析結果でデータリストを検証して検証例数を返す
{
  nd = tp = nn = fn = 0;
  for (lditr_itr lditr = lditrlstv.begin(); lditr != lditrlstv.end(); ++lditr)
  {
    double z = 0.0;
    int gidx = dis_score(z, *(*lditr), dmat);
    if (gidx == 0)              // 疾患群
    {
      ++nd;
      if (z >= 0.0) ++tp;
    }
    else if (gidx == 1)         // 正常群
    {
      ++nn;
      if (z < 0.0) ++fn;
    }
  }
  return nd + nn;               // 合計例数を返す
}

int ldacvclass::dis_score(double& z, datac& cdata, mat<double>& dmat)
// 判別スコアを計算し、群インデックスを返す
{
  size_t didx = dmat.r() - 1;
  z = dmat(0, didx);
  for (size_t i = 1; i < didx; ++i) z += dmat(i, didx) * cdata.fdata(i);
  return (int)(cdata.fdata(0)); // 群インデックスを返す
}

void ldacvclass::set_cvdlst(ldvec& cvdveclst, int vno, int nd, int tp, int nn, int fn, mat<double>& dmat)
// 検証結果を検証結果リストに追加する
{
  size_t ni = id.ni();
  vecdbl cvdvec(ni + 8, 0.0);
  cvdvec[0] = (double)((vno < 1) ? ni + 1 : vno);
  for (size_t i = 0; i < ni; ++i) cvdvec[1 + i] = dmat(i, ni);
  cvdvec[ni + 1] = -dmat(ni, ni);
  cvdvec[ni + 2] = (double)nd;
  cvdvec[ni + 3] = (double)tp;
  cvdvec[ni + 4] = (double)nn;
  cvdvec[ni + 5] = (double)fn;
  cvdvec[ni + 6] = (double)(nd + nn);
  cvdvec[ni + 7] = (double)(tp + fn);
  cvdveclst.push_back(cvdvec);
}

int ldacvclass::cal_cv(ldvec& cvdveclst, int r, int k)
// 判別分析の交差検証を行う
{
  size_t ni = id.ni();
  size_t d = ni + 1;
  lditr lditrlstt, lditrlstv; // トレーニングデータとバリデーションデータ反復子
  for (int i = 0; i < k; ++i)
  {
    int sidx = r * i, eidx = r * (i + 1) - 1;
    int j = 0;
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr, ++j)
    {
      if (sidx <= j && j <= eidx) lditrlstv.push_back(ditr);
      else lditrlstt.push_back(ditr);
    }
    mat<double> dmat(d, d);
    if (cal_lda(dmat, lditrlstt) > 0)
    {
      int nd = 0, tp = 0, nn = 0, fn = 0;
      cv_lda(nd, tp, nn, fn, dmat, lditrlstv);
      set_cvdlst(cvdveclst, i + 1, nd, tp, nn, fn, dmat);
    }
    lditrlstt.clear();  lditrlstv.clear();
  }
  return k;                     // 検証回数を返す
}

void ldacvclass::put_cv(ostream& fout, ldvec& cvdveclst)
// 判別分析の交差検証結果をデータファイル形式で出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  size_t ni = id.ni();
  idd tid(id.title(), ni + 8);
  tid.assign_item(0, "Number of Parameters/Validation Group Number");
  tid.assign_item(1, "a0:Intercept");
  for (size_t i = 1; i < ni; ++i)
  {
    ostringstream ostr;
    ostr << "a" << i << ":" << id.item(i);
    tid.assign_item(1 + i, ostr.str());
  }
  tid.assign_item(ni + 1, "Mahalanobis's Square Distance:D-squared");
  tid.assign_item(ni + 2, "Number of Group1(Disease Group)");
  tid.assign_item(ni + 3, "Number of TP of Group1(Disease Group)");
  tid.assign_item(ni + 4, "Number of Group2(Normal Group)");
  tid.assign_item(ni + 5, "Number of FN of Group2(Normal Group)");
  tid.assign_item(ni + 6, "Total Number of 2 Groups");
  tid.assign_item(ni + 7, "Total Number of AC of 2 Groups");
  tid.put_item(fout, formtag);
  int prec0 = fout.precision(); // 現在の浮動小数点精度(有効数字)を保存する
  for (ldvec_itr cvitr = cvdveclst.begin(); cvitr != cvdveclst.end(); ++cvitr)
  {
    fout << (*cvitr)[0];
    fout.precision(16);
    for (size_t i = 1; i <= ni + 1; ++i) fout << CHRdelm << (*cvitr)[i];
    fout.precision(prec0);
    for (size_t i = ni + 2; i < tid.ni(); ++i) fout << CHRdelm << (*cvitr)[i];
    fout << endl;
  }
}

int ldacvclass::get(const char* c_fname, option& opt) // データを読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.g(0).empty() && opt.g(1).empty()) throw runtime_error("The group code is not specified !");
  if (id.ni() < 2) throw runtime_error("not enough items !");
  nc = (int)(id.nc());          // ケース数変更用
  ldata_itr ditr = datalst.begin();
  while (ditr != datalst.end())
  {
    bool blank = false;
    for (size_t j = 1; blank == false && j < id.ni(); ++j)
    {
      if (ditr->fdata(j) == BLANKf) blank = true;
    }
    if (ditr->sdata(0) == BLANKs || blank) // 欠損のためケースを削除する
    { ditr = datalst.erase(ditr);  --nc; } // 削除した次の要素の位置を返す
    else
    {
      int idx = search_g(opt, ditr->sdata(0));
      if (idx >= 0)             // 群コードをインデックスに変える
      {
        ditr->assign(0, (double)idx);
        ++ditr;
      }
      else                      // ケースを削除する
      { ditr = datalst.erase(ditr);  --nc; }
    }
  }
  if (nc < (int)(id.nc())) id.nc(nc);  // ケース数を減らす
  return nc;                    // 有効なケース数を返す
}

void ldacvclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1) return;      // 有効なデータなし
  int ni = id.ni();
  size_t dim = id.ni() + 1;
  mat<double> dmat(dim);        // 群合計積和行列
  lditr lditrlst;               // データリスト反復子リスト
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr) lditrlst.push_back(ditr);
  int nt = cal_lda(dmat, lditrlst);
  if (nt < 0) return;           // 判別分析できず
  int r = opt.r();              // 検証群のデータ数
  if (r < 1 || nt - r < ni) return; // 交差検証実行せず
  int k = nt / r;               // 検証群数
  if (nt % r > 0) ++k;          // 切り上げる
  int nd0 = 0, tp0 = 0;         // 全体の疾患群の例数と正判別数(TP)
  int nn0 = 0, fn0 = 0;         // 全体の正常群の例数と正判別数(FN)
  ldvec cvdveclst;              // 交差検証法の結果リスト
  cv_lda(nd0, tp0, nn0, fn0, dmat, lditrlst);
  lditrlst.clear();
  set_cvdlst(cvdveclst, 0, nd0, tp0, nn0, fn0, dmat);
  if (cal_cv(cvdveclst, r, k) > 0) put_cv(cout, cvdveclst);
}

int main(int argc, char* argv[]) // メインルーチン
{
  ldacvclass ldacv;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) ldacv.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { ldacv.get(argv[i], opt);  fok = false; }
    }
    ldacv.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
