// lmae.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-29 16:12:14 tac>
//
// 最後の項目を目的変数Y、それ以前の項目を説明変数Xとした
// 線形モデルによる分散分析・共分散分析
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W lmae lmae.C
//              -I/usr/lcoal/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "lmae  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:lma [<OPTION> ...]  [<DATA FILE> ...]",
  "  The ANOVA/ANCOVA by linear models with the first item as criterion",
  "  variable Y and subsequent items as explanatory variables X.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ino=<START NUMBER>[-<END NUMBER>][,...]",
  "    specify the start and end number of explanatory variables of",
  "    interaction items",
  "  -ix=<START NUMBER>[-<END NUMBER>]x<START NUMBER>[-<END NUMBER>][,...]",
  "    specify the start and end number of 2 explanatory variables for",
  "    creating interaction items",
  "  -mf=<NUMBER>:specify the number of explanatory variables to be treated",
  "    as main effects",
  "  -mno=<START NUMBER>[-<END NUMBER>][,...]",
  "    specify the start and end number of explanatory variables to be treated",
  "    as main effects",
  "  -ngp={est|line|pl}[,...]:select the data for Ngraph",
  "    est:estimate of Y/correction of Y  line:partial regression line",
  "    pl:prediction limit  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=lmae_ngp.dat",
  "  -res=[-][{def|all|atab|cc[<C-COEF>]|reg|stat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    atab:ANOVA table  cc[<C-COEF>]:confidence intervals(0<=C-COEF<1)",
  "    reg:multiple regression analysis  stat:basic statistics",
  "  -x<NUMBER>=<VALUE>:specify the number and value of explanatory variables",
  "    whoes value is constant",
  "  -xno=<NUMBER>:specify the number of the explanatory variable whose value",
  "    changes"
};

const string Title = "=== ANOVA/ANCOVA by Linear Models ==="; // 表題
const string NGPNAMEdef = "lmae_ngp.dat"; // 省略時Ngraphファイル名

using lino = list<pairint>;     // 説明変数の番号リストの短名定義
using lino_itr = list<pairint>::iterator; // 反復子の短名定義
using lino_citr = list<pairint>::const_iterator; // 反復子の短名定義

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  lino vinolst;                 // 交互作用の説明変数の番号リスト
  lstr vixstrlst;               // 交互作用項目の説明変数リスト
  lino vixlst;                  // 交互作用項目の説明変数ペアリスト
  int vmf;                      // 主効果の説明変数の個数
  lino vmnolst;                 // 主効果の説明変数の番号リスト
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 解析結果出力リスト
  map<string, bool> vresmap;    // 解析結果出力の有無マップ
  lstr xlst;                    // 説明変数リスト
  vecdbl vxvec;                 // 説明変数ベクトル
  int vxno;                     // 変化説明変数の番号
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  lino& inolst()               // 交互作用の説明変数の番号リストを返す
  { return vinolst; }
  lino& ixlst()              // 交互作用項目の説明変数番号ペアリストを返す
  { return vixlst; }
  int mf() const                // 主効果の説明変数の個数を返す
  { return vmf; }
  lino& mnolst()                // 主効果の説明変数の番号リストを返す
  { return vmnolst; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double x(size_t idx) const    // 説明変数の値を返す
  { return (0 < idx && idx < vxvec.size()) ? vxvec[idx] : BLANKf; }
  int xno() const               // 変化説明変数の番号を返す
  { return vxno; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
  int set_ix(int ni);           // -ixオプションのセット
  int set_mx(int ni);           // -mf,-mnoオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vinolst(), vixstrlst(), vixlst(), vmf(0), vmnolst(), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), xlst(), vxvec(), vxno(0)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ino")     // 交互作用の説明変数の番号指定
  {
    if (getrangelst(vinolst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ix")      // 交互作用項目の説明変数ペア指定
  {
    if (strsplit(vixstrlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "mf")      // 主効果の説明変数の個数指定
  {
    char* ptr = nullptr;
    vmf = strtol(opt[1].c_str(), &ptr, 0);
    if (vmf < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0] == "mno")     // 主効果の説明変数の番号指定
  {
    if (getrangelst(vmnolst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "xno")     // 変化説明変数の番号指定
  {
    char* ptr = nullptr;
    vxno = strtol(opt[1].c_str(), &ptr, 0);
    if (vxno < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0][0] == 'x')    // 説明変数の値指定
  {
    if (strsplit(xlst, optstr, CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 2) return disperr(vpname, "At lease 2 items are required !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3;   // -resオプションのセット
  if (set_ix(ni) < 0) return -4;  // -ixオプションのセット
  if (set_mx(ni) < 0) return -5;  // -mf,-mnoオプションのセット
  if (vxno == 0) return 0;
  if ((int)ni <= vxno) return disperr(vpname, "Wrong number of explanatory variable whose value changes !");
  vxvec.resize(ni);
  vxvec.assign(ni, BLANKf);
  if (xlst.empty()) return 0;
  if (xlst.size() >= ni) return disperr(vpname, "Too many value of the explanatory variables !");
  for (string xstr : xlst)
  {
    string opt[2];
    if (getopt(opt, xstr, OPTdev) > 0)
    {
      char* ptr = nullptr;
      int xno = strtol(opt[0].substr(1).c_str(), &ptr, 0);
      if (xno < 1 || (int)ni <= xno || xno == vxno) return disperr(vpname, "Wrong number of explanatory variable->", xstr);
      double fdata = strtodf(opt[1]);
      if (fdata == BLANKf) return disperr(vpname, "Wrong value of the explanatory variable->", xstr);
      vxvec[xno] = fdata;
    }
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["est"] = vngpmap["line"] = vngpmap["pl"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["est"] = vngpmap["line"] = vngpmap["pl"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["atab"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_ix(int ni)      // -ixオプションのセット
{
  int ixno = ni + 1;            // 交互作用項目の最初の番号
  for (string ixstr : vixstrlst)
  {
    string opt[2];
    if (getopt(opt, ixstr, 'x') > 0)
    {
      int low0 = 0, up0 = 0, low1 = 0, up1 = 0;
      if (getrange(low0, up0, opt[0]) > 0)
      {
        if (getrange(low1, up1, opt[1]) > 0)
        {
          if (low0 < 1 || low0 > up0) return disperr(vpname, "Wrong number of the first explanatory bariable->", ixstr);
          if (low1 < 1 || low1 > up1) return disperr(vpname, "Wrong number of the second explanatory bariable->", ixstr);
          if (ixno <= up0 || ixno <= up1) return disperr(vpname, "Wrong number of explanatory variable->", ixstr);
          for (int j = low1; j <= up1; ++j) // 2番目の番号
          {
            for (int i = low0; i <= up0; ++i) // 1番目の番号
            {
              pairint ixpi {i, j};
              vixlst.push_back(ixpi);
            }
          }
          ixno += (up0 - low0 + 1) * (up1 - low1 + 1);
        }
      }
    }
  }
  return 0;                     // エラーなし
}

int option::set_mx(int ni)      // -mf,-mnoオプションのセット
{
  if (vmf < 1) vmf = ni - 1;  // 全て主効果にする
  else if (ni <= vmf) return disperr(vpname, "Wrong number of explanatry variables to be treated as main effects !");
  double nt = 0;
  for (pairint mnopi : vmnolst)
  {
    if (mnopi.first > mnopi.second || mnopi.second < 1 || vmf < mnopi.first) return disperr(vpname, "Wrong number of the explanatry variable to be treated as main effects !");
    nt += mnopi.second - mnopi.first + 1;
  }
  if (vmf < nt) return disperr(vpname, "Wrong total number of explanatory variables to be treated as main effect !");
  if (vinolst.empty()) return 0;
  int nit = ni + vixlst.size();  // 交互作用を含めた項目数
  for (pairint inopi : vinolst)
  {
    if (inopi.first > inopi.second || inopi.second < 1 || nit <= inopi.second) return disperr(vpname, "Wrong number of the explanatory variable for creating interaction item !");
    for (pairint mnopi : vmnolst)
    {
      if (mnopi.first <= inopi.first && inopi.first <= mnopi.second) return disperr(vpname, "Duplicate numbers of explanatory variables for interactions and main effects !");
      if (mnopi.first <= inopi.second && inopi.second <= mnopi.second) return disperr(vpname, "Duplicate numbers of explanatory variables for interactions and main effects !");
    }
  }
  return 0;                     // エラーなし
}

class lmaclass                  // lmaメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_smat(mat<double>& smat, mat<double>& xmat, vecdbl& mevec, vecdbl& vrvec, vecdbl& ssvec, option& opt);
  // 積和行列と各種統計量を求める
  void cal_lma(mat<double>& smat, vecdbl& fssvec, vecint& fdfvec, option& opt);
  // 線形モデルによる分析
  void put_resid(ostream& fout, option& opt);
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_lma(ostream& fout, option& opt, const vecdbl& ssvec, const vecdbl& fssvec, const vecint& fdfvec, int n, const mat<double>& smat, const mat<double>& xmat);
  // 線形モデルによる分析結果を出力する
  void put_lmatab(ostream& fout, option& opt, int n, const mat<double>& smat, const vecdbl& ssvec, const vecdbl& fssvec, const vecint& fdfvec);
  // 線形モデルによる分析分析表を出力する
  void cal_vif(mat<double>& rmat, const mat<double>& xmat) const;
  // VIFを計算する
  void put_bci(ostream& fout, const option& opt, int n, const mat<double>& smat);
  // 偏回帰係数の信頼区間を出力する
  void plot_ngp(option& opt, int n, const mat<double>& smat, const vecdbl& mevec);
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, const option& opt, const vecdbl& mevec);
  // Ngraph用にデータ識別情報を出力する
  void plot_est(ofstream& fout, option& opt, int n, const mat<double>& smat, const vecdbl& mevec);
  // Yの推定値をNgraph用に出力する
  double cal_estimate(option& opt, datac& cdata, const mat<double>& smat);
  // Yの推定値を計算して返す
  void plot_line(ofstream& fout, option& opt, int n, const mat<double>& smat, const vecdbl& mevec);
  // 回帰直線をNgraph用に出力する
  double search_xrange(double& xmin, double& xmax, option& opt, const mat<double>& smat);
   // x軸の最小値と最大値を検索して範囲を返す
  void plot_pl(ofstream& fout, option& opt, int n, const mat<double>& smat, const vecdbl& mevec);
  // 予測限界をNgraph用に出力する
  double cal_pi(option& opt, datac& cdata, double t, double vr, const mat<double>& smat);
  // 回帰直線の予測限界幅を計算して返す
public:
  lmaclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~lmaclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int lmaclass::cal_smat(mat<double>& smat, mat<double>& xmat, vecdbl& mevec, vecdbl& vrvec, vecdbl& ssvec, option& opt)
// 積和行列と各種統計量を求める
{
  size_t dim = id.ni() + (opt.ixlst()).size() + 1;
  size_t didx = dim - 1;
  vecdbl fdatavec(dim, 0.0);
  fdatavec[0] = 1.0;
  for (auto& cdata : datalst)
  {
    bool blank = false;
    size_t j = 1;
    while (blank == false && j < id.ni())
    {
      fdatavec[j] = cdata.fdata(j);
      if (fdatavec[j] == BLANKf) blank = true;
      else ++j;
    }
    if (blank || cdata.fdata(0) == BLANKf)
    { cdata.assign(0, BLANKs); } // 後の計算のためにYをブランクにする
    else
    {
      j = id.ni();
      for (lino_citr ixitr = (opt.ixlst()).begin(); ixitr != (opt.ixlst()).end(); ++ixitr, ++j) // 交互作用データのセット
      { fdatavec[j] = fdatavec[ixitr->first] * fdatavec[ixitr->second]; }
      fdatavec[didx] = cdata.fdata(0);
      for (size_t i = 0; i <= didx; ++i)
      {
        double x = fdatavec[i];
        for (j = i; j <= didx; ++j) smat.add(i, j, x * fdatavec[j]);
      }
    }
  }
  int n = (int)(smat(0, 0));    // 例数
  int dft = n - 1;
  for (size_t i = 0; i < didx; ++i)
  {
    xmat.assign(i, i, smat(i, i));
    for (size_t j = i + 1; j <= didx; ++j)
    {
      smat.assign(j, i, smat(i, j));
      if (j < didx)
      { xmat.assign(i, j, smat(i, j));  xmat.assign(j, i, smat(i, j)); }
    }
  }
  smat.sweep(0);
  for (size_t i = 0; i <= didx; ++i)
  {
    mevec[i] = smat(0, i);
    vrvec[i] = (n > 1) ? smat(i, i) / (double)dft : 0.0;
    ssvec[i] = smat(i, i);
  }
  return n;                     // 例数を返す
}

void lmaclass::cal_lma(mat<double>& smat, vecdbl& fssvec, vecint& fdfvec, option& opt)
// 線形モデルによる分析
{
  size_t didx = smat.r() - 1;
  size_t iend = opt.mf() + 1;
  for (size_t i = 1; i < iend; ++i) smat.sweep(i);
  int idx = 0;
  if ((opt.mnolst()).empty() == false)
  {
    mat<double> tsmat(smat);
    double sr = tsmat(didx, didx);
    for (lino_citr mnoitr = (opt.mnolst()).begin(); mnoitr != (opt.mnolst()).end(); ++mnoitr, ++idx)
    {
      for (int i = mnoitr->first; i <= mnoitr->second; ++i) tsmat.sweep(i);
      fssvec[idx] = tsmat(didx, didx) - sr;
      fdfvec[idx] = mnoitr->second - mnoitr->first + 1;
      for (int i = mnoitr->first; i <= mnoitr->second; ++i) tsmat.sweep(i);
    }
  }
  for (size_t i = iend; i < didx; ++i) smat.sweep(i);
  if ((opt.inolst()).empty() == false)
  {
    mat<double> tsmat(smat);
    double sr = tsmat(didx, didx);
    for (lino_citr inoitr = (opt.inolst()).begin(); inoitr != (opt.inolst()).end(); ++inoitr, ++idx)
    {
      for (int i = inoitr->first; i <= inoitr->second; ++i) tsmat.sweep(i);
      fssvec[idx] = tsmat(didx, didx) - sr;
      fdfvec[idx] = inoitr->second - inoitr->first + 1;
      for (int i = inoitr->first; i <= inoitr->second; ++i) tsmat.sweep(i);
    }
  }
}

void lmaclass::put_resid(ostream& fout, option& opt)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i;
    fout << ":" << id.item(i) << "\n";
  }
  if ((opt.ixlst()).empty()) return;
  idd tid(id);
  tid.set(id.title(), id.ni() + (opt.ixlst()).size()); // 項目数を変更する
  ostringstream ostr;
  size_t j = id.ni();
  for (lino_citr ixitr = (opt.ixlst()).begin(); ixitr != (opt.ixlst()).end(); ++ixitr, ++j)
  {
    ostr.str("");
    ostr << tid.item(ixitr->first) << " x " << tid.item(ixitr->second);
    tid.assign_item(j, ostr.str());
  }
  for (j = id.ni(); j < tid.ni(); ++j)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << j;
    fout << ":" << tid.item(j) << "\n";
  }
}

void lmaclass::put_lma(ostream& fout, option& opt, const vecdbl& ssvec, const vecdbl& fssvec, const vecint& fdfvec, int n, const mat<double>& smat, const mat<double>& xmat)
// 線形モデルによる分析結果を出力する
{
  constexpr int LENline = 79;   // 区切り線の長さ
  size_t didx = smat.r() - 1;
  mat<double> rmat(didx, didx);
  cal_vif(rmat, xmat);
  fout << "\n<MULTIPLE REGRESSION ANALYSIS>\n";
  fout << "VAR    Partial b  Standard b   Partial r         VIF     F-value     p-value\n";
  putline(fout, LENline);
  double syy = ssvec[didx], sr = smat(didx, didx);
  int dfr = n - didx;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double sii = smat(0, 0), bi = smat(0, didx);
  double vbi = vr * sii;
  double fi = (vbi > 0.0) ? bi * bi / vbi : 0.0;
  ldbl pr = ftop(fi, 1, dfr);
  fout << "ICPT" << setw(12) << bi << setw(36) << "";
  fout << setw(12) << fi << setw(12) << pr << pind(pr) << "\n";
  for (size_t i = 1; i < didx; ++i)
  {
    bi = smat(i, didx);  sii = smat(i, i);
    vbi = vr * sii;
    double bsti = (syy > 0.0) ? bi * sqrt(ssvec[i] / syy) : 0.0;
    double v = sr * sii + bi * bi;
    double r = (v > 0.0) ? bi / sqrt(v) : 0.0;
    fi = (vbi > 0.0) ? bi * bi / vbi : 0.0;
    pr = ftop(fi, 1, dfr);
    fout << "X" << setw(3) << i << setw(12) << bi << setw(12) << bsti;
    fout << setw(12) << r << setw(12) << rmat(i, i) << setw(12) << fi;
    fout << setw(12) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
  if (opt.res("cc")) put_bci(fout, opt, n, smat);
  if (syy <= 0.0) return;       // 重相関係数等計算不要
  double r2 = (syy > 0.0) ? (syy - sr) / syy : 0.0;
  double r2a = (syy > 0.0) ? fabs(1.0 - vr * (double)(n - 1) / syy) : 0.0;
  double aic = (double)n * log(sr / (double)n) + (double)(2 * (didx + 1));
  fout << "Multiple Contribution Rate:R^2=" << r2;
  fout << "  DF-adjusted:R'^2=" << r2a << "\n";
  fout << "Multiple Correlation Coefficient:R=" << sqrt(r2) << "  DF-adjusted:R'=" << sqrt(r2a) << "\n";
  fout << "AIC=" << aic << endl;
}

void lmaclass::cal_vif(mat<double>& rmat, const mat<double>& xmat) const
// VIFを計算する
{
  int d = rmat.r();
  rmat.assign(0, 0, xmat(0, 0));
  for (int j = 1; j < d; ++j)
  { rmat.assign(0, j, xmat(0, j));  rmat.assign(j, 0, xmat(0, j)); }
  for (int i = 1; i < d; ++i)
  {
    rmat.assign(i, i, xmat(i, i));
    for (int j = 1; j < d; ++j)
    { rmat.assign(i, j, xmat(i, j));  rmat.assign(j, i, xmat(i, j)); }
  }
  rmat.sweep(0);                // 積和行列にする
  smat2rmat(rmat);              // 相関行列にする
  mat<double> tmat(rmat.sub(1, 1, d - 1, d - 1));
  tmat.inv();                   // 相関行列部分を逆行列にする
  for (int i = 0; i < d - 1; ++i)
  { for (int j = 0; j < d - 1; ++j) rmat.assign(1 + i, 1 + j, tmat(i, j)); }
}

void lmaclass::put_bci(ostream& fout, const option& opt, int n, const mat<double>& smat)
// 偏回帰係数の信頼区間を出力する
{
  constexpr int LENline = 74;   // 区切り線の長さ
  fout << "\nVAR      Partial b       SE of b" << setw(4) << opt.cc() * 100.0;
  fout << "%Precision  CI:Lower-Lim     Upper-Lim\n";
  putline(fout, LENline);
  size_t didx = smat.r() - 1;
  double sr = smat(didx, didx);
  int dfr = n - didx;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double bi = smat(0, didx), sii = smat(0, 0);
  double vbi = vr * sii;
  double ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vbi);
  fout << "ICPT" << setw(14) << bi << setw(14) << sqrt(vbi);
  fout << setw(14) << ci << setw(14) << bi - ci;
  fout << setw(14) << bi + ci << "\n";
  for (size_t i = 1; i < didx; ++i)
  {
    bi = smat(i, didx);  sii = smat(i, i);
    vbi = vr * sii;
    ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vbi);
    fout << "X" << setw(3) << i << setw(14) << bi;
    fout << setw(14) << sqrt(vbi) << setw(14) << ci;
    fout << setw(14) << bi - ci << setw(14) << bi + ci << "\n";
  }
  putline(fout, LENline);
}

void lmaclass::put_lmatab(ostream& fout, option& opt, int n, const mat<double>& smat, const vecdbl& ssvec, const vecdbl& fssvec, const vecint& fdfvec)
// 線形モデルによる分析分析表を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  size_t didx = smat.r() - 1;
  int dft = n - 1, dfp = didx - 1, dfr = n - didx;
  double syy = ssvec[didx], sr = smat(didx, didx);
  double sp = syy - sr;
  string tstr("ANOVA Table");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  if ((opt.mnolst()).empty() && (opt.inolst()).empty()) // 説明変数の番号指定なし
  {
    calvfp(v, f, pr, sp, dfp, sr, dfr);
    tstr = "REGRESSION";
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
    fout << setw(10) << dfp << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
  }
  else
  {
    int idx = 0;
    for (lino_citr mnoitr = (opt.mnolst()).begin(); mnoitr != (opt.mnolst()).end(); ++mnoitr, ++idx)
    {
      double ss = fssvec[idx];
      int df = fdfvec[idx];
      calvfp(v, f, pr, ss, df, sr, dfr);
      ostringstream ostr;
      ostr << "M:X" << mnoitr->first << "-" << mnoitr->second;
      fout << left << setw(10) << ostr.str() << right << setw(14) << ss;
      fout << setw(10) << df << setw(14) << v << setw(12) << f;
      fout << setw(14) << pr << pind(pr) << "\n";
      sp -= ss;  dfp -= df;
    }
    for (lino_citr inoitr = (opt.inolst()).begin(); inoitr != (opt.inolst()).end(); ++inoitr, ++idx)
    {
      double ss = fssvec[idx];
      int df = fdfvec[idx];
      calvfp(v, f, pr, ss, df, sr, dfr);
      ostringstream ostr;
      ostr << "I:X" << inoitr->first << "-" << inoitr->second;
      fout << left << setw(10) << ostr.str() << right << setw(14) << ss;
      fout << setw(10) << df << setw(14) << v << setw(12) << f;
      fout << setw(14) << pr << pind(pr) << "\n";
      sp -= ss;  dfp -= df;
    }
    if (dfp > 0)
    {
      calvfp(v, f, pr, sp, dfp, sr, dfr);
      tstr = "ETC";
      int npad = tstr.size() - strclen(tstr, CHARwide);
      fout << left << setw(10 + npad) << tstr << right << setw(14) << sp;
      fout << setw(10) << dfp << setw(14) << v << setw(12) << f;
      fout << setw(14) << pr << pind(pr) << "\n";
    }
  }
  tstr = "RESIDUALS";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << sr;
  fout << setw(10) << dfr << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << syy;
  fout << setw(10) << dft << endl;
}

void lmaclass::plot_ngp(option& opt, int n, const mat<double>& smat, const vecdbl& mevec)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("line"))          // 回帰直線
  {
    fout << "-s1 -dline\n";
    plot_ngpid(fout, opt, mevec);
    plot_line(fout, opt, n, smat, mevec);
    delml = true;
  }
  if (opt.ngp("pl"))            // 回帰直線の予測限界
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout, opt, mevec);
    plot_pl(fout, opt, n, smat, mevec);
  }
  if (opt.ngp("est"))            // Yの推測値/補正値
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dmark,00 -o300\n";
    plot_ngpid(fout, opt, mevec);
    plot_est(fout, opt, n, smat, mevec);
  }
}

void lmaclass::plot_ngpid(ofstream& fout, const option& opt, const vecdbl& mevec)
// Ngraph用にデータ識別情報を出力する
{
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%CRITERION VALIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "%EXPLANATORY VARIABLE X" << setw(3) << i << ":" << id.item(i);
    if (opt.xno() > 0)
    {
      if (i == (size_t)(opt.xno())) fout << " Variable Whoes Value Changes";
      else if (opt.x(i) == BLANKf) fout << " Mean=" << mevec[i];
      else fout << " Specified Value=" << opt.x(i);
    }
    fout << "\n";
  }
}

void lmaclass::plot_est(ofstream& fout, option& opt, int n, const mat<double>& smat, const vecdbl& mevec)
// Yの推定値／補正値をNgraph用に出力する
{
  if (opt.xno() < 1)            // Y推定値と実際のYを出力
  {
    fout << "%<ESTIMATE OF Y AND Y>\n";
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
    {
      double yest = cal_estimate(opt, *ditr, smat);
      if (yest != BLANKf) fout << yest << " " << ditr->fdata(0) << "\n";
    }
  }
  else                          // 変化説明変数XとY補正値を出力
  {
    datac cdata(id.ni());
    for (size_t j = 0; j < id.ni(); ++j) cdata.assign(j, (opt.x(j) == BLANKf) ? mevec[j] : opt.x(j));
    fout << "%<VARIABLE WHOES VALUE CHANGES AND CORRECTED Y>\n";
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
    {
      double x = ditr->fdata(opt.xno()), y = ditr->fdata(0);
      double yest = cal_estimate(opt, *ditr, smat);
      if (yest != BLANKf)
      {
        double err = y - yest;
        cdata.assign(opt.xno(), x);
        yest = cal_estimate(opt, cdata, smat);
        fout << x << " " << yest + err << "\n";
      }
    }
  }
}

double lmaclass::cal_estimate(option& opt, datac& cdata, const mat<double>& smat)
// Yの推定値を計算して返す
{
  if (cdata.sdata(0) == BLANKs) return BLANKf; // 欠損があるため計算できず
  size_t didx = smat.r() - 1;
  vecdbl fdatavec(didx, 0.0);
  for (size_t j = 0; j < id.ni(); ++j) fdatavec[j] = cdata.fdata(j);
  size_t j = id.ni();
  for (lino_citr ixitr = (opt.ixlst()).begin(); ixitr != (opt.ixlst()).end(); ++ixitr, ++j) // 交互作用データのセット
  { fdatavec[j] = fdatavec[ixitr->first] * fdatavec[ixitr->second]; }
  double yest = smat(0, didx);
  for (size_t i = 1; i < didx; ++i) yest += smat(i, didx) * fdatavec[i];
  return yest;                  // 推定値を返す
}

void lmaclass::plot_line(ofstream& fout, option& opt, int n, const mat<double>& smat, const vecdbl& mevec)
// 回帰直線をNgraph用に出力する
{
  double xmin = DBL_MAX, xmax = -DBL_MAX;
  double xrange = search_xrange(xmin, xmax, opt, smat);
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  datac cdata(id.ni());
  cdata.assign(0, mevec[0]);
  for (size_t j = 1; j < id.ni(); ++j) cdata.assign(j, (opt.x(j) == BLANKf) ? mevec[j] : opt.x(j));
  double yest = xstart;
  if (opt.xno() > 0)
  {
    cdata.assign(opt.xno(), xstart);
    yest = cal_estimate(opt, cdata, smat);
  }
  fout << "%<REGRESSION LINE>\n";
  fout << xstart << " " << yest << "\n";
  yest = xend;
  if (opt.xno() > 0)
  {
    cdata.assign(opt.xno(), xend);
    yest = cal_estimate(opt, cdata, smat);
  }
  fout << xend << " " << yest << "\n";
}

double lmaclass::search_xrange(double& xmin, double& xmax, option& opt, const mat<double>& smat)
// x軸の最小値と最大値を検索して範囲を返す
{
  xmin = DBL_MAX;  xmax = -DBL_MAX;
  for (auto& cdata : datalst)
  {
    double x = (opt.xno() > 0) ? cdata.fdata(opt.xno()) : cal_estimate(opt, cdata, smat);
    if (x != BLANKf)            // xの最大値と最小値をサーチ
    {
      if (x < xmin) xmin = x;
      if (x > xmax) xmax = x;
    }
  }
  return (xmax - xmin) * 2.0;   // 範囲を返す
}

void lmaclass::plot_pl(ofstream& fout, option& opt, int n, const mat<double>& smat, const vecdbl& mevec)
// 予測限界をNgraph用に出力する
{
  double xmin = DBL_MAX, xmax = -DBL_MAX;
  double xrange = search_xrange(xmin, xmax, opt, smat);
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  constexpr uint RDdef = 50;    // プロットの繰り返し数
  double dx = (xend - xstart) / (double)RDdef;
  size_t didx = smat.r() - 1;
  datac cdata(id.ni());
  cdata.assign(0, mevec[0]);
  for (size_t j = 1; j < id.ni(); ++j) cdata.assign(j, (opt.x(j) == BLANKf) ? mevec[j] : opt.x(j));
  double sr = smat(didx, didx);
  int dfr = n - didx;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double t = ptot(1.0 - opt.cc(), dfr);
  double yest = xstart, xi = xstart;
  fout << "%<" << opt.cc() * 100.0 << "% PREDICTION LIMIT:UPPER>\n";
  for (uint i = 0; i <= RDdef; ++i, xi += dx) // 予測限界上限のデータ出力
  {
    if (opt.xno() > 0)
    {
      cdata.assign(opt.xno(), xi);
      yest = cal_estimate(opt, cdata, smat);
    }
    else
    { yest = xi; }
    double yup = yest + cal_pi(opt, cdata, t, vr, smat);
    fout << xi << " " << yup << "\n";
  }
  fout << "= =\n";
  fout << "%<" << opt.cc() * 100.0 << "% PREDICTION LIMIT:LOWER>\n";
  xi = xstart;
  for (uint i = 0; i <= RDdef; ++i, xi += dx) // 予測限界下限のデータ出力
  {
    if (opt.xno() > 0)
    {
      cdata.assign(opt.xno(), xi);
      yest = cal_estimate(opt, cdata, smat);
    }
    else
    { yest = xi; }
    double ylow = yest - cal_pi(opt, cdata, t, vr, smat);
    fout << xi << " " << ylow << "\n";
  }
}

double lmaclass::cal_pi(option& opt, datac& cdata, double t, double vr, const mat<double>& smat)
// Yの予測限界幅を計算して返す
{
  if (cdata.sdata(0) == BLANKs) return BLANKf; // 欠損があるため計算できず
  size_t didx = smat.r();
  vec<double> xv(didx);
  for (size_t j = 1; j < id.ni(); ++j) xv.assign(j, cdata.fdata(j));
  size_t j = id.ni();
  for (lino_citr ixitr = (opt.ixlst()).begin(); ixitr != (opt.ixlst()).end(); ++ixitr, ++j) // 交互作用データのセット
  { xv.assign(j, xv(ixitr->first) * xv(ixitr->second)); }
  xv.assign(0, 1.0);  xv.assign(didx - 1, 0.0);
  vec<double> tv(xv.dim());
  tv = xv.trn() * smat;
  double vy = vr * (1.0 + inp(tv, xv));
  return t * sqrt(vy);          // Yの予測限界を返す
}

int lmaclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void lmaclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, opt);
      cout << "\n" << endl;
    }
    return;
  }
  size_t dim = id.ni() + (opt.ixlst()).size() + 1; // 積和行列の次元
  int p = dim - 2;
  mat<double> smat(dim);        // 積和行列
  mat<double> xmat(dim - 1);    // 単純積和行列
  vecdbl mevec(dim, 0.0);       // 平均値ベクトル
  vecdbl vrvec(dim, 0.0), ssvec(dim, 0.0); // 分散・積和ベクトル
  int n = cal_smat(smat, xmat, mevec, vrvec, ssvec, opt);
  if (opt.res("res"))
  {
    put_resid(cout, opt);
    if (opt.res("stat"))
    {
      putmsdvec(cout, n, mevec, vrvec, p, 1);
      putmat(cout, "<CORRELATION COEFFICIENT MATRIX>", matr, n, smat, p, 1);
    }
    if (n <= p)                 // 重回帰式計算できず
    {
      cout << "\n" << endl;
      return;
    }
  }
  vecdbl fssvec(dim + 1, 0.0);
  vecint fdfvec(dim + 1, 0);
  cal_lma(smat, fssvec, fdfvec, opt);
  if (opt.res("res"))
  {
    if (opt.res("reg")) put_lma(cout, opt, ssvec, fssvec, fdfvec, n, smat, xmat);
    if (opt.res("atab")) put_lmatab(cout, opt, n, smat, ssvec, fssvec, fdfvec);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, n, smat, mevec);
}

int main(int argc, char* argv[]) // メインルーチン
{
  lmaclass lma;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) lma.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { lma.get(argv[i], opt);  fok = false; }
    }
    lma.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
