// logrege.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 08:29:31 tac>
//
// ロジスティック(ロジット)モデルによる重回帰分析。
// 原則として、前向きデータに適用する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W logrege logrege.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "logrege  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:logrege [<OPTION> ...]  [<DATA FILE> ...]",
  "  The logistic regression analysis with the first item as criterion",
  "  variable and subsequent items as explanatory variables.",
  "<OPTION>",
  "  -b=<VALUE>[,<VALUE>,...]:specify the initial value of regression",
  "    coefficient",
  "  -cin=<Cin>:specify the inclusion criteria Cin(Cin>=0)",
  "  -comp[={-|+}]:complement deficiencies in the calculation of logit scores",
  "    (-comp:-comp=+)",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=logrege_dns.d",
  "  -eps[=<EPS>]:specify the convergence error(0<EPS)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={lci|log|or|score}[,...]:select the data for Ngraph",
  "    lci:CI of logistic curve  log:logistic curve  or:odds ratio",
  "    score:logit scores  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=logrege_ngp.dat",
  "  -pin=<Pin>:specify the inclusion criteria Pin(0<Pin=<1)",
  "  -r{p|n}[=<CODE>]:specify the code of responce yes/no",
  "  -res=[-][{def|all|allx|cc[<C-COEF>]|ltab|selx|stat|v}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    allx:result of selecting all variables",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1 cc:cc0.95)",
  "    ltab:likelihood table  selx:result of selecting variables",
  "    stat:basic statistices  v:initial values of regression coefficient",
  "  -rmax=<NUMBER>:specify the maximum number of repetitions(0=<NUMBER)",
  "  -x<NUMBER>=<VALUE>:specify the value of explanatory variable",
  "    whoes value is constant",
  "  -xno=<NUMBER>:specify the number of explanatory variable whoes value",
  "    changes"
};

const string Title = "=== Logistic Regression Analysis ==="; // 表題
const string DNSNAMEdef = "logrege_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "logrege_ngp.dat"; // 省略時Ngraphファイル名
constexpr double EPSdef = 1.0e-5; // 収束誤差省略時値
constexpr double CIOdef = 2.0; // χ2乗値取り込み・追い出し基準省略時値
constexpr double PIOdef = 0.1; // p値取り込み・追い出し基準省略時値
constexpr int RMAXdef = 100;   // 繰り返し計算数省略時値
constexpr int RDdef = 50;      // グラフの描画範囲分割数省略時値

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  lstr blst;                    // 説明変数の初期値リスト
  vecdbl vbvec;                 // 説明変数の初期値ベクトル
  double vcc;                   // 信頼係数
  double vcin;                  // χ2乗値の取り込み基準
  bool vcomp;                   // 欠損補間の有無
  string vdnsname;              // DANSファイル名
  double veps;                  // 最尤法の収束誤差
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  double vpin;                  // p値の取り込み基準
  vecstr rvec;                  // 反応有・無のコード
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  int vrmax;                    // 最大反復回数
  lstr xlst;                    // 説明変数リスト
  vecdbl vxvec;                 // 説明変数ベクトル
  int vxno;                     // 変化説明変数の番号
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const vecdbl& bvec() const    // 説明変数の初期値ベクトルを返す
  { return vbvec; }
  double b(size_t idx) const    // 説明変数の初期値を返す
  { return (idx < vbvec.size()) ? vbvec[idx] : 0.0; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double cin() const            // χ2乗値の取り込み基準を返す
  { return vcin; }
  bool comp() const             // 欠損補間の有無を返す
  { return vcomp; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  double eps() const            // 最尤法の収束誤差を返す
  { return veps; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  double pin() const            // p値の取り込み基準を返す
  { return vpin; }
  const string& rn() const      // 反応無のコードを返す
  { return rvec[0]; }
  const string& rp() const      // 反応有のコードを返す
  { return rvec[1]; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 最大反復回数を返す
  { return vrmax; }
  double x(size_t no) const     // 説明変数の値を返す
  { return (0 < no && no <= vxvec.size()) ? vxvec[no - 1] : BLANKf; }
  int xno() const               // 変化説明変数の番号を返す
  { return vxno; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), blst(), vbvec(), vcc(CCdef), vcin(CIOdef), vcomp(false), vdnsname(), veps(EPSdef), vngpname(), vngplst(), vngpmap(), vpin(1.0), rvec(2), vreslst(), vresmap(), vrmax(RMAXdef), xlst(), vxvec(), vxno(0)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
  rvec[1] = "1";
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "b")            // 偏回帰係数初期値の指定
  {
    if (strsplit(blst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "cc")      // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "cin")     // χ2乗値の取り込み基準指定
  {
    vcin = opt[1].empty() ? CIOdef : strtodf(opt[1]);
    if (vcin < 0.0) err = badopt(vpname, optstr, USAGE, 0.0, DBL_MAX);
  }
  else if (opt[0] == "comp")    // 欠損補完の有無の指定
  {
    if (opt[1].empty() || opt[1] == "+") vcomp = true;
    else if (opt[1] == "-") vcomp = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 最尤法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < DBL_EPSILON) err = badopt(vpname, optstr, USAGE, DBL_EPSILON, DBL_MAX);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pin")     // p値の取り込み基準指定
  {
    vpin = opt[1].empty() ? PIOdef : strtodf(opt[1]);
    if (vpin <= 0.0 || 1.0 < vpin) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "rn")      // 反応有のコード指定
  { rvec[0] = opt[1]; }
  else if (opt[0] == "rp")      // 反応無のコード指定
  { rvec[1] = opt[1]; }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 最大反復回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "xno")     // 変化説明変数の番号指定
  {
    char* ptr = nullptr;
    vxno = strtol(opt[1].c_str(), &ptr, 0);
    if (vxno < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0][0] == 'x')    // 説明変数の値指定
  {
    if (strsplit(xlst, optstr, CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are requied !");
  if (rvec[0].empty() && rvec[1].empty()) return disperr(vpname, "The code of responce is not specified !");
  int p = ni - 1;
  if (blst.empty() == false)    // 偏回帰係数初期値の指定あり
  {
    if ((int)(blst.size()) > p + 1) return disperr(vpname, "Too many initial values of the regression coefficients !");
    vbvec.resize(p + 1);
    vbvec.assign(p + 1, BLANKf);
    int j = 0;
    for (lstr_citr bitr = blst.begin(); bitr != blst.end(); ++bitr, ++j)
    {
      double fdata = strtodf(*bitr);
      if (fdata == BLANKf) return disperr(vpname, "Wrong initial value->", *bitr);
      vbvec[j] = fdata;
    }
  }
  if (set_ngp() < 0) return -1; // -ngpオプションのセット
  if (set_res() < 0) return -2; // -resオプションのセット
  if (vxno < 1) return 0;       // 説明変数指定なし
  if (p < vxno) return disperr(vpname, "Wrong number of the explanatory variable whoes value are changes !");
  if (xlst.empty()) return 0;   // 説明変数指定なし
  if (int(xlst.size()) > p) return disperr(vpname, "Too many values of the explanatory variables !");
  vxvec.resize(p);
  vxvec.assign(p, BLANKf);
  for (lstr_citr xitr = xlst.begin(); xitr != xlst.end(); ++xitr)
  {
    string opt[2];
    if (getopt(opt, *xitr, OPTdev) > 0)
    {
      char* ptr = nullptr;
      int xno = strtol(opt[0].substr(1).c_str(), &ptr, 0);
      if (xno < 1 || p < xno || xno == vxno) return disperr(vpname, "Wrong number of the explanatory variable->", *xitr);
      double fdata = strtodf(opt[1]);
      if (fdata == BLANKf) return disperr(vpname, "Wrong value of the explanatory variable->", *xitr);
      vxvec[xno - 1] = fdata;
    }
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["lci"] = vngpmap["log"] = vngpmap["or"] = vngpmap["score"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["lci"] = vngpmap["log"] = vngpmap["or"] = vngpmap["score"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["ltab"] = vresmap["selx"] = vresmap["stat"] = true;
  vresmap["allx"] = vresmap["v"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["ltab"] = vresmap["selx"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["ltab"] = vresmap["selx"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["allx"] = vresmap["cc"] = vresmap["ltab"] = vresmap["selx"] = vresmap["stat"] = vresmap["v"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class data_less                 // ケースデータの選択済み説明変数比較用関数オブジェクト
{
private:
  const vecint *checkvec;       // 変数選択ベクトルへのポインター
public:
  data_less(const vecint *checkvec0) : checkvec(checkvec0)
  // コンストラクタ
  {}
  ~data_less()                  // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_less::operator()(datac& cdata1, datac& cdata2) const
// ケースデータの選択済み説明変数を比較する
{
  size_t nx = checkvec->size() - 2;
  int comp = 0;
  for (size_t i = 1; comp == 0 && i <= nx; ++i)
  {
    if (checkvec->at(i))        // 選択済み説明変数だけ比較する
    {
      double fdata1 = cdata1.fdata(i), fdata2 = cdata2.fdata(i);
      if (fdata1 == BLANKf)
      {
        if (fdata2 != BLANKf) comp = 1;
      } else if (fdata2 == BLANKf)
      { comp = -1; }
      else
      { comp = sgn(fdata1 - fdata2); }
    }
  }
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class logregclass               // logregメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_smat(int& nr, mat<double>& smat, vecdbl& midvec, vecdbl& mevec, vecdbl& vrvec, const option& opt);
  // データを集計して積和行列等を求める
  void put_resid(ostream& fout, const option& opt, const vecdbl& midvec);
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, const option& opt, int n, int nr, const mat<double>& smat, const vecdbl& mevec, const vecdbl& vrvec);
  // 統計量を出力する
  int step_sel(double cin, double pin, vecint& checkvec, vec<double>& bv, vec<double>& vbv, double& lb, mat<double>& ifmat, const option& opt);
  // 増減法による変数選択
  int cal_csqr(vecdbl& cvec, const option& opt, const vec<double>& b0v);
  // 変数選択用χ2乗値を計算する
  int sel_invar(vecint& checkvec, vec<double>& btv, vec<double>& vbtv, const vecdbl& cvec, const vec<double>& bv) const;
  // 取り込み変数の選択と各種設定
  int sel_outvar(double cin, double pin, vecint& checkvec, vec<double>& bv, vec<double>& vbv, vec<double>& btv, vec<double>& vbtv) const;
  // 係数ベクトルの代入と追い出し変数の選択
  int newton_log(vec<double>& bv, vec<double>& vbv, double& lb, mat<double>& ifmat, const option& opt, const vecint& checkvec, int rmax);
  // Newton法によるロジットモデルの計算(最尤法)
  int cal_lgh(vec<double>& gv, mat<double>& hmat, double& lb, const vecint& checkvec, const vec<double>& bv);
  // 傾斜ベクトルgとHessian行列Hを計算する
  void set_initb(vec<double>& b0v, vec<double>& vbv, int n, mat<double>& smat, const option& opt);
  // 重回帰分析→判別分析による偏回帰係数初期値のセット
  void put_logreg(ostream& fout, option& opt, int n, int nr, double cin, double pin, int r, const vecdbl& vrvec, const vecint& checkvec, const vec<double>& b0v, const vec<double>& bv, const vec<double>& vbv, double lb, const mat<double>& ifmat);
  // ロジスティック回帰分析結果を出力する
  void put_brci(ostream& fout, const option& opt, const vecint& checkvec, const vec<double>& bv, const vec<double>& vbv) const;
  // 偏回帰係数とオッズ比の信頼区間を出力する
  void put_logtab(ostream& fout, int n, int nr, const vecint& checkvec, double lb);
  // ロジスティック回帰分析の分散分析表を出力する
  int cal_loglf(double& lf, const vecint& checkvec); // 飽和モデルの尤度を求める
  bool data_equal(const vecint& checkvec, datac& cdata1, datac& cdata2);
  // ケースデータの選択済み説明変数を比較する
  void put_dns(option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv);
  // データファイル形式のロジットスコアと反応有の確率を出力する
  void plot_logngp(option& opt, double cin, double pin, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, const vec<double>& vbv, const mat<double>& ifmat);
  // Ngraph用データ出力
  void plot_ngpid(ostream& fout, const option& opt, double cin, double pin, const vecint& checkvec, const vecdbl& midvec);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  double cal_score(double& z, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, datac& cdata);
  // ロジットスコアを計算し、群インデックスを返す
  void plot_score(ostream& fout, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv);
  // Ngraph用ロジットスコアを出力する
  void plot_log(ostream& fout, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, const mat<double>& ifmat);
  // Ngraph用ロジスティック曲線を出力する
  double cal_range(double& xmin, double& xmax, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, datac& cdata);
  // ロジスティック曲線のxの最小値と最大値を計算する
  void plot_lci(ostream& fout, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, const mat<double>& ifmat);
  // Ngraph用ロジスティック曲線の信頼区間を出力する
  double cal_se(const vecint& checkvec, const mat<double>& ifmat, datac& cdata);
  // zのSEを計算して返す
  void plot_or(ostream& fout, const option& opt, const vecint& checkvec, const vec<double>& bv, const vec<double>& vbv) const;
  // Ngraph用にORとその信頼区間を出力する
public:
  logregclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~logregclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int logregclass::cal_smat(int& nr, mat<double>& smat, vecdbl& midvec, vecdbl& mevec, vecdbl& vrvec, const option& opt)
// データを集計して積和行列等を求める
{
  size_t dim = id.ni() + 1, yidx = id.ni();
  vecdbl mevec0(dim, 0.0), mevec1(dim, 0.0), fdatavec(dim, 0.0);
  fdatavec[0] = 1.0;
  for (auto& cdata : datalst)
  {
    if (cdata.sdata(0) == BLANKs)
    { cdata.assign(0, BLANKs); } // 後の集計のためにコードを欠損にする
    else
    {
      double y = -1.0;
      if (opt.rn().empty())     // 反応無しコード指定なし
      { y = (cdata.sdata(0) == opt.rp()) ? 1.0 : 0.0; }
      else if (cdata.sdata(0) == opt.rn())
      { y = 0.0; }
      else if (opt.rp().empty() || cdata.sdata(0) == opt.rp())
      { y = 1.0; }
      if (y < 0.0)              // 指定コード以外
      { cdata.assign(0, BLANKs); }
      else
      {
        bool blank = false;
        for (size_t j = 1; blank == false && j < id.ni(); ++j)
        {
          fdatavec[j] = cdata.fdata(j);
          if (fdatavec[j] == BLANKf) blank = true;
        }
        if (blank)
        { cdata.assign(0, y - 2.0); } // 欠損補完のためにコードを-2/-1にする
        else
        {
          if (y == 1.0) ++nr;   // 反応有の例数を数える
          cdata.assign(0, y);   // 後の集計のためにコードを0/1にする
          fdatavec[yidx] = y;
          for (size_t i = 0; i < dim; ++i)
          {
            double x = fdatavec[i];
            for (size_t j = i; j < dim; ++j) smat.add(i, j, x * fdatavec[j]);
            if (y == 0.0) mevec0[i] += x;
            else mevec1[i] += x;
          }
        }
      }
    }
  }
  for (size_t i = 1; i < dim; ++i) // 反応の有無別平均値とその中央値の計算
  {
    mevec0[i] /= mevec0[0];  mevec1[i] /= mevec1[0];
    midvec[i] = (mevec0[i] + mevec1[i]) * 0.5;
  }
  int n = (int)(smat(0, 0));    // 単純積和行列→積和行列
  for (size_t i = 0; i < dim - 1; ++i)
  {
    for (size_t j = i + 1; j < dim; ++j) smat.assign(j, i, smat(i, j));
  }
  smat.sweep(0);
  for (size_t i = 0; i < dim; ++i)
  {
    mevec[i] = smat(0, i);
    vrvec[i] = (n > 1) ? smat(i, i) / (double)(n - 1) : 0.0;
  }
  return n;                     // 例数を返す
}

void logregclass::put_resid(ostream& fout, const option& opt, const vecdbl& midvec)
// 解析結果のタイトルとデータ識別情報を出力する
{
  idd tid(id.title(), 3);
  tid.assign_item(0, id.item(0));
  if (opt.cin() > 0.0)
  {
    tid.assign_item(1, "Logit Scores(Selection Method)");
    tid.assign_item(2, "Probability of Responce(Selection Method)");
  }
  else
  {
    tid.assign_item(1, "Logit Scores(All Selection)");
    tid.assign_item(2, "Probability of Responce(All Selection)");
  }
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i;
    fout << ":" << id.item(i);
    if (opt.comp()) fout << "  Mean for Missing Complement=" << midvec[i];
    fout << "\n";
  }
}

void logregclass::put_stat(ostream& fout, const option& opt, int n, int nr, const mat<double>& smat, const vecdbl& mevec, const vecdbl& vrvec)
// 統計量を出力する
{
  putmsdvec(fout, n, mevec, vrvec, id.ni() - 1, 1);
  fout << "Responce Yes:CODE=";
  if (opt.rp().empty()) fout << "other than " << opt.rn();
  else fout << opt.rp();
  fout << "  n=" << nr;
  fout << "  No:CODE=";
  if (opt.rn().empty()) fout << "other than " << opt.rp();
  else fout << opt.rn();
  fout << "  n=" << n - nr << "\n";
  putmat(fout, "<CORRELATION COEFFICIENT MATRIX>", matr, n, smat, id.ni() - 1, 1);
}

int logregclass::step_sel(double cin, double pin, vecint& checkvec, vec<double>& bv, vec<double>& vbv, double& lb, mat<double>& ifmat, const option& opt)
// 増減法による変数選択
{
  vecdbl cvec(id.ni(), 0.0);
  vec<double> btv(id.ni()), vbtv(id.ni());
  int iin = 0, iout = 0, r = 0, rm = 0;
  do {
    if (cal_csqr(cvec, opt, bv) != 0)
    { r = -1; }
    else
    {
      iin = sel_invar(checkvec, btv, vbtv, cvec, bv);
      if (iin > 0)                // 取り込み変数あり
      {
        iout = 0;
        int i = 0;
        do {
          r = newton_log(btv, vbtv, lb, ifmat, opt, checkvec, opt.rmax());
          if (r > rm) rm = r;
          i = sel_outvar(cin, pin, checkvec, bv, vbv, btv, vbtv);
          if (i > 0) iout = i;    // 最後に追い出した変数番号を記憶
        } while (r > 0 && i > 0); // 追い出し変数が無くなるまで続行
      }
    }
  } while (r > 0 && iin > 0 && iin != iout && checkvec[id.ni()] < (int)(id.ni()));
  if (r < 0) return r;          // Newton法発散
  return rm;                    // 最大反復数を返す
}

int logregclass::cal_csqr(vecdbl& cvec, const option& opt, const vec<double>& b0v)
// 変数選択用χ2乗値を計算する
{
  vec<double> bv(b0v), gv(id.ni());
  mat<double> hmat(id.ni());
  double lb = 0.0;
  vecint checkvec(id.ni() + 1, 1);
  checkvec[id.ni()] = id.ni();
  if (cal_lgh(gv, hmat, lb, checkvec, bv) != 0) return disperr(opt.pname(), "CANNOT COMPUTE CHI-SQUARED FOR SELECTION VARIABLES !");
  hmat.inv();
  bv -= hmat * gv;
  for (size_t i = 0; i < id.ni(); ++i) cvec[i] = (hmat(i, i) < 0.0) ? bv(i) * bv(i) / (-hmat(i, i)) : 0.0;
  return 0;
}

int logregclass::sel_invar(vecint& checkvec, vec<double>& btv, vec<double>& vbtv, const vecdbl& cvec, const vec<double>& bv) const
// 取り込み変数の選択と各種設定
{
  double cmax = 0.0;
  int imax = 0;
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i] == 0 && cvec[i] > cmax) { cmax = cvec[i];  imax = i; }
  }
  if (imax < 1) return imax;    // 取り込み変数なし
  checkvec[imax] = 1;  checkvec[id.ni()] += 1;
  btv.resize(checkvec[id.ni()]);  vbtv.resize(checkvec[id.ni()]);
  btv = 0.0;
  return imax;                  // 変数番号を返す
}

int logregclass::sel_outvar(double cin, double pin, vecint& checkvec, vec<double>& bv, vec<double>& vbv, vec<double>& btv, vec<double>& vbtv) const
// 係数ベクトルの代入と追い出し変数の選択
{
  double cmin = cin, pmax = pin;
  int imin = 0, j = 1;
  bool csel = (pin < 1.0) ? false : true; // χ2乗値による変数選択
  bv.assign(0, btv(0));  vbv.assign(0, vbtv(0));
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i])
    {
      bv.assign(i, btv(j));  vbv.assign(i, vbtv(j));
      double csqr = (vbtv(j) > 0.0) ? btv(j) * btv(j) / vbtv(j) : 0.0;
      if (csel)
      {
        if (csqr < cmin) { cmin = csqr;  imin = i; }
      }
      else
      {
        double p = ctop(csqr, 1);
        if (p > pmax) { pmax = p;  imin = i; }
      }
      ++j;
    }
  }
  if (imin < 1) return imin;    // 追い出し変数なし
  checkvec[imin] = 0;  checkvec[id.ni()] -= 1;
  btv.resize(checkvec[id.ni()]);  vbtv.resize(checkvec[id.ni()]);
  btv = 0.0;
  return imin;                  // 変数番号を返す
}

int logregclass::newton_log(vec<double>& bv, vec<double>& vbv, double& lb, mat<double>& ifmat, const option& opt, const vecint& checkvec, int rmax)
// Newton法によるロジットモデルの計算(最尤法)
{
  int d = checkvec[id.ni()];
  vec<double> gv(d);
  mat<double> hmat(d);
  double lold = 0.0;
  bool ok = false;
  int r = 0;
  while (ok == false && cal_lgh(gv, hmat, lb, checkvec, bv) == 0)
  {
    ++r;
    hmat.inv();
    if (fabs(lb - lold) < opt.eps() || (rmax > 0 && r >= rmax)) ok = true;
    else { lold = lb;  bv -= hmat * gv; }
  }
  if (ok == false) return -1;   // 発散！
  for (int i = 0; i < d; ++i) vbv.assign(i, -hmat(i, i));
  ifmat.resize(d, d);
  ifmat = hmat;
  ifmat *= -1;                  // 情報行列
  return r;                     // 反復数を返す
}

int logregclass::cal_lgh(vec<double>& gv, mat<double>& hmat, double& lb, const vecint& checkvec, const vec<double>& bv)
// 傾斜ベクトルgとHessian行列Hを計算する
{
  constexpr double Emax = 308.0; // 指数最大値
  int p = checkvec[id.ni()];
  double sbxy = 0.0, sei = 0.0;
  vec<double> sxyv(p), sxv(p), xyv(p + 1);
  xyv.assign(0, 1.0);  gv = 0.0;  hmat = 0.0;  lb = 0.0;
  bool err = false;
  for (ldata_itr ditr = datalst.begin(); !err && ditr != datalst.end(); ++ditr)
  {
    if (ditr->sdata(0) == "0" || ditr->sdata(0) == "1")
    {
      double ei = bv(0);
      int j = 1;
      for (size_t i = 1; i < id.ni(); ++i)
      {
        if (checkvec[i])
        {
          xyv.assign(j, ditr->fdata(i));
          ei += xyv(j) * bv(j);
          ++j;
        }
      }
      if (fabs(ei) > Emax)        // 指数最大値以上
      { err = true; }
      else
      {
        if (ditr->sdata(0) == "0") xyv.assign(p, 0.0);
        else { xyv.assign(p, 1.0);  sbxy += ei; }
        ei = exp(ei);
        sei += log(1.0 + ei);
        double wi = ei / (1.0 + ei), wi2 = ei / (1.0 + ei) / (1.0 + ei);
        for (j = 0; j < p; ++j)
        {
          sxyv.add(j, xyv(j) * xyv(p));
          sxv.add(j, xyv(j) * wi);
          for (int k = j; k < p; ++k) hmat.add(j, k, -xyv(j) * xyv(k) * wi2);
        }
      }
    }
  }
  if (err) return -1;
  lb = sbxy - sei;  gv = sxyv - sxv;
  for (int i = 0; i < p - 1; ++i)
  {
    for (int j = i + 1; j < p; ++j) hmat.assign(j, i, hmat(i, j));
  }
  return 0;
}

void logregclass::set_initb(vec<double>& b0v, vec<double>& vbv, int n, mat<double>& smat, const option& opt)
// 重回帰分析→判別分析による偏回帰係数初期値のセット
{
  double ym = smat(0, id.ni()), st = smat(id.ni(), id.ni());
  for (size_t i = 1; i < id.ni(); ++i) smat.sweep(i);
  double w = 1.0, sr = smat(id.ni(), id.ni());
  if (sr > 0.0) w = (n - 2) / sr;
  int dfr = n - id.ni();
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  const double Beps = DBL_EPSILON * fmax(FDmin, st);
  for (size_t i = 1; i < id.ni(); ++i)
  {
    double bi = smat(i, id.ni());
    if (fabs(bi) < Beps) bi = 0.0;
    b0v.assign(i, bi * w);
    double vbi = vr * smat(i, i);
    vbv.assign(i, vbi * w * w);
  }
  b0v.assign(0, (smat(0, id.ni()) - ym + (ym - 0.5) * (st - sr) / st) * w - log((1.0 - ym) / ym));
  // (1-π)/πの補正
  if (opt.bvec().empty()) return; // 初期値指定なし
  for (size_t i = 0; i < id.ni(); ++i)
  {
    if (opt.b(i) != BLANKf) b0v.assign(i, opt.b(i));
  }
}

void logregclass::put_logreg(ostream& fout, option& opt, int n, int nr, double cin, double pin, int r, const vecdbl& vrvec, const vecint& checkvec, const vec<double>& b0v, const vec<double>& bv, const vec<double>& vbv, double lb, const mat<double>& ifmat)
// ロジスティック回帰分析結果を出力する
{
  if (cin <= 0.0 && pin >= 1.0)
  {
    fout << "\n<RESULT OF SELECTING ALL VARIABLES";
    if (r > 1) fout << "(Repetitions:" << r << ")>\n";
    else if (r == 1) fout << "(Initial Values were used)>\n";
    else fout << "(Newton Method diverged so Initial Values were used)>\n";
  }
  else
  {
    fout << "\n<RESULT OF STEPWISE FORWARD SELECTION METHOD>\n";
    if (pin < 1.0)
    {
      fout << "  Inclusion Criteria:p=<" << pin;
      fout << "  Expulsion Criteria:p>" << pin;
    }
    else
    {
      fout << "  Inclusion Criteria:c^2>=" << cin;
      fout << "  Expulsion Criteria:c^2<" << cin;
    }
    fout << "  Repetitions:" << r << "\n";
  }
  fout << "  Logit Model:p=1/[1+exp(-b0-sum{bj xj})]\n";
  fout << "  p:Probability of y=";
  if (opt.rp().empty()) fout << "other than " << opt.rn();
  else fout << opt.rp();
  fout << "  b0:Intercept  bj:Partial Regression Coefficient of xj\n";
  if (opt.res("v"))
  {
    fout << "  Initial Value:";
    for (size_t i = 0; i < id.ni(); ++i)
    {
      if (i > 0)
      {
        if (i % 4 == 0) fout << "\n" << setw(16) << " ";
        else fout << "  ";
      }
      fout << "b" << i << "=" << b0v(i);
    }
    fout << "\n";
  }
  constexpr int LENline = 79;   // 区切り線の長さ
  fout << "VAR    Partial b     SE of b  Odds Ratio  Standard b  Chi-suared     p-value\n";
  putline(fout, LENline);
  double csqr = (vbv(0) > 0.0) ? bv(0) * bv(0) / vbv(0) : 0.0;
  ldbl pr = ctop(csqr, 1);
  fout << "ICPT" << setw(12) << bv(0) << setw(12) << sqrt(vbv(0));
  fout << setw(36) << csqr << setw(12) << pr << pind(pr) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i])
    {
      csqr = (vbv(i) > 0.0) ? bv(i) * bv(i) / vbv(i) : 0.0;
      pr = ctop(csqr, 1);
      fout << "X" << setw(3) << i << setw(12) << bv(i);
      fout << setw(12) << sqrt(vbv(i)) << setw(12) << exp(bv(i));
      fout << setw(12) << bv(i) * sqrt(vrvec[i]) << setw(12) << csqr;
      fout << setw(12) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
  if (opt.res("cc")) put_brci(fout, opt, checkvec, bv, vbv);
  if (opt.res("ltab")) put_logtab(fout, n, nr, checkvec, lb);
}

void logregclass::put_brci(ostream& fout, const option& opt, const vecint& checkvec, const vec<double>& bv, const vec<double>& vbv) const
// 偏回帰係数とオッズ比の信頼区間を出力する
{
  constexpr int LENline = 76;   // 区切り線の長さ
  fout << "\nVAR    Partial b" << setw(3) << opt.cc() * 100.0;
  fout << "%CI:Lower       Upper  Odds Ratio" << setw(3) << opt.cc() * 100.0;
  fout << "%CI:Lower       Upper\n";
  putline(fout, LENline);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vbv(0));
  double bl = bv(0) - ci, bu = bv(0) + ci;
  fout << "ICPT" << setw(12) << bv(0) << setw(12) << bl;
  fout << setw(12) << bu << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i])
    {
      ci = z * sqrt(vbv(i));
      bl = bv(i) - ci;  bu = bv(i) + ci;
      fout << "X" << setw(3) << i << setw(12) << bv(i);
      fout << setw(12) << bl << setw(12) << bu;
      fout << setw(12) << exp(bv(i)) << setw(12) << exp(bl);
      fout << setw(12) << exp(bu) << "\n";
    }
  }
  putline(fout, LENline);
}

void logregclass::put_logtab(ostream& fout, int n, int nr, const vecint& checkvec, double lb)
// ロジスティック回帰分析の分散分析表を出力する
{
  int dft = n - 1;
  double lf = 0.0;
  dft -= cal_loglf(lf, checkvec);
  int dfb = checkvec[id.ni()] - 1;
  int dfl = dft - dfb;
  double l0 = (double)nr * log((double)nr / (double)n) + (double)(n - nr) * log((double)(n - nr) / (double)n);
  double r2 = 0.0;
  if (sgndiff(lf, l0) > 0) r2 = (lb - l0) / (lf - l0);
  fout << "Log Likelihood:L(Model)=" << lb << "  L(Intercept)=" << l0 << "  Lf=" << lf << "\n";
  fout << "Pseudo-Contribution Rate:R^2=" << r2 << "  AIC=" << 2.0 * (dfb + 1 - lb) << "\n";
  constexpr int LENline = 68;
  const string Head = "Test of Regression and LOF";
  fout << "\n" << setw((LENline - strclen(Head, CHARwide)) / 2 + Head.size()) << Head;
  fout << "\nFACTOR   -(log Likelihood)        DF   Chi-squared       p-value\n";
  putline(fout, LENline);
  double csqr = 2.0 * (lb - l0);
  ldbl pr = ctop(csqr, dfb);
  fout << "REGRESSION" << setw(16) << lb - l0 << setw(10) << dfb;
  fout << setw(14) << csqr << setw(14) << pr << pind(pr) << "\n";
  const double Ceps = DBL_EPSILON * fmax(FDmin, csqr);
  csqr = 2.0 * (lf - lb);
  if (dfl < 1 || csqr < Ceps) csqr = 0.0;
  pr = ctop(csqr, dfl);
  fout << "LOF" << setw(23) << lf - lb << setw(10) << dfl;
  fout << setw(14) << csqr << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  fout << "TOTAL" << setw(21) << lf - l0 << setw(10) << dft << endl;
}

int logregclass::cal_loglf(double& lf, const vecint& checkvec)
// 飽和モデルの尤度を求める
{
  datalst.sort(data_less(&checkvec)); // 選択済み説明変数順にソートする
  int nlf = 0;
  lf = 0.0;
  ldata_itr iitr = datalst.begin();
  while (iitr != datalst.end())
  {
    if (iitr->sdata(0) != "0" && iitr->sdata(0) != "1")
    { ++iitr; }
    else
    {
      int ni = 1, ri = 0;
      if (iitr->sdata(0) == "1") ri = 1; // 反応有
      bool equal = true;
      ldata_itr jitr = iitr;
      ++jitr;
      while (equal && jitr != datalst.end())
      {
        if (jitr->sdata(0) != "0" && jitr->sdata(0) != "1")
        { ++jitr; }
        else
        {
          equal = data_equal(checkvec, *iitr, *jitr);
          if (equal)
          {
            if (jitr->sdata(0) == "1") ++ri;
            ++ni;  ++jitr;
          }
        }
      }
      if (ni > 1)
      {
        nlf += ni - 1;
        if (ri > 0 && ri < ni) lf += (double)ri * log((double)ri / (double)ni) + (double)(ni - ri) * log((double)(ni - ri) / (double)ni);
      }
      iitr = jitr;
    }
  }
  return nlf;                   // 飽和モデルの重複数を返す(自由度減少数)
}

bool logregclass::data_equal(const vecint& checkvec, datac& cdata1, datac& cdata2)
// ケースデータの選択済み説明変数を比較する
{
  size_t nx = checkvec.size() - 2;
  bool equal = true;
  for (size_t i = 1; equal && i <= nx; ++i)
  {
    if (checkvec[i])            // 選択済み説明変数だけ比較する
    {
      double fdata1 = cdata1.fdata(i), fdata2 = cdata2.fdata(i);
      if (fdata1 == BLANKf)
      {
        if (fdata2 != BLANKf) equal = false;
      } else if (fdata2 == BLANKf)
      { equal = false; }
      else
      {
        if (sgndiff(fdata1, fdata2) != 0) equal = false;
      }
    }
  }
  return equal;                 // true:一致, false:不一致
}

void logregclass::put_dns(option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv)
// データファイル形式のロジットスコアと反応有の確率を出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  idd tid(id.title(), 3);
  tid.assign_item(0, id.item(0));
  if (opt.cin() > 0.0)
  {
    tid.assign_item(1, "Logit Scores(Selection Method)");
    tid.assign_item(2, "Probability of Responce(Selection Method)");
  }
  else
  {
    tid.assign_item(1, "Logit Scores(All Selection)");
    tid.assign_item(2, "Probability of Responce(All Selection)");
  }
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  tid.put_item(fout, formtag);
  datac cdata(3);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double z = BLANKf;
    double y = cal_score(z, opt, checkvec, midvec, bv, *ditr);
    cdata.assign(0, y);
    if (y == BLANKf)
    { cdata.assign(1, BLANKs);  cdata.assign(2, BLANKs); }
    else
    { cdata.assign(1, z);  cdata.assign(2, 1.0 / (1.0 + exp(-z))); }
    cdata.put(fout);
  }
}

void logregclass::plot_logngp(option& opt, double cin, double pin, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, const vec<double>& vbv, const mat<double>& ifmat)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("log"))
  {
    fout << "-s1 -dline -w80 -c255,0,0 -zy-0.1,1.1,0.2\n";
    plot_ngpid(fout, opt, cin, pin, checkvec, midvec);
    plot_log(fout, opt, checkvec, midvec, bv, ifmat);
    delml = true;
  }
  if (opt.ngp("lci"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -c255,0,0 -zy-0.1,1.1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, opt, cin, pin, checkvec, midvec);
    plot_lci(fout, opt, checkvec, midvec, bv, ifmat);
  }
  if (opt.ngp("score"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -o300 -zy-0.1,1.1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, opt, cin, pin, checkvec, midvec);
    plot_score(fout, opt, checkvec, midvec, bv);
  }
  if (opt.ngp("or"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -w80 -o400 -zx0,"<< id.ni() << ",1 -zy0.1,100,10 -eylog\n";
      fout << "%when draw CI:-s1 -derrorbar_y -w80\n";
    }
    plot_ngpid(fout, opt, cin, pin, checkvec, midvec);
    plot_or(fout, opt, checkvec, bv, vbv);
  }
}

void logregclass::plot_ngpid(ostream& fout, const option& opt, double cin, double pin, const vecint& checkvec, const vecdbl& midvec)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i])
    {
      fout << "%EXPLANATORY VARIABLE X" << setw(3) << i << ":" << id.item(i);
      if (opt.comp()) fout << "  Mean for Missing Complement=" << midvec[i];
      fout << "\n";
    }
  }
  if (pin < 1.0) fout << "%Inclusion Criteria:p=<Pin=" << pin << "\n";
  else fout << "%Inclusion Criteria:Wald's chi-squared>=Cin=" << cin << "\n";
  if (opt.xno() > 0 && checkvec[opt.xno()]) // 変化説明変数指定あり
  {
    fout << "%Explanatory Variable:" << id.item(opt.xno()) << "\n";
    for (size_t i = 1; i < id.ni(); ++i)
    {
      if (checkvec[i])
      {
        if (opt.x(i) != BLANKf) fout << "%Explanatory Variable:" << id.item(i) << "  Specified Value=" << opt.x(i) << "\n";
      }
    }
  }
}

double logregclass::cal_score(double& z, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, datac& cdata)
// ロジットスコアを計算し、yを返す
{
  double y = cdata.fdata(0);
  if (y == BLANKf) return BLANKf; // 目的変数欠損のため計算できず
  z = bv(0);
  bool blank = false;
  for (size_t i = 1; blank == false && i < id.ni(); ++i)
  {
    if (cdata.fdata(i) == BLANKf)
    {
      if (opt.comp())           // 欠損を平均値で補完する場合
      {
        if (checkvec[i]) z += bv(i) * midvec[i];
      }
      else
      { blank = true; }
    }
    else
    {
      if (checkvec[i]) z += bv(i) * cdata.fdata(i);
    }
  }
  if (y < 0.0) y += 2.0;        // データに欠損がある症例の目的変数
  return blank ? BLANKf : y;    // 計算できたらyを返す
}

void logregclass::plot_score(ostream& fout, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv)
// Ngraph用ロジットスコアを出力する
{
  fout << "%<LOGIT SCORES>\n";
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = BLANKf, z = BLANKf;
    double y = cal_score(z, opt, checkvec, midvec, bv, *ditr);
    if (opt.xno() < 1) x = z;
    else x = ditr->fdata(opt.xno());
    if (x != BLANKf && y != BLANKf) fout << x << " " << y << "\n";
  }
}

void logregclass::plot_log(ostream& fout, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, const mat<double>& ifmat)
// Ngraph用ロジスティック曲線を出力する
{
  fout << "%<LOGISTIC CURVE>\n";
  datac cdata(id.ni());
  cdata.assign(0, bv(0));       // 切片
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (opt.x(i) == BLANKf) cdata.assign(i, midvec[i]);
    else cdata.assign(i, opt.x(i));
  }
  double xmin = 0.0, xmax = 0.0;
  double xrange = cal_range(xmin, xmax, opt, checkvec, midvec, bv, cdata);
  constexpr int RD = 2 * RDdef; // 座標範囲の分解数
  double dx = xrange / (double)RD, z = 0.0, pr = 0.0;
  double x = xmin;
  for (int i = 0; i <= RD; ++i, x += dx)
  {
    z = 0.0;
    if (opt.xno() > 0 && checkvec[opt.xno()])
    {
      cdata.assign(opt.xno(), x);
      cal_score(z, opt, checkvec, midvec, bv, cdata);
    }
    else
    { z = x; }
    pr = 1.0 / (1.0 + exp(-z));
    fout << x << " " << pr << "\n";
  }
}

double logregclass::cal_range(double& xmin, double& xmax, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, datac& cdata)
// ロジスティック曲線のxの最小値と最大値を計算する
{
  double pr = 0.01;             // 最小確率
  double zmin = log(pr / (1 - pr));
  pr = 0.99;                    // 最大確率
  double zmax = log(pr / (1 - pr));
  if (opt.xno() > 0 && checkvec[opt.xno()]) // 変化説明変数が指定されている時
  {
    cdata.assign(opt.xno(), 0.0);
    double z = 0.0;
    cal_score(z, opt, checkvec, midvec, bv, cdata);
    xmin = (zmin - z) / bv(opt.xno());
    xmax = (zmax - z) / bv(opt.xno());
    if (bv(opt.xno()) < 0.0)    // 偏回帰係数が負の時
    {
      z = xmin;
      xmin = xmax;
      xmax = z;
    }
  }
  else
  { xmin = zmin;  xmax = zmax; }
  return xmax - xmin;           // 範囲を返す
}

void logregclass::plot_lci(ostream& fout, const option& opt, const vecint& checkvec, const vecdbl& midvec, const vec<double>& bv, const mat<double>& ifmat)
// Ngraph用ロジスティック曲線の信頼区間を出力する
{
  datac cdata(id.ni());
  cdata.assign(0, bv(0));       // 切片
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (opt.x(i) == BLANKf) cdata.assign(i, midvec[i]);
    else cdata.assign(i, opt.x(i));
  }
  double xmin = 0.0, xmax = 0.0;
  double xrange = cal_range(xmin, xmax, opt, checkvec, midvec, bv, cdata);
  constexpr int RD = 2 * RDdef; // 座標範囲の分解数
  double dx = xrange / (double)RD;
  double t = ptoz(1.0 - opt.cc()), z = 0.0, pr = 0.0, ci = 0.0;
  double x = xmin;
  fout << "%<" << opt.cc() * 100.0 << "%CI OF LOGISTIC CURVE:LOWER>\n";
  for (int i = 0; i <= RD; ++i, x += dx)
  {
    if (opt.xno() > 0 && checkvec[opt.xno()])
    {
      cdata.assign(opt.xno(), x);
      cal_score(z, opt, checkvec, midvec, bv, cdata);
    }
    else
    { z = x; }
    ci = t * cal_se(checkvec, ifmat, cdata);
    pr = 1.0 / (1.0 + exp(-z + ci));
    fout << x << " " << pr << "\n";
  }
  fout << "= =\n";
  fout << "%<" << opt.cc() * 100.0 << "%CI OF LOGISTIC CURVE:UPPER>\n";
  x = xmin;
  for (int i = 0; i <= RD; ++i, x += dx)
  {
    if (opt.xno() > 0 && checkvec[opt.xno()])
    {
      cdata.assign(opt.xno(), x);
      cal_score(z, opt, checkvec, midvec, bv, cdata);
    }
    else
    { z = x; }
    ci = t * cal_se(checkvec, ifmat, cdata);
    pr = 1.0 / (1.0 + exp(-z - ci));
    fout << x << " " << pr << "\n";
  }
}

double logregclass::cal_se(const vecint& checkvec, const mat<double>& ifmat, datac& cdata)
// zのSEを計算して返す
{
  int d = ifmat.r();
  vec<double> xvec(d);
  for (int i = 0; i < d; ++i) xvec.assign(i, cdata.fdata(i));
  vec tvec(xvec.trn() * ifmat);
  double vr = inp(tvec, xvec);
  return sqrt(vr);              // zのSEを返す
}

void logregclass::plot_or(ostream& fout, const option& opt, const vecint& checkvec, const vec<double>& bv, const vec<double>& vbv) const
// Ngraph用にORとその信頼区間を出力する
{
  fout << "%<ODDS RATIO AND " << opt.cc() * 100.0 << "% CONFIDENCE INTERVAL>\n";
  double z = ptoz(1.0 - opt.cc());
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i])
    {
      double ci = z * sqrt(vbv(i));
      double ori = exp(bv(i));
      double orl = exp(bv(i) - ci), oru = exp(bv(i) + ci);
      fout << i << " " << ori << " " << oru - ori << " " << orl - ori << "\n";
    }
  }
}

int logregclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void logregclass::put( option& opt) // 集計結果を出力する
{
  size_t dim = id.ni() + 1;     // 積和行列の次元数
  int yidx = id.ni();           // yのインデックス
  int nr = 0;                   // 反応有の例数
  mat<double> smat(dim);        // 積和行列
  vecdbl midvec(dim, 0.0);      // 反応有無別平均値の中央値ベクトル
  vecdbl mevec(dim, 0.0), vrvec(dim, 0.0); // 平均値・分散ベクトル
  int n = cal_smat(nr, smat, midvec, mevec, vrvec, opt); // 例数
  if (n < 1)                    // 有効なデータなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, opt, midvec);
      cout << "\n" << endl;
    }
    return;
  }
  if (opt.res("res"))
  {
    put_resid(cout, opt, midvec);
    if (opt.res("stat")) put_stat(cout, opt, n, nr, smat, mevec, vrvec);
    if (n < (int)yidx || smat(yidx, yidx) <= 0.0) // 回帰式計算できず
    {
      cout << "\n" << endl;
      return;
    }
  }
  vecint checkvec(dim, 0);
   vec<double> b0v(yidx), bv(yidx), vbv(yidx);
   mat<double> ifmat(yidx);
  double lb = 0.0;
  bool putdns = false;
  if (opt.cin() > 0.0 || (0.0 < opt.pin() && opt.pin() < 1.0)) // 変数選択
  {
    checkvec[0] = 1;  checkvec[yidx] = 1; // 定数は無条件に取り込み
    int r = step_sel(opt.cin(), opt.pin(), checkvec, bv, vbv, lb, ifmat, opt);
    if (opt.res("res") && opt.res("selx")) put_logreg(cout, opt, n, nr, opt.cin(), opt.pin(), r, vrvec, checkvec, b0v, bv, vbv, lb, ifmat);
    if (opt.dnsname() != "") put_dns(opt, checkvec, midvec, bv);
    if (opt.ngpname() != "") plot_logngp(opt, opt.cin(), opt.pin(), checkvec, midvec, bv, vbv, ifmat);
    putdns = true;
    if (checkvec[yidx] >= yidx) // 全変数を選択した時
    {
      if (opt.res("res")) cout << "\n" << endl;
      return;
    }
  }
  set_initb(b0v, vbv, n, smat, opt); // 全変数選択時の初期値設定
  bv = b0v;
  checkvec.assign(dim, 1);      // 全例選択
  checkvec[yidx] = yidx;
  int r =  newton_log(bv, vbv, lb, ifmat, opt, checkvec, opt.rmax());
  if (r < 0)                    // Newton法が発散した時
  {
    bv = b0v;
    newton_log(bv, vbv, lb, ifmat, opt, checkvec, 1); // 初期値を解にする
  }
  if (opt.res("allx") || opt.res("def"))
  {
    put_logreg(cout, opt, n, nr, 0.0, 1.0, r, vrvec, checkvec, b0v, bv, vbv, lb, ifmat);
  }
  if (opt.res("res")) cout << "\n" << endl;
  if (r > 0 && putdns == false)
  {
    if (opt.dnsname() != "") put_dns(opt, checkvec, midvec, bv);
    if (opt.ngpname() != "") plot_logngp(opt, 0.0, 1.0, checkvec, midvec, bv, vbv, ifmat);
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  logregclass logreg;           // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) logreg.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { logreg.get(argv[i], opt);  fok = false; }
    }
    logreg.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
