// mdae.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 08:47:17 tac>
//
// 最初の項目を判別項目、以後の項目を変数とした重判別分析(正準分析)を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W mdae mdae.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "mdae  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:mda  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The multiple discriminant analysis with the first item as discriminant",
  "  item and subsequent items as explanatory variables.",
  "<OPTION>",
  "  -dis={euc|mah}:specify the kind of distance for discrimination",
  "    euc:Euclidean square distance(default)  mah:Mahalanobis pan-distance",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=mdae_dns.d",
  "    output data:canonical scores",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|ca|cs|stat|wl}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    ca:canonical axis  cs:canonical scores  stat:basic statistics",
  "    wl:Wilks's lambda"
};

const string Title = "=== Multiple Discriminant Analysis ==="; // 表題
const string DNSNAMEdef = "mdae_dns.d"; // 省略時DANSファイル名

enum distype                    // 距離の種類
{
  disnon,                       // 指定なし
  diseuc,                       // ユークリッド平方距離
  dismah,                       // Mahalanobisの汎距離
  disetc                        // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  distype vdis;                 // 距離の種類
  string vdnsname;              // DANSファイル名
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                             // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  distype dis() const           // 距離の種類を返す
  { return vdis; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  bool res(const string& code); // 結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vdis(diseuc), vdnsname(), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dis")          // 距離の種類指定
  {
    if (opt[1].empty() || opt[1] == "euc") vdis = diseuc;
    else if (opt[1] == "mah") vdis = dismah;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "項目が2つ以上必要だぜ！");
  if (set_res() < 0) return -3; // -resオプションのエラー
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["ca"] = vresmap["cs"] = vresmap["stat"] = vresmap["wl"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["ca"] = vresmap["cs"] = vresmap["stat"] = vresmap["wl"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all") // 全ての結果を出力する
      {
        vresmap["def"] = vresmap["ca"] = vresmap["cs"] = vresmap["stat"] = vresmap["wl"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群別統計量クラス
{
private:
  mat<ldbl> vsmat;              // 積和行列
  int vn;                       // 例数
  vecldbl vmevec;               // 平均値ベクトル
  vecldbl vvrvec;               // 積和ベクトル
  mat<ldbl> vcsmat;             // 正準スコア積和行列
  mat<ldbl> vcimat;             // 正準スコア共分散逆行列
public:
  gstat() : vsmat(), vn(0), vmevec(), vvrvec(), vcsmat(), vcimat()
  // デフォールトコンストラクタ
  {}
  explicit gstat(size_t p) : vsmat(p + 1), vn(0), vmevec(p + 1), vvrvec(p + 1), vcsmat(1), vcimat(1)
  // コンストラクタ
  {}
  gstat(const gstat& gstat0) : vsmat(gstat0.vsmat), vn(gstat0.vn), vmevec(gstat0.vmevec), vvrvec(gstat0.vvrvec), vcsmat(gstat0.vcsmat), vcimat(gstat0.vcimat)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  void sigma(const vecldbl& fdatavec);         // データを累積する
  int setsmat();                               // 積和行列をセットする
  void calscore(int p, int g, int q, const mat<ldbl>& amat);
  // 群別正準スコアの積和行列を計算する
  const mat<ldbl>& smat() const // 積和行列を返す
  { return vsmat; }
  ldbl smat(int ridx, int cidx) const // 積和行列の要素を返す
  { return vsmat(ridx, cidx); }
  int n() const                 // 例数を返す
  { return vn; }
  const vecldbl& mevec() const  // 平均値ベクトルを返す
  { return vmevec; }
  ldbl me(size_t idx) const     // 平均値を返す
  { return (idx < vmevec.size()) ? vmevec[idx] : 0.0L; }
  const vecldbl& vrvec() const // 分散ベクトルを返す
  { return vvrvec; }
  ldbl vr(size_t idx) const     // 分散を返す
  { return (idx < vvrvec.size()) ? vvrvec[idx] : 0.0L; }
  const mat<ldbl>& csmat() const // 正準スコア積和行列を返す
  { return vcsmat; }
  ldbl csmat(int ridx, int cidx) const // 正準スコア積和行列の要素を返す
  { return vcsmat(ridx, cidx); }
  const mat<ldbl>& cimat() const // 正準スコア共分散逆行列を返す
  { return vcimat; }
  ldbl csm(size_t idx) const    // 正準スコア平均値を返す
  { return (idx < vcsmat.c()) ? vcsmat(0, idx) : 0.0L; }
  ldbl csv(size_t idx) const    // 正準スコア分散を返す
  { return (idx < vcsmat.r() && vn > 1) ? vcsmat(idx, idx) / (ldbl)(vn - 1): 0.0L; }
};

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  vsmat = gstat0.vsmat;  vn = gstat0.vn;
  vmevec = gstat0.vmevec;  vvrvec = gstat0.vvrvec;
  vcsmat = gstat0.vcsmat;  vcsmat = gstat0.vcsmat;
  return *this;
}

void gstat::sigma(const vecldbl& fdatavec) // データを累積する
{
  vmevec[0] = 1.0L;
  for (size_t i = 0; i < fdatavec.size(); ++i) vmevec[i + 1] = fdatavec[i];
  for (size_t i = 0; i < vsmat.r(); ++i)
  {
    double x = vmevec[i];
    for (size_t j = i; j < vsmat.c(); ++j) vsmat.add(i, j, x * vmevec[j]);
  }
}

int gstat::setsmat()            // 積和行列をセットする
{
  vn = (int)(vsmat(0, 0));
  for (size_t i = 0; i < vsmat.r() - 1; ++i)
  { for (size_t j = i + 1; j < vsmat.c(); ++j) vsmat.assign(j, i, vsmat(i, j)); }
  vsmat.sweep(0);
  for (size_t i = 0; i < vsmat.r(); ++i) // 平均と分散をセットする
  {
    vmevec[i] = vsmat(0, i);
    vvrvec[i] = (vn > 1) ? vsmat(i, i) / (ldbl)(vn - 1) : 0.0L;
  }
  return vn;                    // 例数を返す
}

void gstat::calscore(int p, int g, int q, const mat<ldbl>& amat)
// 群別正準スコアの積和行列を計算する
{
  vcsmat.resize(q + 1, q + 1);
  vcsmat.assign(0, 0, (ldbl)(vn)); // 例数
  mat<ldbl> smat(vsmat.sub(1, 1, p, p));
  for (int i = 1; i <= q; ++i)
  {
    ldbl mi = amat(0, i);
    for (int j = 1; j <= p; ++j) mi += vsmat(0, j) * amat(j, i);
    vcsmat.assign(0, i, mi);
    vec<ldbl> aivec(amat.sub(1, i, p, i));
    vec<ldbl> tvec(aivec.trn() * smat);
    for (int j = i; j <= q; ++j)
    {
      vec<ldbl> ajvec(amat.sub(1, j, p, j));
      vcsmat.assign(i, j, inp(tvec, ajvec));
    }
    for (int j = i + 1; j <= q; ++j) vcsmat.assign(j, i, vcsmat(i, j));
  }
  vcimat = vcsmat.sub(1, 1, q, q);
  if (vn > 1) vcimat *= 1.0L / (ldbl)(vn - 1);
  vcimat.inv();                 // 逆行列にする
}

using mapsg = map<string, gstat>;         // 群別統計量マップの別名定義
using mapsg_itr = mapsg::iterator;        // 反復子の別名定義
using mapsg_citr = mapsg::const_iterator; // const反復子の別名定義

class mdaclass                  // mdaメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout); // 解析結果のタイトルとデータ識別情報を出力する
  int set_wbmat(mapsg& gsmap, mat<ldbl>& wmat, mat<ldbl>& bmat);
  // Within(*), Between(*)を計算する
  int cal_mda(int nt, vecldbl& evalvec, mat<ldbl>& amat, mat<ldbl>& wmat, mat<ldbl>& bmat, const mapsg& gsmap);
  // 正準軸を抽出する
  int search_gs(const string& gstr, const mapsg& gsmap) const;
  // 群コードをサーチして群インデクスを返す
  void put_mdadns(option& opt, int nt, int q, const mat<ldbl>& amat, mapsg& gsmap);
  // 正準スコアをデータファイル形式で出力する
  int cal_score(datac& score, datac& cdata, const option& opt, int q, const mat<ldbl>& amat, const mapsg& gsmap);
  // 正準スコアを計算し、データとして返す
  void put_mdatab(ostream& fout, int q, const vecldbl& evalvec) const;
  // 正準軸を出力する
  void put_score(ostream& fout, int nt, int q, const mat<ldbl>& amat, mapsg& gsmap);
  // 群別正準スコアの基礎統計量を出力する
  void put_dis(ostream& fout, const option& opt, int q, const mat<ldbl>& amat, const mapsg& gsmap);
  // 群別判別率を出力する
  void put_wilks(ostream& fout, int nt, int g, int p, int q, const vecldbl& evalvec) const;
  // WilksのΛ統計量を出力する
public:
  mdaclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~mdaclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

void mdaclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "DISCRIMINANT ITEM:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i;
    fout << ":" << id.item(i) << "\n";
  }
}

int mdaclass::set_wbmat(mapsg& gsmap, mat<ldbl>& wmat, mat<ldbl>& bmat)
// gsmap, Within(*), Between(*)を計算する
{
  size_t dim = id.ni(), p = id.ni() - 1;
  vecldbl fdatavec(p);
  for (auto& cdata : datalst)
  {
    bool blank = false;
    for (size_t j = 0; blank == false && j < p; ++j)
    {
      fdatavec[j] = cdata.fdata(j + 1);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (cdata.sdata(0) != BLANKs && blank == false)
    {
      mapsg_itr gitr = gsmap.find(cdata.sdata(0));
      if (gitr == gsmap.end())  // 新規群コード
      {
        gstat gstati(p);
        gstati.sigma(fdatavec);
        gsmap.insert(make_pair(cdata.sdata(0), gstati));
      }
      else                      // すでにある群
      { gitr->second.sigma(fdatavec); }
    }
  }
  int nt = 0;
  for (auto& gmapi : gsmap)
  {
    for (size_t j = 0; j < dim; ++j) bmat.add(0, j, gmapi.second.smat(0, j));
    nt += gmapi.second.setsmat();
    wmat += gmapi.second.smat();
    for (size_t i = 1; i < dim; ++i)
    {
      ldbl sigx = gmapi.second.n() * gmapi.second.me(i);
      for (size_t j = i; j < dim; ++j) bmat.add(i, j, sigx * gmapi.second.me(j));
    }
  }
  wmat.assign(0, 0, bmat(0, 0));
  for (size_t i = 0; i < bmat.r(); ++i)
  {
    for (size_t j = i + 1; j < bmat.c(); ++j) bmat.assign(j, i, bmat(i, j));
  }
  bmat.sweep(0);
  bmat.assign(0, 0, wmat(0, 0));
  for (size_t k = 1; k < dim; ++k) wmat.assign(0, k, bmat(0, k));
  return nt;                    // 総例数を返す
}

int mdaclass::cal_mda(int nt, vecldbl& evalvec, mat<ldbl>& amat, mat<ldbl>& wmat, mat<ldbl>& bmat, const mapsg& gsmap)
// 正準軸を抽出する
{
  int g = gsmap.size();
  if (nt <= g) return -1;       // 計算できず
  ldbl w = sqrt((ldbl)(nt - g));
  int p = id.ni() - 1;
  mat<ldbl> wmatinv = wmat.sub(1, 1, wmat.r() - 1, wmat.c() - 1);
  wmatinv.inv();
  mat<ldbl> wbmat = wmatinv * bmat.sub(1, 1, bmat.r() - 1, bmat.c() - 1);
  evalvec[0] = 0.0L;
  for (int i = 0; i < p; ++i) evalvec[0] += wbmat(i, i); // トレースの計算
  int q = 0, qmax = min(g - 1, p);
  vec<ldbl> axv(p), ayv(p);
  while (q < qmax && (evalvec[q + 1] = eigenud(wbmat, axv, ayv)) > 0.0L)
  {
    ++q;
    vec<ldbl> tv = wmat.sub(1, 1, wmat.r() - 1, wmat.c() - 1) * axv;
    ldbl norm = inp(axv, tv);
    axv *= (norm > 0.0L) ? sqrt(1.0L / norm) : 0.0L;
    amat.assign(0, q, 0.0L);
    for (int i = 1; i <= p; ++i)
    {
      amat.assign(i, q, w * axv(i - 1));
      amat.add(0, q, -amat(i, q) * wmat(0, i));
    }
  }
  return q;                     // 正準軸数を返す
}

int mdaclass::search_gs(const string& gstr, const mapsg& gsmap) const
// 群コードをサーチして群インデクスを返す
{
  mapsg_citr gitr = gsmap.begin();
  int gidx = 0;
  bool hit = false;
  while (hit == false && gitr != gsmap.end())
  {
    if (gstr == gitr->first) hit = true;
    else { ++gitr;  ++gidx; }
  }
  return hit ? gidx : -1;       // 群インデクスを返す
}

void mdaclass::put_mdadns(option& opt, int nt, int q, const mat<ldbl>& amat, mapsg& gsmap)
// 正準スコアをデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  int p = id.ni() - 1, g = gsmap.size();
  idd tid(id.title(), 1 + q + g);
  tid.assign_item(0, id.item(0));
  ostringstream ostr;
  for (int i = 1; i <= q; ++i)
  {
    ostr << "Canonical Score" << i;
    tid.assign_item(i, ostr.str());
    ostr.str("");
  }
  int i = q + 1, gno = 1;
  for (mapsg_itr gitr = gsmap.begin(); gitr != gsmap.end(); ++gitr, ++gno, ++i)
  {
    if (opt.dis() == diseuc) ostr << "Euclidean Square Distance";
    else ostr << "Mahalanobis Pan-distance";
    ostr << " from centre of gravity in GROUP" << gno;
    tid.assign_item(i, ostr.str());
    ostr.str("");
    gitr->second.calscore(p, g, q, amat);
  }
  tid.put_item(fout, formtag);
  datac score(1 + q + g);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    cal_score(score, *ditr, opt, q, amat, gsmap);
    score.put(fout);
  }
}

int mdaclass::cal_score(datac& score, datac& cdata, const option& opt, int q, const mat<ldbl>& amat, const mapsg& gsmap)
// 正準スコアを計算し、データとして返す
{
  int p = id.ni() - 1;
  score.assign(0, cdata.sdata(0));
  vecdbl fdatavec(p, 0.0);
  bool blank = false;
  for (int j = 0; blank == false && j < p; ++j)
  {
    fdatavec[j] = cdata.fdata(j + 1);
    if (fdatavec[j] == BLANKf) blank = true;
  }
  if (cdata.sdata(0) == BLANKs || blank) // スコアの計算ができない
  {
    for (size_t i = 1; i < score.size(); ++i) score.assign(i, BLANKs);
    return 0;                   // スコア計算できず
  }
  for (int i = 1; i <= q; ++i)
  {
    double z = (double)(amat(0, i));
    for (int j = 0; j < p; ++j) z += amat(j + 1, i) * fdatavec[j];
    score.assign(i, z);
  }
  int i = q + 1;
  for (mapsg_citr gitr = gsmap.begin(); gitr != gsmap.end(); ++gitr, ++i)
  {
    vec<ldbl> dvec(q);
    for (int j = 1; j <= q; ++j) dvec.assign(j - 1, score.fdata(j) - gitr->second.csm(j));
    vec<ldbl> dtvec(dvec);
    if (opt.dis() == dismah) dtvec = dvec.trn() * gitr->second.cimat(); // Mahalanobisの汎距離
    score.assign(i, inp(dtvec, dvec));
  }
  return q;                     // 計算できたスコア数を返す
}

void mdaclass::put_mdatab(ostream& fout, int q, const vecldbl& evalvec) const
// 正準軸を出力する
{
  constexpr int LENline = 76;   // 区切り線の長さ
  fout << "\n<MULTIPLE DISCRIMINANT ANALYSIS>\n";
  fout << "C-Axis     Eigen Value     Cumulative EV  ContributionRate     Cumulative CR\n";
  putline(fout, LENline);
  ldbl slam = 0.0L, tr = evalvec[0];
  for (int i = 1; i <= q; ++i)
  {
    ldbl lam = evalvec[i];
    slam += lam;
    fout << "Z" << setw(3) << i << setw(18) << lam << setw(18) << slam;
    fout << setw(18) << lam / tr << setw(18) << slam / tr << "\n";
  }
  putline(fout, LENline);
}

void mdaclass::put_score(ostream& fout, int nt, int q, const mat<ldbl>& amat, mapsg& gsmap)
// 群別正準スコアの基礎統計量を出力する
{
  int p = id.ni() - 1, g = gsmap.size();
  int df = nt - g;
  constexpr int LENline = 77;   // 区切り線の長さ
  for (int i = 1; i <= q; ++i)
  {
    fout << "\n<CANONICAL SCORE:Z" << i << ">\n";
    putline(fout, LENline);
    ldbl ss = (ldbl)df;
    int gno = 1;
    for (mapsg_itr gitr = gsmap.begin(); gitr != gsmap.end(); ++gitr, ++gno)
    {
      gitr->second.calscore(p, g, q, amat);
      ldbl m = gitr->second.csm(i), v = gitr->second.csv(i);
      fout << "GROUP" << setw(2) << gno << ":";
      putmsd(fout, gitr->second.n(), m, v);
      ss += gitr->second.n() * m * m;
    }
    putline(fout, LENline);
    fout << "TOTAL  :";
    putmsd(fout, nt, 0.0, ss / (ldbl)(nt - 1));
  }
}

void mdaclass::put_dis(ostream& fout, const option& opt, int q, const mat<ldbl>& amat, const mapsg& gsmap)
// 判別結果を出力する
{
  int g = gsmap.size();
  datac score(1 + q + g);
  ftable ftab(g, g);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    int gidx = search_gs(ditr->sdata(0), gsmap);
    if (gidx >= 0 && cal_score(score, *ditr, opt, q, amat, gsmap) > 0)
    {
      double dmin = DBL_MAX;
      int midx = 0;
      for (int i = 1 + q; i < 1 + q + g; ++i)
      {
        if (score.fdata(i) < dmin) { dmin = score.fdata(i);  midx = i - 1 - q; }
      }
      ftab.f(gidx + 1, midx + 1, ftab.f(gidx + 1, midx + 1) + 1);
    }
  }
  ostringstream ostr;
  string tstr("GRP\\DIS");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  int wr = 8, wc = 5, tn = 0;
  ostr << left << setw(wr + npad) << tstr << right;
  ftab.hed(ostr.str());
  for (int i = 1; i <= g; ++i)
  {
    ostr.str("");
    tstr = "GROUP";
    npad = tstr.size() - strclen(tstr, CHARwide);
    ostr << setw(wc + npad) << tstr << i;
    ftab.ccode(i, ostr.str());
    ostr.str("");
    ostr << tstr << i;
    ftab.rcode(i, ostr.str());
    tn += ftab.f(i, i);
  }
  ftab.set();
  fout << "\n<DISCRIMINATION RESULTS WITH CANONICAL SCORES>";
  if (opt.dis() == diseuc) fout << "(Euclidean Square Distance)>";
  else fout << "(Mahalanobis pan-distance)>";
  putftable(fout, ftab, tabfprr);
  fout << "Overall Discrimination Rate=" << (double)tn / (double)(ftab.ftotal()) * 100.0 << "%(" << tn << "/" << ftab.ftotal() << ")\n";
}

void mdaclass::put_wilks(ostream& fout, int nt, int g, int p, int q, const vecldbl& evalvec) const
// WilksのΛ統計量を出力する
{
  ldbl lamd = 1.0L, f = 0.0L;
  for (int i = 1; i <= q; ++i) lamd *= (evalvec[i] + 1.0L);
  lamd = 1.0 / lamd;
  int df1 = 0, df2 = 0, w = nt - g;
  if (g == 2)
  { df1 = p;  df2 = w - p + 1;  f = (1.0 - lamd) / lamd * (w - p + 1) / p; }
  else if (g == 3)
  {
    df1 = 2 * p;  df2 = 2 * (w - p + 1);
    f = (1.0 - sqrt(lamd)) / sqrt(lamd) * (w - p + 1) / p;
  }
  else if (p == 2 && g > 3)
  {
    df1 = 2 * (g - 1);  df2 = 2 * (w - 1);
    f = (1.0L - sqrt(lamd)) / sqrt(lamd) * (ldbl)(w - 1) / (ldbl)(g - 1);
  }
  else
  { df1 = (g - 1) * p;  df2 = 0;  f = -((ldbl)w - (ldbl)(p - g) * 0.5L) * log(lamd); }
  fout << "\nWilks's Lambda=" << lamd << "  Correlation Ratio:r=" << sqrt(1.0L - lamd) << "\n";
  if (df2 > 0)
  {
    ldbl pr = ftop(f, df1, df2);
    fout << "  F=" << f << "  DF1=" << df1 << "  DF2=" << df2 << "  p=" << pr << pind(pr) << endl;
  }
  else
  {
    ldbl pr = ctop(f, df1);
    fout << "  chi-squared=" << f << "  DF=" << df1 << "  p=" << pr << pind(pr) << endl;
  }
}

int mdaclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void mdaclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  size_t dim = id.ni();         // 積和行列の次数
  int p = id.ni() - 1;          // 変数の数
  mapsg gsmap;                  // 群統計量マップ
  mat<ldbl> wmat(dim), bmat(dim); // 群内行列と群間行列
  int nt = set_wbmat(gsmap, wmat, bmat); // 総例数
  if (opt.res("res"))
  {
    put_resid(cout);
    if (opt.res("stat"))
    {
      int gno = 1;
      for (mapsg_itr gitr = gsmap.begin(); gitr != gsmap.end(); ++gitr, ++gno)
      {
        cout << "\nGROUP" << gno << ":" << id.item(0) << "=" << gitr->first;
        putmsdvec(cout, gitr->second.n(), gitr->second.mevec(), gitr->second.vrvec());
      }
      mat<ldbl> stmat(wmat + bmat);
      vecldbl mtvec(dim, 0.0L), vtvec(dim, 0.0L);
      for (int i = 0; i <= p; ++i)
      { mtvec[i] = wmat(0, i);  vtvec[i] = stmat(i, i) / (ldbl)(nt - 1); }
      cout << "\nOVERALL";
      putmsdvec(cout, nt, mtvec, vtvec);
      putmat(cout, "<CORRELATION MATIRX(WITHIN)>", matr, nt, wmat);
    }
  }
  int g = gsmap.size();         // 群数
  int qmax = min(g - 1, p);
  vecldbl evalvec(qmax + 1, 0.0L);
  mat<ldbl> amat(dim, qmax + 1);
  int q = cal_mda(nt, evalvec, amat, wmat, bmat, gsmap);
  if (q > 0)                    // 正準軸あり
  {
    if (opt.res("res"))
    {
      if (opt.res("ca"))
      {
        put_mdatab(cout, q, evalvec);
        putcoef(cout, "<CANONICAL COEFFICIENT>", "C-Axis", "X", amat);
        mat<ldbl> samat(dim, q + 1);
        for (int i = 1; i <= p; ++i)
        {
          ldbl sdi = sqrt(wmat(i, i) / (ldbl)(nt - g));
          for (int j = 1; j <= q; ++j) samat.assign(i, j, amat(i, j) * sdi);
        }
        putcoef(cout, "<STANDARD CANONICAL COEFFICIENT>", "C-Axis", "X", samat);
      }
      if (opt.res("cs"))
      {
        put_score(cout, nt, q, amat, gsmap);
        put_dis(cout, opt, q, amat, gsmap);
      }
      if (opt.res("wl")) put_wilks(cout, nt, g, p, q, evalvec);
      cout << "\n" << endl;
    }
    if (opt.dnsname() != "") put_mdadns(opt, nt, q, amat, gsmap);
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  mdaclass mda;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) mda.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { mda.get(argv[i], opt);  fok = false; }
    }
    mda.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
