// mloge.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 08:58:41 tac>
//
// 累積ロジットモデル(比例オッズモデル)による順序ロジスティック回帰分析。
// ロジスティック回帰分析における群項目を順序のある多分類に拡張した手法。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W mloge mloge.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "mloge  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:mlog [<OPTION> ...]  [<DATA FILE> ...]",
  "  The ordinal logistic regression analysis with the first item as criterion",
  "  variable and subsequent items as explanatory variables.",
  "<OPTION>",
  "  -b=<VALUE>[,<VALUE>,...]:specify the initial value of regression",
  "    coefficient  default:auto calculation",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=mloge_dns.d",
  "  -eps[=<EPS>]:specify the convergence error(0<EPS)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={log|score}[,...]:select the data for Ngraph",
  "    log:logistic curve  score:logit scores",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=mloge_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|ltab|olra|stat|v}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1 cc:cc0.95)",
  "    ltab:likelihood table  olra:ordinal logsitic regression analysis",
  "    stat:basic statistics  v:initial values of regression coefficient",
  "  -rmax=<NUMBER>:specify the maximum number of repetitions(0=<NUMBER)"
};

const string Title = "=== Ordinal Logistic Regression Analysis ==="; // 表題
const string DNSNAMEdef = "mloge_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "mloge_ngp.dat"; // 省略時Ngraphファイル名
constexpr double EPSdef = 1.0e-5; // 収束誤差省略時値
constexpr int RMAXdef = 100;      // 繰り返し系算数省略時値

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  lstr blst;                    // 説明変数の初期値リスト
  vecdbl vbvec;                 // 説明変数の初期値ベクトル
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  double veps;                  // 最尤法の収束誤差
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  int vrmax;                    // 最大反復回数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const vecdbl& bvec() const    // 説明変数の初期値ベクトルを返す
  { return vbvec; }
  double b(size_t idx) const    // 説明変数の初期値を返す
  { return (idx < vbvec.size()) ? vbvec[idx] : 0.0; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  double eps() const            // 最尤法の収束誤差を返す
  { return veps; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 最大反復回数を返す
  { return vrmax; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int p);             // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), blst(), vbvec(), vcc(CCdef), vdnsname(), veps(EPSdef), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vrmax(RMAXdef)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "b")            // 偏回帰係数初期値の指定
  {
    if (strsplit(blst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "cc")      // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 最尤法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < DBL_EPSILON) err = badopt(vpname, optstr, USAGE, DBL_EPSILON, DBL_MAX);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 最大反復回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(int p)        // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (blst.empty()) return 0;   // 偏回帰係数初期値の指定なし
  if ((int)(blst.size()) > p) return disperr(vpname, "Too many initial values of the regression coefficients !");
  vbvec.resize(p);
  vbvec.assign(p, BLANKf);
  int i = 0;
  for (lstr_citr bitr = blst.begin(); bitr != blst.end(); ++bitr, ++i)
  {
    double fdata = strtodf(*bitr);
    if (fdata == BLANKf) return disperr(vpname, "Wrong initial value->", *bitr);
    vbvec[i] = fdata;
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["log"] = vngpmap["score"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["log"] = vngpmap["score"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["ltab"] = vresmap["olra"] = vresmap["stat"] = true;
  vresmap["v"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["ltab"] = vresmap["olra"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["ltab"] = vresmap["olra"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["ltab"] = vresmap["olra"] = vresmap["stat"] = vresmap["v"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群別統計量クラス
{
private:
  mat<double> vsmat;            // 積和行列
  int vn;                       // 例数
  double vmy;                   // yの平均値
  double vsyy;                  // yの平方和
  double vsr;                   // 回帰残差
public:
  gstat() : vsmat(), vn(0), vmy(BLANKf), vsyy(0.0), vsr(0.0)
  // デフォールトコンストラクタ
  {}
  explicit gstat(size_t ni) : vsmat(ni + 1), vn(0), vmy(BLANKf), vsyy(0.0), vsr(0.0)
  // コンストラクタ
  {}
  gstat(const gstat& gstat0) : vsmat(gstat0.vsmat), vn(gstat0.vn), vmy(gstat0.vmy), vsyy(gstat0.vsyy), vsr(gstat0.vsr)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  const gstat& operator+=(const gstat& gstat0); // 加算代入演算子
  const mat<double>& smat() const               // 積和行列を返す
  { return vsmat; }
  double smat(int ridx, int cidx) const // 積和行列の要素を返す
  { return vsmat(ridx, cidx); }
  int n() const                 // 例数を返す
  { return vn; }
  double my() const             // yの平均値を返す
  { return vmy; }
  double syy() const            // yの平方和を返す
  { return vsyy; }
  double sr() const             // 回帰残差を返す
  { return vsr; }
  void sigma(const vecdbl& fdata); // データを累積する
  int setsmat(vecdbl& mevec, vecdbl& vrvec);
  // 積和行列・平均値ベクトル・分散ベクトルを設定して例数を返す
};

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  vsmat = gstat0.vsmat;  vn = gstat0.vn;  vmy = gstat0.vmy;
  vsyy = gstat0.vsyy;  vsr = gstat0.vsr;
  return *this;
}

const gstat& gstat::operator+=(const gstat& gstat0) // 加算代入演算子
{
  vsmat += gstat0.vsmat;  vn += gstat0.vn;
  return *this;
}

void gstat::sigma(const vecdbl& fdatavec) // データを累積する
{
  for (size_t i = 0; i < fdatavec.size(); ++i)
  {
    double x = fdatavec[i];
    for (size_t j = i; j < fdatavec.size(); ++j) vsmat.add(i, j, x * fdatavec[j]);
  }
}

int gstat::setsmat(vecdbl& mevec, vecdbl& vrvec)
// 積和行列・平均値ベクトル・分散ベクトルを設定して例数を返す
{
  vn = (int)(vsmat(0, 0));
  for (size_t i = 0; i < vsmat.r() - 1; ++i)
  { for (size_t j = i + 1; j < vsmat.c(); ++j) vsmat.assign(j, i, vsmat(i, j)); }
  vsmat.sweep(0);
  for (size_t i = 0; i < vsmat.r(); ++i)
  {
    mevec[i] = vsmat(0, i);
    vrvec[i] = (vn > 1) ? vsmat(i, i) / (double)(vn - 1) : 0.0;
  }
  vmy = vsmat(0, vsmat.r() -1);
  vsyy = vsr = vsmat(vsmat.r() - 1, vsmat.c() - 1);
  return vn;                    // 例数を返す
}

using mapdg = map<double, gstat>;  // 群別統計量マップの別名定義
using mapdg_itr = mapdg::iterator; // 反復子の別名定義
using mapdg_citr = mapdg::const_iterator; // const反復子の別名定義

class mlogclass                 // mlogメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_stat(mapdg& gsmap, gstat& gstatall, mat<double>& smat0, mat<double>& smat1, vecdbl& mevec, vecdbl& vrvec);
  // データを集計して各種統計量を求める
  void put_resid(ostream& fout, const mapdg& gsmap);
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, mapdg& gsmap); // 群ごとの統計量を出力する
  void set_initb(vec<double>& b0v, option& opt, const mapdg& gsmap, const gstat& gstatall, const mat<double>& smat0, const mat<double>& smat1) const;
  // 重回帰分析→判別分析近似による偏回帰係数初期値のセット
  int newton_mlog(vec<double>& bv, vec<double>& vbv, double& lb, const option& opt, const mapdg& gsmap);
  // Newton法によるパラメーターの計算(最尤法)
  int cal_lgh(vec<double>& gv, mat<double>& hmat, double& lb, const vec<double>& bv, const mapdg& gsmap);
  // 傾斜ベクトルgとHessian行列Hの計算
  int search_index(double yg, const mapdg& gsmap) const;
  // 目的変数データを検索し、インデックスを返す
  void put_mlogreg(ostream& fout, option& opt, int r, int n, const mapdg& gsmap, const vecdbl& vrvec, const vec<double>& b0v, const vec<double>& bv, const vec<double>& vbv, double lb);
  // 順序ロジスティック回帰分析結果を出力する
  void put_mlogtab(ostream& fout, int n, double lb, const mapdg& gsmap);
  // 順序ロジスティック回帰分析の分散分析表を出力する
  int cal_mloglf(double& lf, const mapdg& gsmap); // 飽和モデルの尤度を求める
  void put_brci(ostream& fout, const option& opt, int g, const vec<double>& bv, const vec<double>& vbv) const;
  // 偏回帰係数と相対リスクの信頼区間を出力する
  void put_dns(option& opt, const mapdg& gsmap, const vec<double>& bv);
  // データファイル形式のロジットスコアと確率を出力する
  void put_dnsid(ostream& fout, const mapdg& gsmap);
  // DANSデータのタイトルとデータ識別情報を出力する
  void plot_logngp(option& opt, const mapdg& gsmap, const vec<double>& bv);
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  double cal_zscore(int bidx, int g, const vec<double>& bv, datac& cdata);
  // ロジットスコアを計算して返す
  void plot_score(ofstream& fout, const mapdg& gsmap, const vec<double>& bv);
  // Ngraph用ロジットスコアを出力する
  void plot_log(ofstream& fout, const mapdg& gsmap, const vec<double>& bv);
  // Ngraph用ロジスティック曲線を出力する
public:
  mlogclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~mlogclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int mlogclass::cal_stat(mapdg& gsmap, gstat& gstatall, mat<double>& smat0, mat<double>& smat1, vecdbl& mevec, vecdbl& vrvec)
// データを集計して各種統計量を求める
{
  vecdbl fdatavec(id.ni() + 1, 0.0);
  fdatavec[0] = 1.0;
  for (auto& cdata : datalst)
  {
    bool blank = false;
    for (size_t j = 1; blank == false && j < id.ni(); ++j)
    {
      fdatavec[j] = cdata.fdata(j);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (cdata.fdata(0) == BLANKf || blank)
    { cdata.assign(0, BLANKf); } // 後の集計のために欠損にする
    else
    {
      fdatavec[id.ni()] = cdata.fdata(0);
      mapdg_itr gitr = gsmap.find(cdata.fdata(0));
      if (gitr == gsmap.end())  // 新規群コード
      {
        gstat gstati(id.ni());
        gstati.sigma(fdatavec);
        gsmap.insert(make_pair(cdata.fdata(0), gstati));
      }
      else                      // すでにある群
      { gitr->second.sigma(fdatavec); }
    }
  }
  int i = 1;
  for (auto& gmapi : gsmap)
  {
    gstatall += gmapi.second;
    if (i > 1) smat1 += gmapi.second.smat(); // 2番目以後の単純積和行列を累積する
    gmapi.second.setsmat(mevec, vrvec); // smatを積和行列にする
    if (i == 1) smat0 = gmapi.second.smat(); // 最初の群の積和行列を保存する
    ++i;
  }
  int nt = gstatall.setsmat(mevec, vrvec);
  for (size_t i = 0; i < smat1.r() - 1; ++i)
  {
    for (size_t j = i + 1; j < smat1.c(); ++j) smat1.assign(j, i, smat1(i, j));
  }
  smat1.sweep(0);
  return nt;                    // 総例数を返す
}

void mlogclass::put_resid(ostream& fout, const mapdg& gsmap)
// 解析結果のタイトルとデータ識別情報を出力する
{
  int g = gsmap.size();
  idd tid(id.title(), 2 * g);
  tid.assign_item(0, id.item(0));
  for (int i = 1; i < g; ++i)
  {
    ostringstream ostr;
    ostr << "Logit Score" << i;
    tid.assign_item(i, ostr.str());
  }
  int i = 0;
  for (mapdg_citr gitr = gsmap.begin(); gitr != gsmap.end(); ++gitr, ++i)
  {
    ostringstream ostr;
    ostr << "Probability:" << id.item(0) << "=" << gitr->first;
    tid.assign_item(g + i, ostr.str());
  }
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i;
    fout << ":" << id.item(i) << "\n";
  }
}

void mlogclass::put_stat(ostream& fout, mapdg& gsmap)
// 群ごとの統計量を出力する
{
  size_t dim = id.ni() + 1;       // 次数
  vecdbl meveci(dim, 0.0), vrveci(dim, 0.0);
  int i = 1;
  for (auto& gmapi : gsmap)
  {
    int n = gmapi.second.n();
    for (size_t i = 0; i < dim; ++i)
    {
      meveci[i] = gmapi.second.smat(0, i);
      vrveci[i] = (n > 1) ? gmapi.second.smat(i, i) / (double)(n - 1) : 0.0;
    }
    cout << "\nORDER" << i << ":" << id.item(0) << "=" << gmapi.first;
    putmsdvec(cout, n, meveci, vrveci, id.ni() - 1);
    ++i;
  }
}

void mlogclass::set_initb(vec<double>& b0v, option& opt, const mapdg& gsmap, const gstat& gstatall, const mat<double>& smat0, const mat<double>& smat1) const
// 判別分析による偏回帰係数初期値のセット
{
  mat<double> dmat(smat0 + smat1);
  dmat *= 1.0 / (double)(gstatall.n() - 2);
  dmat.assign(0, 0, 0.0);  dmat.assign(0, id.ni(), 0.0);
  dmat.assign(id.ni(), 0, 0.0);  dmat.assign(id.ni(), id.ni(), 0.0);
  for (size_t i = 1; i < id.ni(); ++i)
  {
    dmat.assign(0, i, (smat0(0, i) + smat1(0, i)) * 0.5);
    dmat.assign(i, 0, -dmat(0, i));
    dmat.assign(i, id.ni(), smat1(0, i) - smat0(0, i));
    dmat.assign(id.ni(), i, dmat(i, id.ni()));
  }
  for (size_t i = 1; i < id.ni(); ++i) dmat.sweep(i);
  int g = gsmap.size();
  mapdg_citr gitr = gsmap.begin();
  int n = gstatall.n(), nk = gitr->second.n();
  double cf = log((double)(n - nk) / (double)nk);
  b0v.assign(0, cf + dmat(0, id.ni()));
  const double Beps = DBL_EPSILON * fmax(FDmin, dmat.maxabse());
  ++gitr;
  for (int i = 1; i < g - 1; ++i, ++gitr)
  {
    nk += gitr->second.n();
    double cfk = log((double)(n - nk) / (double)nk);
    b0v.assign(i, b0v(i - 1) + (cfk - cf));
    if (fabs(b0v(i)) < Beps) b0v.assign(i, 0.0);
    cf = cfk;
  }
  for (size_t i = g - 1; i < b0v.dim(); ++i)
  {
    b0v.assign(i, dmat(i - g + 2, id.ni()));
    if (fabs(b0v(i)) < Beps) b0v.assign(i, 0.0);
  }
  if (opt.bvec().empty()) return; // 初期値指定なし
  for (size_t i = 0; i < b0v.dim(); ++i)
  {
    if (opt.b(i) != BLANKf) b0v.assign(i, opt.b(i));
  }
}

int mlogclass::newton_mlog(vec<double>& bv, vec<double>& vbv, double& lb, const option& opt, const mapdg& gsmap)
// Newton法によるパラメーターの計算(最尤法)
{
  vec<double> gv(bv.dim());
  mat<double> hmat(bv.dim());
  double lold = 0.0;
  bool ok = false;
  int r = 0;
  while (ok == false && cal_lgh(gv, hmat, lb, bv, gsmap) == 0)
  {
    ++r;
    hmat.inv();
    if (fabs(lb - lold) < opt.eps() || (opt.rmax() > 0 && r >= opt.rmax())) ok = true;
    else { lold = lb;  bv -= hmat * gv; }
  }
  if (ok == false) throw runtime_error("NEWTON METHOD DIVERGED !");
  for (size_t i = 0; i < bv.dim(); ++i) vbv.assign(i, -hmat(i, i));
  return r;                     // 反復数を返す
}

int mlogclass::cal_lgh(vec<double>& gv, mat<double>& hmat, double& lb, const vec<double>& bv, const mapdg& gsmap)
// 傾斜ベクトルgとHessian行列Hの計算
{
  constexpr double Emax = 308.0; // 指数最大値
  int p = (int)(bv.dim()), g = (int)(gsmap.size());
  vec<double> xiv(id.ni() - 1);
  gv = 0.0;  hmat = 0.0;  lb = 0.0;
  bool err = false;
  ldata_itr ditr = datalst.begin();
  while (err == false && ditr != datalst.end())
  {
    if (ditr->fdata(0) != BLANKf)
    {
      double bxi = 0.0;
      for (size_t j = 0; j < id.ni() - 1; ++j)
      {
        xiv.assign(j, ditr->fdata(j + 1));
        bxi += bv(g - 1 + j) * xiv(j);
      }
      int k = search_index(ditr->fdata(0), gsmap);
      double bxik1 = (k == 0) ? 0.0 : bv(k - 1) + bxi;
      double bxik = (k == g - 1) ? 0.0 : bv(k) + bxi;
      if (fabs(bv(k)) > Emax || fabs(bxik1) > Emax || fabs(bxik) > Emax)
      { err = true; }
      else
      {
        double eik1 = exp(bxik1), eik = exp(bxik), ei = 0.0, ei2 = 0.0;
        if (k == 0)
        {
          lb -= log(1.0 + eik);
          ei = eik / (1.0 + eik);  ei2 = eik / (1.0 + eik) / (1.0 + eik);
          gv.add(k, -ei);  hmat.add(k, k, -ei2);
          for (int j = g - 1; j < p; ++j) hmat.add(k, j, -xiv(j - g + 1) * ei2);
        }
        else if (k == g - 1)
        {
          lb += bxik1 - log(1.0 + eik1);
          ei = -1.0 / (1.0 + eik1);  ei2 = eik1 / (1.0 + eik1) / (1.0 + eik1);
          gv.add(k - 1, -ei);
          hmat.add(k - 1, k - 1, -ei2);
          for (int j = g - 1; j < p; ++j) hmat.add(k - 1, j, -xiv(j - g + 1) * ei2);
        }
        else
        {
          double d = eik1 / (1.0 + eik1) - eik / (1.0 + eik);
          lb += (d > 0.0) ? log(d) : 0.0;
          d = bv(k - 1) - bv(k);
          gv.add(k - 1, ((d != 0.0) ? 1.0 / (exp(d) - 1.0) : 0.0) + 1.0 / (1.0 + eik1));
          gv.add(k, ((d != 0.0) ? 1.0 / (exp(-d) - 1.0) : 0.0) + 1.0 / (1.0 + eik));
          d = exp(-bv(k - 1)) - exp(-bv(k));
          d = (d != 0.0) ? exp(-bv(k - 1) - bv(k)) / d / d : 0.0;
          hmat.add(k - 1, k - 1, -(d + eik1 / (1.0 + eik1) / (1.0 + eik1)));
          hmat.add(k, k, -(d + eik / (1.0 + eik) / (1.0 + eik)));
          hmat.add(k - 1, k, d);
          for (int j = g - 1; j < p; ++j)
          {
            hmat.add(k - 1, j, -xiv(j - g + 1) * eik1 / (1.0 + eik1) / (1.0 + eik1));
            hmat.add(k, j, -xiv(j - g + 1) * eik / (1.0 + eik) / (1.0 + eik));
          }
          ei = 1.0 - 1.0 / (1.0 + eik1) - 1.0 / (1.0 + eik);
          ei2 = eik1 / (1.0 + eik1) / (1.0 + eik1) + eik / (1.0 + eik) / (1.0 + eik);
        }
        for (int j = g - 1; j < p; ++j)
        {
          double xij = xiv(j - g + 1);
          gv.add(j, -xij * ei);
          for (int l = j; l < p; ++l) hmat.add(j, l, -xij * xiv(l - g + 1) * ei2);
        }
      }
    }
    ++ditr;
  }
  if (err) return -1;
  for (int i = 0; i < p - 1; ++i)
  { for (int j = i + 1; j < p; ++j) hmat.assign(j, i, hmat(i, j)); }
  return 0;
}

int mlogclass::search_index(double yg, const mapdg& gsmap) const
// 目的変数データを検索し、インデックスを返す
{
  mapdg_citr gitr = gsmap.find(yg);
  if (gitr == gsmap.end()) return -1; // データなし
  return (int)(distance(gsmap.begin(), gitr)); // 先頭と要素の位置の差を返す
}

void mlogclass::put_mlogreg(ostream& fout, option& opt, int r, int n, const mapdg& gsmap, const vecdbl& vrvec, const vec<double>& b0v, const vec<double>& bv, const vec<double>& vbv, double lb)
// 順序ロジスティック回帰分析結果を出力する
{
  int g = gsmap.size();
  fout << "\nCumulative Logit Model:pk=1/[1+exp(-b0k-sum{bj xj})]  Repetitions:" << r << "\n";
  fout << "  pk:Cumulative probability of order (k+1) or more\n";
  fout << "  b0k:Intercept of k  bj:Partial Regression Coefficient of xj\n";
  if (opt.res("v"))
  {
    fout << "  Initial Value:";
    for (int i = 0; i < g - 1; ++i)
    {
      if (i > 0)
      {
        if (i % 4 == 0) fout << "\n" << setw(16) << " ";
        else fout << "  ";
      }
      fout << "b0" << i + 1 << "=" << b0v(i);
    }
    for (size_t i = g - 1; i < b0v.dim(); ++i)
    {
      size_t j = i - (g - 1);
      if (j % 4 == 0) fout << "\n" << setw(14) << " ";
      fout << "  b" << i - g + 2 << "=" << b0v(i);
    }
    fout << "\n";
  }
  constexpr int LENline = 79;   // 区切り線の長さ
  fout << "VAR    Partial b     SE of b  Odds Ratio  Standard b  Chi-suared     p-value\n";
  putline(fout, LENline);
  for (int i = 0; i < g - 1; ++i)
  {
    double csqr = (vbv(i) > 0.0) ? bv(i) * bv(i) / vbv(i) : 0.0;
    ldbl pr = ctop(csqr, 1);
    fout << "ICPT" << setw(2) << i + 1 << setw(10) << bv(i);
    fout << setw(12) << sqrt(vbv(i)) << setw(36) << csqr;
    fout << setw(12) << pr << pind(pr) << "\n";
  }
  for (size_t i = g - 1; i < bv.dim(); ++i)
  {
    double csqr = (vbv(i) > 0.0) ? bv(i) * bv(i) / vbv(i) : 0.0;
    double pr = ctop(csqr, 1);
    fout << "X" << setw(3) << i - g + 2 << setw(12) << bv(i);
    fout << setw(12) << sqrt(vbv(i)) << setw(12) << exp(bv(i));
    fout << setw(12) << bv(i) * sqrt(vrvec[i - g + 2]);
    fout << setw(12) << csqr << setw(12) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
  if (opt.res("cc")) put_brci(fout, opt, g, bv, vbv);
  if (opt.res("ltab")) put_mlogtab(fout, n, lb, gsmap);
}

void mlogclass::put_mlogtab(ostream& fout, int n, double lb, const mapdg& gsmap)
// 順序ロジスティック回帰分析の分散分析表を出力する
{
  double l0 = 0.0;
  for (auto& gmapi : gsmap)
  {
    int nk = gmapi.second.n();
    if (nk > 0 && nk < n) l0 += (double)nk * log((double)nk / (double)n);
  }
  int dft = n - 1;
  double lf = 0.0;
  dft -= cal_mloglf(lf, gsmap);
  int dfb = (int)(id.ni()) - 1;
  int dfl = dft - dfb;
  double r2 = 0.0;
  if (sgndiff(lf, l0) != 0) r2 = (lb - l0) / (lf - l0);
  fout << "Log Likelihood:L(Model)=" << lb << "  L(Intercept)=" << l0 << "  Lf=" << lf << "\n";
  fout << "Psedo-Contrivution Rate:R^2=" << r2 << "  AIC=" << 2.0 * (dfb + 1 - lb) << "\n";
  constexpr int LENline = 68;
  const string Head = "Test of Regression and LOF";
  fout << "\n" << setw((LENline - strclen(Head, CHARwide)) / 2 + Head.size()) << Head;
  fout << "\nFACTOR   -(log Likelihood)        DF   Chi-squared       p-value\n";
  putline(fout, LENline);
  double csqr = 2.0 * (lb - l0);
  ldbl pr = ctop(csqr, dfb);
  fout << "REGRESSION" << setw(16) << lb - l0 << setw(10) << dfb;
  fout << setw(14) << csqr << setw(14) << pr << pind(pr) << "\n";
  csqr = 2.0 * (lf - lb);
  const double Ceps = DBL_EPSILON * fmax(FDmin, fabs(lb));
  if (dfl < 1 || csqr < Ceps) csqr = 0.0;
  pr = ctop(csqr, dfl);
  fout << "LOF" << setw(23) << lf - lb << setw(10) << dfl;
  fout << setw(14) << csqr << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  fout << "TOTAL" << setw(21) << lf - l0 << setw(10) << dft << endl;
}

int mlogclass::cal_mloglf(double& lf, const mapdg& gsmap)
// 飽和モデルの尤度を求める
{
  vector<bool> dcvec(datalst.size(), true);
  size_t p = id.ni() - 1, na = gsmap.size();
  vec<double> xiv(p), xjv(p);
  int nlf = 0;
  lf = 0.0;
  size_t dci = 0;
  for (ldata_itr iitr = datalst.begin(); iitr != datalst.end(); ++iitr, ++dci)
  {
    if (dcvec[dci] == true && iitr->fdata(0) != BLANKf)
    {
      int ni = 1;
      vecint nivec(na, 0);
      int k = search_index(iitr->fdata(0), gsmap);
      nivec[k] = 1;
      for (size_t j = 0; j < p; ++j) xiv.assign(j, iitr->fdata(j + 1));
      ldata_itr jitr = iitr;
      ++jitr;
      size_t dcj = dci + 1;
      while (jitr != datalst.end())
      {
        if (dcvec[dcj] == true && jitr->fdata(0) != BLANKf)
        {
          for (size_t j = 0; j < p; ++j) xjv.assign(j, jitr->fdata(j + 1));
          if (xiv == xjv)       // xが等しい
          {
            dcvec[dcj] = false; // 1度しか数えない
            ++ni;
            k = search_index(jitr->fdata(0), gsmap);
            ++(nivec[k]);
          }
        }
        ++jitr;  ++dcj;
      }
      if (ni > 1)
      {
        nlf += ni - 1;
        for (size_t k = 0; k < na; ++k)
        {
          int nk = nivec[k];
          if (nk > 0 && nk < ni) lf += (double)nk * log((double)nk / (double)ni);
        }
      }
    }
  }
  return nlf;                   // 飽和モデルの重複数を返す(自由度減少数)
}

void mlogclass::put_brci(ostream& fout, const option& opt, int g, const vec<double>& bv, const vec<double>& vbv) const
// 偏回帰係数と相対リスクの信頼区間を出力する
{
  constexpr int LENline = 76;   // 区切り線の長さ
  fout << "\nVAR    Partial b" << setw(3) << opt.cc() * 100.0;
  fout << "%CI:Lower       Upper  Odds Ratio" << setw(3) << opt.cc() * 100.0;
  fout << "%CI:Lower       Upper\n";
  putline(fout, LENline);
  double z = ptoz(1.0 - opt.cc());
  for (int i = 0; i < g - 1; ++i)
  {
    double ci = z * sqrt(vbv(i));
    double bl = bv(i) - ci, bu = bv(i) + ci;
    fout << "ICPT" << setw(2) << i + 1 << setw(10) << bv(i);
    fout << setw(12) << bl << setw(12) << bu << "\n";
  }
  for (size_t i = g - 1; i < bv.dim(); ++i)
  {
    double ci = z * sqrt(vbv(i));
    double bl = bv(i) - ci, bu = bv(i) + ci;
    fout << "X" << setw(3) << i - g + 2 << setw(12) << bv(i);
    fout << setw(12) << bl << setw(12) << bu;
    fout << setw(12) << exp(bv(i)) << setw(12) << exp(bl);
    fout << setw(12) << exp(bu) << "\n";
  }
  putline(fout, LENline);
}

void mlogclass::put_dns(option& opt, const mapdg& gsmap, const vec<double>& bv)
// データファイル形式のロジットスコアと反応確率を出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout, gsmap);
  int g = gsmap.size();
  datac cdata(2 * g);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double y = ditr->fdata(0);
    cdata.assign(0, y);
    if (y == BLANKf)
    { for (int i = 1; i < 2 * g; ++i) cdata.assign(i, BLANKs); }
    else
    {
      for (int i = 1; i < g; ++i) cdata.assign(i, cal_zscore(i - 1, g, bv, *ditr));
      double zk = cdata.fdata(1);
      if (zk == BLANKf)
      { for (int k = 0; k < g; ++k) cdata.assign(g + k, BLANKf); }
      else
      {
        double pk = 1.0 / (1.0 + exp(-zk));
        cdata.assign(g, 1.0 - pk);
        double ppre = pk;
        for (int k = 2; k < g; ++k)
        {
          zk = cdata.fdata(k);
          pk = 1.0 / (1.0 + exp(-zk));
          cdata.assign(g + k - 1, ppre - pk);
          ppre = pk;
        }
        cdata.assign(2 * g - 1, pk);
      }
    }
    cdata.put(fout);
  }
}

void mlogclass::put_dnsid(ostream& fout, const mapdg& gsmap)
// DANSデータのタイトルとデータ識別情報を出力する
{
  int g = gsmap.size();
  idd tid(id.title(), 2 * g);
  tid.assign_item(0, id.item(0));
  for (int i = 1; i < g; ++i)
  {
    ostringstream ostr;
    ostr << "Logit Score" << i;
    tid.assign_item(i, ostr.str());
  }
  int i = 0;
  for (mapdg_citr gitr = gsmap.begin(); gitr != gsmap.end(); ++gitr, ++i)
  {
    ostringstream ostr;
    ostr << "Probability:" << id.item(0) << "=" << gitr->first;
    tid.assign_item(g + i, ostr.str());
  }
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  tid.put_item(fout, formtag);
}

void mlogclass::plot_logngp(option& opt, const mapdg& gsmap, const vec<double>& bv)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("log"))
  {
    fout << "-s1 -dline -w80 -c255,0,0 -zy-0.1,1.1,0.2\n";
    plot_ngpid(fout);
    plot_log(fout, gsmap, bv);
    delml = true;
  }
  if (opt.ngp("score"))
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dmark,00 -o300 -zy-0.1,1.1,0.2\n";
    plot_ngpid(fout);
    plot_score(fout, gsmap, bv);
  }
}

void mlogclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  { fout << "%EXPLANATORY VARIABLE X" << setw(3) << i << ":" << id.item(i) << "\n"; }
}

double mlogclass::cal_zscore(int bidx, int g, const vec<double>& bv, datac& cdata)
// ロジットスコアを計算して返す
{
  double z = bv(bidx);
  bool blank = false;
  for (size_t i = 1; blank == false && i < id.ni(); ++i)
  {
    double x = cdata.fdata(i);
    if (x == BLANKf) blank = true;
    else z += bv(i + g - 2) * x;
  }
  return blank ? BLANKf : z;    // 計算できたらzを返す
}

void mlogclass::plot_score(ofstream& fout, const mapdg& gsmap, const vec<double>& bv)
// Ngraph用ロジットスコアを出力する
// プロット用に本来のロジットスコアの定数を定数1に固定する
{
  int g = gsmap.size();
  for (int i = 1; i < g; ++i)
  {
    if (i > 1) fout << NGPdelm << "\n";
    fout << "%<LOGIT SCORE" << i << ">\n";
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
    {
      double y = ditr->fdata(0);
      if (y != BLANKf)
      {
        int k = search_index(y, gsmap);
        y = (k < i) ? 0.0 : 1.0;
        double z = cal_zscore(0, g, bv, *ditr);
        if (z != BLANKf) fout << z << " " << y << "\n";
      }
    }
  }
}

void mlogclass::plot_log(ofstream& fout, const mapdg& gsmap, const vec<double>& bv)
// Ngraph用ロジスティック曲線を出力する
// プロット用にロジットスコアの定数を定数1に固定し、x座標を定数の差分だけずらす
{
  constexpr int RDdef = 50;     // ロジスティック曲線の分解数
  double z = 0.0, zmin = -5.0, zmax = 5.0;
  int g = gsmap.size();
  for (auto& cdata : datalst)
  {
    double y = cdata.fdata(0);
    if (y != BLANKf)
    {
      z = cal_zscore(0, g, bv, cdata);
      if (z != BLANKf)
      {
        if (z < zmin) zmin = z;
        if (z > zmax) zmax = z;
      }
    }
  }
  double xdiff = bv(0) - bv(g - 2);
  if (xdiff > 0.0) zmax += xdiff;
  else zmin += xdiff;
  for (int i = 1; i < g; ++i)
  {
    if (i > 1) fout << NGPdelm << "\n";
    fout << "%<LOGISTIC CURVE" << i << ">\n";
    double range = zmax - zmin;
    double xstart = zmin - range * 0.1, xend = zmax + range * 0.1;
    double dx = (xend - xstart) / (double)RDdef;
    xdiff = bv(0) - bv(i + g - 4);
    double x = xstart;
    for (int i = 0; i <= RDdef; ++i, x += dx)
    {
      double pr = 1.0 / (1.0 + exp(-x));
      fout << x + xdiff << " " << pr << "\n";
    }
  }
}

int mlogclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (id.ni() < 2) throw runtime_error("At least 2 items are requied !");
  return nc;
}

void mlogclass::put(option& opt) // 集計結果を出力する
{
  size_t dim = id.ni() + 1;       // 行列の次数
  mapdg gsmap;                    // 群統計量マップ
  gstat gstatall(id.ni());        // 全体の統計量
  mat<double> smat0(dim), smat1(dim); // 判別分析用積和行列
  vecdbl mevec(dim, 0.0), vrvec(dim, 0.0);
  int nt = cal_stat(gsmap, gstatall, smat0, smat1, mevec, vrvec);
  if (nt < 1)                   // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, gsmap);
      cout << "\n" << endl;
    }
    return;
  }
  int g = gsmap.size();         // 群数
  int p = id.ni() + g - 2;      // 初期値の数(変数+群ごとの定数-1)
  if (opt.check(p) != 0) throw option_error("Run-time option error !");
  if (opt.res("res"))
  {
    put_resid(cout, gsmap);
    if (opt.res("stat")) put_stat(cout, gsmap);
  }
  if (opt.res("stat"))
  {
    cout << "\nOVERALL";
    putmsdvec(cout, nt, mevec, vrvec, id.ni() - 1, 1);
    putmat(cout, "<CORRELATION COEFFICIENT MATRIX>", matr, nt, gstatall.smat(), id.ni() - 1, 1);
    if (nt < p || gstatall.syy() <= 0.0) // 回帰式計算できず
    {
      cout << "\n" << endl;
      return;
    }
  }
  double lb = 0.0;
  vec<double> b0v(p), bv(p), vbv(p);
  set_initb(b0v, opt, gsmap, gstatall, smat0, smat1);
  bv = b0v;
  int r = newton_mlog(bv, vbv, lb, opt, gsmap);
  if (opt.res("res"))
  {
    if (opt.res("olra")) put_mlogreg(cout, opt, r, nt, gsmap, vrvec, b0v, bv, vbv, lb);
    cout << "\n" << endl;
  }
  if (opt.dnsname() != "") put_dns(opt, gsmap, bv);
  if (opt.ngpname() != "") plot_logngp(opt, gsmap, bv);
}

int main(int argc, char* argv[]) // メインルーチン
{
  mlogclass mlog;               // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) mlog.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { mlog.get(argv[i], opt);  fok = false; }
    }
    mlog.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
