// mracve.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 09:03:05 tac>
//
// 最初の項目を目的変数、以後の項目を説明変数とした重回帰分析の交差検証を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W mracve mracve.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "mracve  Version 7.41e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:mracv [<OPTION> ...] [<DATA FILE> ...]",
  "  The cross validation of multiple regression analysis with the first item",
  "  as criterion variable and subsequent items as explanatory variables.",
  "<OPTION>",
  "  -cc[=<C-COEF>]:specify the confidence coefficient(0=<C-COEF<1)",
  "    default:-cc=0.5",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -r=<NUMBER>:specify the number of cases in validation group"
};

constexpr double CCmra = 0.5;   // 省略時信頼係数(CCdefとは別の値)

using lditr = list<ldata_itr>; // データリスト反復子リストの短縮名定義
using lditr_itr = list<ldata_itr>::iterator; // 反復子リスト反復子の短縮名定義
using ldvec = list<vecdbl>;     // ベクトルデータの短縮名定義
using ldvec_itr = list<vecdbl>::iterator; // 反復子の短縮名定義

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  int vr;                       // 検証群のデータ数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  int r() const                 // 検証群データ数を返す
  { return vr; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCmra), vr(1)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数
  {
    vcc = opt[1].empty() ? CCmra : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "r")       // 検証群データ数の指定
  {
    char* ptr = nullptr;
    if (opt[1].empty()) vr = 1;
    else vr = (int)(strtol(opt[1].c_str(), &ptr, 10));
    if (vr < 0) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

class mracvclass                // mracvメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_mra(mat<double>& smat, double& syy, lditr& lditrlstt);
  // 重回帰分析の計算をして総例数を返す
  int cv_mra(const option& opt, mat<double>& smat, int dfr, lditr& lditrlstv);
  // 重回帰分析結果でデータリストを検証してyが予測限界内に入った例数を返す
  double cal_yest(datac& cdata, mat<double>& smat) const;
  // y推定値を計算して返す
  double cal_pi(datac& cdata, mat<double>& smat, double vr, double tcc) const;
  // yの予測限界幅を計算して返す
  int set_cvdlst(ldvec& cvdveclst, int vno, int ny, int yfit, mat<double>& smat, double syy) const;
  // 検証結果を検証結果リストに追加して要素数を返す
  int cal_cv(const option& opt, ldvec& cvdveclst, int r, int k);
  // 重回帰分析の交差検証を行う
  void put_cv(ostream& fout, const option& opt, ldvec& cvdveclst);
  // 重回帰分析の交差検証結果をデータファイル形式で出力する
public:
  mracvclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~mracvclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int mracvclass::cal_mra(mat<double>& smat, double& syy, lditr& lditrlstt)
// 重回帰分析の計算をして総例数を返す
{
  size_t ni = id.ni();
  size_t d = ni + 1;
  vecdbl fdatavec(d, 0.0);
  fdatavec[0] = 1.0;
  for (lditr_itr lditr = lditrlstt.begin(); lditr != lditrlstt.end(); ++lditr)
  {
    for (size_t i = 1; i < ni; ++i) fdatavec[i] = (*lditr)->fdata(i);
    fdatavec[ni] = (*lditr)->fdata(0); // y
    for (size_t i = 0; i < d; ++i)
    {
      double xyi = fdatavec[i];
      for (size_t j = i; j < d; ++j) smat.add(i, j, xyi * fdatavec[j]);
    }
  }
  int nt = (int)(smat(0, 0));
  if (nt < 1) return 0;         // データなし
  for (size_t i = 0; i < ni; ++i)
  {
    for (size_t j = i + 1; j < d; ++j) smat.assign(j, i, smat(i, j));
  }
  smat.sweep(0);                // 単純積和行列→積和行列
  syy = smat(ni, ni);
  for (size_t i = 1; i < ni; ++i) smat.sweep(i); // 行列を掃き出す
  return nt;                    // 総例数を返す
}

int mracvclass::cv_mra(const option& opt, mat<double>& smat, int dfr, lditr& lditrlstv)
// 重回帰分析結果でデータリストを検証してyが予測限界内に入った例数を返す
{
  int ni = smat.r() - 1;
  double vr = (dfr > 0) ? smat(ni, ni) / (double)dfr : 0.0;
  double tcc = ptot(1.0 - opt.cc(), dfr);
  int yfit = 0;
  for (lditr_itr lditr = lditrlstv.begin(); lditr != lditrlstv.end(); ++lditr)
  {
    double y = (*lditr)->fdata(0);
    double yest = cal_yest(*(*lditr), smat);
    double pi = cal_pi(*(*lditr), smat, vr, tcc);
    double ylow = yest - pi, yup = yest + pi;
    if (ylow <= y && y <= yup) ++yfit;
  }
  return yfit;                  // yが予測限界内に入った例数を返す
}

double mracvclass::cal_yest(datac& cdata, mat<double>& smat) const
// y推定値を計算して返す
{
  size_t ni = smat.r() - 1;
  double yest = smat(0, ni);
  for (size_t i = 1; i < ni; ++i) yest += smat(i, ni) * cdata.fdata(i);
  return yest;                  // y推定値を返す
}

double mracvclass::cal_pi(datac& cdata, mat<double>& smat, double vr, double tcc) const
// yの予測限界幅を計算して返す
{
  size_t d = smat.r();
  size_t ni = d - 1;
  vec<double> xv(d);
  for (size_t i = 1; i < ni; ++i) xv.assign(i, cdata.fdata(i));
  xv.assign(0, 1.0);  xv.assign(ni, 0.0);
  vec<double> tv(xv.trn() * smat);
  double vy = vr * (1.0 + inp(tv, xv));
  return tcc * sqrt(vy);        // yの予測限界幅を返す
}

int mracvclass::set_cvdlst(ldvec& cvdveclst, int vno, int ny, int yfit, mat<double>& smat, double syy) const
// 検証結果を検証結果リストに追加して要素数を返す
{
  size_t ni = id.ni();
  size_t nd = ni + 4;
  vector<double> cvdvec(nd, 0.0);
  cvdvec[0] = (double)((vno < 1) ? ni + 1 : vno);
  for (size_t i = 0; i < ni; ++i) cvdvec[i + 1] = smat(i, ni);
  cvdvec[ni + 1] = 1.0 - smat(ni, ni) / syy;
  cvdvec[ni + 2] = (double)ny;
  cvdvec[ni + 3] = (double)yfit;
  cvdveclst.push_back(cvdvec);
  return nd;                    // 要素数を返す
}

int mracvclass::cal_cv(const option& opt, ldvec& cvdveclst, int r, int k)
// 重回帰分析の交差検証を行う
{
  int ni = id.ni();
  int d = ni + 1;
  lditr lditrlstt, lditrlstv; // トレーニングデータとバリデーションデータ反復子
  for (int i = 0; i < k; ++i)
  {
    int sidx = r * i, eidx = r * (i + 1) - 1;
    int j = 0;
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr, ++j)
    {
      if (sidx <= j && j <= eidx) lditrlstv.push_back(ditr);
      else lditrlstt.push_back(ditr);
    }
    mat<double> smat(d, d);     // 積和行列
    double syy = 0.0;           // yの平方和
    int nt = cal_mra(smat, syy, lditrlstt);
    if (nt > 0)
    {
      int yfit = cv_mra(opt, smat, nt - ni, lditrlstv);
      int ny = lditrlstv.size();
      set_cvdlst(cvdveclst, i + 1, ny, yfit, smat, syy);
    }
    lditrlstt.clear();  lditrlstv.clear();
  }
  return k;                     // 検証回数を返す
}

void mracvclass::put_cv(ostream& fout, const option& opt, ldvec& cvdveclst)
// 重回帰分析の交差検証結果をデータファイル形式で出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  size_t ni = id.ni();
  idd tid(id.title(), ni + 4);
  tid.assign_item(0, "Number of Parameters/Validation Group Number");
  tid.assign_item(1, "b0:Intercept");
  for (size_t i = 1; i < ni; ++i)
  {
    ostringstream ostr;
    ostr << "b" << i << ":" << id.item(i);
    tid.assign_item(1 + i, ostr.str());
  }
  ostringstream ostr;
  ostr << "R-squared of " << id.item(0);
  tid.assign_item(ni + 1, ostr.str());
  tid.assign_item(ni + 2, "Validation Group Number");
  ostr.str("");
  ostr << "Number of Cases included in the " << opt.cc() * 100.0 << "% Prediction Limit";
  tid.assign_item(ni + 3, ostr.str());
  tid.put_item(fout, formtag);
  int prec0 = fout.precision(); // 現在の浮動小数点精度(有効数字)を保存する
  for (ldvec_itr cvitr = cvdveclst.begin(); cvitr != cvdveclst.end(); ++cvitr)
  {
    fout << (*cvitr)[0];
    fout.precision(16);
    for (size_t i = 1; i <= ni + 1; ++i) fout << CHRdelm << (*cvitr)[i];
    fout.precision(prec0);
    for (size_t i = ni + 2; i < tid.ni(); ++i) fout << CHRdelm << (*cvitr)[i];
    fout << endl;
  }
}

int mracvclass::get(const char* c_fname, option& opt) // データを読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (id.ni() < 2) throw runtime_error("at least 2 items are requied !");
  nc = id.nc();                 // ケース数変更用
  ldata_itr ditr = datalst.begin();
  while (ditr != datalst.end())
  {
    bool blank = false;
    for (size_t j = 0; blank == false && j < id.ni(); ++j)
    {
      if (ditr->fdata(j) == BLANKf) blank = true;
    }
    if (blank)                  // ケースを削除する
    { ditr = datalst.erase(ditr);  --nc; }
    else
    { ++ditr; }
  }
  if (nc < (int)(id.nc())) id.nc(nc);  // ケース数を減らす
  return nc;                           // 有効なケース数を返す
}

void mracvclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1) return;      // 有効なデータなし
  int ni = id.ni();
  size_t dim = id.ni() + 1;
  mat<double> smat(dim);        // 積和行列
  double syy = 0.0;             // yの平方和
  lditr lditrlst;               // データリスト反復子リスト
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr) lditrlst.push_back(ditr);
  int nt = cal_mra(smat, syy, lditrlst);
  int yfit0 = cv_mra(opt, smat, nt - ni, lditrlst); // yが予測限界内に入った例数
  lditrlst.clear();
  int r = opt.r();              // 検証群のデータ数
  if (r < 1 || nt - r < ni) return; // 交差検証実行せず
  int k = nt / r;               // 検証群数
  if (nt % r > 0) ++k;          // 切り上げる
  ldvec cvdveclst;              // 交差検証法の結果リスト
  set_cvdlst(cvdveclst, 0, nt, yfit0, smat, syy);
  if (cal_cv(opt, cvdveclst, r, k) > 0) put_cv(cout, opt, cvdveclst);
}

int main(int argc, char* argv[]) // メインルーチン
{
  mracvclass mracv;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) mracv.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { mracv.get(argv[i], opt);  fok = false; }
    }
    mracv.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
