// pcae.C Version 7.41e   Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 09:15:43 tac>
//
// 指定した項目について主成分分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W pcae pcae.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "pcae  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:pcae [<OPTION> ...] [<DATA FILE> ...]",
  "  The principal component analysis based on correlation, covariance,",
  "  and sum-of-products matrices.",
  "<OPTION>",
  "  -c=<CCR>:specify cumulative contribution rate(0=<CCR=<1)",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=pcae_dns.d",
  "    output data:principal component scores",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -m={r|s|v|x}:select the type of matrix to start calculating",
  "    r:correation matrix(default)  s:SS matrix",
  "    v:covariance matrix x:simple SS matrix",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=pcae_ngp.dat",
  "    output data:principal component loadings",
  "  -q=<MN>:specify the maximum number of principal components",
  "    (0<MN=<number of variables)",
  "  -res=[-][{def|all|pc|pcc|pcl|stat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    pc:principal component  pcc:coefficient for principal component",
  "    pcl:principal component lodings  stat:basic statistics",
  "  -std[={-|+}]:specify whether to standardize principal component scores",
  "    (-std:std=+)"
};

const string Title = "=== Principal Component Analysis ==="; // 表題
const string DNSNAMEdef = "pcae_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "pcae_ngp.dat"; // 省略時Ngraphファイル名
constexpr double Cdef = 0.8;    // 省略時累積寄与率

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vc;                    // 累積寄与率基準値
  string vdnsname;              // DANSファイル名
  mattype vm;                   // 計算開始行列の種類
  string vngpname;              // Ngraphファイル名
  int vq;                       // 主成分抽出用最大数
  lstr vreslst;                 // 解析結果出力リスト
  map<string, bool> vresmap;    // 解析結果出力の有無マップ
  bool vstd;                    // 主成分スコアの標準化フラッグ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double c() const              // 累積寄与率基準値を返す
  { return vc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  mattype m() const             // 計算開始行列の種類を返す
  { return vm; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  int q() const                 // 主成分抽出用最大数を返す
  { return vq; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  bool std() const              // 主成分スコアの標準化フラッグを返す
  { return vstd; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vc(Cdef), vdnsname(), vm(matr), vngpname(), vq(0), vreslst(), vresmap(), vstd(true)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "c")            // 累積寄与率基準値の指定
  {
    vc = opt[1].empty() ? Cdef : strtodf(opt[1]);
    if (vc < 0.0 || 1.0 < vc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "m")       // 計算開始行列指定
  {
    if (opt[1].empty() || opt[1] == "r") vm = matr;
    else if (opt[1] == "s") vm = mats;
    else if (opt[1] == "v") vm = matv;
    else if (opt[1] == "x") vm = matetc;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "q")       // 最大主成分数指定
  {
    char* ptr = nullptr;
    vq = strtol(opt[1].c_str(), &ptr, 0);
    if (vq < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "std")     // 主成分スコアの標準化指定
  {
    if (opt[1].empty() || opt[1] == "+") vstd = true;
    else if (opt[1] == "-") vstd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if (set_res() < 0) return -3; // -resオプションのエラー
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["pc"] = vresmap["pcc"] = vresmap["pcl"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["pc"] = vresmap["pcc"] = vresmap["pcl"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["pc"] = vresmap["pcc"] = vresmap["pcl"] = vresmap["stat"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class pcaclass                  // pcaメインクラス
{
private:
  string fname;                 // ファイル名
  int cal_smat(const option& opt, const idd& id, ldata& datalst, mat<ldbl>& smat, vecldbl& mevec, vecldbl& vrvec) const;
  // データを集計して積和行列を求める
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void plot_ngpid(ostream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  int cal_pca(const option& opt, const idd& id, vecldbl& evalvec, mat<ldbl>& amat, int n, const mat<ldbl>& smat) const;
  // 主成分を抽出する
  void put_pcadns(const option& opt, idd& id, ldata& datalst, int n, const mat<ldbl>& smat, const vecldbl& evalvec, const mat<ldbl>& amat) const;
  // 主成分スコアをデータファイル形式で出力する
  void amat2cmat(const option& opt, const idd& id, mat<ldbl>& amat, int n, const mat<ldbl>& smat, const vecldbl& evalvec) const;
  // 固有ベクトルを主成分計算用係数に変換する
  void plot_plplot(const option& opt, idd& id, int n, const mat<ldbl>& smat, const vecldbl& evalvec, const mat<ldbl>& amat) const;
  // Ngraph用主成分負荷量のプロットデータを出力する
  void amat2lmat(const option& opt, mat<ldbl>& amat, int n, const mat<ldbl>& smat, const vecldbl& evalvec) const;
  // 固有ベクトルを主成分負荷量に変換する
  void put_pca(ostream& fout, option& opt, idd& id, int n, const mat<ldbl>& smat, const vecldbl& evalvec, const mat<ldbl>& amat) const;
  // 主成分計算結果を出力する
  void put_pload(ostream& fout, const mat<ldbl>& lmat) const;
  // 主成分負荷量を出力する。
public:
  pcaclass() : fname()          // コンストラクタ
  {}
  ~pcaclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

int pcaclass::cal_smat(const option& opt, const idd& id, ldata& datalst, mat<ldbl>& smat, vecldbl& mevec, vecldbl& vrvec) const
// データを集計して積和行列を求める
{
  size_t dim = id.ni() + 1;
  vecdbl fdatavec(dim, 0.0);
  fdatavec[0] = 1.0;
  for (auto& cdata : datalst)
  {
    bool blank = false;
    for (size_t j = 1; blank == false && j < dim; ++j)
    {
      fdatavec[j] = cdata.fdata(j - 1);
      if (fdatavec[j] == BLANKf) blank = true;
    }
    if (blank)
    { cdata.sdata(0, BLANKs); } // 後の計算のために最初の項目をブランクにする
    else
    {
      for (size_t i = 0; i < dim; ++i)
      {
        double x = fdatavec[i];
        for (size_t j = i; j < dim; ++j) smat.add(i, j, (ldbl)(x * fdatavec[j]));
      }
    }
  }
  for (size_t i = 0; i < dim - 1; ++i)
  { for (size_t j = i + 1; j < dim; ++j) smat.assign(j, i, smat(i, j)); }
  int n = (int)(smat(0, 0));
  mat<ldbl> amat(smat);         // 一時行列
  amat.sweep(0);                // 積和行列にする
  for (size_t i = 0; i < dim; ++i)
  {
    mevec[i] = amat(0, i);
    vrvec[i] = (n > 1) ? amat(i, i) / (ldbl)(n - 1) : 0.0L;
  }
  if (opt.m() != matetc) smat.sweep(0);
  return n;                     // 例数を返す
}

void pcaclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout,  Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  { fout << "X" << setw(3) << i + 1 << ":" << id.item(i) << "\n"; }
}

int pcaclass::cal_pca(const option& opt, const idd& id, vecldbl& evalvec, mat<ldbl>& amat, int n, const mat<ldbl>& smat) const
// 主成分を抽出する
{
  int p = id.ni();
  ldbl w = (n > 1) ? 1.0L / (ldbl)(n - 1) : 0.0L;
  mat<ldbl> tmat(smat.sub(1, 1, p, p));
  switch (opt.m())
  {
  case matr : smat2rmat(tmat);
    break;
  case matv : tmat *= w;
  }
  evalvec[0] = 0.0;             // 計算開始行列のトレース(寄与率の分母)
  for (int i = 0; i < p; ++i) evalvec[0] += tmat(i, i);
  ldbl seval = 0.0L;            // 累積寄与率
  int q = 0;                    // 主成分数
  vec<ldbl> evecv(p);           // 固有ベクトル
  int qmax = (opt.q() > 0) ? opt.q() : p;
  while (seval / evalvec[0] < (ldbl)(opt.c()) && q < qmax && (evalvec[q + 1] = eigenud(tmat, evecv)) > 0.0L)
  {
    ++q;
    for (int i = 0; i < p; ++i) amat.assign(i + 1, q, evecv(i));
    seval += evalvec[q];
  }
  return q;                     // 抽出した主成分数を返す
}

void pcaclass::put_pcadns(const option& opt, idd& id, ldata& datalst, int n, const mat<ldbl>& smat, const vecldbl& evalvec, const mat<ldbl>& amat) const
// 主成分スコアをデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  int q = amat.c() - 1;
  idd tid(id.title(), q);
  ostringstream ostr;
  for (int i = 0; i < q; ++i)
  {
    ostr << "Principal Componet Score " << i + 1;
    tid.assign_item(i, ostr.str());
    ostr.str("");
  }
  tid.put_item(fout, formtag);
  mat<ldbl> cmat(amat);
  amat2cmat(opt, id, cmat, n, smat, evalvec);
  datac score(q);
  for (auto& cdata : datalst)
  {
    if (cdata.sdata(0) == BLANKs)
    { score = BLANKs; }
    else
    {
      vecdbl fdatavec(id.ni(), 0.0);
      for (size_t i = 0; i < id.ni(); ++i) fdatavec[i] = cdata.fdata(i);
      for (size_t i = 1; i <= score.size(); ++i)
      {
        double z = (double)(cmat(0, i));
        for (size_t j = 0; j < id.ni(); ++j) z += cmat(j + 1, i) * fdatavec[j];
        score.assign(i - 1, z);
      }
    }
    score.put(fout);
  }
}

void pcaclass::amat2cmat(const option& opt, const idd& id, mat<ldbl>& amat, int n, const mat<ldbl>& smat, const vecldbl& evalvec) const
// 固有ベクトルを主成分計算用係数に変換する
{
  int df = n - 1, p = id.ni(), q = amat.c() - 1;
  for (int j = 1; j <= q; ++j)
  {
    ldbl a0 = 0.0, lamda = evalvec[j];
    for (int i = 1; i <= p; ++i)
    {
      ldbl w = 1.0L;            // 共分散行列の時
      switch (opt.m())
      {
      case matr : w = (smat(i, i) > 0.0L) ? (ldbl)df / smat(i, i) : 0.0L;
        break;
      case mats : w = (ldbl)df;
        break;
      case matetc : w = (ldbl)n;
      }
      if (opt.std()) amat.assign(i, j, amat(i, j) * sqrt(w / lamda));
      else amat.assign(i, j, amat(i, j) * sqrt(w));
      a0 -= amat(i, j) * smat(0, i);
    }
    amat.assign(0, j, a0);
  }
}

void pcaclass::plot_plplot(const option& opt, idd& id, int n, const mat<ldbl>& smat, const vecldbl& evalvec, const mat<ldbl>& amat) const
// Ngraph用主成分負荷量のプロットデータを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  constexpr int RD = 60;        // 円の分解数
  mat<ldbl> lmat(amat);
  amat2lmat(opt, lmat, n, smat, evalvec); // 固有ベクトルを主成分負荷量に変換する
  fout << "-s1 -zx-1,1,0.5 -zy-1,1,0.5\n";
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  for (size_t i = 0; i < id.ni(); ++i)
  { fout << "%x" << setw(3) << i + 1 << ":" << id.item(i) << "\n"; }
  fout << "%When drawing a circle:-dline, when drawing a plot:-dmark,00 -o300\n";
  double dh = 2.0 * M_PI / (double)RD;
  double h = 0.0;
  for (int i = 0; i <= RD; ++i, h += dh) fout << cos(h) << " " <<  sin(h) << "\n";
  int p = lmat.r() - 1, q = lmat.c() - 1;
  for (int j = 1; j < q; ++j)
  {
    for (int k = j + 1; k <= q; ++k)
    {
      fout << NGPdelm << "\n";
      fout << "%<X:Z" << j << "  Y:Z" << k << ">\n";
      for (int i = 1; i <= p; ++i) fout << lmat(i, j) << " " << lmat(i, k) << "\n";
    }
  }
}

void pcaclass::amat2lmat(const option& opt, mat<ldbl>& amat, int n, const mat<ldbl>& smat, const vecldbl& evalvec) const
// 固有ベクトルを主成分負荷量に変換する
{
  int df = n - 1, p = amat.r() - 1, q = amat.c() - 1;
  for (int i = 1; i <= p; ++i)
  {
    ldbl v = 1.0L;              // 相関行列の場合
    switch (opt.m())
    {
    case matv : v = (df > 0) ? smat(i, i) / (ldbl)df : 0.0L;
      break;
    case matetc :
    case mats : v = smat(i, i);
    }
    if (v <= 0.0L)
    { for (int j = 1; j <= q; ++j) amat.assign(i, j, 0.0); }
    else
    {
      for (int j = 1; j <= q; ++j) amat.assign(i, j, amat(i, j) * sqrt(evalvec[j] / v));
    }
  }
}

void pcaclass::put_pca(ostream& fout, option& opt, idd& id, int n, const mat<ldbl>& smat, const vecldbl& evalvec, const mat<ldbl>& amat) const
// 主成分計算結果を出力する
{
  constexpr int LENline = 76;   // 区切り線の長さ
  int p = amat.r() - 1, q = amat.c() - 1;
  if (opt.res("pc"))
  {
    fout << "\n<PRICIPAL COMPONENT>\n";
    fout << "  Conditions for Extraction:";
    if (0 < opt.q() && opt.q() == q) fout << "Maximum Number of PC=" << opt.q();
    else fout << "Cumulative Contribution Rate cum-r^2=" << opt.c();
    fout << "\nPC   Eigen Value      cum-EV         r^2     cum-r^2       Alpha          SD\n";
    putline(fout, LENline);
  }
  mat<ldbl> cmat(amat);
  amat2cmat(opt, id, cmat, n, smat, evalvec);
  mat<ldbl> ctmat(cmat);
  for (int i = 0; i < p; ++i) ctmat.assign(i, 0, 0.0L);
  for (int j = 1; j < q; ++j) ctmat.assign(0, j, 0.0L);
  mat<ldbl> tmat(ctmat.trn() * smat);
  mat<ldbl> zsmat(tmat * ctmat.trn());
  int df = n - 1;
  if (opt.res("pc"))
  {
    ldbl slamda = 0.0L;
    for (int k = 1; k <= q; ++k)
    {
      ldbl lamda = evalvec[k];
      ldbl vzz = (opt.m() == matr) ? 1.0L : evalvec[k];
      slamda += lamda;
      if (opt.m() == mats) vzz = lamda / (ldbl)df;
      ldbl sxx = 0.0L;
      for (int i = 1; i <= p; ++i) sxx += cmat(i, k) * smat(i, i) * cmat(i, k);
      ldbl alpha = (zsmat(k, k) > 0.0L) ? (ldbl)p / (ldbl)(p - 1) * (1.0L - sxx / zsmat(k, k)) : 0.0L;
      fout << "Z" << setw(3) << k << setw(12) << lamda;
      fout << setw(12) << slamda << setw(12) << lamda / evalvec[0];
      fout << setw(12) << slamda / evalvec[0] << setw(12) << alpha;
      fout << setw(12) << sqrt(vzz) << "\n";
    }
    putline(fout, LENline);
    putcoef(fout, "<EIGNE VECTOR>", "Z", "X", amat);
  }
  if (opt.res("pcc")) putcoef(fout, "<COEFFICINET FOR PRINCIPAL COMPONENT>", "Z", "X", cmat);
  mat<ldbl> lmat(amat);
  amat2lmat(opt, lmat, n, smat, evalvec);
  if (opt.res("pcl")) putfl(fout, "<PRINCIPAL COMPONENT LOADINGS>", "Z", "X", lmat);
}

int pcaclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time opion error !");
  return nc;
}

void pcaclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  int dim = id.ni() + 1;        // 積和行列の次数
  mat<ldbl> smat(dim);          // 積和行列
  vecldbl mevec(dim, 0.0L), vrvec(dim, 0.0L); // 平均値ベクトルと分散ベクトル
  int n = cal_smat(opt, id, datalst, smat, mevec, vrvec); // 例数
  if (opt.res("res"))
  {
    put_resid(cout, id);
    if (opt.res("stat"))
    {
      putmsdvec(cout, n, mevec, vrvec);
      string head = "<START MATRIX:CORRERATION MATRIX>";
      switch (opt.m())
      {
      case mats : head = "<START MATRIX:SUM OF PRODUCTS MATRIX>";
        break;
      case matv : head = "<START MATRIX:COVARIANCE MATRIX>";
        break;
      case matetc : head = "<START MATRIX:SIMPLE SUM OF PRODUCTS MATRIX>";
      }
      if (opt.m() == matetc) putmat(cout, head, mats, n, smat);
      else putmat(cout, head, opt.m(), n, smat);
    }
  }
  vecldbl evalvec(dim, 0.0L);   // 固有値ベクトル
  mat<ldbl> amat(dim);          // 因子スコア計算用係数行列
  int q = cal_pca(opt, id, evalvec, amat, n, smat);
  if (q < 1)                    // 主成分なし
  {
    if (opt.res("res")) cout << "\n" << endl;
    return;
  }
  amat.resize(dim, q + 1);
  if (opt.res("res"))
  {
    put_pca(cout, opt, id, n, smat, evalvec, amat);
    cout << "\n" << endl;
  }
  if (opt.dnsname() != "") put_pcadns(opt, id, datalst, n, smat, evalvec, amat);
  if (opt.ngpname() != "") plot_plplot(opt, id, n, smat, evalvec, amat);
}

int main(int argc, char* argv[]) // メインルーチン
{
  pcaclass pca;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) pca.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { pca.get(argv[i], opt, id, datalst);  fok = false; }
    }
    pca.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
