// poirege.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 09:22:35 tac>
//
// Poisson回帰分析。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W poirege poirege.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "poirege  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:poireg [<OPTION> ...]  [<DATA FILE> ...]",
  "  The Poisson regression analysis with the first item as criterion variable",
  "  and subsequent items as explanatory variables.",
  "<OPTION>",
  "  -b=<VALUE>[,<VALUE>,...]:specify the initial value of regression",
  "    coefficient",
  "  -eps[=<EPS>]:specify the convergence error(0<EPS)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -meth[={c|x}]:select the caluculation method(-meth:-meth=x)",
  "  -model[={a|l|p}]:select the model(-model:-model=a)",
  "  -ngp={ex|lamda|rr}[,...]:select the data for Ngraph",
  "    ex:exponential function  lamda:ln(lambda) and y  rr:risk ratio",
  "  -res=[-][{def|all|cc[<C-COEF>]|reg|stat|v}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1  cc:cc0.95)",
  "    reg:result of regression analysis  stat:basic statistics",
  "    v:initial values of regression coefficient",
  "  -rmax=<NUMBER>:specify the maximum number of repetitions(0=<NUMBER)",
  "  -x<NUMBER>=<VALUE>:specify the value of explanatory variable",
  "    whoes value is constant",
  "  -xno=<NUMBER>:specify the number of explanatory variable whoes value",
  "    changes"
};

const string Title = "=== Poisson Regression Snalysis ==="; // 表題
const string NGPNAMEdef = "poirege_ngp.dat"; // 省略時Ngraphファイル名
constexpr double EPSdef = 1.0e-5; // 収束誤差省略時値
constexpr int RMAXdef = 200;      // 省略時繰り返し計算数

enum methtype                   // 計算方法
{
  methnon,                      // 指定なし
  methc,                        // ケース単位
  methx,                        // 説明変数単位
  methetc                       // その他
};

enum modeltype                  // モデル
{
  modelnon,                     // 指定なし
  modela,                       // 最適モデルを自動選択
  modell,                       // λ(出現件数)モデル
  modelp,                       // p=λ/n(出現率)モデル
  modeletc                      // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  lstr blst;                    // 説明変数の初期値リスト
  vecdbl vbvec;                 // 説明変数の初期値ベクトル
  double vcc;                   // 信頼係数
  double veps;                  // 最尤法の収束誤差
  methtype vmeth;               // 計算方法
  modeltype vmodel;             // モデル
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  int vrmax;                    // 最大反復回数
  lstr xlst;                    // 説明変数リスト
  vecdbl vxvec;                 // 説明変数ベクトル
  int vxno;                     // 変化説明変数の番号
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const vecdbl& bvec() const    // 説明変数の初期値ベクトルを返す
  { return vbvec; }
  double b(size_t idx) const    // 説明変数の初期値を返す
  { return (idx < vbvec.size()) ? vbvec[idx] : 0.0; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double eps() const            // 最尤法の収束誤差を返す
  { return veps; }
  methtype meth() const         // 計算方法を返す
  { return vmeth; }
  modeltype model() const       // モデルを返す
  { return vmodel; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 最大反復回数を返す
  { return vrmax; }
  double x(size_t no) const     // 説明変数の値を返す
  { return (0 < no && no <= vxvec.size()) ? vxvec[no - 1] : BLANKf; }
  int xno() const               // 変化説明変数の番号を返す
  { return vxno; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), blst(), vbvec(), vcc(CCdef), veps(EPSdef), vmeth(methx), vmodel(modela), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vrmax(RMAXdef), xlst(), vxvec(), vxno(0)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "b")            // 偏回帰係数初期値の指定
  {
    if (strsplit(blst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 最尤法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < DBL_EPSILON) err = badopt(vpname, optstr, USAGE, DBL_EPSILON, DBL_MAX);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "meth")    // 計算方法指定
  {
    if (opt[1].empty() || opt[1] == "x") vmeth = methx;
    else if (opt[1] == "c") vmeth = methc;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "model")   // モデル指定
  {
    if (opt[1].empty() || opt[1] == "a") vmodel = modela;
    else if (opt[1] == "l") vmodel = modell;
    else if (opt[1] == "p") vmodel = modelp;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 最大反復回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "xno")     // 変化説明変数の番号指定
  {
    char* ptr = nullptr;
    vxno = strtol(opt[1].c_str(), &ptr, 0);
    if (vxno < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0][0] == 'x')    // 説明変数の値指定
  {
    if (strsplit(xlst, optstr, CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are requied !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (blst.empty() == false)    // 偏回帰係数初期値の指定あり
  {
    if (blst.size() > ni) return disperr(vpname, "Too many initial values of the regression coefficients !");
    vbvec.resize(ni);
    vbvec.assign(ni, BLANKf);
    int j = 0;
    for (lstr_citr bitr = blst.begin(); bitr != blst.end(); ++bitr, ++j)
    {
      double fdata = strtodf(*bitr);
      if (fdata == BLANKf) return disperr(vpname, "Wrong initial value->", *bitr);
      vbvec[j] = fdata;
    }
  }
  if (vxno < 1) return 0;       // 説明変数指定なし
  int nx = ni - 1;
  if (nx < vxno) return disperr(vpname, "Wrong number of the explanatory variable whoes value are changes !");
  if (xlst.empty()) return 0;   // 説明変数指定なし
  if ((int)(xlst.size()) > nx) return disperr(vpname, "Too many values of the explanatory variables !");
  vxvec.resize(nx);
  vxvec.assign(nx, BLANKf);
  for (lstr_citr xitr = xlst.begin(); xitr != xlst.end(); ++xitr)
  {
    string opt[2];
    if (getopt(opt, *xitr, OPTdev) > 0)
    {
      char* ptr = nullptr;
      int xno = strtol(opt[0].substr(1).c_str(), &ptr, 0);
      if (xno < 1 || nx < xno || xno == vxno) return disperr(vpname, "Wrong number of the explanatory variable->", *xitr);
      double fdata = strtodf(opt[1]);
      if (fdata == BLANKf) return disperr(vpname, "Wrong value of the explanatory variable->", *xitr);
      vxvec[xno - 1] = fdata;
    }
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["ex"] = vngpmap["lamda"] = vngpmap["rr"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["ex"] = vngpmap["lamda"] = vngpmap["rr"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = true;
  vresmap["v"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = vresmap["v"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class data_less // ケースデータの選択済み説明変数比較用関数オブジェクト
{
public:
  data_less()                   // コンストラクタ
  {}
  ~data_less()                  // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_less::operator()(datac& cdata1, datac& cdata2) const
// ケースデータの説明変数を比較する
{
  int nx = (int)(cdata1.size()) - 1;
  int comp = 0;
  for (int i = 1; comp == 0 && i <= nx; ++i)
  {
    double fdata1 = cdata1.fdata(i), fdata2 = cdata2.fdata(i);
    if (fdata1 == BLANKf)
    {
      if (fdata2 != BLANKf) comp = 1;
    } else if (fdata2 == BLANKf)
    { comp = -1; }
    else
    { comp = sgn(fdata1 - fdata2); }
  }
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class poiregclass               // poiregメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_smat(int& nr, int& n0, mat<double>& smat, vecdbl& mevec, vecdbl& vrvec);
  // データを集計して積和行列を求める
  void put_resid(ostream& fout); // 解析結果のタイトルとデータ識別情報を出力する
  void set_initb(vec<double>& b0v, vec<double>& vbv, int n, int nr, mat<double>& smat, const option& opt, const vecdbl& mevec);
  // 重回帰分析→判別分析による偏回帰係数初期値のセット
  int newton_poi(vec<double>& bv, vec<double>& vbv, double& lb, const option& opt, modeltype modelt);
  // Newton法によるPoisson回帰の計算(最尤法)
  int cal_lgh(vec<double>& gv, mat<double>& hmat, double& lb, const option& opt, const vec<double>& bv, modeltype modelt);
  // 傾斜ベクトルgとHessian行列Hを計算する
  void put_poireg(ostream& fout, option& opt, int nr, modeltype modelt, const vecdbl& vrvec, int r, const vec<double>& b0v, const vec<double>& bv, const vec<double>& vbv, double lb); // Poisson回帰分析結果を出力する
  void put_brci(ostream& fout, const option& opt, const vec<double>& bv, const vec<double>& vbv) const;
  // 偏回帰係数とリスク比の信頼区間を出力する
  void put_poitab(ostream& fout, const option& opt, int nr, modeltype modelt, double lb);
  // Poisson回帰分析の分散分析表を出力する
  int cal_poilf(int& np, double& lf, const option& opt);
  // 飽和モデルのグループ数と例数と尤度を求める
  bool data_equal(datac& cdata1, datac& cdata2);
  // ケースデータの説明変数を比較する
  void plot_poingp(option& opt, modeltype modelt, const vecdbl& mevec, const vec<double>& bv, const vec<double>& vbv);
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, const option& opt);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  double cal_lamda(const vec<double>& bv, datac& cdata); // λ推測値を返す
  void plot_ex(ofstream& fout, const option& opt, modeltype modelt, const vecdbl& mevec, const vec<double>& bv);
  // Ngraph用に指数曲線を出力する
  void plot_lamda(ofstream& fout, const option& opt, modeltype modelt, const vec<double>& bv);
  // Ngraph用にln(λ)とyを出力する
  void plot_rr(ofstream& fout, const option& opt, modeltype modelt, const vec<double>& bv, const vec<double>& vbv) const;
  // Ngraph用にリスク比とその信頼区間を出力する
public:
  poiregclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~poiregclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int poiregclass::cal_smat(int& nr, int& n0, mat<double>& smat, vecdbl& mevec, vecdbl& vrvec)
// データを集計して積和行列を求める
{
  size_t dim = id.ni() + 1, yidx = id.ni();
  nr = 0;  n0 = 0;
  vecdbl fdatavec(dim, 0.0);
  fdatavec[0] = 1.0L;
  for (auto& cdata : datalst)
  {
    if (cdata.fdata(0) != BLANKf) // 目的変数欠損
    {
      bool blank = false;
      for (size_t j = 1; blank == false && j < id.ni(); ++j)
      {
        fdatavec[j] = cdata.fdata(j);
        if (fdatavec[j] == BLANKf) blank = true;
      }
      if (blank)
      { cdata.assign(0, BLANKf); } // 後の集計のために目的変数を欠損にする
      else
      {
        if (cdata.fdata(0) <= 0.0) ++n0; // 目的変数が0
        fdatavec[yidx] = cdata.fdata(0);
        nr += (int)(fdatavec[yidx]); // 合計個数を数える
        for (size_t i = 0; i < dim; ++i)
        {
          double x = fdatavec[i];
          for (size_t j = i; j < dim; ++j) smat.add(i, j, x * fdatavec[j]);
        }
      }
    }
  }
  int n = (int)(smat(0, 0));
  for (size_t i = 0; i < yidx; ++i)
  {
    for (size_t j = i + 1; j < dim; ++j) smat.assign(j, i, smat(i, j));
  }
  smat.sweep(0);
  for (size_t i = 0; i < dim; ++i)
  {
    mevec[i] = smat(0, i);
    vrvec[i] = (n > 1) ? smat(i, i) / (double)(n - 1) : 0.0;
  }
  return n;                     // 例数を返す
}

void poiregclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i;
    fout << ":" << id.item(i) << "\n";
  }
}

void poiregclass::set_initb(vec<double>& b0v, vec<double>& vbv, int n, int nr, mat<double>& smat, const option& opt, const vecdbl& mevec)
// 重回帰分析→判別分析による偏回帰係数初期値のセット
{
  int p = id.ni();
  for (int i = 1; i < p; ++i) smat.sweep(i);
  double sr = smat(p, p);
  int dfr = n - p;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double lamda = (double)nr / (double)n;
  b0v.assign(0, log(lamda));
  const double Beps = DBL_EPSILON * fmax(FDmin, smat.maxabse());
  for (int i = 1; i < p; ++i)
  {
    double bi = smat(i, p);
    if (fabs(bi) < Beps) bi = 0.0;
    b0v.assign(i, bi);
    vbv.assign(i, vr * smat(i, i));
    b0v.assign(0, b0v(0) - bi * mevec[i]);
  }
  if (opt.bvec().empty()) return; // 初期値指定なし
  for (int i = 0; i < p; ++i)
  {
    if (opt.b(i) != BLANKf) b0v.assign(i, opt.b(i));
  }
}

int poiregclass::newton_poi(vec<double>& bv, vec<double>& vbv, double& lb, const option& opt, modeltype modelt)
// Newton法によるPoisson回帰の計算(最尤法)
{
  int p = id.ni();
  vec<double> gv(p);
  mat<double> hmat(p);
  double lold = 0.0;
  bool ok = false;
  int r = 0;
  datalst.sort(data_less());    // 説明変数順にソートする
  while (ok == false && cal_lgh(gv, hmat, lb, opt, bv, modelt) == 0)
  {
    ++r;
    hmat.inv();
    if (fabs(lb - lold) < opt.eps() || (opt.rmax() > 0 && r >= opt.rmax())) ok = true;
    else { lold = lb;  bv -= hmat * gv; }
  }
  if (!ok) return -1;
  for (int i = 0; i < p; ++i) vbv.assign(i, -hmat(i, i));
  return r;                     // 反復数を返す
}

int poiregclass::cal_lgh(vec<double>& gv, mat<double>& hmat, double& lb, const option& opt, const vec<double>& bv, modeltype modelt)
// 傾斜ベクトルgとHessian行列Hを計算する
{
  constexpr double Emax = 308.0; // 指数最大値
  int p = id.ni();
  vec<double> xv(p);
  xv.assign(0, 1.0);  gv = 0.0;  hmat = 0.0;  lb = 0.0;
  bool err = false;
  ldata_itr iitr = datalst.begin();
  while (err == false && iitr != datalst.end())
  {
    if (iitr->fdata(0) == BLANKf)
    { ++iitr; }
    else
    {
      int ri = (int)(iitr->fdata(0)); // 反応有の個数
      int ni = (ri > 0) ? ri : 1;   // 全体の例数
      ldata_itr jitr = iitr;
      ++jitr;
      if (opt.meth() != methc)  // 説明変数単位の計算方法
      {
        bool equal = true;
        while (equal && jitr != datalst.end())
        {
          if (jitr->fdata(0) == BLANKf)
          { ++jitr; }
          else
          {
            equal = data_equal(*iitr, *jitr);
            if (equal)
            {
              int rj = (int)(jitr->fdata(0));
              ri += rj;
              ni += (rj > 0) ? rj : 1;
              ++jitr;
            }
          }
        }
      }
      for (int i = 1; i < p; ++i) xv.assign(i, iitr->fdata(i));
      double sbxi = inp(bv, xv);
      if (fabs(sbxi) > Emax)    // 指数最大値以上
      { err = true; }
      else
      {
        double wi = exp(sbxi), ri_f = (double)ri, ni_f = (double)ni;
        if (modelt == modelp)   // 出現率モデル
        {
          double nwi = ni_f * wi;
          lb += ri_f * (log(ni_f) + sbxi) - lgamma(ri_f + 1.0) - nwi;
          for (int j = 0; j < p; ++j)
          {
            gv.add(j, (ri_f - nwi) * xv(j));
            for (int k = j; k < p; ++k) hmat.add(j, k, -nwi * xv(j) * xv(k));
          }
        }
        else                    // 出現例数モデル
        {
          lb += ri_f * sbxi - lgamma(ri_f + 1.0) - wi; // ri=0も含む
          for (int j = 0; j < p; ++j)
          {
            gv.add(j, (ri_f - wi) * xv(j));
            for (int k = j; k < p; ++k) hmat.add(j, k, -wi * xv(j) * xv(k));
          }
        }
      }
      iitr = jitr;
    }
  }
  if (err) return -1;
  for (int i = 0; i < p - 1; ++i)
  { for (int j = i + 1; j < p; ++j) hmat.assign(j, i, hmat(i, j)); }
  return 0;
}

void poiregclass::put_poireg(ostream& fout, option& opt, int nr, modeltype modelt, const vecdbl& vrvec, int r, const vec<double>& b0v, const vec<double>& bv, const vec<double>& vbv, double lb)
// Poisson回帰分析結果を出力する
{
  constexpr int LENline = 79;   // 区切り線の長さ
  fout << "\n<RESULT BY MAXMUM LIKELIHOOD METHOD(Repetitions:" << r << ")>\n";
  fout << "  Method:";
  if (opt.meth() == methc) fout << "Case-based(same as SAS)";
  else fout << "Explanatory Variable-based";
  fout << "\n  Model:";
  if (modelt == modelp) fout << "p=lambda/n=exp(b0+sum{bj xj})  n:Total Cases";
  else fout << "lambda=exp(b0+sum{bj xj})";
  fout << "\n  b0:Intercept  bj:Partial Regression Coefficient of xj\n";
  int p = (int)(id.ni());
  if (opt.res("v"))
  {
    fout << "  Initial Value:";
    for (int i = 0; i < p; ++i)
    {
      if (i > 0)
      {
        if (i % 4 == 0) fout << "\n" << setw(16) << " ";
        else fout << "  ";
      }
      fout << "b" << i << "=" << b0v(i);
    }
    fout << "\n";
  }
  if (r < 0)
  {
    fout << "  NEWTON METHOD DIVERGED !\n";
    return;
  }
  fout << "VAR    Partial b     SE of b      exp(b)  Standard b  Chi-suared     p-value\n";
  putline(fout, LENline);
  double csqr = (vbv(0) > 0.0) ? bv(0) * bv(0) / vbv(0) : 0.0;
  ldbl pr = ctop(csqr, 1);
  fout << "ICPT" << setw(12) << bv(0) << setw(12) << sqrt(vbv(0));
  fout << setw(36) << csqr << setw(12) << pr << pind(pr) << "\n";
  for (int i = 1; i < p; ++i)
  {
    csqr = (vbv(i) > 0.0) ? bv(i) * bv(i) / vbv(i) : 0.0;
    pr = ctop(csqr, 1);
    fout << "X" << setw(3) << i << setw(12) << bv(i);
    fout << setw(12) << sqrt(vbv(i)) << setw(12) << exp(bv(i));
    fout << setw(12) << bv(i) * sqrt(vrvec[i]) << setw(12) << csqr;
    fout << setw(12) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
  if (opt.res("cc")) put_brci(fout, opt, bv, vbv);
  put_poitab(fout, opt, nr, modelt, lb);
}

void poiregclass::put_brci(ostream& fout, const option& opt, const vec<double>& bv, const vec<double>& vbv) const
// 偏回帰係数とリスク比の信頼区間を出力する
{
  constexpr int LENline = 76;   // 区切り線の長さ
  fout << "\nVAR    Partial b" << setw(3) << opt.cc() * 100.0;
  fout << "%CI:Lower       Upper  Odds Ratio" << setw(3) << opt.cc() * 100.0;
  fout << "%CI:Lower       Upper\n";
  putline(fout, LENline);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vbv(0));
  double bl = bv(0) - ci, bu = bv(0) + ci;
  fout << "ICPT" << setw(12) << bv(0) << setw(12) << bl;
  fout << setw(12) << bu << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    ci = z * sqrt(vbv(i));
    bl = bv(i) - ci;  bu = bv(i) + ci;
    fout << "X" << setw(3) << i << setw(12) << bv(i);
    fout << setw(12) << bl << setw(12) << bu;
    fout << setw(12) << exp(bv(i)) << setw(12) << exp(bl);
    fout << setw(12) << exp(bu) << "\n";
  }
  putline(fout, LENline);
}

void poiregclass::put_poitab(ostream& fout, const option& opt, int nr, modeltype modelt, double lb)
// Poisson回帰分析の分散分析表を出力する
{
  double l0 = 0.0, lf = 0.0;
  int np = 0;
  int m = cal_poilf(np, lf, opt);
  int dft = m - 1, p = id.ni();
  int dfb = p - 1;
  int dflof = dft - dfb;
  vec<double> gv(p), bv(p);
  mat<double> hmat(p);
  double lamda0 = (modelt == modelp) ? (double)nr / (double)np : (double)nr / (double)m;
  bv.assign(0, log(lamda0));
  cal_lgh(gv, hmat, l0, opt, bv, modelt);
  double sb = lb - l0, st = lf - l0;
  double slof = st - sb;
  const double Seps = DBL_EPSILON * fmax(FDmin, st);
  if (dft < 1 || st < Seps) st = 0.0;
  if (dflof < 1 || slof < Seps) slof = 0.0;
  double r2 = (st > 0.0) ? sb / st : 0.0;
  fout << "Log Likelihood:L(Model)=" << lb << "  L(Intercept)=" << l0 << "  Lf=" << lf << "\n";
  fout << "Pseudo-Contribution Rate:R^2=" << r2 << "  AIC=" << 2.0 * (p - lb) << "\n";
  constexpr int LENline = 68;
  const string Head = "Test of Regression and LOF";
  fout << "\n" << setw((LENline - strclen(Head, CHARwide)) / 2 + Head.size()) << Head;
  fout << "\nFACTOR   -(log Likelihood)        DF   Chi-squared       p-value\n";
  putline(fout, LENline);
  double csqr = 2.0 * sb;
  ldbl pr = ctop(csqr, dfb);
  fout << "REGRESSION" << setw(16) << sb << setw(10) << dfb << setw(14) << csqr;
  fout << setw(14) << pr << pind(pr) << "\n";
  csqr = (dflof < 1) ? 0.0 : 2.0 * slof;
  pr = ctop(csqr, dflof);
  fout << "LOF" << setw(23) << slof << setw(10) << dflof;
  fout << setw(14) << csqr << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  fout << "TOTAL" << setw(21) << st << setw(10) << dft << endl;
}

int poiregclass::cal_poilf(int& np, double& lf, const option& opt)
// 飽和モデルのグループ数と例数と尤度を求める
{
  int m = 0;
  np = 0;  lf = 0.0;
  ldata_itr iitr = datalst.begin();
  while (iitr != datalst.end())
  {
    if (iitr->fdata(0) == BLANKf)
    { ++iitr; }
    else
    {
      int ri = (int)(iitr->fdata(0)); // 反応有の個数
      int ni = (ri > 0) ? ri : 1;   // 全体の例数
      ldata_itr jitr = iitr;
      ++jitr;
      if (opt.meth() != methc)  // 説明変数単位の計算方法
      {
        bool equal = true;
        while (equal && jitr != datalst.end())
        {
          if (jitr->fdata(0) == BLANKf)
          { ++jitr; }
          else
          {
            equal = data_equal(*iitr, *jitr);
            if (equal)
            {
              int rj = int(jitr->fdata(0));
              ri += rj;
              ni += (rj > 0) ? rj : 1;
              ++jitr;
            }
          }
        }
      }
      double ri_f = (double)ri, ni_f = (double)ni;
      double lamdai = ri_f;
      if (ri > 0)               // 発生件数がある時だけ尤度計算
      { lf += ri_f * log(lamdai) - lgamma(ri_f + 1.0) - lamdai; }
      np += ni;  ++m;
      iitr = jitr;
    }
  }
  return m;                     // グループ数を返す
}

bool poiregclass::data_equal(datac& cdata1, datac& cdata2)
// ケースデータの説明変数を比較する
{
  int nx = int(cdata1.size()) - 1;
  bool equal = true;
  for (int i = 1; equal && i <= nx; ++i)
  {
    double fdata1 = cdata1.fdata(i), fdata2 = cdata2.fdata(i);
    if (fdata1 == BLANKf)
    {
      if (fdata2 != BLANKf) equal = false;
    } else if (fdata2 == BLANKf)
    { equal = false; }
    else
    {
      if (sgn(fdata1 - fdata2) != 0) equal = false;
    }
  }
  return equal;                 // true:一致, false:不一致
}

void poiregclass::plot_poingp(option& opt, modeltype modelt, const vecdbl& mevec, const vec<double>& bv, const vec<double>& vbv)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("ex"))
  {
    fout << "-s1 -dline -w80 -c255,0,0\n";
    plot_ngpid(fout, opt);
    plot_ex(fout, opt, modelt, mevec, bv);
    delml = true;
  }
  if (opt.ngp("lamda"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -o300\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_lamda(fout, opt, modelt, bv);
  }
  if (opt.ngp("rr"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -w80 -o400 -zx0,"<< id.ni() << ",1 -zy0.1,100,10 -eypoi\n";
      fout << "%When draw CI:-s1 -derrorbar_y -w80\n";
    }
    plot_ngpid(fout, opt);
    plot_rr(fout, opt, modelt, bv, vbv);
  }
}

void poiregclass::plot_ngpid(ofstream& fout, const option& opt)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  { fout << "%EXPLANATORY VARIABLE X" << setw(3) << i << ":" << id.item(i) << "\n"; }
  if (opt.xno() > 0)            // 変化説明変数指定あり
  {
    fout << "%Explanatory Variable:" << id.item(opt.xno()) << "\n";
    for (size_t i = 1; i < id.ni(); ++i)
    {
      if (opt.x(i) != BLANKf) fout << "%Explanatory Variable:" << id.item(i) << " Specified Value=" << opt.x(i) << "\n";
    }
  }
}

double poiregclass::cal_lamda(const vec<double>& bv, datac& cdata)
// λまたはλ/n推測値を返す
{
  if (cdata.fdata(0) == BLANKf) return BLANKf; // 目的変数欠損のため計算できず
  double bx = bv(0);
  for (size_t i = 1; i < id.ni(); ++i) bx += bv(i) * cdata.fdata(i);
  return exp(bx);               // λまたはλ/n推測値を返す
}

void poiregclass::plot_ex(ofstream& fout, const option& opt, modeltype modelt, const vecdbl& mevec, const vec<double>& bv)
// Ngraph用指数曲線を出力する
{
  fout << "%<EXPONENTIAL CURVE:";
  if (modelt == modelp) fout << "p=lambda/n=exp(b0+sum{bj xj})";
  else fout << "lambda=exp(b0+sum{bj xj})";
  fout << "　Method:";
  if (opt.meth() == methc) fout << "Case-based>\n";
  else fout << "Explanatory Variable-based>\n";
  double xmin = DBL_MAX, xmax = -DBL_MAX;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = BLANKf;
    if (opt.xno() > 0)
    { x = ditr->fdata(opt.xno()); }
    else
    {
      double lamda = cal_lamda(bv, *ditr);
      x = (lamda == BLANKf) ? BLANKf : log(lamda);
    }
    if (x != BLANKf)
    {
      if (x < xmin) xmin = x;
      if (x > xmax) xmax = x;
    }
  }
  constexpr int RDdef = 50;     // 曲線の分解数
  double range = xmax - xmin;
  double xstart = xmin - range * 0.1, xend = xmax + range * 0.1;
  double dx = (xend - xstart) / (double)RDdef;
  datac cdata(id.ni());
  cdata.assign(0, 0.0);
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (opt.x(i) == BLANKf) cdata.assign(i, mevec[i]);
    else cdata.assign(i, opt.x(i));
  }
  double x = xstart;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    double lamda = BLANKf;
    if (opt.xno() < 1)
    { lamda = exp(x); }
    else
    {
      cdata.assign(opt.xno(), x);
      lamda = cal_lamda(bv, cdata);
    }
    if (lamda != BLANKf) fout << x << " " << lamda << "\n";
  }
}

void poiregclass::plot_lamda(ofstream& fout, const option& opt, modeltype modelt, const vec<double>& bv)
// Ngraph用にln(λ)とyを出力する
{
  if (modelt == modelp) fout << "%<ln(LAMBDA/N) AND Y";
  else fout << "%<ln(LAMBDA) AND Y";
  fout << "　Method:";
  if (opt.meth() == methc) fout << "Case-based>\n";
  else fout << "Explanatory Variable-based>\n";
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = BLANKf, y = ditr->fdata(0);
    if (opt.xno() > 0)
    { x = ditr->fdata(opt.xno()); }
    else
    {
      double lamda = cal_lamda(bv, *ditr);
      x = (lamda == BLANKf) ? BLANKf : log(lamda);
    }
    if (x != BLANKf && y != BLANKf) fout << x << " " << y << "\n";
  }
}

void poiregclass::plot_rr(ofstream& fout, const option& opt, modeltype modelt, const vec<double>& bv, const vec<double>& vbv) const
// Ngraph用にリスク比とその信頼区間を出力する
{
  fout << "%<RISK RATIO AND CI　Model:";
  if (modelt == modelp) fout << "p=lambda/n";
  else fout << "lambda";
  fout << "　Method:";
  if (opt.meth() == methc) fout << "Case-based>\n";
  else fout << "Explanatory Variable-based>\n";
  double z = ptoz(1.0 - opt.cc());
  for (size_t i = 1; i < id.ni(); ++i)
  {
    double ci = z * sqrt(vbv(i));
    double rri = exp(bv(i));
    double rrl = exp(bv(i) - ci), rru = exp(bv(i) + ci);
    fout << i << " " << rri << " " << rru - rri << " " << rrl - rri << "\n";
  }
}

int poiregclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void poiregclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  int dim = id.ni() + 1;        // 積和行列の次数
  mat<double> smat(dim);        // 積和行列
  int nr = 0;                   // 目的変数の合計個数
  int n0 = 0;                   // 0件数の例数
  vecdbl mevec(dim, 0.0L);      // 平均値ベクトル
  vecdbl vrvec(dim, 0.0L);      // 分散ベクトル
  int n = cal_smat(nr, n0, smat, mevec, vrvec); // 例数
  modeltype modelt = modela;    // モデルの種類
  if (n0 < 1) modelt = modell;  // 発生件数0がない時は発生件数モデル
  else if (opt.model() == modela) modelt = modelp;
  else modelt = opt.model();
  if (opt.res("res"))
  {
    put_resid(cout);
    int nx = id.ni() - 1;
    if (opt.res("stat"))
    {
      putmsdvec(cout, n, mevec, vrvec, nx, 1);
      putmat(cout, "<CORRELATION COEFFICIENT MATRIX>", matr, n, smat, nx, 1);
    }
    if (n <= nx || nr < 1)      // 回帰式計算できず
    {
      if (opt.res("res")) cout << "\n" << endl;
      return;
    }
  }
  int p = id.ni();              // パラメータ数
  vec<double> b0v(p), bv(p), vbv(p);
  double lb = 0.0;
  set_initb(b0v, vbv, n, nr, smat, opt, mevec);
  bv = b0v;
  int r = newton_poi(bv, vbv, lb, opt, modelt);
  if (opt.res("res"))
  {
    if (opt.res("reg")) put_poireg(cout, opt, nr, modelt, vrvec, r, b0v, bv, vbv, lb);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_poingp(opt, modelt, mevec, bv, vbv);
}

int main(int argc, char* argv[]) // メインルーチン
{
  poiregclass poireg;           // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) poireg.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { poireg.get(argv[i], opt);  fok = false; }
    }
    poireg.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
