// regane.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 09:46:19 tac>
//
// 最後の項目を目的変数Y、以後の項目を説明変数Xとした
// 逐次変数選択法による重回帰分析。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W regane regane.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "regane  Version 7.41e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:regane [<OPTION> ...]  [<DATA FILE> ...]",
  "  The multiple regression analysis with the first item as criterion",
  "  variable and subsequent items as explanatory variables.",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=regane_dns.d",
  "  -f=<FILE>:reads options from the specified file",
  "  -fin=<Fin>:specify the inclusion criteria Fin(Fin>=0)",
  "  -fout=<Fout>:specify the expulsion criteria Fout(0=<Fout=<Fin)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={est|line|pl}[,...]:select the data for Ngraph",
  "    est:estimate Y/corrected Y  line:partial regression line",
  "    pl:prediction limit  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=regane_ngp.dat",
  "  -pin=<Pin>:specify the inclusion criteria Pin(0<Pin=<1)",
  "  -pout=<Pout>:specify the expulsion criteria Pout(1>=Pout>=Pin)",
  "  -res=[-][{def|all|allx|atab|cc[<C-COEF>]|selx|stat}][,...]",
  "    :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    allx:result of selecting all variables  atab:ANOVA table",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1  cc:cc0.95)",
  "    selx:result of selecting all variables  stat:basic statistics",
  "  -x<NUMBER>=<VALUE>:specify the number and value of explanatory variable",
  "    whoes value is constant",
  "  -xno=<NUMBER>:specify the number of explanatory variable whose value",
  "    changes"
};

const string Title = "=== Multiple Regression Analysis ==="; // 表題
const string DNSNAMEdef = "regane_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "regane_ngp.dat"; // 省略時Ngraphファイル名
constexpr double FINdef = 2.0;  // 省略時F値取り込み基準
constexpr double FOUTdef = 2.0; // 省略時F値追い出し基準
constexpr double PINdef = 0.1;  // 省略時p値取り込み基準
constexpr double POUTdef = 0.1; // 省略時p値追い出し基準

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  pairdbl flevel;               // F値の取り込み基準と追い出し基準
  pairdbl plevel;               // P値の取り込み基準と追い出し基準
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  lstr xlst;                    // 説明変数リスト
  vecdbl vxvec;                 // 説明変数ベクトル
  int vxno;                     // 変化説明変数の番号
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  double fin() const            // F値の取り込み基準を返す
  { return flevel.first; }
  double fout() const           // F値の追い出し基準を返す
  { return flevel.second; }
  double pin() const            // P値の取り込み基準を返す
  { return plevel.first; }
  double pout() const           // P値の追い出し基準を返す
  { return plevel.second; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double x(size_t idx) const    // 説明変数の値を返す
  { return (0 < idx && idx < vxvec.size()) ? vxvec[idx] : BLANKf; }
  int xno() const               // 変化説明変数の番号を返す
  { return vxno; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdnsname(), flevel(FINdef, FOUTdef), plevel(1.0, 1.0), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), xlst(), vxvec(), vxno(0)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "fin")     // F値の取り込み基準指定
  {
    flevel.first = opt[1].empty() ? FINdef : strtodf(opt[1]);
    if (flevel.first < 0.0) err = badopt(vpname, optstr, USAGE, 0.0, DBL_MAX);
    if (flevel.second > flevel.first) flevel.second = flevel.first;
  }
  else if (opt[0] == "fout")    // F値の追い出し基準指定
  {
    flevel.second = opt[1].empty() ? flevel.first : strtodf(opt[1]);
    if (flevel.second < 0.0 || flevel.first < flevel.second) err = badopt(vpname, optstr, USAGE, 0.0, flevel.first);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pin")     // p値の取り込み基準指定
  {
    plevel.first = opt[1].empty() ? PINdef : strtodf(opt[1]);
    if (plevel.first <= 0.0 || 1.0 < plevel.first) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "pout")    // p値の追い出し基準指定
  {
    plevel.second = opt[1].empty() ? plevel.first : strtodf(opt[1]);
    if (plevel.second < plevel.first || 1.0 < plevel.second) err = badopt(vpname, optstr, USAGE, plevel.first, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "xno")     // 変化説明変数の番号指定
  {
    char* ptr = nullptr;
    vxno = strtol(opt[1].c_str(), &ptr, 0);
    if (vxno < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0][0] == 'x')    // 説明変数の値指定
  {
    if (strsplit(xlst, optstr, CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 2) return disperr(vpname, "At least 2 items are requied !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (vxno == 0) return 0;
  if (vxno < 0 || (int)ni <= vxno) return disperr(vpname, "Wrong number of the explanatory variable whose value changes !");
  vxvec.resize(ni);
  vxvec.assign(ni, BLANKf);
  if (xlst.empty()) return 0;
  if (xlst.size() >= ni) return disperr(vpname, "Too many values of the explanatory variables !");
  for (string xstr : xlst)
  {
    string opt[2];
    if (getopt(opt, xstr, OPTdev) > 0)
    {
      char* ptr = nullptr;
      int xno = strtol(opt[0].substr(1).c_str(), &ptr, 0);
      if (xno < 1 || (int)ni <= xno || xno == vxno) return disperr(vpname, "Wrong number of the explanatory variable->", xstr);
      double fdata = strtodf(opt[1]);
      if (fdata == BLANKf) return disperr(vpname, "Wrong value of the explanatory variable->", xstr);
      vxvec[xno] = fdata;
    }
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["est"] = vngpmap["line"] = vngpmap["pl"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["est"] = vngpmap["line"] = vngpmap["pl"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["allx"] = vresmap["atab"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["allx"] = vresmap["atab"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["allx"] = vresmap["atab"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class reganclass                // reganメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_smat(mat<double>& smat, mat<double>& xmat, statv<double>& ystat, vecdbl& mevec, vecdbl& vrvec, vecdbl& ssvec);
  // データを集計して積和行列を求める
  void put_resid(ostream& fout); // 解析結果のタイトルとデータ識別情報を出力する
  void step_sel(int n, double fin, double fout, double pin, double pout, mat<double>& smat, vecint& checkvec);
  // 変数増減法による重回帰分析の計算
  int sel_var(int in, int n, double fin, double fout, double pin, double pout, const mat<double>& smat, vecint& checkvec) const;
  // 変数増減法による変数の選択
  void put_regtab(ostream& fout, option& opt, int n, double fin, double flout, double pin, double pout, const mat<double>& smat, const mat<double>& xmat, const vecdbl& ssvec, const vecint& checkvec);
  // 重回帰分析結果を出力する
  void cal_vif(mat<double>& rmat, const mat<double>& xmat, const vecint& checkvec) const;
  // VIFを計算する
  void put_bci(ostream& fout, const option& opt, int n, const mat<double>& smat, const vecint& checkvec) const;
  // 偏回帰係数の信頼区間を出力する
  void put_dns(option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec);
  // Yの推定値をデータファイル形式で出力する
  void put_dnsid(ofstream& fout, const option& opt, const vecdbl& mevec, const vecint& checkvec);
  // データファイル形式のデータ識別情報を出力する
  void plot_regngp(option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec);
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, const option& opt, const vecdbl& mevec, const vecint& checkvec);
  // Ngraph用にデータ識別情報を出力する
  void plot_est(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec);
  // Yの推定値をNgraph用に出力する
  double cal_estimate(datac& cdata, const mat<double>& smat, const vecint& checkvec) const;
  // Yの推定値を計算して返す
  void plot_line(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec);
  // 回帰直線をNgraph用に出力する
  double search_xrange(double& xmin, double& xmax, const option& opt, const mat<double>& smat, const vecint& checkvec);
   // x軸の最小値と最大値を検索して範囲を返す
  void plot_pl(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec);
  // 予測限界をNgraph用に出力する
  double cal_pi(datac& cdata, double t, double vr, const mat<double>& smat) const;
  // 回帰直線の予測限界幅を計算して返す
public:
  reganclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~reganclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int reganclass::cal_smat(mat<double>& smat, mat<double>& xmat, statv<double>& ystat, vecdbl& mevec, vecdbl& vrvec, vecdbl& ssvec)
// データを集計して積和行列を求める
{
  size_t dim = id.ni() + 1, yidx = id.ni();
  vecdbl fdatavec(dim, 0.0);
  fdatavec[0] = 1.0;
  for (auto& cdata : datalst)
  {
    bool blank = false;
    size_t j = 1;
    while (blank == false && j < id.ni())
    {
      if (cdata.fdata(j) == BLANKf)
      { blank = true; }
      else
      {
        fdatavec[j] = cdata.fdata(j);
        ++j;
      }
    }
    if (blank || cdata.fdata(0) == BLANKf)
    { cdata.assign(0, BLANKs); } // 後の計算のためにYをブランクにする
    else
    {
      ystat.sigma(cdata.fdata(0)); // yを累積する
      fdatavec[yidx] = cdata.fdata(0);
      for (size_t i = 0; i < dim; ++i)
      {
        double xi = fdatavec[i];
        for (size_t j = i; j < dim; ++j) smat.add(i, j, xi * fdatavec[j]);
      }
    }
  }
  int n = ystat.n();
  for (size_t i = 0; i < yidx; ++i)
  {
    xmat.assign(i, i, smat(i, i));
    for (size_t j = i + 1; j <= yidx; ++j)
    {
      smat.assign(j, i, smat(i, j));
      if (j < yidx)
      { xmat.assign(i, j, smat(i, j));  xmat.assign(j, i, smat(i, j)); }
    }
  }
  smat.sweep(0);
  for (size_t i = 0; i <= yidx; ++i)
  {
    mevec[i] = smat(0, i);
    vrvec[i] = (n > 1) ? smat(i, i) / (double)(n - 1) : 0.0;
    ssvec[i] = smat(i, i);
  }
  return n;                     // 例数を返す
}

void reganclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i;
    fout << ":" << id.item(i) << "\n";
  }
}

void reganclass::step_sel(int n, double fin, double fout, double pin, double pout, mat<double>& smat, vecint& checkvec)
// 変数増減法による重回帰分析の計算
{
  int in = 1, i = 0;
  do {
    in = 1;
    while ((i = sel_var(in, n, fin, fout, pin, pout, smat, checkvec)) > 0)
    {
      smat.sweep(i);  checkvec[i] = in ? 1 : 0;
      checkvec[0] += in ? 1 : -1;
      in = !in;
    }
  } while (!in);                // 取り込む変数がなくなるまで続ける
}

int reganclass::sel_var(int in, int n, double fin, double fout, double pin, double pout, const mat<double>& smat, vecint& checkvec) const
// 変数増減法による変数の選択
{
  size_t d = smat.r() - 1, i = 0;
  const double Seps = DBL_EPSILON * fmax(FDmin, smat.maxabse());
  double dsel = in ? Seps : DBL_MAX;
  bool fsel = true;             // F値による変数選択
  if (pin < 1.0) fsel = false;  // 有意確率pによる変数選択
  for (size_t j = 1; j < d; ++j)
  {
    if (checkvec[j] != in && fabs(smat(j, j)) >= Seps)
    {
      double dj = smat(j, d) * smat(j, d) / smat(j, j);
      if ((in && dj > dsel) || (!in && dj < dsel)) { dsel = dj;  i = j; }
    }
  }
  if (i < 1) return i;          // 選択すべき変数なし
  if (in)                       // 変数取り込み
  {
    int df = n - checkvec[0] - 2;
    double f = dsel * (double)df / (smat(d, d) - dsel);
    if (fsel)
    {
      if (f < fin) i = 0;
    }
    else
    {
      double p = ftop(f, 1, df);
      if (p > pin) i = 0;
    }
  }
  else                          // 変数追い出し
  {
    int df = n - checkvec[0] - 1;
    double f = dsel * (double)df / smat(d, d);
    if (fsel)
    {
      if (f >= fout) i = 0;
    }
    else
    {
      double p = ftop(f, 1, df);
      if (p <= pout) i = 0;
    }
  }
  return i;                     // 選択すべき変数のインデックスを返す
}

void reganclass::put_regtab(ostream& fout, option& opt, int n, double fin, double flout, double pin, double pout, const mat<double>& smat, const mat<double>& xmat, const vecdbl& ssvec, const vecint& checkvec)
// 重回帰分析結果を出力する
{
  constexpr int LENline = 79;   // 区切り線の長さ
  int p = checkvec[0];
  mat<double> rmat(p + 1, p + 1);
  if (p > 0) cal_vif(rmat, xmat, checkvec); // 変数が選択されている時だけrmatを計算
  if (fin <= 0.0 && pin >= 1.0)
  { fout << "\n<RESULT OF SELECTING ALL VARIABLES>\n"; }
  else
  {
    fout << "\n<RESULT OF STEPWISE FORWARD SELECTION METHOD>\n";
    if (pin < 1.0L) fout << "  Inclusion Criteria:p=<Pin=" << pin << "  Expulsion Criteria:p>Pout=" << pout << "\n";
    else fout << "  Inclusion Criteria:F>=Fin=" << fin << "  Expulsion Criteria:F<Fout=" << flout << "\n";
  }
  fout << "VAR    Partial b  Standard b   Partial r         VIF     F-value     p-value\n";
  putline(fout, LENline);
  size_t d = smat.r() - 1;
  double syy = ssvec[d], sr = smat(d, d);
  const double Seps = DBL_EPSILON * fmax(FDmin, syy);
  if (sr < Seps) sr = 0.0;
  int dfr = n - 1 - checkvec[0];
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double sii = smat(0, 0), bi = smat(0, d);
  double vbi = vr * sii;
  double fi = (vbi > 0.0) ? bi * bi / vbi : 0.0;
  double pr = ftop(fi, 1, dfr);
  fout << "ICPT" << setw(12) << bi << setw(36) << "" << setw(12) << fi;
  fout << setw(12) << pr << pind(pr) << "\n";
  size_t k = 1;
  for (size_t i = 1; i < d; ++i)
  {
    if (checkvec[i])
    {
      bi = smat(i, d);  sii = smat(i, i);
      vbi = vr * sii;
      double bsti = (syy > 0.0) ? bi * sqrt(ssvec[i] / syy) : 0.0;
      double v = sr * sii + bi * bi;
      double r = (v < Seps) ? 0.0 : bi / sqrt(v);
      fi = (vbi < Seps) ? 0.0 : bi * bi / vbi;
      pr = ftop(fi, 1, dfr);
      fout << "X" << setw(3) << i << setw(12) << bi << setw(12) << bsti;
      fout << setw(12) << r << setw(12) << rmat(k, k) << setw(12) << fi;
      fout << setw(12) << pr << pind(pr) << "\n";
      ++k;
    }
  }
  putline(fout, LENline);
  if (opt.res("cc")) put_bci(fout, opt, n, smat, checkvec);
  if (checkvec[0] < 1 || syy <= 0.0) return; // 重相関係数等計算不要
  double r2 = (syy < Seps) ? 0.0 : (syy - sr) / syy;
  double r2a = (syy < Seps) ? 0.0 : fabs(1.0 - vr * (double)(n - 1) / syy);
  double aic = (sr < Seps) ? 0.0 : (double)n * log(sr / (double)n) + (double)(2 * (p + 2));
  fout << "Multiple Contribution Rate:R^2=" << r2;
  fout << "  DF-adjusted:R'^2=" << r2a << "\n";
  fout << "Multiple Correlation Coefficient:R=" << sqrt(r2) << "  DF-adjusted:R'=" << sqrt(r2a) << "\n";
  fout << "AIC=" << aic << "\n\n";
  if (opt.res("atab")) putatab(fout, "REGRESSION", syy - sr, p, sr, dfr, syy, n - 1);
}

void reganclass::cal_vif(mat<double>& rmat, const mat<double>& xmat, const vecint& checkvec) const
// VIFを計算する
{
  size_t p = checkvec[0], d = xmat.r() - 1;
  rmat.assign(0, 0, xmat(0, 0));
  size_t l = 1;
  for (size_t j = 1; j < d; ++j)
  {
    if (checkvec[j])
    {
      rmat.assign(0, l, xmat(0, j));  rmat.assign(l, 0, xmat(0, j));
      ++l;
    }
  }
  size_t k = 1;
  for (size_t i = 1; i < d; ++i)
  {
    if (checkvec[i])
    {
      rmat.assign(k, k, xmat(i, i));
      l = 1;
      for (size_t j = 1; j < d; ++j)
      {
        if (checkvec[j])
        {
          rmat.assign(k, l, xmat(i, j));  rmat.assign(l, k, xmat(i, j));
          ++l;
        }
      }
      ++k;
    }
  }
  rmat.sweep(0);                // 積和行列にする
  smat2rmat(rmat);              // 相関行列にする
  mat<double> tmat(p);
  tmat = rmat.sub(1, 1, p, p);
  tmat.inv();                   // 相関行列部分を逆行列にする
  for (size_t i = 0; i < p; ++i)
  { for (size_t j = 0; j < p; ++j) rmat.assign(1 + i, 1 + j, tmat(i, j)); }
}

void reganclass::put_bci(ostream& fout, const option& opt, int n, const mat<double>& smat, const vecint& checkvec) const
// 偏回帰係数の信頼区間を出力する
{
  constexpr int LENline = 74;   // 区切り線の長さ
  fout << "\nVAR      Partial b       SE of b" << setw(4) << opt.cc() * 100.0;
  fout << "%Precision  CI:Lower-Lim     Upper-Lim\n";
  putline(fout, LENline);
  size_t d = smat.r() - 1;
  double sr = smat(d, d);
  int dfr = n - 1 - checkvec[0];
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double bi = smat(0, d), sii = smat(0, 0);
  double vbi = vr * sii;
  double ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vbi);
  fout << "ICPT" << setw(14) << bi << setw(14) << sqrt(vbi);
  fout << setw(14) << ci << setw(14) << bi - ci << setw(14) << bi + ci;
  fout << "\n";
  for (size_t i = 1; i < d; ++i)
  {
    if (checkvec[i])
    {
      bi = smat(i, d);  sii = smat(i, i);
      vbi = vr * sii;
      ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vbi);
      fout << "X" << setw(3) << i << setw(14) << bi << setw(14) << sqrt(vbi);
      fout << setw(14) << ci << setw(14) << bi - ci << setw(14) << bi + ci;
      fout << "\n";
    }
  }
  putline(fout, LENline);
}

void reganclass::put_dns(option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec)
// Yの推定値をデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout, opt, mevec, checkvec);
  datac tdata(checkvec[0] + 4);
  int dfr = n - 1 - checkvec[0];
  size_t yidx = smat.r() - 1;
  double vr = (dfr > 0) ? smat(yidx, yidx) / (double)dfr : 0.0;
  double tcc = ptot(1.0 - opt.cc(), dfr);
  if (opt.xno() < 1)            // 変化説明変数がない時
  {
    for (auto& cdata : datalst)
    {
      tdata.assign(0, cdata.fdata(0));
      size_t j = 1;
      for (size_t i = 1; i < id.ni(); ++i)
      {
        if (checkvec[i])
        {
          tdata.assign(j, cdata.fdata(i));
          ++j;
        }
      }
      double yest = cal_estimate(tdata, smat, checkvec);
      if (yest == BLANKf)
      {
        tdata.assign(j, BLANKs);
        tdata.assign(j + 1, BLANKs);  tdata.assign(j + 2, BLANKs);
      }
      else
      {
        double pi = cal_pi(cdata, tcc, vr, smat);
        tdata.assign(j, yest);
        tdata.assign(j + 1, yest - pi);  tdata.assign(j + 2, yest + pi);
      }
      tdata.put(fout);
    }
  }
  else                          // 変化説明変数がある時
  {
    size_t d = smat.r() - 1;
    datac ttdata(d);
    for (size_t i = 0; i < d; ++i)
    { ttdata.assign(i, (opt.x(i) == BLANKf) ? mevec[i] : opt.x(i)); }
    for (auto& cdata : datalst)
    {
      double y = cdata.fdata(0);
      tdata.assign(0, y);
      size_t j = 1;
      for (size_t i = 1; i < id.ni(); ++i)
      {
        if (checkvec[i])
        {
          if (i == size_t(opt.xno())) tdata.assign(j, cdata.fdata(i));
          else if (opt.x(i) == BLANKf) tdata.assign(j, mevec[i]);
          else tdata.assign(j, opt.x(i));
          ++j;
        }
      }
      double yest = cal_estimate(cdata, smat, checkvec);
      if (yest == BLANKf)
      {
        tdata.assign(j, BLANKs);
        tdata.assign(j + 1, BLANKs);  tdata.assign(j + 2, BLANKs);
      }
      else
      {
        double err = y - yest;
        ttdata.assign(opt.xno(), cdata.fdata(opt.xno()));
        yest = cal_estimate(ttdata, smat, checkvec) + err;
        double pi = cal_pi(ttdata, tcc, vr, smat);
        tdata.assign(j, yest);
        tdata.assign(j + 1, yest - pi);  tdata.assign(j + 2, yest + pi);
      }
      tdata.put(fout);
    }
  }
}

void reganclass::put_dnsid(ofstream& fout, const option& opt, const vecdbl& mevec, const vecint& checkvec)
// データファイル形式のデータ識別情報を出力する
{
  idd tid(id);
  tid.set(id.title(), checkvec[0] + 4); // 項目数だけ変更する
  ostringstream ostr;
  if (opt.pin() < 1.0) ostr << "Inclusion Criteria:p=<Pin=" << opt.pin() << "  Expulsion Criteria:p>Pout=" << opt.pout();
  else ostr << "Inclusion Criteria:F>=Fin=" << opt.fin() << "  Expulsion Criteria:F<Fout=" << opt.fout();
  tid.append_stra(ostr.str(), attnon, avalnon); // 層別情報として追加
  tid.assign_item(0, id.item(0));
  size_t j = 1;
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i])
    {
      ostr.str("");
      ostr << id.item(i);
      if (opt.xno() > 0)
      {
        if (i == (size_t)(opt.xno())) ostr << "  Variable Whoes Value Changes";
        else if (opt.x(i) == BLANKf) ostr << "  Mean=" << mevec[i];
        else ostr << "  Specified Value=" << opt.x(i);
      }
      tid.assign_item(j, ostr.str());
      ++j;
    }
  }
  ostr.str("");
  ostr << "Estimate Value of "<< id.item(0);
  tid.assign_item(j, ostr.str());
  ostr.str("");
  ostr << opt.cc() * 100.0 << "% PL of " << id.item(0) << ":Lower";
  tid.assign_item(j + 1, ostr.str());
  ostr.str("");
  ostr << opt.cc() * 100.0 << "% PL of " << id.item(0) << ":Upper";
  tid.assign_item(j + 2, ostr.str());
  tid.put(fout, formtag);
}

void reganclass::plot_regngp(option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("line"))          // 回帰直線
  {
    fout << "-s1 -dline\n";
    plot_ngpid(fout, opt, mevec, checkvec);
    plot_line(fout, opt, n, smat, mevec, checkvec);
    delml = true;
  }
  if (opt.ngp("pl"))            // 回帰直線の予測限界
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
    delml = true;
    }
    plot_ngpid(fout, opt, mevec, checkvec);
    plot_pl(fout, opt, n, smat, mevec, checkvec);
  }
  if (opt.ngp("est"))            // Yの推測値/補正値
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dmark,00 -o300\n";
    plot_ngpid(fout, opt, mevec, checkvec);
    plot_est(fout, opt, n, smat, mevec, checkvec);
  }
}

void reganclass::plot_ngpid(ofstream& fout, const option& opt, const vecdbl& mevec, const vecint& checkvec)
// Ngraph用にデータ識別情報を出力する
{
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%CRITERION VARIABLE Y:" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (checkvec[i])
    {
      fout << "%EXPLANATORY VARIABLE X" << setw(3) << i << ":" << id.item(i);
      if (opt.xno() > 0)
      {
        if (i == (size_t)(opt.xno())) fout << "  Variable Whoes Value Changes";
        else if (opt.x(i) == BLANKf) fout << "  Mean=" << mevec[i];
        else fout << "  Specified Value=" << opt.x(i);
      }
      fout << "\n";
    }
  }
  if (opt.pin() < 1.0) fout << "%Inclusion Criteria:p=<Pin=" << opt.pin() << "  Expulsion Criteria:p>Pout=" << opt.pout() << "\n";
  else fout << "%Inclusion Criteria:F>=Fin=" << opt.fin() << "  Expulsion Criteria:F<Fout=" << opt.fout() << "\n";
}

void reganclass::plot_est(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec)
// Yの推定値／補正値をNgraph用に出力する
{
  if (opt.xno() < 1)            // Y推定値と実際のYを出力
  {
    fout << "%<ESTIMATE OF Y AND Y>\n";
    for (datac ldata : datalst)
    {
      double yest = cal_estimate(ldata, smat, checkvec);
      if (yest != BLANKf) fout << yest << " " << ldata.fdata(0) << "\n";
    }
  }
  else                          // 変化説明変数XとY補正値を出力
  {
    size_t d = smat.r() - 1;
    datac cdata(d);
    for (size_t i = 0; i < d; ++i)
    { cdata.assign(i, (opt.x(i) == BLANKf) ? mevec[i] : opt.x(i)); }
    fout << "%<VARIABLE WHOES VALUE CHANGES AND CORRECTED Y>\n";
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
    {
      double x = ditr->fdata(opt.xno()), y = ditr->fdata(0);
      double yest = cal_estimate(*ditr, smat, checkvec);
      if (yest != BLANKf)
      {
        double err = y - yest;
        cdata.assign(opt.xno(), x);
        yest = cal_estimate(cdata, smat, checkvec);
        fout << x << " " << yest + err << "\n";
      }
    }
  }
}

double reganclass::cal_estimate(datac& cdata, const mat<double>& smat, const vecint& checkvec) const
// Yの推定値を計算して返す
{
  if (cdata.sdata(0) == BLANKs) return BLANKf; // 欠損があるため計算できず
  size_t d = smat.r() - 1;
  double yest = smat(0, d);
  for (size_t i = 1; i < d; ++i)
  {
    if (checkvec[i]) yest += smat(i, d) * cdata.fdata(i);
  }
  return yest;                  // 推定値を返す
}

void reganclass::plot_line(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec)
// 回帰直線をNgraph用に出力する
{
  double xmin = DBL_MAX, xmax = -DBL_MAX;
  double xrange = search_xrange(xmin, xmax, opt, smat, checkvec);
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  size_t d = smat.r() - 1;
  datac cdata(d);
  cdata.assign(0, mevec[0]);
  for (size_t i = 1; i < d; ++i)
  {
    if (checkvec[i]) cdata.assign(i, (opt.x(i) == BLANKf) ? mevec[i] : opt.x(i));
    else cdata.assign(i, 0.0);
  }
  double yest = xstart;
  if (opt.xno() > 0)
  {
    cdata.assign(opt.xno(), xstart);
    yest = cal_estimate(cdata, smat, checkvec);
  }
  fout << "%<REGRESSION LINE>\n";
  fout << xstart << " " << yest << "\n";
  yest = xend;
  if (opt.xno() > 0)
  {
    cdata.assign(opt.xno(), xend);
    yest = cal_estimate(cdata, smat, checkvec);
  }
  fout << xend << " " << yest << "\n";
}

double reganclass::search_xrange(double& xmin, double& xmax, const option& opt, const mat<double>& smat, const vecint& checkvec)
// x軸の最小値と最大値を検索して範囲を返す
{
  xmin = DBL_MAX;  xmax = -DBL_MAX;
  for (auto& cdata : datalst)
  {
    double x = (opt.xno() > 0) ? cdata.fdata(opt.xno()) : cal_estimate(cdata, smat, checkvec);
    if (x != BLANKf)            // xの最大値と最小値をサーチ
    {
      if (x < xmin) xmin = x;
      if (x > xmax) xmax = x;
    }
  }
  return (xmax - xmin) * 2.0;   // 範囲を返す
}

void reganclass::plot_pl(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec)
// 予測限界をNgraph用に出力する
{
  double xmin = DBL_MAX, xmax = -DBL_MAX;
  double xrange = search_xrange(xmin, xmax, opt, smat, checkvec);
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  constexpr uint RDdef = 50;    // プロットの繰り返し数
  double dx = (xend - xstart) / (double)RDdef;
  size_t d = smat.r() - 1;
  datac cdata(d);
  cdata.assign(0, mevec[0]);
  for (size_t i = 1; i < d; ++i)
  {
    if (checkvec[i]) cdata.assign(i, (opt.x(i) == BLANKf) ? mevec[i] : opt.x(i));
    else cdata.assign(i, 0.0);
  }
  double sr = smat(d, d);
  int dfr = n - 1 - checkvec[0];
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double t = ptot(1.0 - opt.cc(), dfr);
  double yest = xstart, xi = xstart;
  fout << "%<" << opt.cc() * 100.0 << "% PREDICTION LIMIT:UPPER>\n";
  for (uint i = 0; i <= RDdef; ++i, xi += dx) // 予測限界上限のデータ出力
  {
    if (opt.xno() > 0)
    {
      cdata.assign(opt.xno(), xi);
      yest = cal_estimate(cdata, smat, checkvec);
    }
    else
    { yest = xi; }
    double yup = yest + cal_pi(cdata, t, vr, smat);
    fout << xi << " " << yup << "\n";
  }
  fout << "= =\n";
  fout << "%<" << opt.cc() * 100.0 << "% PREDICTION LIMIT:LOWER>\n";
  xi = xstart;
  for (uint i = 0; i <= RDdef; ++i, xi += dx) // 予測限界下限のデータ出力
  {
    if (opt.xno() > 0)
    {
      cdata.assign(opt.xno(), xi);
      yest = cal_estimate(cdata, smat, checkvec);
    }
    else
    { yest = xi; }
    double ylow = yest - cal_pi(cdata, t, vr, smat);
    fout << xi << " " << ylow << "\n";
  }
}

double reganclass::cal_pi(datac& cdata, double t, double vr, const mat<double>& smat) const
// Yの予測限界幅を計算して返す
{
  if (cdata.sdata(0) == BLANKs) return BLANKf; // 欠損があるため計算できず
  size_t d = smat.r();
  vec<double> xv(d);
  for (size_t i = 1; i < d - 1; ++i) xv.assign(i, cdata.fdata(i));
  xv.assign(0, 1.0);  xv.assign(d - 1, 0.0);
  vec<double> tv(xv.dim());
  tv = xv.trn() * smat;
  double vy = vr * (1.0 + inp(tv, xv));
  return t * sqrt(vy);          // Yの予測限界を返す
}

int reganclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void reganclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  size_t dim = id.ni() + 1;     // 積和行列の次数
  mat<double> smat(dim);        // 積和行列
  mat<double> xmat(dim);        // 1とXの単純積和行列
  statv<double> ystat;          // yの基礎統計量
  vecdbl mevec(dim + 1, 0.0);   // 平均値ベクトル
  vecdbl vrvec(dim + 1, 0.0);   // 分散ベクトル
  vecdbl ssvec(dim + 1, 0.0);   // 積和ベクトル
  vecint checkvec(dim, 0);      // 変数選択用ベクトル
  int n = cal_smat(smat, xmat, ystat, mevec, vrvec, ssvec); // 例数
  int p = dim - 2;              // 説明変数の数
  if (opt.res("res"))
  {
    put_resid(cout);
    if (opt.res("stat"))
    {
      putmsdvec(cout, n, mevec, vrvec, p, 1);
      putmat(cout, "<CORRELATION COEFFICIENT MATRIX>", matr, n, smat, p, 1);
    }
  }
  if (n < p + 1)                // 重回帰式計算できず
  {
    if (opt.res("res")) cout << "\n" << endl;
    return;
  }
  step_sel(n, opt.fin(), opt.fout(), opt.pin(), opt.pout(), smat, checkvec);
  if (opt.res("res") && opt.res("selx")) put_regtab(cout, opt, n, opt.fin(), opt.fout(), opt.pin(), opt.pout(), smat, xmat, ssvec, checkvec);
  if (opt.dnsname() != "") put_dns(opt, n, smat, mevec, checkvec);
  if (opt.ngpname() != "") plot_regngp(opt, n, smat, mevec, checkvec);
  if (opt.res("res"))
  {
    if (opt.res("allx") && (opt.res("selx") == false || checkvec[0] < p))
    {                           // 全変数選択結果出力
      if (checkvec[0] < p) step_sel(n, 0.0, 0.0, 1.0, 1.0, smat, checkvec);
      put_regtab(cout, opt, n, 0.0, 0.0, 1.0, 1.0, smat, xmat, ssvec, checkvec);
    }
    cout << "\n" << endl;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  reganclass regan;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) regan.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { regan.get(argv[i], opt);  fok = false; }
    }
    regan.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
