// rmate.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 10:34:15 tac>
//
// 項目間の相関行列とその検定を計算する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W rmate rmate.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "rmate  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:rmat  [<OPTION> ...]  [<DATA FILE> ...]",
  "  Output matix of correlation coefficient, significant probability,",
  "  significant symbol.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -m={i|l|p|r}[{i|l|p|r}][{i|l|p|r}][{i|l|p|r}]:select output results",
  "    i:significant symbol  l:statistics of multicollinearity",
  "    p:significant probability  r:correlation coefficient",
  "    default:-m=ipr"
};

const string Title = "=== Correlation Coefficient Matix ==="; // 表題

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  bool vputi;                   // 有意記号行列出力フラッグ
  bool vputl;                   // 多重共線性指標出力フラッグ
  bool vputp;                   // 有意確率行列出力フラッグ
  bool vputr;                   // 相関行列出力フラッグ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  bool puti() const             // 有意記号行列出力フラッグを返す
  { return vputi; }
  bool putl() const             // 多重共線性指標出力フラッグを返す
  { return vputl; }
  bool putp() const             // 有意確率行列出力フラッグを返す
  { return vputp; }
  bool putr() const             // 相関行列出力フラッグを返す
  { return vputr; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vputi(true), vputl(false), vputp(true), vputr(true)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "m")       // 出力行列の指定
  {
    if (opt[1].empty())         // -mと指定された時
    { vputi = vputl = vputp = vputr = true; }
    else
    {
      vputi = vputl = vputp = vputr = false;
      for (size_t i = 0; i < opt[1].size(); ++i)
      {
        switch (opt[1][i])
        {
        case 'i' : vputi = true;
          break;
        case 'l' : vputl = true;
          break;
        case 'p' : vputp = true;
          break;
        case 'r' : vputr = true;
          break;
        default : err = badopt(vpname, optstr, USAGE);
        }
      }
    }
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

class rmatclass                 // rmatメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_smat(mat<double>& smat, vecdbl& mevec, vecdbl& vrvec);
  // データを集計して積和行列を求める
  void put_id(ostream& fout);   // タイトルとデータ識別情報を出力する
  void put_pmat(ostream& fout, int n, const mat<double>& smat, mattype mtype) const;
  // 有意確率・有意記号行列を出力する
  void put_ml(ostream& fout, int n, mat<double>& smat);
  // 多重共線性指標を出力する
public:
  rmatclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~rmatclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int rmatclass::cal_smat(mat<double>& smat, vecdbl& mevec, vecdbl& vrvec)
// データを集計して積和行列を求める
{
  size_t dim = id.ni() + 1;
  vecdbl fdatavec(dim, 0.0);
  fdatavec[0] = 1.0;
  for (auto& cdata : datalst)
  {
    bool blank = false;
    size_t j = 1;
    while (blank == false && j < dim)
    {
      if (cdata.fdata(j - 1) == BLANKf)
      { blank = true; }
      else
      {
        fdatavec[j] = cdata.fdata(j - 1);
        ++j;
      }
    }
    if (blank == false)
    {
      for (size_t i = 0; i < dim; ++i)
      {
        double x = fdatavec[i];
        for (size_t j = i; j < dim; ++j) smat.add(i, j, x * fdatavec[j]);
      }
    }
  }
  int n = (int)(smat(0, 0));
  size_t p = dim - 1;
  for (size_t i = 0; i < p; ++i)
  {
    for (size_t j = i + 1; j <= p; ++j) smat.assign(j, i, smat(i, j));
  }
  smat.sweep(0);
  for (size_t i = 0; i <= p; ++i)
  {
    mevec[i] = smat(0, i);
    vrvec[i] = (n > 1) ? smat(i, i) / (double)(n - 1) : 0.0;
  }
  return n;                     // 例数を返す
}

void rmatclass::put_id(ostream& fout) // タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    fout << "VARIABLE X" << setw(3) << i + 1;
    fout << ":" << id.item(i);
    fout << "\n";
  }
}

void rmatclass::put_pmat(ostream& fout, int n, const mat<double>& smat, mattype mtype) const
// 有意確率・有意記号行列を出力する
{
  constexpr int Cmax = 9;       // 行列表横の最大数
  int df = n - 2;
  int p = smat.r() - 1;
  if (df < 0 || p < 1) return;  // 出力なし
  int prec0 = fout.precision(); // 現在の有効桁数を保存する
  int low = 1;
  fout << "\n<SIGNIFICANT ";
  if (mtype == matr) fout << "PROBABILITY MATIX(p-value)>";
  else fout << "SYMBOL MATRIX(+:p<0.1  *:p<0.05  **:p<0.01  ***:p<0.001)>";
  while (low <= p)
  {
    int up = (low + Cmax - 1 < p) ? low + Cmax - 1 : p;
    fout << "\n    ";
    for (int i = low; i <= up; ++i) fout << "    X" << setw(3) << i;
    fout << "\n";
    int llen = 4 + (up - low + 1) * 8;
    putline(fout, llen);
    for (int i = 1; i <= p; ++i)
    {
      fout << "X" << setw(3) << i;
      if (mtype == matr) { fout.precision(4);  fout.setf(ios::fixed); }
      for (int j = low; j <= up; ++j)
      {
        if (i == j)
        { fout << "    -   "; }
        else
        {
          double r = smat(i, j);
          double f = (sgn(1.0 - r * r) == 0) ? 0.0 : r * r / (1.0 - r * r) * (double)df;
          ldbl pr = ftop(f, 1, df);
          if (mtype == matr)
          { fout << setw(8) << pr; }
          else
          {
            if (pr > 0.1L) fout << setw(8) << " ";
            else fout << "    " << left << setw(4) << pind(pr) << right;
          }
        }
      }
      if (mtype == matr) { fout.precision(prec0);  fout.unsetf(ios::fixed); }
      fout << "\n";
    }
    putline(fout, llen);
    low += Cmax;
  }
}

void rmatclass::put_ml(ostream& fout, int n, mat<double>& smat)
// 多重共線性指標を出力する
{
  int p = smat.r() - 1, i = 0;
  mat<double> tmat(p);
  tmat = smat.sub(1, 1, p, p);
  tmat.inv();
  for (i = 0; i < p; ++i)
  { for (int j = 0; j < p; ++j) smat.assign(1 + i, 1 + j, tmat(i, j)); }
  putmat(fout, "<INVERSE CORRELATION COEFFICIENT MATRIX>", mats, n, smat);
  constexpr int LENline = 40;
  fout << "\n<STATISTICS OF MULTICOLLINEARITY>";
  fout << "\nVAR          VIF   Tolerance  Multiple R\n";
  putline(fout, LENline);
  const double Reps = DBL_EPSILON * fmax(FDmin, tmat.maxabse());
  for (i = 0; i < p; ++i)
  {
    double rii = tmat(i, i);
    double tii = (fabs(rii) < Reps) ? 0.0 : 1.0 / rii;
    fout << "x" << setw(3) << 1 + i << setw(12) << rii << setw(12) << tii;
    fout << setw(12) << sqrt(1.0L - tii) << "\n";
  }
  putline(fout, LENline);
}

int rmatclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  return nc;
}

void rmatclass::put(option& opt) // 集計結果を出力する
{
  put_id(cout);
  if (id.nc() < 1)              // データなし
  {
    cout << "\n" << endl;
    return;
  }
  size_t dim = id.ni() + 1;     // 積和行列の次数
  mat<double> smat(dim);        // 積和行列
  vecdbl mevec(dim, 0.0), vrvec(dim, 0.0); // 平均値・分散ベクトル
  int n = cal_smat(smat, mevec, vrvec);    // 例数
  putmsdvec(cout, n, mevec, vrvec);
  if (opt.putr()) putmat(cout, "<CORRELATION COEFFICIENT MATRIX>", matr, n, smat);
  smat2rmat(smat);              // 相関行列に変換する
  if (opt.putp()) put_pmat(cout, n, smat, matr);
  if (opt.puti()) put_pmat(cout, n, smat, matetc);
  if (opt.putl()) put_ml(cout, n, smat);
  cout << "\n" << endl;
}

int main(int argc, char* argv[]) // メインルーチン
{
  rmatclass rmat;               // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) rmat.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { rmat.get(argv[i], opt);  fok = false; }
    }
    rmat.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
