// wmrae.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-30 10:42:25 tac>
//
// 最後の項目を重みw、次の項目を目的変数Y、以後の項目を説明変数Xとした
// 重み付け重回帰分析。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W wmrae wmrae.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "wmrae  Version 7.41e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:wmrae [<OPTION> ...]  [<DATA FILE> ...]",
  "  The weighted multiple regression analysis with the first item as weight",
  "  item and the second item as criterion variable and subsequent items as",
  "  explanatory variables.",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=wmrae_dns.d",
  "  -f=<FILE>:reads options from the specified file",
  "  -fin=<Fin>:specify the inclusion criteria Fin(Fin>=0)",
  "  -fout=<Fout>:specify the expulsion criteria Fout(0=<Fout=<Fin)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={est|line|pl}[,...]:select the data for Ngraph",
  "    est:estimate Y/corrected Y  line:partial regression line",
  "    pl:prediction limit  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=wmrae_ngp.dat",
  "  -pin=<Pin>:specify the inclusion criteria Pin(0<Pin=<1)",
  "  -pout=<Pout>:specify the expulsion criteria Pout(1>=Pout>=Pin)",
  "  -res=[-][{def|all|allx|atab|cc[<C-COEF>]|selx|stat}][,...]",
  "    :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    allx:result of selecting all variables  atab:ANOVA table",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1  cc:cc0.95)",
  "    selx:result of selecting all variables  stat:basic statistics",
  "  -tstat[={c|f}]:select the test statistics(-tstat:tstat=f)",
  "  -x<NUMBER>=<VALUE>:specify the number and value of explanatory variable",
  "    whoes value is constant",
  "  -xno=<NUMBER>:specify the number of explanatory variable whose value",
  "    changes"
};

const string Title = "=== Weighted Multiple Regression Analysis ==="; // 表題
const string DNSNAMEdef = "wmrae_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "wmrae_ngp.dat"; // 省略時Ngraphファイル名
constexpr double FIOdef = 2.0;  // 省略時F値の取り込み・追い出し基準
constexpr double PIOdef = 0.1;  // 省略時p値の取り込み・追い出し基準

enum tstattype                  // 検定統計量の種類
{
  tstatnon,                     // 指定なし
  tstatc,                       // χ2乗値
  tstatf,                       // F値
  tstatetc                      // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  pairdbl flevel;               // F値の取り込み基準と追い出し基準
  pairdbl plevel;               // P値の取り込み基準と追い出し基準
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  tstattype vtstat;             // 検定統計量の種類
  lstr xlst;                    // 説明変数リスト
  vecdbl vxvec;                // 説明変数ベクトル
  int vxno;                     // 変化説明変数の番号
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  double fin() const            // F値の取り込み基準を返す
  { return flevel.first; }
  double fout() const           // F値の追い出し基準を返す
  { return flevel.second; }
  double pin() const            // P値の取り込み基準を返す
  { return plevel.first; }
  double pout() const           // P値の追い出し基準を返す
  { return plevel.second; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  tstattype tstat() const       // 検定統計量の種類を返す
  { return vtstat; }
  double x(size_t idx) const    // 説明変数の値を返す
  { return (0 < idx && idx < vxvec.size()) ? vxvec[idx] : BLANKf; }
  int xno() const               // 変化説明変数の番号を返す
  { return vxno; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdnsname(), flevel(FIOdef, FIOdef), plevel(1.0, 1.0), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vtstat(tstatf), xlst(), vxvec(), vxno(0)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "fin")     // F値の取り込み基準指定
  {
    flevel.first = opt[1].empty() ? FIOdef : strtodf(opt[1]);
    if (flevel.first < 0.0) err = badopt(vpname, optstr, USAGE, 0.0, DBL_MAX);
    if (flevel.second > flevel.first) flevel.second = flevel.first;
  }
  else if (opt[0] == "fout")    // F値の追い出し基準指定
  {
    flevel.second = opt[1].empty() ? flevel.first : strtodf(opt[1]);
    if (flevel.second < 0.0 || flevel.first < flevel.second) err = badopt(vpname, optstr, USAGE, 0.0, flevel.first);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pin")     // p値の取り込み基準指定
  {
    plevel.first = opt[1].empty() ? PIOdef : strtodf(opt[1]);
    if (plevel.first <= 0.0 || 1.0 < plevel.first) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "pout")    // p値の追い出し基準指定
  {
    plevel.second = opt[1].empty() ? plevel.first : strtodf(opt[1]);
    if (plevel.second < plevel.first || 1.0 < plevel.second) err = badopt(vpname, optstr, USAGE, plevel.first, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "tstat")   // 検定統計量の種類指定
  {
    if (opt[1].empty() || opt[1] == "f") vtstat = tstatf;
    else if (opt[1] == "c") vtstat = tstatc;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "xno")     // 変化説明変数の番号指定
  {
    char* ptr = nullptr;
    vxno = strtol(opt[1].c_str(), &ptr, 0);
    if (vxno < 1) err = badopt(vpname, optstr, USAGE, 1, INT_MAX);
  }
  else if (opt[0][0] == 'x')    // 説明変数の値指定
  {
    if (strsplit(xlst, optstr, CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (ni < 3) return disperr(vpname, "At least 3 items are requied !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (vxno == 0) return 0;
  if (vxno < 0 || (int)ni - 1 <= vxno) return disperr(vpname, "Wrong number of the explanatory variable whose value changes !");
  vxvec.resize(ni);
  vxvec.assign(ni, BLANKf);
  if (xlst.empty()) return 0;
  if (xlst.size() >= ni) return disperr(vpname, "Too many values of the explanatory variables !");
  for (lstr_citr xitr = xlst.begin(); xitr != xlst.end(); ++xitr)
  {
    string opt[2];
    if (getopt(opt, *xitr, OPTdev) > 0)
    {
      char* ptr = nullptr;
      int xno = strtol(opt[0].substr(1).c_str(), &ptr, 0);
      if (xno < 1 || (int)ni <= xno || xno == vxno) return disperr(vpname, "Wrong number of the explanatory variable->", *xitr);
      double fdata = strtodf(opt[1]);
      if (fdata == BLANKf) return disperr(vpname, "Wrong value of the explanatory variable->", *xitr);
      vxvec[xno] = fdata;
    }
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["est"] = vngpmap["line"] = vngpmap["pl"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["est"] = vngpmap["line"] = vngpmap["pl"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["allx"] = vresmap["atab"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["allx"] = vresmap["atab"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["allx"] = vresmap["atab"] = vresmap["cc"] = vresmap["selx"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class wmraclass                // reganメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_smat(mat<double>& smat, mat<double>& xmat, vector<statv<double>>& wyxstatvec, vecdbl& mevec, vecdbl& vrvec, vecdbl& ssvec);
  // データを集計して積和行列を求める
  void put_resid(ostream& fout); // 解析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, int n, const vector<statv<double>>& wyxstatvec, const vecdbl& mevec, const vecdbl& vrvec);
  // 変数の基礎統計量を出力する
  void step_sel(int n, double fin, double fout, double pin, double pout, mat<double>& smat, vecint& checkvec);
  // 変数増減法による重回帰分析の計算
  int sel_var(int in, int n, double fin, double fout, double pin, double pout, const mat<double>& smat, vecint& checkvec) const;
  // 変数増減法による変数の選択
  void put_regtab(ostream& fout, option& opt, int n, double fin, double flout, double pin, double pout, const mat<double>& smat, const mat<double>& xmat, const vecdbl& ssvec, const vecint& checkvec);
  // 重回帰分析結果を出力する
  void cal_vif(mat<double>& rmat, const mat<double>& xmat, const vecint& checkvec) const;
  // VIFを計算する
  void put_bci(ostream& fout, const option& opt, int n, const mat<double>& smat, const vecint& checkvec) const;
  // 偏回帰係数の信頼区間を出力する
  void put_catab(ostream& fout, double sbw, int dfb, double srw, int dfr, double syyw, int dft) const;
  // χ2乗値による分散分析表を出力する
  void put_dns(option& opt, int n, const mat<double>& smat, const vector<statv<double>>& wyxstatvec, const vecdbl& mevec, const vecint& checkvec);
  // Yの推定値をデータファイル形式で出力する
  void put_dnsid(ofstream& fout, const option& opt, const vecdbl& mevec, const vecint& checkvec);
  // データファイル形式のデータ識別情報を出力する
  void plot_regngp(option& opt, int n, const mat<double>& smat, const vector<statv<double>>& wyxstatvec, const vecdbl& mevec, const vecint& checkvec);
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, const option& opt, const vecdbl& mevec, const vecint& checkvec);
  // Ngraph用にデータ識別情報を出力する
  void plot_est(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec);
  // Yの推定値をNgraph用に出力する
  double cal_estimate(datac& cdata, const mat<double>& smat, const vecint& checkvec) const;
  // Yの推定値を計算して返す
  void plot_line(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec);
  // 回帰直線をNgraph用に出力する
  double search_xrange(double& xmin, double& xmax, const option& opt, const mat<double>& smat, const vecint& checkvec);
  // x軸の最小値と最大値を検索して範囲を返す
  void plot_pl(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vector<statv<double>>& wyxstat, const vecdbl& mevec, const vecint& checkvec);
  // 予測限界をNgraph用に出力する
  void plot_pllu(ofstream& fout, const option& opt, datac& tdata, const mat<double>& smat, const vector<statv<double>>& wyxstat, const vecint& checkvec, double t, double vr, double xstart, double xend, int rd, bool cup) const;
  // 下限／上限の予測限界をNgraph用に出力する
  double cal_pi(datac& cdata, double tcc, double vr, const mat<double>& smat, const vector<statv<double>>& wyxstatvec) const;
  // 回帰直線の予測限界幅を計算して返す
public:
  wmraclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~wmraclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

void wmraclass::put_resid(ostream& fout)
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "WEIGHT W:" << id.item(0) << "\n";
  fout << "CRITERION VARIABLE Y:" << id.item(1) << "\n";
  for (size_t i = 2; i < id.ni(); ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i - 1;
    fout << ":" << id.item(i) << "\n";
  }
}

int wmraclass::cal_smat(mat<double>& smat, mat<double>& xmat, vector<statv<double>>& wyxstatvec, vecdbl& mevec, vecdbl& vrvec, vecdbl& ssvec)
// データを集計して積和行列を求める
{
  size_t nd = id.ni();          // 項目数
  vecdbl fdatavec(nd, 0.0);
  const double Deps = DBL_EPSILON * fmax(FDmin, smat.maxabse());
  size_t yidx = nd - 1;         // fdatavec中のyのインデックス
  for (auto& cdata : datalst)
  {
    bool blank = false;
    double w = 0.0;
    if (cdata.fdata(0) == BLANKf || cdata.fdata(0) < Deps) blank = true;
    else w = cdata.fdata(0);
    double yw = 0.0;
    if (cdata.fdata(1) == BLANKf) blank = true;
    else yw = w * cdata.fdata(1);
    size_t i = 2;
    while (blank == false && i < nd)
    {
      if (cdata.fdata(i) == BLANKf)
      { blank = true; }
      else
      {
        fdatavec[i - 1] = w * cdata.fdata(i);
        ++i;
      }
    }
    if (blank)
    { cdata.assign(0, BLANKs); } // 後の計算のためにwをブランクにする
    else
    {
      wyxstatvec[0].sigma(w);   // wの基礎統計量
      fdatavec[0] = w;  fdatavec[yidx] = yw;
      for (i = 0; i < nd; ++i)
      {
        if (i > 0) wyxstatvec[i].sigma(cdata.fdata(i)); // yとxの基礎統計量
        double fwi = fdatavec[i];
        for (size_t j = i; j < nd; ++j) smat.add(i, j, fwi * fdatavec[j]);
      }
    }
  }
  int n = wyxstatvec[0].n();
  xmat = smat;
  for (size_t i = 0; i < nd - 1; ++i)
  {
    for (size_t j = i + 1; j < nd; ++j)
    { smat.assign(j, i, smat(i, j));  xmat.assign(j, i, xmat(i, j)); }
  }
  smat.sweep(0);
  for (size_t i = 0; i < nd; ++i)
  {
    mevec[i] = smat(0, i);
    vrvec[i] = (n > 1) ? smat(i, i) / (double)(n - 1) : 0.0;
    ssvec[i] = smat(i, i);
  }
  return n;                     // 例数を返す
}

void wmraclass::put_stat(ostream& fout, int n, const vector<statv<double>>& wyxstatvec, const vecdbl& mevec, const vecdbl& vrvec)
// 変数の基礎統計量を出力する
{
  size_t nd = id.ni();
  size_t yidx = nd - 1;
  fout << "\n<BASIC STATISTICS AND WEIGHTED STATISTICS>\n";
  constexpr int LENline = 77;
  putline(fout, LENline);
  fout << "W   :";
  putmsd(fout, wyxstatvec[0].n(), wyxstatvec[0].mean(), wyxstatvec[0].vr());
  fout << "Y   :";
  putmsd(fout, n, wyxstatvec[1].mean(), wyxstatvec[1].vr());
  for (size_t j = 2; j < nd; ++j)
  {
    fout << "X" << setw(3) << j - 1 << ":";
    putmsd(fout, n, wyxstatvec[j].mean(), wyxstatvec[j].vr());
  }
  putline(fout, LENline);
  fout << "YW  :";
  putmsd(fout, n, mevec[yidx], vrvec[yidx]);
  for (size_t j = 1; j < yidx; ++j)
  {
    fout << "XW" << setw(2) << j << ":";
    putmsd(fout, n, mevec[j], vrvec[j]);
  }
  putline(fout, LENline);
}

void wmraclass::step_sel(int n, double fin, double fout, double pin, double pout, mat<double>& smat, vecint& checkvec)
// 変数増減法による重回帰分析の計算
{
  int in = 1, i = 0;
  do {
    in = 1;
    while ((i = sel_var(in, n, fin, fout, pin, pout, smat, checkvec)) > 0)
    {
      smat.sweep(i);  checkvec[i] = in ? 1 : 0;
      checkvec[0] += in ? 1 : -1;
      in = !in;
    }
  } while (!in);                // 取り込む変数がなくなるまで続ける
}

int wmraclass::sel_var(int in, int n, double fin, double fout, double pin, double pout, const mat<double>& smat, vecint& checkvec) const
// 変数増減法による変数の選択
{
  int yidx = id.ni() - 1, i = 0;
  double dsel = in ? FDmin : FDmax;
  const double Seps = DBL_EPSILON * fmax(FDmin, smat.maxabse());
  bool fsel = true;             // F値による変数選択
  if (pin < 1.0) fsel = false;  // 有意確率pによる変数選択
  for (int j = 1; j < yidx; ++j)
  {
    if (checkvec[j] != in && fabs(smat(j, j)) > Seps)
    {
      double dj = smat(j, yidx) * smat(j, yidx) / smat(j, j);
      if ((in && dj > dsel) || (!in && dj < dsel)) { dsel = dj;  i = j; }
    }
  }
  if (i < 1) return i;          // 選択すべき変数なし
  if (in)                       // 変数取り込み
  {
    int df = n - checkvec[0] - 2;
    double f = dsel * (double)df / (smat(yidx, yidx) - dsel);
    if (fsel)
    {
      if (f < fin) i = 0;
    }
    else
    {
      ldbl p = ftop(f, 1, df);
      if (p > (ldbl)pin) i = 0;
    }
  }
  else                          // 変数追い出し
  {
    int df = n - checkvec[0] - 1;
    double f = dsel * (double)df / smat(yidx, yidx);
    if (fsel)
    {
      if (f >= fout) i = 0;
    }
    else
    {
      ldbl p = ftop(f, 1, df);
      if (p <= (ldbl)pout) i = 0;
    }
  }
  return i;                     // 選択すべき変数のインデックスを返す
}

void wmraclass::put_regtab(ostream& fout, option& opt, int n, double fin, double flout, double pin, double pout, const mat<double>& smat, const mat<double>& xmat, const vecdbl& ssvec, const vecint& checkvec)
// 重回帰分析結果を出力する
{
  constexpr int LENline = 79;   // 区切り線の長さ
  int p = checkvec[0];
  mat<double> rmat(p + 1, p + 1);
  if (p > 0) cal_vif(rmat, xmat, checkvec); // 変数が選択されている時だけrmatを計算
  if (fin <= 0.0 && pin >= 1.0)
  { fout << "\n<RESULT OF SELECTING ALL VARIABLES>\n"; }
  else
  {
    fout << "\n<STEPWISE FORWARD SELECTION METHOD>\n";
    if (pin < 1.0) fout << "  Inclusion Criterion:p=<Pin=" << pin << "  Expulsion Criteria:p>Pout=" << pout << "\n";
    else fout << "  Inclusion Criteria:F>=Fin=" << fin << "  Expulsion Criteria:F<Fout=" << flout << "\n";
  }
  fout << "VAR    Partial b  Standard b   Partial r         VIF";
  if (opt.tstat() == tstatc) fout << " Chi-squared";
  else fout << "     F-value";
  fout << "     p-value\n";
  putline(fout, LENline);
  int yidx = id.ni() - 1;
  double syy = ssvec[yidx], sr = smat(yidx, yidx);
  int dfr = n - 1 - checkvec[0];
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  if (opt.tstat() == tstatc) vr = 1.0; // 検定統計量としてχ2乗値を使う
  double sii = smat(0, 0), bi = smat(0, yidx);
  double vbi = vr * sii;
  double fi = (vbi > 0.0) ? bi * bi / vbi : 0.0;
  double pr = ftop(fi, 1, dfr);
  fout << "ICPT" << setw(12) << bi << setw(36) << "" << setw(12) << fi;
  fout << setw(12) << pr << pind(pr) << "\n";
  int k = 1;
  for (int i = 1; i < yidx; ++i)
  {
    if (checkvec[i])
    {
      bi = smat(i, yidx);  sii = smat(i, i);
      vbi = vr * sii;
      double bsti = (syy > 0.0) ? bi * sqrt(ssvec[i] / syy) : 0.0;
      double v = sr * sii + bi * bi;
      double r = (v > 0.0) ? bi / sqrt(v) : 0.0;
      fi = (vbi > 0.0) ? bi * bi / vbi : 0.0;
      pr = (opt.tstat() == tstatc) ? ctop(fi, 1) : ftop(fi, 1, dfr);
      fout << "X" << setw(3) << i << setw(12) << bi << setw(12) << bsti;
      fout << setw(12) << r << setw(12) << rmat(k, k) << setw(12) << fi;
      fout << setw(12) << pr << pind(pr) << "\n";
      ++k;
    }
  }
  putline(fout, LENline);
  if (opt.res("cc")) put_bci(fout, opt, n, smat, checkvec);
  if (checkvec[0] < 1 || syy <= 0.0) return; // 重相関係数等計算不要
  double r2 = (syy > 0.0) ? (syy - sr) / syy : 0.0;
  double r2a = (syy > 0.0) ? fabs(1.0 - vr * (double)(n - 1) / syy) : 0.0;
  double aic = (double)n * log(sr / (double)n) + (double)(2 * (p + 2));
  fout << "Multiple Contribution Rate:R^2=" << r2;
  fout << "  DF-adjusted:R'^2=" << r2a << "\n";
  fout << "Multiple Correlation Coefficient:R=" << sqrt(r2) << "  DF-adjusted:R'=" << sqrt(r2a) << "\n";
  fout << "AIC=" << aic << "\n\n";
  if (opt.res("atab"))
  {
    if (opt.tstat() == tstatc) put_catab(fout, syy - sr, p, sr, dfr, syy, n - 1);
    else putatab(fout, "REGRESSION", syy - sr, p, sr, dfr, syy, n - 1);
  }
}

void wmraclass::cal_vif(mat<double>& rmat, const mat<double>& xmat, const vecint& checkvec) const
// VIFを計算する
{
  int p = checkvec[0], yidx = id.ni() - 1;
  rmat.assign(0, 0, xmat(0, 0));
  int l = 1;
  for (int j = 1; j < yidx; ++j)
  {
    if (checkvec[j])
    {
      rmat.assign(0, l, xmat(0, j));  rmat.assign(l, 0, xmat(0, j));
      ++l;
    }
  }
  int k = 1;
  for (int i = 1; i < yidx; ++i)
  {
    if (checkvec[i])
    {
      rmat.assign(k, k, xmat(i, i));
      l = 1;
      for (int j = 1; j < yidx; ++j)
      {
        if (checkvec[j])
        {
          rmat.assign(k, l, xmat(i, j));  rmat.assign(l, k, xmat(i, j));
          ++l;
        }
      }
      ++k;
    }
  }
  rmat.sweep(0);                // 積和行列にする
  smat2rmat(rmat);              // 相関行列にする
  mat<double> tmat(rmat.sub(1, 1, p, p));
  tmat.inv();                   // 相関行列部分を逆行列にする
  for (int i = 0; i < p; ++i)
  { for (int j = 0; j < p; ++j) rmat.assign(1 + i, 1 + j, tmat(i, j)); }
}

void wmraclass::put_bci(ostream& fout, const option& opt, int n, const mat<double>& smat, const vecint& checkvec) const
// 偏回帰係数の信頼区間を出力する
{
  constexpr int LENline = 74;   // 区切り線の長さ
  fout << "\nVAR      Partial b       SE of b" << setw(4) << opt.cc() * 100.0;
  fout << "%Precision  CI:Lower-Lim     Upper-Lim\n";
  putline(fout, LENline);
  int yidx = id.ni() - 1;
  double sr = smat(yidx, yidx);
  int dfr = n - 1 - checkvec[0];
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  if (opt.tstat() == tstatc) vr = 1.0; // 検定統計量としてχ2乗値を使う
  double bi = smat(0, yidx), sii = smat(0, 0);
  double vbi = vr * sii;
  double ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vbi);
  fout << "ICPT" << setw(14) << bi << setw(14) << sqrt(vbi);
  fout << setw(14) << ci << setw(14) << bi - ci;
  fout << setw(14) << bi + ci << "\n";
  for (int i = 1; i < yidx; ++i)
  {
    if (checkvec[i])
    {
      bi = smat(i, yidx);  sii = smat(i, i);
      vbi = vr * sii;
      ci = (opt.tstat() == tstatc) ? calmcci(mcfis, opt.cc(), 1, 1, 0, 1, vbi) :  calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vbi);
      fout << "X" << setw(3) << i << setw(14) << bi;
      fout << setw(14) << sqrt(vbi) << setw(14) << ci;
      fout << setw(14) << bi - ci << setw(14) << bi + ci << "\n";
    }
  }
  putline(fout, LENline);
}

void wmraclass::put_catab(ostream& fout, double sbw, int dfb, double srw, int dfr, double syyw, int dft) const
// χ2乗値による分散分析表を出力する
{
  constexpr int LENline = 68;
  const string Head = "ANOVA table";
  fout << setw((LENline - strclen(Head, CHARwide)) / 2 + Head.size()) << Head;
  fout << "\n VAR        Sum of Squares        DF    Chi-squred       p-value\n";
  putline(fout, LENline);
  ldbl pr = ctop(sbw, dfb);
  fout << "REGRESSION" << setw(16) << sbw << setw(10) << dfb << setw(14) << sbw;
  fout << setw(14) << pr << pind(pr) << "\n";
  pr = ctop(srw, dfr);
  fout << "RESIDUALS" << setw(17) << srw << setw(10) << dfr << setw(14) << srw;
  fout << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  pr = ctop(syyw, dft);
  fout << "TOTAL" << setw(21) << syyw << setw(10) << dft << setw(14) << syyw;
  fout << setw(14) << pr << pind(pr) << endl;
}

void wmraclass::put_dns(option& opt, int n, const mat<double>& smat, const vector<statv<double>>& wyxstatvec, const vecdbl& mevec, const vecint& checkvec)
// Yの推定値をデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout, opt, mevec, checkvec);
  datac cdata(checkvec[0] + 5);
  double dfr = n - 1 - checkvec[0];
  int yidx = smat.r() - 1;
  double vr = (dfr > 0) ? smat(yidx, yidx) / (double)dfr : 0.0;
  double tcc = ptot(1.0 - opt.cc(), dfr);
  if (opt.tstat() == tstatc) // 検定統計量としてχ2乗値を使う
  {
    vr = 1.0;
    tcc = ptoz(1.0 - opt.cc());
  }
  if (opt.xno() < 1)            // Y推定値と実際のYを出力
  {
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
    {
      cdata.assign(0, ditr->fdata(0));
      cdata.assign(1, ditr->fdata(1));
      size_t j = 2;
      for (size_t i = 2; i < id.ni(); ++i)
      {
        if (checkvec[i - 1])
        {
          cdata.assign(j, ditr->fdata(i));
          ++j;
        }
      }
      double yest = cal_estimate(*ditr, smat, checkvec);
      if (yest == BLANKf)
      {
        cdata.assign(j, BLANKs);
        cdata.assign(j + 1, BLANKs);  cdata.assign(j + 2, BLANKs);
      }
      else
      {
        double pi = cal_pi(*ditr, tcc, vr, smat, wyxstatvec);
        cdata.assign(j, yest);
        cdata.assign(j + 1, yest - pi);  cdata.assign(j + 2, yest + pi);
      }
      cdata.put(fout);
    }
  }
  else                          // 変化説明変数XとY補正値を出力
  {
    size_t nd = id.ni();
    size_t yidx = nd - 1;
    datac tdata(nd);
    tdata.assign(0, mevec[0]);  tdata.assign(1, mevec[yidx]);
    for (size_t i = 2; i < nd; ++i)
    {
      size_t xno = i - 1;
      tdata.assign(i, (opt.x(xno) == BLANKf) ? mevec[xno] : opt.x(xno));
    }
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
    {
      double w = ditr->fdata(0), y = ditr->fdata(1);
      cdata.assign(0, w);  cdata.assign(1, y);
      size_t j = 2;
      for (size_t i = 2; i < id.ni(); ++i)
      {
        size_t xno = i - 1;
        if (checkvec[xno])
        {
          if (xno == size_t(opt.xno())) cdata.assign(j, ditr->fdata(i));
          else if (opt.x(xno) == BLANKf) cdata.assign(j, mevec[xno]);
          else cdata.assign(j, opt.x(xno));
          ++j;
        }
      }
      double yest = cal_estimate(*ditr, smat, checkvec);
      if (yest == BLANKf)
      {
        cdata.assign(j, BLANKs);
        cdata.assign(j + 1, BLANKs);  cdata.assign(j + 2, BLANKs);
      }
      else
      {
        double err = y - yest;
        size_t xidx = 1 + opt.xno();
        tdata.assign(xidx, ditr->fdata(xidx));
        yest = cal_estimate(tdata, smat, checkvec) + err;
        double pi = cal_pi(tdata, tcc, vr, smat, wyxstatvec);
        cdata.assign(j, yest);
        cdata.assign(j + 1, yest - pi);  cdata.assign(j + 2, yest + pi);
      }
      cdata.put(fout);
    }
  }
}

void wmraclass::put_dnsid(ofstream& fout, const option& opt, const vecdbl& mevec, const vecint& checkvec)
// データファイル形式のデータ識別情報を出力する
{
  idd tid(id);
  tid.set(id.title(), checkvec[0] + 5); // 項目数だけ変更する
  ostringstream ostr;
  if (opt.pin() < 1.0) ostr << "Inclusion Criteria:p=<Pin=" << opt.pin() << "  Expulsion Criteria:p>Pout=" << opt.pout();
  else ostr << "Inclusion Criteria:F>=Fin=" << opt.fin() << "  Expulsion Criteria:F<Fout=" << opt.fout();
  tid.append_stra(ostr.str(), attnon, avalnon); // 層別情報として追加する
  tid.assign_item(0, id.item(0));
  tid.assign_item(1, id.item(1));
  size_t j = 2;
  for (size_t i = 2; i < id.ni(); ++i)
  {
    int xno = i - 1;
    if (checkvec[xno])
    {
      ostr.str("");
      ostr << id.item(i);
      if (opt.xno() > 0)
      {
        if (xno == opt.xno()) ostr << "  Variable Whoes Value Changes";
        else if (opt.x(xno) == BLANKf) ostr << "  Mean=" << mevec[xno];
        else ostr << "  Specified Value=" << opt.x(xno);
      }
      tid.assign_item(j, ostr.str());
      ++j;
    }
  }
  ostr.str("");
  ostr << "Estimate Value of " << id.item(1);
  tid.assign_item(j, ostr.str());
  ostr.str("");
  ostr << opt.cc() * 100.0 << "% PL of " << id.item(0) << ":Lower";
  tid.assign_item(j + 1, ostr.str());
  ostr.str("");
  ostr << opt.cc() * 100.0 << "% PL of " << id.item(0) << ":Upper";
  tid.assign_item(j + 2, ostr.str());
  tid.put(fout, formtag);
}

void wmraclass::plot_regngp(option& opt, int n, const mat<double>& smat, const vector<statv<double>>& wyxstatvec, const vecdbl& mevec, const vecint& checkvec)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("line"))          // 回帰直線
  {
    fout << "-s1 -dline\n";
    plot_ngpid(fout, opt, mevec, checkvec);
    plot_line(fout, opt, n, smat, mevec, checkvec);
    delml = true;
  }
  if (opt.ngp("pl"))            // 回帰直線の予測限界
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout, opt, mevec, checkvec);
    plot_pl(fout, opt, n, smat, wyxstatvec, mevec, checkvec);
  }
  if (opt.ngp("est"))            // Yの推測値/補正値
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dmark,00 -o300\n";
    plot_ngpid(fout, opt, mevec, checkvec);
    plot_est(fout, opt, n, smat, mevec, checkvec);
  }
}

void wmraclass::plot_ngpid(ofstream& fout, const option& opt, const vecdbl& mevec, const vecint& checkvec)
// Ngraph用にデータ識別情報を出力する
{
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%WEIGHT W:" << id.item(0) << "\n";
  fout << "%CRITERION VARIABLE Y:" << id.item(1) << "\n";
  for (size_t i = 2; i < id.ni(); ++i)
  {
    int xno = i - 1;
    if (checkvec[xno])
    {
      fout << "%EXPLANATORY VARIABLE X" << setw(3) << xno << ":" << id.item(i);
      if (opt.xno() > 0)
      {
        if (xno == (opt.xno())) fout << "  Variable Whoes Value Changes";
        else if (opt.x(xno) == BLANKf) fout << "  Mean=" << mevec[xno];
        else fout << "  Specified Value=" << opt.x(xno);
      }
      fout << "\n";
    }
  }
  if (opt.pin() < 1.0) fout << "%Inclusion Criteria:p=<Pin=" << opt.pin() << "  Expulsion Criteria:p>Pout=" << opt.pout() << "\n";
  else fout << "%Inclusion Criteria:F>=Fin=" << opt.fin() << "  Expulsion Criteria:F<Fout=" << opt.fout() << "\n";
}

void wmraclass::plot_est(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec)
// Yの推定値／補正値をNgraph用に出力する
{
  if (opt.xno() < 1)            // Y推定値と実際のYを出力
  {
    fout << "%<ESTIMATE OF Y AND Y>\n";
    for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
    {
      double yest = cal_estimate(*ditr, smat, checkvec);
      if (yest != BLANKf) fout << yest << " " << ditr->fdata(1) << "\n";
    }
  }
  else                          // 変化説明変数XとY補正値を出力
  {
    size_t nd = id.ni();
    size_t yidx = nd - 1;
    datac tdata(nd);
    tdata.assign(0, mevec[0]);  tdata.assign(1, mevec[yidx]);
    for (size_t i = 2; i < nd; ++i)
    {
      size_t xno = i - 1;
      tdata.assign(i, (opt.x(xno) == BLANKf) ? mevec[xno] : opt.x(xno));
    }
    fout << "%<VARIABLE WHOES VALUE CHANGES AND CORRECTED Y>\n";
    for (auto& cdata : datalst)
    {
      double yest = cal_estimate(cdata, smat, checkvec);
      if (yest != BLANKf)
      {
        double err = cdata.fdata(1) - yest;
        size_t xidx = 1 + opt.xno();
        tdata.assign(xidx, cdata.fdata(xidx));
        yest = cal_estimate(tdata, smat, checkvec);
        fout << cdata.fdata(xidx) << " " << yest + err << "\n";
      }
    }
  }
}

double wmraclass::cal_estimate(datac& cdata, const mat<double>& smat, const vecint& checkvec) const
// Yの推定値を計算して返す
{
  if (cdata.sdata(0) == BLANKs) return BLANKf; // 欠損があるため計算できず
  size_t yidx = id.ni() - 1;
  double yest = smat(0, yidx);
  for (size_t i = 1; i < yidx; ++i)
  {
    if (checkvec[i]) yest += smat(i, yidx) * cdata.fdata(1 + i);
  }
  return yest;                  // 推定値を返す
}

void wmraclass::plot_line(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vecdbl& mevec, const vecint& checkvec)
// 回帰直線をNgraph用に出力する
{
  double xmin = DBL_MAX, xmax = -DBL_MAX;
  double xrange = search_xrange(xmin, xmax, opt, smat, checkvec);
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  size_t nd = id.ni();
  size_t yidx = nd - 1;
  datac tdata(nd);
  tdata.assign(0, mevec[0]);  tdata.assign(1, mevec[yidx]);
  for (size_t i = 2; i < nd; ++i)
  {
    size_t xno = i - 1;
    tdata.assign(i, (opt.x(xno) == BLANKf) ? mevec[xno] : opt.x(xno));
  }
  double yest = xstart;
  if (opt.xno() > 0)
  {
    tdata.assign(1 + opt.xno(), xstart);
    yest = cal_estimate(tdata, smat, checkvec);
  }
  fout << "%<REGRESSION LINE>\n";
  fout << xstart << " " << yest << "\n";
  yest = xend;
  if (opt.xno() > 0)
  {
    tdata.assign(1 + opt.xno(), xend);
    yest = cal_estimate(tdata, smat, checkvec);
  }
  fout << xend << " " << yest << endl;
}

double wmraclass::search_xrange(double& xmin, double& xmax, const option& opt, const mat<double>& smat, const vecint& checkvec)
// x軸の最小値と最大値を検索して範囲を返す
{
  xmin = DBL_MAX;  xmax = -DBL_MAX;
  for (auto& cdata : datalst)
  {
    double x = (opt.xno() > 0) ? cdata.fdata(1 + opt.xno()) : cal_estimate(cdata, smat, checkvec);
    if (x != BLANKf)            // xの最大値と最小値をサーチ
    {
      if (x < xmin) xmin = x;
      if (x > xmax) xmax = x;
    }
  }
  return (xmax - xmin) * 2.0;   // 範囲を返す
}

void wmraclass::plot_pl(ofstream& fout, const option& opt, int n, const mat<double>& smat, const vector<statv<double>>& wyxstatvec, const vecdbl& mevec, const vecint& checkvec)
// 予測限界をNgraph用に出力する
{
  double xmin = DBL_MAX, xmax = -DBL_MAX;
  double xrange = search_xrange(xmin, xmax, opt, smat, checkvec);
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  constexpr int RDdef = 50;     // プロットの繰り返し数
  size_t nd = id.ni();
  size_t yidx = nd - 1;
  datac tdata(nd);
  tdata.assign(0, wyxstatvec[0].mean());  tdata.assign(1, mevec[yidx]);
  for (size_t i = 2; i < nd; ++i)
  {
    size_t xno = i - 1;
    if (checkvec[xno]) tdata.assign(i, (opt.x(xno) == BLANKf) ? mevec[xno] : opt.x(xno));
    else tdata.assign(i, 0.0);
  }
  double sr = smat(yidx, yidx);
  int dfr = n - 1 - checkvec[0];
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double tcc = ptot(1.0 - opt.cc(), dfr);
  if (opt.tstat() == tstatc) // 検定統計量としてχ2乗値を使う
  {
    vr = 1.0;
    tcc = ptoz(1.0 - opt.cc());
  }
  fout << "%<" << opt.cc() * 100.0 << "% PREDICTION LIMIT:UPPER>\n";
  plot_pllu(fout, opt, tdata, smat, wyxstatvec, checkvec, tcc, vr, xstart, xend, RDdef, true);
  fout << "= =\n";
  fout << "%<" << opt.cc() * 100.0 << "% PREDICTION LIMIT:LOWER>\n";
  plot_pllu(fout, opt, tdata, smat, wyxstatvec, checkvec, tcc, vr, xstart, xend, RDdef, false);
}

void wmraclass::plot_pllu(ofstream& fout, const option& opt, datac& cdata, const mat<double>& smat, const vector<statv<double>>& wyxstatvec, const vecint& checkvec, double t, double vr, double xstart, double xend, int rd, bool cup) const
// 下限／上限の予測限界をNgraph用に出力する
{
  double dx = (xend - xstart) / (double)rd;
  double yest = xstart, xi = xstart;
  for (int i = 0; i <= rd; ++i, xi += dx)
  {
    if (opt.xno() > 0)
    {
      cdata.assign(1 + opt.xno(), xi);
      yest = cal_estimate(cdata, smat, checkvec);
    }
    else
    { yest = xi; }
    double pi = cal_pi(cdata, t, vr, smat, wyxstatvec);
    if (cup) yest += pi;
    else yest -= pi;
    fout << xi << " " << yest << "\n";
  }
}

double wmraclass::cal_pi(datac& cdata, double tcc, double vr, const mat<double>& smat, const vector<statv<double>>& wyxstatvec) const
// Yの予測限界幅を計算して返す
{
  if (cdata.sdata(0) == BLANKs) return BLANKf; // 欠損があるため計算できず
  size_t nd = id.ni();
  size_t yidx = nd - 1;
  vec<double> xv(nd);
  double w = cdata.fdata(0);
  for (size_t i = 1; i < yidx; ++i) xv.assign(i, w * cdata.fdata(1 + i));
  xv.assign(0, w);  xv.assign(yidx, 0.0);
  vec<double> tv(xv.trn() * smat);
  double vwy = vr * (1.0 + inp(tv, xv));
  const double Veps = DBL_EPSILON * fmax(FDmin, wyxstatvec[0].vr());
  if (vwy < Veps) vwy = 0.0;
  return tcc * sqrt(vwy) / wyxstatvec[0].mean(); // Yの予測限界を返す
}

int wmraclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void wmraclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
      return;
    }
  }
  size_t nd = id.ni();          // 項目数
  mat<double> smat(nd);         // 積和行列
  mat<double> xmat(nd);         // 単純積和行列
  vector<statv<double>> wyxstatvec(nd); // wとyとxの基礎統計量
  vecdbl mevec(nd, 0.0);        // 平均値ベクトル
  vecdbl vrvec(nd, 0.0);        // 分散ベクトル
  vecdbl ssvec(nd, 0.0);        // 積和ベクトル
  vecint checkvec(nd - 1, 0);   // 変数選択用ベクトル
  int n = cal_smat(smat, xmat, wyxstatvec, mevec, vrvec, ssvec);
  int p = nd - 2;               // 説明変数の数
  if (opt.res("res"))
  {
    put_resid(cout);
    if (opt.res("stat"))
    {
      put_stat(cout, n, wyxstatvec, mevec, vrvec);
      putmat(cout, "<CORRELATION COEFFICIENT MATRIX>", matr, n, smat, p, 1);
    }
  }
  if (n < p + 1)                // 重回帰式計算できず
  {
    if (opt.res("res")) cout << "\n" << endl;
    return;
  }
  step_sel(n, opt.fin(), opt.fout(), opt.pin(), opt.pout(), smat, checkvec);
  if (opt.res("res") && opt.res("selx")) put_regtab(cout, opt, n, opt.fin(), opt.fout(), opt.pin(), opt.pout(), smat, xmat, ssvec, checkvec);
  if (opt.dnsname() != "") put_dns(opt, n, smat, wyxstatvec, mevec, checkvec);
  if (opt.ngpname() != "") plot_regngp(opt, n, smat, wyxstatvec, mevec, checkvec);
  if (opt.res("res"))
  {
    if (opt.res("allx") && (opt.res("selx") == false || checkvec[0] < p))
    {                           // 全変数選択結果出力
      if (checkvec[0] < p) step_sel(n, 0.0, 0.0, 1.0, 1.0, smat, checkvec);
      put_regtab(cout, opt, n, 0.0, 0.0, 1.0, 1.0, smat, xmat, ssvec, checkvec);
    }
    cout << "\n" << endl;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  wmraclass wmra;               // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) wmra.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { wmra.get(argv[i], opt);  fok = false; }
    }
    wmra.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
